import time, datetime
import threading
import os
import json
import werkzeug

from threading import Event

import m_utils
import m_tree
import m_api
import m_rest

ACTIONS_API_PATH = 'api.actions'
ACTIONS_URL = '/actions'

ACTION_NODE_TYPE = 'Action'
ACTION_BROWSE_FORM_TYPE = 'Action'
ACTION_EDIT_FORM_TYPE = 'Action_edit'

#=========================================================================
#   Инициализация модуля
#    - регистрация интерфейса в api.actions
#    - регистрация endpoint /actions
#=========================================================================
api = None
_debug = False

def init(_api, debug=False):                                     # Инициализация модуля
    global api
    api = _api

    global _debug
    _debug = debug

    # Регистрация модуля в структуре API
    api.add_module_api(ACTIONS_API_PATH, ApiActions())

    api.repo.add('add_action', repo_add_action)
    api.repo.provide(['add_action'])

    api.add_url(ACTIONS_URL, rest_endpoint)


#=========================================================================
#   Контейнер действия
#=========================================================================

class ActionThread:                                 # Контейнер для выполняемого действия
    STATUS_WAIT = 'wait'
    STATUS_ACTIVE = 'active'
    STATUS_COMPLETED = 'completed'
    STATUS_CANCELED = 'canceled'
    STATUS_ERROR = 'error'

    def __init__(self, func, kwargs={}):
        self.proc_id = None             # Код процесса в таблице actions
        self.progress = 0
        self.progress2 = None
        self.status = ActionThread.STATUS_WAIT
        self.status_msg = ''
        self.status_msg2 = ''
        self.status_start2 = 0
        self.status_last_msg2 = ''
        self.log_msgs = []
        self.result = None              # Результат выполнения действия
        self.thread_id = None           # Идентификатор потока выполнения
        self.data = {}

        self.data = []

        self.done = Event()  # ключ координации потоков для прерывания потока

        all_kwargs = {
            'api'    : api,
            'func'   : func,
            'request': api.context.get('request'),
            'user'   : api.context.get('user'),
            'done'   : self.done,
        }
        all_kwargs.update(kwargs)

        self.thread = threading.Thread(target=self.thread_func, kwargs=all_kwargs, daemon=True)
        #
        #

    def thread_func(self, **kwargs):                # Обертка для запуска функции в отдельном потоке
        try:
            api = kwargs['api']

            # Создание контекста потока
            api.build_context()                     # Создаем отдельное подключение к БД для потока
            
            # Регистрация контектса потока
            self.thread_id = api.context._get_th_id()       # Идентификатор потока

            # Установка параметров передаваемых через контекст
            api.context['user'] = kwargs.pop('user')
            api.context['request'] = kwargs.pop('request')
            api.context['action'] = self

            func = kwargs.pop('func')

            self.set_message('Выполняется')

            '''
                kwargs:
                    api         - (__init__)

                context:
                    user        - (__init__)
                    request     - (__init__)
                    action      - (this)
            '''
            self.result = func( **kwargs )

            if self.is_active():
                self.complete()

                if api.context.find('db_connect'):
                    api.context.db_connect.commit()

            else:
                if api.context.find('db_connect'):
                    api.context.db_connect.rollback()

        except Exception as e:
            self.result = None
            self.status = ActionThread.STATUS_ERROR

            if api.context.find('db_connect'):
                api.context.db_connect.rollback()

            if isinstance(e, m_api.ApiException):
                msg = '\n'.join(e.args)
                self.set_message(msg)
            else:
                msg = m_utils.exception_text(e)
                self.set_message('Ошибка при выполнении') 
                self.add_log( msg, '-'*40 )

            api.log.error( msg )

        if api.context.find('db_connect'):
            api.context.db_connect.disconnect()

        api.context.clear()
            #
            #
    def complete(self):                             # Установка состояния "действие выполнено"
        self.status = ActionThread.STATUS_COMPLETED
        self.set_progress(100)
        try:
            self.done.set()  # прерываем поток
            self.thread.join()
            import gc
            gc.collect()
        except:
            pass
        self.set_message('Выполнено')
            #
            #
    def start(self):                                # Запуск действия
        self.status = ActionThread.STATUS_ACTIVE
        self.thread.start()
        #
        #
    def stop(self):                                 # Принудительная остановка действия
        self.status = ActionThread.STATUS_CANCELED
        self.done.set()  # прерываем поток
        self.thread.join()
        import gc
        gc.collect()
        self.set_message('Прервано пользователем')

    def get_status2_time_str(self):
        if self.status_msg2 != self.status_last_msg2:
            self.status_last_msg2 = self.status_msg2
            self.status_start2 = time.time()

        delta = datetime.timedelta(seconds = time.time() - self.status_start2)
        delta_str = str(delta).split('.')[0]
        return f' ({delta_str})'


    def frontend_status(self):                      # Создает структуру данных для передачи на frontend
        res = {
            'proc_id'    : self.proc_id,
            'progress'   : self.progress,
            'progress2'  : self.progress2,
            'status'     : self.status,
            'status_msg' : self.status_msg,
            'status_msg2': self.status_msg2 + self.get_status2_time_str(),
            'log_msgs'   : self.log_msgs,
            'data'       : self.data,
            'result'     : self.result, 
        }

        if self.status == 'error':
            res['status_msg'] = '\n'.join(self.log_msgs)

        return res

    def is_active(self):                            # Флаг активности текущего действия
        return self.status == ActionThread.STATUS_ACTIVE
            #
            #
    def set_progress(self, progress):               # Устанавливает индикатор выплнения действия
        '''
        Устанавливает значение прогресса выполнения

        progress - прогресс выполнения от 0 до 100
        '''
        self.progress = progress
        self.progress2 = 0
        self.status_msg2 = ''
            #
            #
    def set_message(self, message:str):             # Устанавливаем сообщение для текущего статуса
        self.status_msg = message
            #
            #
    def add_log(self, *args):                       # Добавляет сообщение в лог передаваемый со статусом
        '''
        Добавляет строку в лог 

        *args - произвольные аргументы для лога, будет выведены в отдельных строках
        '''
        for arg in args:
            self.log_msgs.append(f'{arg}')
            #
            #
    def add_data(self, data):                       # Добавляет данные для передачи вмесете со статусом действия
        if type(data) == list:
            self.data.extend(data)
        else:
            self.data.append(data)
            #
            #
    def data_ack(self, data_len):                   # Удаляем данные для которых подтверждено получение (frontend: api2.action.data_ack)
        try:
            if type(data_len)==str:
                data_len = int(data_len)
            self.data = self.data[data_len:]
        except Exception as e:
            api.log.error(e, {'data_len':data_len})
            #
            #


#=========================================================================
#   Интерфейс для API
#=========================================================================
actions = {}            # Словарь активных действий {proc_id : Action}
action_last_id = 0

class ActionNodeData:                               # Класс описывающий действие в структуре API
    def __init__(self, func=None):
        self.name = None                # Имя действия для дерева API
        self.func = func                # Функция исполняющая действие
        self.frontend_func = None       # Имя функции на Frontend
        self.caption = None             # Заголовок для кнопки
        self.url = None
        self.icon = None
        self.form_type = None
        self.progress = True            
        self.kwargs = {}
        self._node = None               # Ссылка на ноду к которой прикреплено действий
        self.index = ''                 # Индекс в списке действий
        self.frontend_props = None
        self.actions = []               # Подчиненные действия
        #
        #
    def __repr__(self):
        return f'Action({self.name} : {self.caption}, form_type={self.form_type})'
        #
        #
    def frontend_model(self):           # Модель действия для Frontend
        if self._node != None and self._node._parent != None:
            repo_path = self._node._parent.path()
        else:
            repo_path = ''

        res = {
                'action_name'   : self.name,
                'caption'       : self.caption,
                'icon'          : self.icon,
                'props'         : self.frontend_props
            }

        if self.url:
            res['url'] = self.url

        elif self.frontend_func:
            res['action_func'] = self.frontend_func
            
        elif self.func:
            res['url'] = f'/api{ACTIONS_URL}?action=exec&action_name={self.name}&path={repo_path}',

        if hasattr(self, 'actions') and self.actions:
            res['actions'] = [child.frontend_model() for child in self.actions]

        return res
            #
            #


def repo_add_action(action_name:str, action_func=None, caption:str='', icon=None, form_type=None, frontend_func=None, url=None, index='5', progress=True, _path:m_tree.Node=None, frontend_props=None, **kwargs):
    api.actions.add_action(
        _path.path(), 
        action_name, 
        action_func, 
        caption=caption, 
        frontend_func=frontend_func,
        url=url,
        icon=icon, 
        form_type=form_type, 
        index=index,
        progress=progress, 
        frontend_props=frontend_props,
        **kwargs
    )


class ApiActions(m_api.ModuleApi):                  # Интерфейс модуля доступный через api.actions
    def __init__(self):
        super().__init__()

        self.aliases = {}
        #
        #
    def node_type(self, form_type):
        if form_type == 'edit':
            node_type = ACTION_EDIT_FORM_TYPE
        else:
            node_type = ACTION_NODE_TYPE

        return node_type
            #
            #
    def add_action(self, repo_path:str, action_name:str, action_func, caption:str='', frontend_func=None, url=None, icon=None, form_type=None, index='', progress=True, frontend_props=None, **kwargs):
        '''
        Добавляет действие в дерево API\n
        repo_path - маршрут в дереве API относительно api.repo
        '''
        if isinstance(repo_path, m_tree.Node):
            node = repo_path
            repo_path = repo_path.path()

        else:
            if repo_path:
                if 'repo.' not in repo_path:
                    repo_path = 'repo.' + repo_path
            else:
                repo_path = ACTIONS_API_PATH

            node = api.find(repo_path, is_assert=True)

        action = ActionNodeData()
        action.name = action_name               # Имя деййствия для дерева API
        action.frontend_func = frontend_func    # Функция выполняющая действие на Frontend
        action.func = action_func               # Функция выполняющая действие на Backend
        action.caption = caption                # Заголовок для кнопки на панели действий
        action.url = url
        action.icon = icon                      # Пиктограмма для кнопки на панели действий
        action.form_type = form_type            # Тип форм для котрых доступно действие (browse, edit)
        action.index = index                    # Индекс действия в списке
        action.progress = progress              # 
        action.frontend_props = frontend_props
        action.kwargs = kwargs

        action_node = node.add(action_name, action, node_type=self.node_type(form_type))
        action._node = action_node
        action_node._desc = action.caption

        return action_node
        #
        #
    def actions_list(self, repo_path:str, journal:str=None, form_type=None):         # Возвращает список доступных действий
        '''
        Возвращает список действий доступных для маршрута в репозитории api.repo
        '''

        # Трансляция имени класса через алиасы
        class_name = repo_path.split('.')[-1]
        if class_name in self.aliases:
            class_name = self.aliases[class_name]
            base_path = '.'.join(repo_path.split('.')[:-1])
            if base_path: base_path += '.'
            repo_path = base_path + class_name

        if journal: repo_path += '.' + journal

        # Определяем точку API от которой будем собирать действия
        if not repo_path:
            node = api.find(ACTIONS_API_PATH)

        elif repo_path.startswith('api.'):
            node = api.find(repo_path)

        else:
            cur_path = repo_path
            while repo_path:
                node = api.repo.find(cur_path) 
                if node: break

                cur_path = '.'.join( cur_path.split('.')[:-1] )
            
        if not node:
            api.log.warning(f'action_list: Не определен маршрут "api.repo.{repo_path}"')
            return []

        # Собираем действия от заданного маршрута и выше
        if form_type==None:
            node_type = [ACTION_NODE_TYPE, ACTION_EDIT_FORM_TYPE]
        else:
            node_type = self.node_type(form_type) 

        actions = node.gather_data(node_type=node_type)

        if not actions: return None

        actions.reverse()

        # Удаляем определения действий перекрытие на более низком уровне
        res = []
        actions_set = []
        for action in actions:
            if not action or not isinstance(action, ActionNodeData) or action.name in actions_set:
                continue

            # Добавляем описания действий дочерних к текущему
            if action._node != None:
                _, children_actions = action._node.children_find_down(node_type=node_type)
                if children_actions:
                    children_actions.sort(key=lambda action: action.index)
                    action.actions = children_actions

            res.append(action)
            actions_set.append(action.name)

        res.sort(key=lambda action: action.index)

        return res
        #
        #
    def get_action(self, repo_path, action_name)->ActionNodeData:       # Возвращает объект действия
        actions = self.actions_list(repo_path, form_type=None)

        if actions:
            for action in actions:
                if action.name == action_name:
                    return action

        if action_name == '_test':
            action = ActionNodeData()
            action.func = action_test
            action.name = '_test'
            action.caption = '_test'
            action.icon = None
            action.kwargs = {}
            return action

        if not repo_path: repo_path = ACTIONS_API_PATH
        raise m_api.ApiException(f'Не найдено действие {action_name} по маршруту {repo_path}')
            #

    def run(self, action_node:ActionNodeData, action_context)->ActionThread:  # Запуск функции действия в отдельном потоке
        args = {
            'context' : action_context,
        }
        action_func = action_node.func

        if action_node.progress:
            # Определение индентификтора действия
            global action_last_id
            action_last_id += 1
            proc_id = str(action_last_id)

            # Создание действия в отдельном потоке и запуск
            args.update(action_node.kwargs)
            action_thread = ActionThread(action_func, args)
            action_thread.proc_id = proc_id
            actions[proc_id] = action_thread
            action_thread.start()

        else:
            action_thread = ActionThread(action_func)
            action_thread.proc_id = 0
            api.context['action'] = action_thread
            action_thread.result = action_func(api=api, **args)
            action_thread.complete()

        return action_thread


def actions_api():
    node = api.find(ACTIONS_API_PATH)
    if node:
        return node
    else:
        raise Exception(f'm_actions: actions_api не найден по маршруту {ACTIONS_API_PATH}')


#=========================================================================
#   actions REST API
#=========================================================================

def action_test(context, api=None,**kwargs):                # Действие _test
    action = api.context.action
    action.log_msgs = str(context)

    for i in range(10):
        action.set_progress(i*10)
        action.set_message(f'Tick {i}')
        action.add_log(f'log for tick {i}')
        time.sleep(0.5)

    return 'Test action completed'


def rest_list(repo_path:str, journal:str=None, form_type=None):                           # Список действий для заданного маршрута
    actions = actions_api().actions_list(repo_path, journal, form_type=form_type)
    if not actions: 
        api.log.error(f'rest_list: Действия для маршрута {repo_path} не определены')
        return []

    res = [action.frontend_model() for action in actions]

    return res
    #
    #
def rest_exec(repo_path:str, action_name:str, action_context:dict):     # Запуск действия на выполнение
    #api.log.var(['action_name', 'action_context'])

    # Определение функции выполняющей действие пользователя
    action_node_data = actions_api().get_action(repo_path, action_name)

    action_thread:ActionThread = actions_api().run(action_node_data, action_context)

    if api.actions.is_debug(): 
        api.log.dump_add(f'm_actions: Запущено действие {action_name} по маршруту {repo_path}')

    if action_thread.status == ActionThread.STATUS_COMPLETED:
        return m_rest.ApiRest.result(action_thread.result)
    else:
        return m_rest.ApiRest.result( action_thread.frontend_status() )
        #
        #
def rest_progress(action_thread:ActionThread):                          # Структура текущего состояния действия
    return m_rest.ApiRest.result( action_thread.frontend_status() )
        #
        #
def rest_cancel(action_thread:ActionThread):
    action_thread.stop()
    return m_rest.ApiRest.result( action_thread.frontend_status() )
        #
        #
def rest_result(action, proc_id, request):                              # Результата выполнения действия
    if action.status == ActionThread.STATUS_COMPLETED :
        if type(action.result) == dict and '_redirect' in action.result:         # Результат - редирект
            res = {'url': action.result['_redirect']}

        elif type(action.result) == dict and '_send_file' in action.result:        # Результат - файл для отправки
            file_path = action.result['_send_file']

            file_path = os.path.abspath(file_path)
            if os.path.isfile(file_path):
                res = werkzeug.utils.send_file(file_path, request.environ, download_name=action.result.get('download_name'))
            else:
                res = m_rest.ApiRest.result_notify('Ошибка сервера', f'Файл {file_path} не найден')

        else:
            res = action.result

        del actions[proc_id]
        return res

    elif action.status == ActionThread.STATUS_ERROR :
        text = '<BR>'.join([
                action.status_msg.replace('\n', '<BR>'), 
                '<b>Ошибки:</b>', api.log.errors_str(action.thread_id).replace('\n', '<BR>'),
                '<b>Сообщения:</b>', api.log.msgs_str(action.thread_id).replace('\n', '<BR>'),
                '<b>Переменные:</b>', api.log.vars_str(action.thread_id, inline=True, CR='\n\n').replace('\n', '<BR>'),
            ]) 
        del actions[proc_id]
        return m_rest.ApiRest.result_notify('Ошибка сервера', text)

    else:
        return None
    #
    #
def rest_endpoint(api, context, **kwargs):                             # Точка входа для REST API действий пользователя
    try:
        args = api.lib.rest.get_args(context)
        
        if 'action' not in args:
            raise Exception('Не указан параметр action')

        action = args['action']
        if action == 'list':
            path = args.get('path', '')
            journal = args.get('journal','')

            form_type = 'edit' if 'report_form_id' in args else 'browse'
            return rest_list(path, journal, form_type=form_type)

        elif action == 'exec':
            action_name = args['action_name']
            repo_path = args.get('path', '')
            action_context = args.get('context', {})
            if type(action_context)==str:
                action_context = json.loads(action_context)
                
            return rest_exec(repo_path, action_name, action_context)

        else:
            proc_id = args['proc_id']
            if proc_id not in actions:
                return {}
            
            action_thread = actions[proc_id]

            if action == 'progress':
                return rest_progress(action_thread)

            elif action == 'data_ack':
                action_thread.data_ack(args['data_len'])

            elif action == 'cancel':
                return rest_cancel(action_thread)

            elif action == 'result':
                res = rest_result(action_thread, proc_id, context['request'])
                return  res

        return f'Неизвестное действие {action}'
    except Exception as e:
        return m_utils.exception_text(e)
    #
    #

