import re, json, datetime

import m_utils, m_rec

api = None
_debug = False

class DbException(Exception): pass

# --- Работа со справочником
def dict_remove(src_dict, keys):                                    # Удаляет из справоничка заданные ключи
    if type(keys) == dict:
        keys = list(keys.keys())

    if type(keys) != list:
        keys = [keys]

    res = dict(src_dict)
    for key in keys:
        if key in res:
            del res[key]

    return res
    #
    #
def dict_fit(src_dict, keys):                                       # Оставляет в справочнике только перечисленные ключи
    to_del = [key for key in src_dict.keys() if key not in keys]
    for key in to_del:
        del src_dict[key]
    return src_dict

class DbAbstractConnect:pass


class Field:                                    # Описание поля 
    serial = 'serial'
    int = 'int'
    str = 'str'
    float = 'float'
    date = 'date'
    timestamp = 'timestamp'
    json = 'json'

    def __init__(self, field_name, field_type, **kwargs):
        self.field_name = field_name
        self.field_type = field_type
        self.repo_data = kwargs

    def _load_ftype(self, db):          # Определение типа поля для ForeignKey
        pass

    def sql_str(self, db:DbAbstractConnect):
        return f'{db.field_sql_str(self.field_name)} {db._db_ftype(self.field_type)}'

    def sql2_str(self, db:DbAbstractConnect):
        return ''


class PK_Field(Field):                          # Описание первичного ключа
    def __init__(self, field_name, **kwargs):
        super().__init__(field_name, Field.serial, **kwargs)

    def sql_str(self, db):
        return f'{db.field_sql_str(self.field_name)} SERIAL PRIMARY KEY'


class FK_Field(Field):                          # Описание ссылки
    def __init__(self, field_name, ref_table_name, **kwargs):
        super().__init__(field_name, None, **kwargs)

        self.ref_table_name = ref_table_name


    def _load_ftype(self, db):
        if not self.field_type:
            ref_table = db._load_table_model(self.ref_table_name, is_assert=True)
            self.ref_field_name = ref_table.primary_key
            ref_table_pk_type = ref_table.fields[self.ref_field_name]

            assert ref_table_pk_type, f'Не определен первичный ключ для таблицы {self.ref_table_name}'

            self.field_type = ref_table_pk_type


    def sql2_str(self, db):
        self._load_ftype(db)
        return f'FOREIGN KEY({db.field_sql_str(self.field_name)}) REFERENCES {db.field_sql_str(self.ref_table_name)}({db.field_sql_str(self.ref_field_name)}) ON DELETE CASCADE ON UPDATE CASCADE'


class TableModel:
    PK_Field    = lambda : PK_Field(None)
    FK_Field    = lambda ref_table_name: FK_Field(None, ref_table_name=ref_table_name)
    IntField    = lambda : Field(None, Field.int)
    FloatField  = lambda : Field(None, Field.float) 
    DateField   = lambda : Field(None, Field.date)
    TimestampField = lambda : Field(None, Field.timestamp)
    StrField    = lambda : Field(None, Field.str)
    JsonField   = lambda : Field(None, Field.json)

    def __init__(self, db, table_name, *args, ensure=False, check_only=False):
        self.table_name = table_name
        self.fields = []

        self._load_from_class(db)

        for arg in args:
            if isinstance(arg, Field):
                arg._load_ftype(db)
                self.fields.append(arg)

        if ensure:
            db.ensure_table(table_name, self, check_only=check_only)
            db.commit()


    def _load_from_class(self, db):
        for k, v in self.__class__.__dict__.items():
            if k.startswith('_'): continue

            if isinstance(v, Field):
                v.field_name = k
                v._load_ftype(db)
                self.fields.append(v)
            else:
                print(f'Не корректный тип атрибута {k} : {v}')


    def fields_dict(self):
        res = {}
        for field in self.fields:
            res[field.field_name] = field.field_type

        return res

    
    def create_sql(self, db):
        sql = f'CREATE TABLE {db.field_sql_str(self.table_name)} (\n'

        fields_sql_list = [field.sql_str(db) for field in self.fields]
        fields_sql_list.extend([field.sql2_str(db) for field in self.fields if field.sql2_str(db)])
        sql += '    '
        sql +=  ',\n    '.join(fields_sql_list)
        sql += '\n)'

        return sql


class TableInfo:
    '''
    .db_connect
    .name_space
    .table_name
    .db_table_name
    .fields: dict
    .primary_key: str
    .value_convertors: dict
    '''
    def __init__(self, db_connect, table:str, fields={}, primary_key=[], def_namespace=''):
        self.db_connect = db_connect

        if type(table) == str:
            words = table.split('.')
            if len(words) == 1:
                self.name_space = def_namespace
                self.table_name = words[0]

            elif len(words) == 2:
                self.name_space = words[0]
                self.table_name = words[1]

            else:
                raise DbException(f'Неправильное имя таблицы {table}')

        elif isinstance(table, TableInfo):
            self.name_space = table.name_space
            self.table_name = table.table_name

        else:
            print(f'TableInfo: self_type={type(self)}  table_type={type(table)}')
            raise DbException(f'Неправильный тип данных параметра table {type(table)} table={table}')

        if self.table_name[0] in ('"', "'"): 
            self.table_name = self.table_name[1:-1]

        if self.name_space:
            self.db_table_name = f'{self.name_space}.{self.db_connect.field_sql_str(self.table_name)}'
        else:
            self.db_table_name = f'{self.db_connect.field_sql_str(self.table_name)}'

        self.fields = fields
        self.primary_key = primary_key

        self.value_convertors = {}
        #
        #
    def __repr__(self):
        return f'TableInfo(table_name={self.table_name}, pk={self.primary_key}, fields={list(self.fields.keys())})'
        #
        #


class DbAbstractConnect:                        # Абстрактный класс для работы с БД
    def __init__(self, core_db_connect=None):
        self.debug = _debug
        self.db_type = 'abstract'
        self.try_only = False           # Флаг - не выполнять SQL 
        self.print_sql = False          # Флаг - выводить SQL перед выполнением

        self.autocommit = False
        self.def_namespace = ''
        self.conn = core_db_connect
        self.conn_self_created = not bool(core_db_connect)
        self.cursor_kwargs = {}
        self.insert_cache = {}
        self.insert_cache_active = False

        self.value_convertors = {}      # Справочник для преобразования типов данных в строку для SQL-запроса
        #
        #
    def __repr__(self):
        return f'DbAbstractConnect()'
        #

    # --- Преобразование в правильное SQL выражение
    def _str_encode(self, value):               # Преобразование строки для SQL выражения
        return f'{value}'.replace("'", "''")
            #

    def _date_decode(self, value):              # Декодирование даты
        return value
            #
            #
    def _date_encode(self, value):              # Кодирование даты в строку для SQL выражения
        if type(value)==str: 
            value = m_utils.str2date(value)
            if not value: return 'NULL'
        
        value = value.strftime('%Y-%m-%d') 
        return f"'{value}'" 
            #
            #
    def _datetime_encode(self, value):              # Кодирование даты в строку для SQL выражения
        if type(value)==str: 
            value = m_utils.str2datetime(value, isAssert=False)
            if not value: return 'NULL'
        
        value = value.strftime('%Y-%m-%d %H:%M:%S') 
        return f"'{value}'" 
            #
            #
    def _json_decode(self, value):              # Декодирование поля типа JSON             
        if value=='{}': return {}
        if value=='[]': return []
        return value
            #
            #
    def _json_encode(self, value):              # Кодирование поля типа JSON в строку SQL выражения
        if not value:
            return "'{}'"

        try:
            value = json.dumps(value, ensure_ascii=False)
        except Exception as e:
            raise Exception(f'Ошибка преобразования в JSON. value_type={type(value)} value={value}')

        value = self._str_encode(value)
        return f"'{value}'"
            #

    def field_sql_str(self, field_name):                                      # Возвращает имя атрибута в виде корректной строки для SQL
        if '->' in field_name or ' as ' in field_name:
            return field_name

        for ch in field_name:
            if ch>='A' and ch<='Z':
                return f'"{field_name}"'
                
        return field_name
        #
        #
    def value_sql_str(self, value, res_type=None, value_convertors={}):       # Значение -> SQL строка значения
        if type(value) in (list, dict) or isinstance(value, m_rec.Record) or isinstance(value, m_rec.RecordSet):
            value = m_utils.simplify(value)

        if value == None:
            return 'NULL'

        # Если значение типа контеж, распаковываем значение и его тип
        if type(value) == tuple:
            res_type = value[1]             # Тип определяется вторым элементом кортежа со значением
            value = value[0]

            if res_type == 'raw':
                return value

        # Если тип результата не определен, извлекаем его из значения
        if res_type == None or res_type == type(None):      
            res_type = type(value)

        # Кодирование через конверторы
        if res_type in value_convertors:
            if callable(value_convertors[res_type]):
                return value_convertors[res_type](value)
            else:
                raise Exception(f'Конвертор для типа {res_type} должен быть функцией')

        if value == None:
            return 'NULL'

        if res_type == 'date' or hasattr(value, 'strftime'):
            return self._date_encode(value)

        if res_type == 'timestamp':
            return self._datetime_encode(value)

        if res_type in (list, dict, 'jsonb', 'json'):
            return self._json_encode(value)

        if res_type == str or res_type == 'str':
            value = self._str_encode(value)
            return f"'{value}'"

        if res_type in (int, 'int'):
            if type(value) in (int,float, str):
                try:
                    return f'{int(value)}'
                except:
                    return 'NULL'

        if res_type in (float, 'float'):
            try:
                return f"{float(value)}"
            except:
                return 'NULL'

        raise DbException(f'Неизвестный тип данных {res_type} для значения "{value}"')
        #
        #
    def sql_pairs(self, data, delimiter, fields={}, value_convertors={}):     # Создание WHERE части запроса
        if type(data) == str:
            return data

        elif type(data) == list:
            return delimiter.join(data)

        elif hasattr(data, 'to_dict'):
            data = data.to_dict()
            
        elif type(data)==dict:
            pass

        else:
            raise DbException(f'sql_pairs: Некорректный тип данных {type(data)}')


        conds = []
        for field in data.keys():
            if fields and field not in fields: continue

            field_str = self.field_sql_str(field)
            value = data[field]

            if type(value) == tuple:        # Значение указано с типом данных
                value, value_type = value
            else:
                value_type = fields.get(field) if type(fields) == dict else None

            cond_str = '='
            if type(value) == str:
                cond_info = re.search('^\s*[!=<>]+|LIKE|like', value)
                if cond_info:
                    split_pos = cond_info.span()[1]
                    cond_str = value[:split_pos].strip()
                    value = value[split_pos:].strip()
            
            value_str = self.value_sql_str(value, value_type, value_convertors)

            conds.append(f' {field_str} {cond_str} {value_str}')

        return delimiter.join(conds)
        #
        #
    def sql_order(self, orders):                                              # Создание ORDER BY части запроса
        if type(orders) == str:
            return orders

        elif type(orders) == list:
            for idx in range(len(orders)):
                orders[idx] = self.field_sql_str(orders[idx])

            return ', '.join(orders)

        else:
            raise DbException(f'Некорректный тип данных сортировки {type(orders)}')

    def sql_fields_str(self, data):
        fields_str = ''

        if type(data) == dict:
            data = data.keys() 

        elif type(data) != list:
            data = [data]

        for field in data:
            fields_str += f' {self.field_sql_str(field)},'

        fields_str = fields_str[:-1]        # Убираем последнюю ,

        return fields_str
        #
        #
    def sql_values_str(self, table, data):
        values_str = ''

        for field, value in data.items():
            value_type = table.fields.get(field)
            if value_type:
                values_str += f' {self.value_sql_str(value, table.fields.get(field), table.value_convertors)},'
            else:
                api.log.error(f'Не определен тип данных для поля {field} доступны поля {table.fields}')

        values_str = values_str[:-1]        # Убираем последнюю ,

        return values_str
        #
        #
    def sql_set_str(self, table:TableInfo, data:dict):
        data = dict_remove(data, table.primary_key)
        return self.sql_pairs(data, ', ', table.fields, table.value_convertors)
        #
        #

    # --- Подключение и управление транзакциями
    def is_closed(self) -> bool :
        raise DbException('DbAbstactConnect.is_closed is abstract method')
            #
            #
    def connect(self):
        raise DbException('DbAbstactConnect.connect is abstract method')
            #
            #
    def connect_ensure(self):
        pass
            #
            #
    def disconnect(self):
        raise DbException('DbAbstactConnect.disconnect is abstract method')
        #
        #
    def commit(self):
        self.conn.commit()
        if self.debug: api.log.msg('Transaction commit')
        #
        #
    def rollback(self):
        self.conn.rollback()
        if self.debug: api.log.msg('Transaction rollback')
        #

    def is_transaction_active(self):
        raise Exception('m_db.is_transaction_active is abstract')
            #
            #
    def get_transaction_info(self):
        raise Exception('m_db.get_transaction_info is abstract')

        
    # --- Выполнение SQL команды на сервере
    def sql_cursor(self, sql) -> dict:
        cur = self.conn.cursor(**self.cursor_kwargs)
        return cur.execute(sql)
        #
        #
    def sql_exec(self, sql:str, commit:any=None, res_type='RecordSet', **kwargs) -> m_rec.RecordSet:    # Вовзращает результат выполнения SQL-запроса
        '''
        Возвращает результат выполнения SQL-запроса
        '''
        if self.debug: api.log.msg(f'sql_exec: sql={sql}')

        self.connect_ensure()

        if 'delete' in sql.lower() or 'insert' in sql.lower() or 'update' in sql.lower():

            #api.log.msg(f'sql_exec: sql={sql}\nConnection: {self.conn}')

            # Отображение списка конкурирующих активных транзакций
            t_list = []
            for th, context in api.context._data._data.items():
                if not 'db_connect' in context: continue
                if th == api.log._th_id(): continue

                conn = context['db_connect']
                if not conn.is_transaction_active(): continue

                try:
                    t_list.append(f'поток {th} {conn.get_transaction_info()}')
                except:
                    pass
            
            if self.debug and t_list:
                api.log.warning('открыты транзации:<BR>' + "<BR>".join(t_list))


        cur = self.conn.cursor(**self.cursor_kwargs)
        try:
            fields = None
            cur.execute(sql)
            if res_type != None:
                rows = cur.fetchall()
                if cur.description:
                    fields = [desc[0] for desc in cur.description]
        except Exception as e:
            msg = f'Ошибка выполнения запроса {e} \nSQL={sql}'
            api.log.error(msg)
            if not self.is_closed():
                self.conn.rollback()
            raise Exception(msg)
        finally:
            cur.close()

        if commit or (commit==None and self.autocommit):
            self.conn.commit()

        #print(f'sql_exec: fields={fields}\nrows={rows}')
        if res_type==None or not fields:
            return None

        elif res_type == 'RecordSet':
            if rows and type(rows[0]).__name__=='Row':
                rows = [dict(row) for row in rows]
            try:
                return m_rec.RecordSet((fields, rows))
            except Exception as e:
                raise Exception(e, f'sql={sql}  fields={fields}  rows={rows}')

        elif res_type == '':
            return fields, rows
        #
        #
    
    def is_table_exist(self, db_table_name) -> bool:
        raise DbException('DbAbstactConnect.abstract method called')

    def _load_table_model(self, db_table_name, is_assert=False) -> TableInfo:                           # Возвращает список полей таблицы 
        '''
        Если таблица не найдена, возвращает None
        '''
        raise DbException('DbAbstactConnect.table_fields is abstract method')
        #
        #
    def _check_fields(self, table_name:str=None, table:TableInfo=None) -> TableInfo:                    # Автоматическая загрузка списка полей таблицы
        if table and table.fields:
            return table

        if not table_name and table:
            table_name = table.db_table_name

        table = self._load_table_model(table_name, is_assert=True)        
        table.value_convertors = self.value_convertors

        return table
        #
        #
    def get_table_info(self, db_table_name, primary_key=[]):
        table = self._check_fields(table_name = db_table_name)
        table.primary_key = primary_key
        return table

    # --- Работа со структурой БД
    db_types = {
        'serial'   : 'SERIAL',
        'int'      : 'INTEGER',
        'float'    : 'FLOAT',
        'str'      : 'VARCHAR',
        'date'     : 'DATE',
        'timestamp': 'TIMESTAMP', 
        'json'     : 'JSONB'
    }
    def _db_ftype(self, field_type):
        if field_type not in self.db_types:
            raise Exception(f'Не корректный тип поля "{field_type}"')

        return self.db_types[field_type]
            #
            #
    def _get_ftype(self, data):                                                 # Извлекает SQL-тип поля из описания в модели
        return data[0] if type(data) in (tuple, list) else data
            #
            #
    def _get_ftype_str(self, data):                                             # Извлекает SQL-определение поля из описания в модели
        if type(data) in (tuple, list):
            field_type = data[0]
            data = ' '.join(data[1:])
        else:
            field_type = data
            data = ''

        return self._db_ftype(field_type) + ' ' + data 
            #

    def check_model(self, table_name, table_model) -> bool:                     # Проверяет таблицу в БД на соответствие модели
        table = self.get_table_info(table_name)
        table_fields = table.fields

        if len(table_fields) != len(table_model): return False
        for fname, ftype in table_model.items():
            field_type = self._get_ftype(ftype)

            if fname not in table_fields or table_fields[fname] != field_type: return False

        return True
            #
            #
    def create_table(self, table_name, table_model, check_only=True) -> str:    # Создает таблицу в БД по модели
        sql = table_model.create_sql(self)
        
        if not check_only:
            self.sql_exec(sql, res_type=None)
            self.commit()
            api.log.msg(f'create_table: table_name={table_name}\nsql={sql}')
        else:
            print(f'create_table: sql={sql}')

        return sql
            #
            #
    def drop_table(self, table_name):                                           # Удаляет таблицу из БД
        table_name = self.field_sql_str(table_name)
        sql = f'DROP TABLE {table_name}'
        try:
            self.sql_exec(sql, res_type=None)
            self.commit()
            api.log.msg(f'drop_table: table_name={table_name}')

        except Exception as e:
            if 'no such table' in e.args[0]: 
                print(f'Таблица {table_name} не найдена')
                return
            raise
            #
            #
    def modify_table(self, table_name, table_model, check_only=True) -> dict:   # Приводит таблицу в БД к соответствию модели
        table = self.get_table_info(table_name)

        result = {'add':[], 'modify':[], 'drop':[]}

        # Добавляем и изменяем атрибуты 
        for fname, ftype in table_model.fields_dict().items():

            try:
                fname_db_str = self.field_sql_str(fname)
                ftype_db_str = self._get_ftype_str(ftype)
                table_sql_str = self.field_sql_str(table_name)

                sql = None 
                if fname not in table.fields:
                    sql = f'ALTER TABLE {table_sql_str} ADD COLUMN {fname_db_str} {ftype_db_str}'
                    result['add'].append(sql)

                elif self._get_ftype(ftype) != table.fields[fname]:
                    if 'serial' not in ftype_db_str.lower():
                        sql = f'-- Change field type from {table.fields[fname]} to {ftype}\n'
                        sql += f'ALTER TABLE {table_sql_str} ALTER COLUMN {fname_db_str} TYPE {ftype_db_str}'
                        result['modify'].append(sql)

                if sql:
                    if not check_only:
                        self.sql_exec(sql, res_type=None)
                    else:
                        print(f'modify_table: sql={sql}')
            except Exception as e:
                raise Exception(f'Ошибка изменения поля "{fname}" {ftype}. {e}')
            

        for fname in table.fields.keys():
            if fname not in table_model.fields_dict():
                sql = f'ALTER TABLE {table_sql_str} DROP COLUMN {self.field_sql_str(fname)}'
                result['drop'].append(sql)

                if not check_only:
                    self.sql_exec(sql, res_type=None)
                else:
                    print(f'modify_table: sql={sql}')


        if not check_only and (result['add'] or result['modify'] or result['drop']):
            api.log.msg(f'modify_table: table_name={table_name} result={result}')
            self.commit()

        return result
            #
    def ensure_table(self, table_name, table_model, check_only=True):
        if not self.is_table_exist(table_name):
            res = self.create_table(table_name, table_model, check_only=check_only)

        else:
            res = self.modify_table(table_name, table_model, check_only=check_only)
        
        return res


    def load_field_values(self, table_name, field_name) -> list:                # Возвращает список значений поля
        table = self.get_table_info(table_name)

        field_name = self.field_sql_str( field_name )
        table_name = self.field_sql_str( table_name )

        sql = f'SELECT DISTINCT {field_name} FROM {table_name}'
        rs = self.sql_exec(sql)

        return rs.attr_values(table.primary_key)

    # --- Работа с кэшем 
    def insert_cache_start(self):
        self.insert_cache = {}
        self.insert_cache_active = True
        #
        #
    def insert_cache_add(self, sql):
        phrases = sql.split('VALUES')
        key = phrases[0]
        value = phrases[1]

        if key not in self.insert_cache:
            self.insert_cache[key] = []

        self.insert_cache[key].append(value)
        #
        #
    def insert_cache_flush(self):
        self.insert_cache_active = False

        sql = ''
        for key in self.insert_cache.keys():
            sql += key + ' VALUES '
            sql += ','.join(self.insert_cache[key])
            sql += ';'

        self.sql_exec(sql, res_type=None)
        self.insert_cache = {}
        #
        
    # --- Базовые операции с таблицей
    def returning_str(self, fields):
        return f' RETURNING { fields }'


    def insert(self, table:TableInfo, data:dict, return_id=False, **kwargs) -> m_rec.Record:
        '''
        Добалвяет запись в таблицу table.\n
        data - данные для добавления в справочнике {поле: значение, ...}\n
        return_id - флаг для возвращения идентификатора созданной записи
        '''
        if hasattr(data, 'to_dict'):
            data = data.to_dict()
        assert(type(data)==dict)

        table = self._check_fields(table=table)

        # Удаляем поле первичного ключа, если оно пустое
        if table.primary_key and table.primary_key[0] in data and not data[table.primary_key[0]]:
            data.pop(table.primary_key[0])

        # Формирование списков полей и значений
        fields_str = self.sql_fields_str(data)
        values_str = self.sql_values_str(table, data)

        sql = f'INSERT INTO {table.db_table_name} ({fields_str}) VALUES ({values_str})'
        if self.print_sql: print(sql)

        if self.insert_cache_active:
            if return_id:
                raise DbException('Невозможно вернуть ID записи при включенном кэше')
            self.insert_cache_add(sql)
        
        else:
            if return_id:
                if not table.primary_key:
                    raise Exception(f'Не определен первичный ключ для таблицы {table}')

                sql += self.returning_str(self.sql_fields_str(table.primary_key))
                res = self.sql_exec(sql) if not self.try_only else None
                return res[0]
            else:
                self.sql_exec(sql, res_type=None) if not self.try_only else None
        #
        #
    def update(self, table:TableInfo, data:dict, filters:dict, **kwargs) -> None:
        '''
        Обновляет значения записей в таблице table.\n
        data - данные для записи в справочнике {поле: значение, ...}\n
        filters - фильтры
            :dict - пары {поле: значение}\n
        '''
        if hasattr(data, 'to_dict'):
            data = data.to_dict()
        assert(type(data)==dict)

        table = self._check_fields(table=table)

        data = dict_remove(data, filters)
        set_str = self.sql_set_str(table, data)
        where_str = self.sql_pairs(filters, ' AND ', table.fields)

        sql = f'UPDATE {table.db_table_name} SET {set_str} WHERE {where_str}'

        if self.print_sql: print(sql)
        self.sql_exec(sql, res_type=None) if not self.try_only else None
        #
        #
    def delete(self, table:TableInfo, filters:dict, **kwargs) -> None:
        '''
        Удаляет записи из таблицы table.\n
        filters - фильтры
            :dict - пары {поле: значение}\n
        '''
        table = self._check_fields(table=table)

        where_str = self.sql_pairs(filters, ' AND ', table.fields)

        sql = f'''DELETE FROM {table.db_table_name} WHERE {where_str}'''

        if self.print_sql: print(sql)
        self.sql_exec(sql, res_type=None) if not self.try_only else None
        #
        #
    def select(self, table:TableInfo, filters=[], orders=[], offset=0, limit=None, fields='*', try_only=False, debug=False, **kwargs) -> m_rec.RecordSet:
        '''
        Возвращает результат запроса SELECT к таблице table.\n
        filters - фильтры
            :dict - пары {поле: значение}\n
        orders - порядок сортировки 
            :list - список полей
            :str  - строка добавляемая после ORDER BY\n

        offset  - номер первой записи из результата запроса\n
        limit   - максимальное кол-во записей\n
        fields  - список полей или строка со списком. По-умолчанию все. 
        '''
        if type(table) == str:
            if debug or _debug: print(f'm_db.select: Retriving table info for {table}')
            table = self.get_table_info(table)

        table = self._check_fields(table=table)

        if type(fields) == list:
            fields = self.sql_fields_str(fields)

        sql = f'SELECT {fields} FROM {table.db_table_name}'

        if filters:
            filter_sql = self.sql_pairs(filters, ' AND ', table.fields)
            if filter_sql:
                sql  += f' WHERE {filter_sql}'

        if orders:
            order_sql = self.sql_fields_str(orders)
            if order_sql:
                sql += f' ORDER BY {order_sql}'

        if offset:
            sql += f' OFFSET {offset}'

        if limit:
            sql += f' LIMIT {limit}'

        if debug:
            print(f'DbAbstractConnect.select: sql={sql}')

        if not try_only:
            recordset = self.sql_exec(sql, **kwargs)
            if type(recordset) == m_rec.RecordSet:
                recordset._fields = table.fields
                recordset._table = table
                
            return recordset
        #    
        #

    def save(self, rs:m_rec.RecordSet):
        pass

def init(_api, debug=False):
    global api
    api = _api

    global _debug
    _debug = debug