import psycopg2
from psycopg2.extras import DictCursor

import m_db

api = None

class DbPgConnect(m_db.DbAbstractConnect):             # Класс для работы с БД PostgreSQL
    def __init__(self, core_db_connect=None):
        super().__init__(core_db_connect=core_db_connect)
        self.db_type = 'postgresql'

        db_user = api.get_config('DB_USER',{})
        self.LOGIN = db_user.get('user', 'rvrao')
        self.PWD = db_user.get('password', 'rvrao')
        self.DB_NAME = api.get_config('POSTGRES_DATABASE', 'dev')
        self.DB_HOST = api.get_config('DB_HOST', 'wiki.robolab.msk.ru')
        self.DB_PORT = api.get_config('DB_PORT', '5432')
        self.def_namespace = 'public'

        self.cursor_kwargs = {'cursor_factory':DictCursor}
        #
        #
    def __repr__(self):
        return f'DbPgConnect({self.DB_NAME}@{self.DB_HOST})'
            #
            #
    def is_closed(self):
        return self.conn.closed
        
    def _date_encode(self, value):
        res = super()._date_encode(value)
        return res + '::DATE'

    def _json_encode(self, value):
        res = super()._json_encode(value)
        return res + '::JSONB'

    def connect(self, pg_conn=None, db_name=None):     # Подключается в БД
        '''
        Если задан db_name - подключается к указанной БД
        Если не определен self.conn
            Если задан conn - возвращает его 
            
        '''
        if db_name:
            self.DB_NAME = db_name
            self.conn = None

        if self.conn == None or self.conn.closed:
            if pg_conn:
                self.conn = pg_conn
                if self.debug: api.log.msg(f'Используется подключение к БД преданное как параметр {self}')
            else:    
                connect_node = api.find('system.pg_connect')

                if connect_node:
                    self.conn = connect_node()
                    msg = 'DbPgConnect.connect: used system.pg_connect'
                    #api.log.msg(msg)
                    #print(msg)
                    if self.debug: api.log.msg(f'Создано подключение к БД через api.system.pg_connect {self}')
                else:
                    self.conn = psycopg2.connect(
                            user=self.LOGIN, 
                            password=self.PWD, 
                            database=self.DB_NAME, 
                            host=self.DB_HOST, 
                            port=self.DB_PORT
                        )
                    if self.debug: api.log.msg(f'Создано подключение к БД через psycopg2.connect {self}')
                    #print('DbPgConnect.connect: used default connection')

                    self.conn.autocommit = False
                    self.conn.set_client_encoding('UTF8')

        return self.conn        
        #
        #
    def disconnect(self):
        api.log.msg(f'DbPgConnect.disconnect: connection={self.conn} is_closed={self.conn.is_closed()}')
        try:
            api.system.pg_release(self.conn)
        finally:
            self.conn = None
        #
        #
    def is_transaction_active(self)->bool :
        if self.conn:
            return bool( self.conn.get_transaction_status()!=0 )
        else:
            return False

    def get_transaction_info(self)->str :
        tr_status = self.conn.get_transaction_status()
        if tr_status < 4:
            status = ['IDLE', 'ACTIVE', 'INTRAS','INERROR','UNKNOWN'][tr_status]
        else:
            status = tr_status
        
        return f'транзакция {self.conn.fileno()} статус транзакции {status}'


    def _load_table_pk(self, table_name) -> str:
        sql = f'''
            SELECT 
                t.relname, a.attname, format_type(a.atttypid, a.atttypmod) AS data_type, 
                i.indisprimary,
                pg_get_expr(d.adbin, d.adrelid) adsrc
            FROM   pg_catalog.pg_class as t 
            JOIN   pg_attribute a ON a.attrelid = t.oid
            LEFT OUTER JOIN pg_index i on t.oid = i.indrelid AND a.attnum = ANY(i.indkey)                     
            LEFT OUTER JOIN pg_attrdef AS d ON d.adrelid = t.oid AND d.adnum=a.attnum

            WHERE t.relname = '{table_name}'
            AND (i.indisprimary OR pg_get_expr(d.adbin, d.adrelid) LIKE 'nextval%')
            ORDER BY i.indisprimary
        '''
        rs = self.sql_exec(sql)
        if not rs.is_empty():
            return rs[0].attname
        else:
            return None


    def is_table_exist(self, db_table_name) -> bool:
        table = m_db.TableInfo(self, db_table_name, def_namespace=self.def_namespace)
        sql = f'''
            select t.relname
            from
                pg_catalog.pg_namespace n
                inner join pg_catalog.pg_class as t on t.relnamespace=n.oid
            where t.relname='{table.table_name}' and
                  n.nspname='{table.name_space}' 
            '''
        return not self.sql_exec(sql).is_empty()


    def _load_table_model(self, db_table_name, is_assert=False):          # Возвращает список полей таблицы 
        table = m_db.TableInfo(self, db_table_name, def_namespace=self.def_namespace)

        sql = f'''
            select attname, tp.typname, tp.typcategory
            from
                pg_catalog.pg_namespace n
                inner join pg_catalog.pg_class as t on t.relnamespace=n.oid
                inner join pg_catalog.pg_attribute as a on a.attrelid=t.oid
                inner join pg_catalog.pg_type as tp on tp.oid = a.atttypid
            where t.relname='{table.table_name}' and
                  n.nspname='{table.name_space}' and 
                  attname not in ('tableoid','cmax','xmax','cmin','xmin', 'ctid')    
            '''
        data = self.sql_exec(sql)

        type_translate = {
            'serial'  : 'serial',
            'float'   : 'float',
            'float8'  : 'float',
            'int4'    : 'int',
            'integer' : 'int',
            'varchar' : 'str',
            'jsonb'   : 'json',
            'date'    : 'date',
            'timestamp' : 'timestamp',
        }

        for rec in data:
            item = rec._data
            if item['typcategory'] == 'E':
                item['typname'] = 'str'
            else:
                item['typname'] = type_translate.get(item['typname'], item['typname'])

        table.primary_key = self._load_table_pk(table.table_name)
        table.fields = {}
        for rec in data:
            table.fields[rec.attname] = rec.typname

        #if is_assert: assert table.fields, f'Не найдена таблица "{db_table_name}"'

        return table #if table.fields else None
        #
        #


def init(_api, debug=False):
    global api
    api = _api
