'''
Модуль источников данных
'''
import m_rec
import m_db
import m_triggers


DS_BEFORE_SHOW = 'DS_BEFORE_SHOW'
DS_BEFORE_SELECT = 'DS_BEFORE_SELECT'
DS_BEFORE_READ = 'DS_BEFORE_READ'
DS_BEFORE_CREATE = 'DS_BEFORE_CREATE'
DS_AFTER_CREATE = 'DS_AFTER_CREATE'
DS_BEFORE_UPDATE = 'DS_BEFORE_UPDATE'
DS_AFTER_UPDATE = 'DS_AFTER_UPDATE'
DS_BEFORE_DELETE = 'DS_BEFORE_DELETE'
DS_AFTER_DELETE = 'DS_AFTER_DELETE'
 
api = None
_debug = False

class DataSource:
    def __init__(self, aliases={}):
        assert(type(aliases)==dict)

        self.aliases = aliases

    # --- Внутренние методы 
    def _emit(self, event_name, class_path, context):       # Запуск обработки события
        if api.find('triggers'):
            api.triggers.emit(event_name, class_path, context)
        #
        #
    def _get_connect(self, use_system_connect=False, allow_system_connect=False) -> m_db.DbAbstractConnect:     # Возвращает DbAbstractConnect для методов обращения к хранилищу
        if use_system_connect:
            conn = api.system.system_db_connect

        else:
            try:
                conn = api.context.db_connect
            except:
                if allow_system_connect:
                    conn = api.system.system_db_connect
                else:
                    raise

        if conn.is_closed():
            print('Повторное подключение к серверу БД')
            conn.connect()
            if conn.is_closed():
                raise Exception('Ошибка подключения к БД')

        return conn
        #
        #
    def _add_aliases(self, records):
        if self.aliases:
            for key, value in self.aliases.items():
                records.add_alias(key, value)
        #
        #
    def db_select(self, **kwargs) -> m_rec.RecordSet:
        pass

    def on_select(self, rs):
        pass

    
    def decode_record(self, rec):
        pass
            #
            #
    def decode_recordset(self, rs):
        pass
            #

    def select(self, **kwargs) -> m_rec.RecordSet:      # Возвращает список (аргументы передаются в m_ds.DbAbstractConnect.select)
        # Получение данных
        rs = self.db_select(**kwargs)

        if rs:
            self.decode_recordset(rs)        

        # Добавление алиасов для атрибутов
        if type(rs) == m_rec.RecordSet:
            try:
                self._add_aliases(rs)
            except m_rec.RecAliasException:
                pass

            self.on_select(rs)

        return rs
        #
        #


class SqlDataSource(DataSource):
    def __init__(self, sql, aliases={}, **kwargs):
        super().__init__(aliases=aliases)

        self.sql = sql

    def db_select(self, **kwargs):
        db = self._get_connect()
        return db.sql_exec(self.sql, **kwargs)
        

class TableDataSource(DataSource):
    def __init__(self, type_name:str, table_name:str, pk_field_name:str, class_field_name:str=None, aliases:dict={}, container=None):
        if not aliases and container:
            aliases = {'': container}
        
        super().__init__(aliases=aliases)

        self.type_name = type_name                      # Имя хранилища данных
        self.table_name = table_name                    # Имя таблицы в хранилище
        self.pk_field_name = pk_field_name              # Имя поля с первичным ключём
        self.class_field_name = class_field_name        # Имя поля с указанием класса объекта
        self.fields = None                              # Список обновляемыъ полей 
        self.container = container                      # JSON атрибут для хранения атрибутов отсутствующих в таблице
            
        db_connect = self._get_connect(use_system_connect=True)

        self.table = db_connect.get_table_info(table_name, primary_key=[pk_field_name])
        #
        
    def record_type(self, rec):                             # Возвращает тип объекта данных
        return self.type_name
        #
        #
    def record_class(self, rec):                            # Возвращает класс объекта данных
        if self.class_field_name: 
            if type(self.class_field_name) != str:
                raise Exception(f'Некорректное значение class_field_name ="{self.class_field_name}"')
            if self.class_field_name in rec:
                return rec[self.class_field_name]

        return None
        #
        #
    def record_id(self, rec):                               # Возвращает ID экземпляра объекта данных
        return rec[self.pk_field_name]
        #
        
    # --- Упаковка атрибутов в контейнер
    def pack_data(self, data):                              # Упаковка атрибутов в контейнер
        if not self.container: return data

        res = {}
        res[self.container] = {}

        for attr, value in data.items():
            if attr == self.container:
                pass

            elif attr in self.table.fields:
                res[attr] = value

            else:
                res[self.container][attr] = value

        if isinstance(data, m_rec.Record):
            return m_rec.Record(res)
        else:
            return res

    # --- Методы обращения к хранилищу, для переопределения в наследниках
    def db_select(self, **kwargs) -> m_rec.RecordSet:       # Реализация выборки списка экземпляров объектов
        db = self._get_connect()
        return db.select(self.table, **kwargs)
        #
        #
    def on_select(self, rs):
        super().on_select(rs)

        # Фильтрация данных по исключениям генерируемым триггерами
        i = 0
        while i < rs.count():
            rec = rs[i]
            try:
                self._emit('DS_BEFORE_SELECT', self.record_class(rec), rec) 
                i += 1
            except m_triggers.TriggerExceptionNotAllowed as e:
                rs.delete(i)
            #

    def db_read(self, id, **kwargs) -> m_rec.Record:        # Реализация чтения экземпляра из БД

        filters = kwargs.pop('filters', {})
        filters[self.pk_field_name] = id

        if kwargs.get('debug'):
            print(f'm_ds.DataSource.db_read: filters={filters}') 

        res = self.db_select(filters=filters, **kwargs)

        if not res.is_empty():
            return res[0]
        else:
            return None 
        #
        #
    def db_create(self, data):                              # Реализация чтения создание экземпляра в БД
        data = self.pack_data(data)

        db = self._get_connect()

        new_row = db.insert(self.table, data, return_id=True)
        new_id = new_row[self.table.primary_key[0]]
        rec = self.db_read(new_id)
        if rec==None:
            api.log.warning(f'Ошибка чтения созданой записи. id={new_id}:{type(new_id)} table={self.table} data={data}')
            rec = m_rec.Record(new_row)

        return rec
        #
        #
    def db_update(self, rec:m_rec.Record, filters={}):      # Реализация обновления экземпляра в БД
        rec = self.pack_data(rec)

        db = self._get_connect()

        filters[self.pk_field_name] = self.record_id(rec)

        db.update(self.table, rec, filters)
        #
        #
    def db_delete(self, id):                                # Реализация удаления экземпляра в БД
        db = self._get_connect()
        table_sql_name = db.field_sql_str(self.table_name)
        filters = {self.pk_field_name: id}
        where_str = db.sql_pairs(filters, ' AND ')

        sql = f'''
            DELETE FROM {table_sql_name} WHERE {where_str}
        '''
        db.sql_exec(sql, res_type=None)
        #
        #

    # --- Публичные методы
    def decode_record(self, rec):
        db_connect = self._get_connect(use_system_connect=True)

        for field in self.table.fields.keys():
            if self.table.fields[field] in ('json', 'jsonb'):
                if field in rec:
                    if type(rec[field]) == str:
                        value = db_connect._json_decode(rec[field])
                        if type(value) == dict:
                            rec[field] = m_rec.Record(value)
            #
            #
    def decode_recordset(self, rs):
        for rec in rs:
            self.decode_record(rec)
            #

    def create(self, class_name):                       # Создание экземпляра заданного класса
        data = {self.class_field_name : class_name}

        return self.insert(data)
        #
        #
    def insert(self, data):
        class_name = data.get(self.class_field_name) if self.class_field_name else None
        class_path = api.repo.find_class(self.type_name, class_name) if class_name else None

        # Триггеры перед созданием
        if class_path:
            self._emit(DS_BEFORE_CREATE, class_path, {'NEW':{'class_name': class_name}})

        # Создание записи в БД
        data = self.db_create(data)

        try:
            self._add_aliases(data)
        except m_rec.RecAliasException:
            pass

        # Триггеры после создания
        if class_path:
            self._emit(DS_AFTER_CREATE, class_path, {'NEW':data})

        return data
            #
            #
    def read(self, id, **kwargs) -> m_rec.Record:       # Чтение экземпляра
        if 'debug' in kwargs:
            print(f'm_ds.DataSource.read: id={id} kwargs={kwargs}')

        # Чтение данных
        rec = self.db_read(id, **kwargs)
        if rec == None: rec = m_rec.Record({})

        self.decode_record(rec)

        # Изменение и фильтрация данных триггерами 
        if not rec.is_empty():
            self._emit('DS_BEFORE_READ', self.record_class(rec), rec)

            # Добавление алиасов для атрибутов
            self._add_aliases(rec)

        return rec
        #
        #
    def update(self, data):                             # Обновление экземпляра
        if type(data) != m_rec.Record:
            rec = m_rec.Record(data)
            try:
                self._add_aliases(rec)
            except:
                pass
        else:
            rec = data

        self._emit('DS_BEFORE_UPDATE', self.record_class(rec), {'NEW':rec})

        self.db_update(rec)

        self._emit('DS_AFTER_UPDATE', self.record_class(rec), {'NEW':rec})
        #
        #
    def delete(self, id):                               # Удаление экземпляра
        self.db_delete(id)


def init(_api, debug=False):
    global api
    api = _api

    if not api.find('context'):
        print('m_ds.init: api не содержит context')

    global _debug
    _debug = debug

    # Регистрация событий для триггеров
    if api.find('triggers'):    
        triggers = api.triggers
        
        triggers.add_event(DS_BEFORE_SHOW)
        triggers.add_event(DS_BEFORE_READ)
        triggers.add_event(DS_BEFORE_SELECT)

        triggers.add_event(DS_BEFORE_CREATE)
        triggers.add_event(DS_AFTER_CREATE)

        triggers.add_event(DS_BEFORE_UPDATE)
        triggers.add_event(DS_AFTER_UPDATE)

        triggers.add_event(DS_BEFORE_DELETE)    
        triggers.add_event(DS_AFTER_DELETE)    