'''
Модуль хранилища файлов
Confluence: https://robolabmsk.atlassian.net/wiki/spaces/ADM/pages/223871960/m+files.py

Параметры передаваемые методу add_download
 - mimetype
 - download_name
'''

import shutil
import os
import werkzeug
import tempfile

import m_api

FILES_API_PATH = 'lib.files'
FILES_URL = '/files'


#=================================================================================
#   Инициализация модуля
#=================================================================================
api = None
_debug = False
fs = None

def init(_api, debug=False):
    global api
    api = _api

    global _debug
    _debug = debug

    global fs
    fs = FileStorageInterface()

    api.add_module_api(FILES_API_PATH, fs)

    api.add_url(FILES_URL, fs.endpoint)
    

#=================================================================================
#   Интерфейс API
#=================================================================================
import m_rest

class FileStorageInterface(m_api.ModuleApi):
    def __init__(self):
        super().__init__()

        tmp_dir = api.get_config('API2_FILES_TMP')

        if tmp_dir == None:
            tmp_dir = api.get_config('API2_PATH')
            if tmp_dir == None:
                raise Exception('Не задана переменная окружения API2_PATH')
            tmp_dir = os.path.join(tmp_dir, 'TMP')

        self.tmp_dir = os.path.abspath(tmp_dir)
        print(f'm_file initialized. tmp_dir={self.tmp_dir}')

        self.download_list = {}
        #
        #
    def tmp_file_full_path(self, file_path):                # Возвращает полный путь к временному файлу
        dir, file_name = os.path.split(file_path)

        if not dir:
            dir = self.tmp_dir

        full_name = os.path.join(dir, file_name)

        return full_name
        #
        #
    def new_tmp_file(self, extension=''):                   # Создает новый временный файл
        while True:
            temp_name = next(tempfile._get_candidate_names()) + extension
            temp_name = self.tmp_file_full_path(temp_name)

            if not os.path.isfile(temp_name): break

        return temp_name
        #
        #
    def copy(self, src_path, dst_path=None):                # Копирует временный файл, по-умолчанию из каталога временных файлов
        src_path = self.tmp_file_full_path(src_path)
        _, src_ext = os.path.splitext(src_path)

        if dst_path:
            dst_path = self.tmp_file_full_path(dst_path)
        else:
            dst_path = self.new_tmp_file(src_ext)

        self.log(f'files.copy: src={src_path} dst={dst_path}')

        if not os.path.isfile(src_path):
            msg = f'Не найден файл шаблона {src_path}'
            api.log.error(msg)
            raise Exception(msg)

        try:
            shutil.copy(src_path, dst_path)
        except Exception as e:
            msg = f'Ошибка копирования файла из {src_path} в {dst_path}\n{e}'
            api.log.error(msg)
            raise Exception(msg)

        return dst_path
        #
        #
    def delete(self, file_path):                            # Удалаяет файл, по-умолчанию из каталога временных файлов
        file_path = self.tmp_file_full_path(file_path)

        self.log(f'files.delete: {file_path}')

        os.remove(file_path)
        #
        #
    
    def get_download_url(self, download_key):               # Вовзаращает ссылку для загрузки файла
        return f'/api/files?action=download&file={download_key}'
            #
            #
    def add_download(self, download_key, file_path, **kwargs):  # Добавляет файл для загрузки
        file_desc = {'file':file_path, 'params':kwargs}
        self.download_list[download_key] = file_desc

        self.log(f'files.add_download: {file_desc}')

        return self.get_download_url(download_key)
        #
        #
    def send_file_response(self, file_id, request):         # Возвращает файл для загрузки
        if file_id == '_test':
            file_desc = {'file':os.path.join(self.tmp_dir,'temp_file'), 'params':{}}
        else:
            file_desc = self.download_list.get(file_id)

        file_path = file_desc['file']
        
        if not file_path:
            return f'Неправильный идентификатор файла {file_id}' 

        params = file_desc['params']

        if 'download_name' not in params:
            params['download_name'] = file_id

        self.log(f'files_endpoint: download {file_path} params={params}')

        resp = werkzeug.utils.send_file(file_path, request.environ, **params)
        resp.headers['Access-Control-Allow-Origin'] = '*'

        return resp
            #
            #

    def endpoint(self, api, context, **kwargs):            # Обработчик REST API  запросов 
        request = context['request']
        args = request.args.to_dict()
        
        action = args.get('action')

        if action == 'download':
            file_id = args['file']

            return self.send_file_response(file_id, request)

        elif action == 'upload':

            if not request.files:
                return m_rest.ApiRest.result_error('Отсутствует файл')

            file_list = {}
            for file_name, file_obj in request.files.items():
                tmp_file = self.new_tmp_file()
                file_obj.save(tmp_file)
                self.log(f'Uploaded file: {file_name}, file_obj={file_obj}')
                file_list[file_name] = tmp_file

            return m_rest.ApiRest.result(file_list)

        elif action == 'list':
            pass

        else:
            return m_rest.ApiRest.result_error( f'Неизвестная команда action={action}' )

