'''
Загружает файлы расширений

В каталоге EXT_PATH или API2_PATH/../extensions должны находится файлы *.ext в которых содержатся маршруты к файлам-манифестам загружаемых модулей

Файлы манифеста являются YAML-файлами со структурой:
    # Требуемые модули
    required_modules : 
    - m_api

    # Требуемые маршруты API
    required_api :
    - api.ext2

    # Предоставляемые маршруты API
    provided_api :
    - api.ext1

    # Список файлов модулей для загрузки относительно файла-манифеста
    modules:
    - ext1_m1.py

Каждый модуль из секции modules загружается через функция api.add_module
'''


EXT_PATH_PARAM = 'EXT_PATH'                 # Название параметра конфигурации содержащего путь к каталогу расширений
DEFAULT_EXT_PATH = ['..', 'extensions']     # Пусть к расширения по-умолчанию, относительно API2_PATH

import os.path
import os
import yaml
import importlib.util
import importlib

import m_utils

api = None

def init(_api):
    global api
    api = _api

    api.add('load_extensions', load_extensions)
    api.add('load_extension', load_extension)


extract_file_name = lambda file : os.path.split(file)[-1].split('.')[0]


def _load_manifest(mfile: str):
    '''Загрузка файла манифеста'''
    manifest_data = {}
    manifest_data['path'], manifest_data['name'] = os.path.split(os.path.abspath(mfile))

    with open(mfile) as f:
        content = '\n'.join( f.readlines() )
        data = yaml.load(content, Loader=yaml.SafeLoader)
        manifest_data['provided_api'] = data.get('provided_api') or []
        manifest_data['required_api'] = data.get('required_api') or []
        manifest_data['modules'] = data.get('modules') or []

        for module in manifest_data['modules']:
            module_name = extract_file_name(module)
            manifest_data['provided_api'].append('api.modules.'+module_name)

    return manifest_data


def _load_extension_modules(manifest):
    '''Загрузка программных модулей'''

    ext_path = manifest['path']
    ext_name = manifest['name']

    for module_file in manifest['modules']:

        module_path = os.path.abspath( os.path.join(ext_path, module_file ) )
        #print(f'Loading: {module_path}')
        if not os.path.isfile(module_path):
            raise Exception(f'Надстройка {ext_name}: Не найден модуль {module_path}')

        for api_path in manifest['required_api']:
            if not api.find(api_path):
                raise Exception(f'Надстройка {ext_name}: Не найден маршрут {api_path}')

        module = _import_module_from_file(module_path)

        module_name = extract_file_name(module_path)
        api.add_module(module_name)



def load_extension(mfile: str):
    '''Загрузка расширения через файл манифеста'''
    manifest = _load_manifest(mfile)
    _load_extension_modules(manifest)


# Функция импорта модуля по полному имени файла
import sys
def _import_module_from_file(file_name):
    module_name = extract_file_name(file_name)
    spec = importlib.util.spec_from_file_location(module_name, file_name)
    #print(f'{module_name} : {file_name}\n{spec}')
    if not spec:
        raise Exception(f'Не найден файл {file_name} для модуля {module_name}')

    m = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = m

    sys.path.append( os.path.split(file_name)[0] )
    try:
        spec.loader.exec_module(m)
    finally:
        sys.path.pop()

    return m


def load_extensions():
    api.stdout(f'\n{"-"*40}\nЗагрузка расширений\n{"-"*40}\n')

    try:
        _load_extensions(api)
    except Exception as e:
        print(m_utils.exception_text(e))


def _load_extensions(api):
    # Определяем путь к каталогу с раширениями
    ext_path = api.get_config(EXT_PATH_PARAM)
    api_path = api.get_config('API2_PATH')

    if ext_path:
        pass

    elif api_path:
        ext_path = os.path.abspath( os.path.join(api_path, *DEFAULT_EXT_PATH) )
        msg = f'Не определен путь к расширениям {EXT_PATH_PARAM} используется API2_PATH => {ext_path}'
        print(msg)
        api.log.msg(msg)
        
    else:
        msg = 'm_loader: Не определены параметры конфигурации EXT_PATH или API2_PATH'
        print(msg)
        api.log.error(msg)
        return
        
    ext_dir = os.path.abspath(ext_path)
    if not os.path.isdir(ext_dir):
        msg = f'Не найден путь к расширениям {ext_dir}'
        print(msg)
        api.log.error(msg)
        return

    print(f'm_loader: Каталог расширений {ext_dir}')

    # Определяем список файлов-ссылок на манифесты модулей
    dir_items = os.listdir(ext_dir)
    mfiles = []
    for item in dir_items:
        file = os.path.join(ext_dir, item)
        if not file.endswith('.ext'): continue
        if not os.path.isfile(file): continue
        try:
            if extract_file_name(file)[0] in ('.', '~'): continue
        except Exception:
            continue

        # Определяем список файлов-манифестов
        with open(file) as f:
            for line in f.readlines():
                try:
                    if not line.strip() or line[0]=='#': continue

                    mfile = os.path.abspath(os.path.join(ext_dir, line.strip()))
                    if not os.path.isfile(mfile):
                        raise Exception(f'Надстройка {file}: Не найден файл-манифеста {mfile}')

                    mfiles.append(mfile)
                except Exception as e:
                    print('[ERROR]', e.args)

    # Определяем список надстроек
    exts = {}
    for mfile in mfiles:
        ext_path, mfile = os.path.split(mfile)
        ext_name = mfile.split('.')[0]
        exts[ext_name] = {'path':ext_path, 'mfile':mfile}


    # Загружаем информацию о надстройках из файлов-манифестов (*.yaml)
    for ext_name, ext_data in exts.items():
        mfile = os.path.join(ext_data['path'], ext_data['mfile'])
        ext_data.update(_load_manifest(mfile))
            

    # Функция определяющая необходимость загрузки e1 после e2
    def is_after(e1, e2):
        required = exts[e1].get('required_api') or []
        provided = exts[e2].get('provided_api') or []
        for api_path in required:
            if api_path in provided:
                return True
        return False


    # Расчитываем порядок загрузки настроек
    ext_list = list(exts.keys())
    history = [list(ext_list)]
    cur_pos = 0
    while cur_pos < len(ext_list):
        for i in range(len(ext_list)-1, cur_pos, -1 ):
            if is_after(ext_list[cur_pos], ext_list[i]):
                ext_list.insert(i+1, ext_list[cur_pos])
                del ext_list[cur_pos]
                
                if ext_list in history:
                    raise Exception('Loop found')
                history.append(list(ext_list))
        
        else:
            cur_pos += 1

    print(f'm_loader: Загружаем надстройки {ext_list}')
            

    # Загрузка модулей и их инициализация
    for ext_name in ext_list:
        module_path = None
        try:
            _load_extension_modules( exts[ext_name] )

            '''
            ext_path = exts[ext_name]['path']
            for module_file in exts[ext_name]['modules']:

                module_path = os.path.abspath( os.path.join(ext_path, module_file ) )
                #print(f'Loading: {module_path}')
                if not os.path.isfile(module_path):
                    raise Exception(f'Надстройка {ext_name}: Не найден модуль {module_path}')

                for api_path in exts[ext_name]['required_api']:
                    if not api.find(api_path):
                        raise Exception(f'Надстройка {ext_name}: Не найден маршрут {api_path}')

                module = _import_module_from_file(module_path)
                #api.log.msg(f'm_loader: Загружен модуль {module_path}')

                module_name = extract_file_name(module_path)
                api.add_module(module_name)
            '''

            api.add(f'extensions.{ext_name}', exts[ext_name])
            print(f'{m_utils.datetime_str()}: Надстрока {ext_name} загружена')

        except Exception as e:
            msg = f'Ошибка загрузка расширения extension={ext_name}\n\n'
            msg += m_utils.exception_text(e)
            api.log.error(msg)
            print(msg)
            print(f'{m_utils.datetime_str()}: [ERROR] Надстрока {ext_name} не загружена')
