LOG_PATH = 'api.log'

import time
import inspect
import threading
import os

import m_utils

api = None

def init(_api):
    global api
    api = _api

# Все категории должны быть учтены в Log._block
ERRORS = 'errors'
WARNINGS = 'warnings'
MSGS = 'msgs'
VARS = 'vars'
SOURCES = 'sources'
STACK = 'stack'

class Log:
    def __init__(self, _api, level=0):
        self.api=_api
        self.level = 0
        self.node = _api.set(LOG_PATH, self, ensure=True)
        self.auto_print = []
        self.data = {}          # Данные 
        self.dump = []          # Единый список соообщений

        self.start_time_str = self.time_str()

        self.set_level(level)
        self.clear()
            #
            #
    def time_str(self):
        ts=time.time() 
        return time.strftime('%H:%M:%S')+('%.3f'%(ts%1))[1:]
            #
            #
    def dump_add(self, *args, data=None, level=None):
        line = {}
        line['ts'] = self.time_str()
        line['level'] = level

        try:
            if self.api and 'request' in self.api.context:
                line['request'] = self.api.context.request.url
        except:
            pass

        args = [f'{arg}' for arg in args]

        line['msg'] = '\n'.join(args)
        line['data'] = data
        line['thread'] = self._th_id()
        line['caller'] = m_utils.CallerInfo().outer_func()

        self.dump.append(line)

        if len(self.dump) > 100: self.dump[-100:]
            #
            #
    def get_dump_html(self):
        dump = list(self.dump)
        dump.reverse()

        res = []
        for _line in dump:
            line = dict(_line)
            line['msg'] = m_utils.to_html(line['msg'])
            res.append(line)

        return res
            #
            #
    def clear_dump(self):
        self.dump = []
            #
            #

    def _th_id(self):                           # Возвращает идентификатор текущего потока выполнения
        return threading.get_ident()
            #
            #
    def _block(self, thread_id=None):           # Возвращает ссылку на словарь данных потока
        if not thread_id:
            thread_id = self._th_id()

        if thread_id not in self.data:
            self.data[thread_id] = {ERRORS:[], WARNINGS:[], MSGS:[], VARS:{}, STACK:{}, SOURCES:{}}
            self._clean(thread_id)


        return self.data[thread_id]
        #
        #
    def _clean(self, ignore_thread_id):         # Очищает данные от несуществующих потоков
        threads = [th.ident for th in threading.enumerate()]
        for th_id in list(self.data.keys()):
            if th_id not in threads and th_id != ignore_thread_id:
                del self.data[th_id]
        #
        #
    def _list(self, list_name):                 # Возвращает объект списка в который будет добавляться информация
        block = self._block()
        return block[list_name]
        #
        #
    def clear(self):                            # Очищает лог текущего потока
        th_id = self._th_id()
        if th_id in self.data:
            del self.data[th_id]        
        #
        #
    @property
    def errors(self):                           # Список ошибок текущего потока
        return self._block()[ERRORS]
        #
        #
    @property
    def warnings(self):                         # Список предупреждений текущего потока
        return self._block()[WARNINGS]
        #
        #
    @property
    def msgs(self):                             # Список сообщений текущего потока
        return self._block()[MSGS]
        #
        #
    @property
    def vars(self):                             # Список переменных текущего потока
        return self._block()[VARS]
        #
        #
    @property
    def sources(self):                          # Источники данных
        return self._block()[SOURCES]
        #
        #
    @property
    def stack(self):                            # Стек вызовов текущего потока
        return self._block()[STACK]
        #
        #
    def set_level(self, level):                 # Задает текущий уровень до которого сохраняются сообщения
        self.level = level
        #
        #
    def _auto_print(self, level, *args):        # Автоматическая печать добавляемого в log значения
        self.dump_add(*args, level=level)

        if not level in self.auto_print: return

        prefix = {
            ERRORS   : '\nERROR:   ',
            WARNINGS :   'WARN:    ',
            MSGS     :   '    MSG: ',
            VARS     :   '    VAR: '
        }[level]

        for data in args:
            if type(data) == list:
                for line in data:
                    print(prefix, line)

            if type(data) == dict:
                for k, v in data.items():
                    rec_key = level + '.' + k
                    print(f'{k} = {v} {level + "." + self.sources[rec_key]}')

            else:
                print(prefix, data)
        #
        #

    # === Работа со стеком вызова ==========================================================================
    priority = [ERRORS, WARNINGS, MSGS]         # Приоритеты сообщений

    def _frame_id(self, frame_info, lineno=True):
        module_name = inspect.getmodulename(frame_info.filename)
        res = f'{module_name}.{frame_info.function}'
        if lineno:
            res += f'[{frame_info.lineno}]'
        return res
        #
        #
    def _call_stack(self, lineno=True):
        stack = [self._frame_id(f, lineno) for f in inspect.stack() if inspect.getmodulename(f.filename) != __name__]
        return stack
        #
        #
    def _caller_id(self, lineno=True):
        return self._call_stack(lineno=lineno)[0]
        #
        #
    def flag_set(self, level):
        stack = self._call_stack(lineno=False)
        stack.reverse()                             # Делаем порядок наиболее старший уровень в начале

        # Удаляем из стека устаревшие элементы
        saved_funcs = list(self.stack.keys())
        for saved_func in saved_funcs:
            if saved_func not in stack:
                del self.stack[saved_func]

        for func in stack:
            saved_level = self.stack.get(func)
            if not saved_level or Log.priority.index(level) < Log.priority.index(saved_level):
                self.stack[func] = level
        #
        #
    def flag_check(self, level):
        caller = self._caller_id(lineno=False)
        saved_level = self.stack.get(caller)
        return bool(saved_level and Log.priority.index(level) >= Log.priority.index(saved_level))
        #
        #
    def flag_clear(self):
        caller = self._caller_id(lineno=False)
        clear_flag = False
        saved_funcs = list(self.stack.keys())
        for key in saved_funcs:
            if key == caller: 
                clear_flag = True

            if clear_flag:
                del self.stack[key]
        #
        #
    def error(self, *args):                     # Добавляет информацию об ошибке
        for arg in args:
            if not arg: continue
            self.sources[ERRORS+f'[{len(self.errors)}]'] = self._caller_id()
            self.errors.append(arg)

        self.flag_set(ERRORS)
        self._auto_print(ERRORS, *args)
        #
        #
    def exception(self, e, data={}):
        self.error(m_utils.exception_text(e), data)
        #
        #
    def warning(self, *args):                   # Добавляет предупреждение
        args = [arg for arg in args if arg]     # Удаление пустых аргументов
        for arg in args:
            self.sources[WARNINGS+f'[{len(self.warnings)}]'] = self._caller_id()
            self.warnings.append(arg)

        self.flag_set(WARNINGS)
        self._auto_print(WARNINGS, *args)
        #
        #
    def msg(self, *args, level=0):              # Добавляет сообщение
        if self.level < level: return 

        self.msgs.extend(args)
        self.flag_set(MSGS)
        self._auto_print(MSGS, *args)
        #
        #
    def var(self, *args, level=1, for_err=False, for_warn=False, **kwargs):    # Сохранает переменные с их значением в api.log.vars
        '''
        Сохраняет значения переменных в api.log.vars

        var('var1', ...)                - перечисленные переменные
        var(['var1', ...])              - перечисленные переменные
        var('locals')                   - все локальный переменные
        var({<var_name>:<value>, ...})  - пары ключ-значение
        '''
        if for_err and not self.flag_check(ERRORS): return
        if for_warn and not self.flag_check(WARNINGS): return

        def add(var_name, var_val):
            self.sources[VARS+'.'+var_name] = self._caller_id()
            self.vars[var_name] = var_val
            self._auto_print(VARS, {var_name:var_val})

        frame = inspect.stack()[level].frame

        for idx, arg in enumerate(args):
            if type(arg) == dict:
                for k,v in arg.items():
                    add(k ,v)

            elif type(arg) == list:
                var_names = arg
                for var_name in var_names:                    
                    if var_name in frame.f_locals:
                        add(var_name, frame.f_locals[var_name])

                    elif var_name in frame.f_globals:
                        add(var_name+' @global', frame.f_globals[var_name])

                    else:
                        add(var_name, '<undefined>')

            elif arg == 'locals':
                for var_name, var_val in frame.f_locals.items():
                    add(var_name, var_val)

            elif arg in frame.f_locals:
                add(arg, frame.f_locals[arg])                

            elif arg in frame.f_globals:
                add(arg+ ' @global', frame.f_globals[arg])                

            else:
                add(f'args[{idx}]', arg)

        for key, value in kwargs.items():
            add(key, value)
        #
        #
    def errors_str(self, thread_id=None, CR='\n\n'):            # Текст ошибок текущего потока
        errors = self._block(thread_id)[ERRORS]
        return CR.join(errors) if errors else ''
        #
        #
    def warnings_str(self, thread_id=None, CR='\n'):            # Текст предупреждений текущего потока
        warnings = self._block(thread_id)[WARNINGS]
        return CR.join(warnings) if warnings else ''
        #
        #
    def msgs_str(self, thread_id=None, CR='\n\n'):              # Текст сообщений текущего потока
        msgs = self._block(thread_id)[MSGS]

        res = f'Start time: {self.start_time_str}{CR}'
        if msgs:
            res +=  CR.join(msgs) + CR

        return res
        #
        #
    def vars_str(self, thread_id=None, CR='\n', inline=False):  # Текст с описанием переменных и их значений
        vars = self._block(thread_id)[VARS]
        sources = self._block(thread_id)[SOURCES]

        if vars:
            res = ''
            for name, val in vars.items():
                res += f'{name} = '
                res += f'{val}' if inline else m_utils.pretty(val)
                res += f'({sources[VARS+"."+name]}){CR}'
            return res
        else:
            return ''
        #
        #
    def dumps(self, exception=None, thread_id=None):            # Диагностические данные для frontend
        caption = 'Уведомление'
        text = ''

        if self.warnings:
            caption = 'Предпреждение'

            text += self.warnings_str(thread_id) 

        if self.errors:
            caption = 'Ошибка'

            if text: text += '\n\n'
            text = self.errors_str(thread_id) + text
            if exception:
                text += m_utils.exception_text(exception)

        if self.msgs:
            if text: text += '\n\n'
            text +=  self.msgs_str(thread_id)
            
        if self.vars:
            if text: text += '\n\n'
            text +=  self.vars_str(thread_id)

        return caption, text
        #
        #
    def print_all(self):
        print(f'{self.node.path()}')

        if self.errors:
            print('=== Ошибки =================================================')
            print(self.errors_str())

        if self.warnings:
            print('=== Предупреждения ==========================================')
            print(self.warnings_str())

        if self.msgs:
            print('=== Сообщения ===============================================')
            print(self.msgs_str())

        if self.vars:
            print('=== Переменные ==============================================')
            print(self.vars_str())

        if self.sources:
            print('=== Источники ===============================================')
            for k, v in self.sources.items():
                print(k, ':', v)
