'''
Модуль отображения атрибутов по карте отображения
 - хранит карты отображения

Карта экспорта задается в виде 
{
    <export_attr_name> : <attr_name>,                                       # Экспорт атрибута с изменением имени
    <export_attr_name> : [<const>],                                         # Экспорт константы
    <export_attr_name> : (<attr_name>, <convert_func(value, row_dict)>),    # Экспорт через вызов функции конвертации
    <export_attr_name> : (<attr_name>, <ref_name>, *ref_args),              # Экспорт атрибута содержащего ссылку
                                                                                для конвертации используется self.export_reference
}

Методы 
    add_map(map_name, map)  - Добавить карту отображения. 
        Карта задается в виде словаря в форма описанному выше.
    
    map_record(source, map_name, **export_reference_kwargs)   - Отобразить данные в соответствии с картой

    export_reference(map_name, value, *ref_args, **export_reference_kwargs)  - Вызывается при экспорте атрибута содержащего ссылку
        Вызывается когда отображается поле в описании отображения которого указан <action>
        Возвращает id записи в связанной таблице, соответствующей переданному value
        Метод должен переопределяться для конкретных применений

Пример:
    Инициализация:
        maps = Mapping()
        maps.add_map('test', {
                'target_attr':'source_attr'}
            )

    Использование:
        source = {'source_attr': 1}
        target = maps.map_record(source, 'test')

'''

import m_utils
import m_rec
    
class MapNotFound(Exception):
    pass

class MapSourceAttrNotFoundException(Exception):
    pass


# TODO 
# - Выделить из Mapping.map_record работу со словарем
# - Перенести метод Mapping.map_record без работы со словарем
class Map:
    def __init__(self, map={}, method_cb=None):
        self._map = map
        self._method_cb = None      # Callback для обработки действия заданного вторым аргументом
        #
        #
    def map(self, source):
        pass


class Mapping:                                                          # Хранилище правил отображения
    '''
    Карта задается в виде словаря с записями вида
        <поле в таблице-получателе> : <поле в таблице-источнике>
        <поле в таблице-получателе> : [<константа>]
        <поле в таблице-получателе> : [<поле в таблице-источнике>, <имя словаря>]
        <поле в таблице-получателе> : [<поле в таблице-источнике>, <функция преобразования>, <параметры функции>]
    '''
    def __init__(self, debug=False):
        self.maps = {}
        self.params = {}
        self.voc_values = {}                                            # Инициализация списков использованных значений справочников

        self._debug = debug
        self.map_warns = []
        #
        #
    def add_map(self, map_name:str, map:dict, params={}):               # Добавление карты отображения атрибутов
        '''
        map_name    - имя карты
        map         - карта переноса
        '''
        self.maps[map_name] = map
        self.params[map_name] = params
        #
        #
    def find_map(self, map_name):                                       # Поиск карты для отображения атрибутов
        for key in self.maps.keys():
            if m_utils.is_match(map_name, key):
                return self.maps[key], self.params.get(key, {})

        raise MapNotFound(f'Не найдена карта трансляции {map_name}\nдоступны карты {self.maps.keys()}')
        #
        #
    def export_reference(ref_name, value, *args, **kwargs):             # Экспорт ссылки на связанные данные
        return value
        #
        #
    def map_record(self, source, map_name, **export_reference_kwargs):  # Отображает запись в соответствии с картой
        self.map_warns = []

        if not source: 
            return {}

        map = self.find_map(map_name)[0]
        if not map:
            raise Exception(f'Не найдена карта отображения {map_name}')

        #if type(map) != dict:
        #    raise Exception(f'Не корректная карта {map_name} : {map}')

        if not isinstance(source, m_rec.Record):
            source = m_rec.Record(source)

        row = {}

        for new_attr, data in map.items():

            attr_name = None
            params = None

            # Извлекаем имя поля или константу
            if not data:
                value = None
                
            elif type(data) in (tuple, list):
                if len(data) == 1:          # Если одно значение - это константа для записи                 
                    value = data[0]
                else:                               
                    attr_name = data[0]
                    params = list(data[1:])

            else:
                attr_name = data


            # Извлекаем значение поля
            if attr_name:
                if attr_name in source:
                    value = source[attr_name]  
                else:
                    msg = f'Не найден атрибут {attr_name} для {map_name}'
                    self.map_warns.append(msg)
                    value = None


            # Если второй элементе строка, то это имя словаря
            if params and type(params[0]) == str:             
                voc_name = params.pop(0)
                if voc_name not in self.voc_values:                # Проверяет было ли такое значение уже добавлено
                    self.voc_values[voc_name] = {}

                if value:
                    if value not in self.voc_values[voc_name]:     # Экспортируем запись в словарь, если не было
                        try:
                            value_id = self.export_reference(voc_name, value, *params, **export_reference_kwargs)
                        except Exception as e:
                            print(f'm_map.export_reference exception voc_name={voc_name} value={value}')
                            raise e
                        self.voc_values[voc_name][value] = value_id 
                    
                    value =  self.voc_values[voc_name][value]       # Замечаем значений на код из словаря значений справочников                     


            # Если второй элемент функция, делаем преобразование
            elif params and callable(params[0]):
                convert_func = params.pop(0)
                try:
                    value = convert_func(value, source, *params, **export_reference_kwargs)

                except Exception as e:
                    print(f'Ошибка функции {convert_func.__name__} для параметра {value} типа {type(value)}')
                    print(f'    атрибут {new_attr} карта {map_name}')
                    print(m_utils.exception_text(e))
                    raise e

            row[new_attr] = value

        return row
