'''
Модуль для взаимодействия с Frontend
    - предоставляет компоненты для загрузки на Frontend
    - предоставляет модели данных для работы компонентов

api
+---frontend_models : API для интерфейсов данных к Frontend
+---modals          : API для хранения компонентов модальных страниц
+---pages           : API для хранения HTML-страниц

URLs:
/api/frontend-model     данные для Frontend-объекта
/api/frontend-actions   список действий для 
/api/pages              раздача html-страниц
/api/modals             раздача компонентов для модальных форм
/api/frontend           раздача JS-модулей для Frontend
'''
API_URL = '/api'

FRONTEND_MODELS_API_PATH = 'api.frontend_models'
FRONTEND_MODELS_URL = '/frontend-model'
FRONTEND_ACTIONS_URL = '/frontend-actions'
MODEL_NODE_TYPE = 'model'

PAGES_API_PATH  = 'api.pages'
PAGES_URL       = '/pages'
PAGE_NODE_TYPE  = 'page'

MODALS_API_PATH = 'api.modals'
MODALS_URL      = '/modals'
MODAL_NODE_TYPE = 'modal'

LIB_VUE_URL     = '/api2/vue'
LIB_API2_URL    = '/frontend'
LIB_COMPONENTS_URL  = '/api2/components'    

def init(_api, debug=False):
    global api
    api = _api

    # Регистрация модуля в структуре API 
    init_frontend_models()

    api.add_module_api('pages', ApiPages())
    api.add_url(PAGES_URL, ep_pages)

    api.add_module_api('modals', ApiModals())
    api.add_url(MODALS_URL, ep_modals)

    # Статические ресурсы для Frontend
    api.add_url(LIB_VUE_URL, ep_lib_vue)
    api.add_url(LIB_API2_URL, ep_lib_api2)
    api.add_url(LIB_COMPONENTS_URL, ep_lib_components)


api = None

import os
from werkzeug.wrappers.response import Response

import m_api

def get_file_path(file_name, base_path=None):       # Абсолютный путь к файлу

    if base_path and (file_name[0] == '.' or not os.path.split(file_name)[0]):
        path = os.path.split(os.path.abspath(base_path))[0]
        file_name = file_name.split('/')
        res_file_name = os.path.join(path, *file_name)

    else:
        res_file_name = os.path.abspath(file_name)

    return res_file_name


def load_content(file_name):                      # Загружает содержимое файла
    if not os.path.isfile(file_name):
        api.log.error(f'm_pages.load_content: Не найден файл {file_name}')
        raise Exception(f'Не найден файл {file_name}')

    with open(file_name, encoding='utf-8') as f:
        return f.readlines()


#===========================================================================================
#   Модели для страниц Frontend
#
#       добавление модели
#           model = api.frontend_models.FrontendModel(...)
#           api.repo.<repo_path>.add_frontend_model( model )
#
#       метод add_frontend_model реализуется ApiFrontendModels через provide
#           добавляет ноду с node_type = <имя класса потомка от FrontendModel>
#===========================================================================================
from typing import NewType
Url = NewType('Url', str)

import m_utils, m_tree, m_rest

# Интерфейс для работы с Frontend-объектом (странице, модальной формой)
# для каждой страницы на frontend создается индивидуальный класс-потомок от FrontendModel
# для каждой отдельной реализации на Frontend создается класс-потомок
# для каждого представления класса документов создается экземпляр в репозитории
class FrontendModelWrapper:
    def get_model(self): pass

class FrontendModel:
    '''
    Описание модели
    - собирает описание атрибутов
    - собирает список доступных действий
    - определяет URL REST API к источнику данных
    '''
    def __init__(self, path:m_tree.Node, model_name:str, title:str, data:dict=None):
        '''
        title - заголовок страницы
        '''

        self._node: m_tree.Node = path.add(model_name, self)
        self._title = title
        self._data = data if data!=None else {}
            #
            #
    def _gather_attrs(self, node):     # Собирает словарь атрибутов
        # TODO реализовать через repo API
        pass
            #
            #
    def _find_datasource(self, node):  # Возвращает источник данных
        # TODO взять из demo_browse_edit
        pass
            #
            #
    def _get_actions_list(self, node): # Возвращает список доступных действий
        # TODO реализовать через repo API
        pass
            #
            #
    def frontend_model(self, *args) -> dict:
        '''
        Генерирует модель для передачи на Frontend. Устанавливает атрибуты:
            - caption
        '''
        res = {'caption': self._title}
        res.update(self._data)
        return res
            #
            #
    def get_model_url(self) -> Url:
        '''
        Возвращает URL для загрузки данных этой Frontend-модели
        '''
        return f'{API_URL}{FRONTEND_MODELS_URL}?pname={self._node.path()}'


    def get_action_model(self):
        '''
        Возвращает модель Action для открытия страницы
        '''
        raise Exception('get_action_result is abstract')


    def get_menu_item_model(self, caption:str):
        '''
        Вовзращает описание элемента меню
            caption - заголовок элемента меню
        '''
        raise Exception('get_menu_item_model is abstract')


    def print(self):
        data = self.frontend_model()
        api.lib.debug.debug_print(m_utils.pretty(data))


    def update_context(self, context:dict) -> None:
        '''
        Добавляет в контекст вызова параметры передаваемые открывающемуся объекту
        '''
        if context == None: context = {}
        context['title'] = self._title
        context['model_url'] = self.get_model_url()


    def get_action_model(self, caption:str):
        '''
        Возвращает action=функцию для открытия frontend-объекта
        '''
        return lambda context, **kwargs: self.get_action_open(context, **kwargs)


    def get_action_open(self, context, **kwargs):
        raise Exception('FronendModel.action_open is abstract')


class ModalForm(FrontendModel):
    def __init__(self, 
                 path: m_tree.Node, 
                 model_name: str, 
                 title: str,
                 data: dict = {}, 
                 frontend_url:str=''):
        super().__init__(path, model_name, title, data)

        self._frontend_url = frontend_url        # URL для запуска модальной формы на frontend


    def get_frontend_url(self):
        return self._frontend_url

    def get_action_open(self, caption, context):
        '''
        Возвращает результат действия для открытия модальной формы
        '''
        self.update_context(context)
        data = m_rest.ApiRest.result_modal(self._frontend_url, context)
        url = api.actions.get_result_url() 
        return {'caption':caption, 'url':url, 'data':{'result':data}} 


    def get_menu_item_model(self, caption:str, context:dict=None):
        '''
        Возвращает описание элемента меню для вызова модальной формы
        '''
        if context==None: context = {}
        self.update_context(context)
        model = {
            'caption' : caption,
            'modal'   : self._frontend_url,
            'context' : context
        }
        return model




import m_actions

class ApiFrontendModels(m_api.ModuleApi):
    def add_page_type(self, page_type:str, model_type) -> m_tree.Node:
        return self._node.add(page_type, model_type, node_type=MODEL_NODE_TYPE)
            #
            #
    def get_actions_list(self, page_url, repo_path):
        node: m_tree.Node = api.repo.find_node(repo_path)
        page_node = api.frontend_models.find(page_url)
        model_node: m_tree.Node = None

        if isinstance(node, FrontendModel):
            model_node = node
        else:
            models, _ = node.children_find_down(instance=page_node._data)
            if not models: return []
            model_node = models[0]
        
        actionsApi :m_actions.ApiActions = api.actions
        actions = actionsApi.actions_list(model_node.path())

        if actions:
            return [action.frontend_model() for action in actions]
        else:
            return []
            #
            #
    def get_model(self, repo_path, args):
        '''
        Возвращает реализацию модели для конкретного 
        '''
        page_url = args.get('ptype')
 
        
        node:m_tree.Node = api.repo.find_node(repo_path, is_assert=True)

        if isinstance(node._data, FrontendModel):
            model = node._data

        elif isinstance(node._data, FrontendModelWrapper):
            model = node._data.get_model()

        else:
            _, node_datas = node.children_find_up(node_name=page_url, node_type=MODEL_NODE_TYPE)

            if not node_datas:
                raise Exception(f'Не найдена модель {page_url} для пути api.repo.{repo_path}')

            model:FrontendModel = node_datas[0]

        return model.frontend_model(args)
            #
            #
    def get_page_url(self, page_url, repo_path):       # URL для вызова страницы на сайте
        return f'/{page_url}/{repo_path}'

    
class FrontendModelsNode(ApiFrontendModels, m_tree.Node): pass


def add_frontend_model( page_type:m_tree.Node, *args, _path:m_tree.Node=None, **kwargs ):
    '''
    Добавление в репозиторий экземпляра потомка FrontendModel для отображения класса документов
    '''
    node_name = page_type._name

    node = _path.add(node_name, node_type=MODEL_NODE_TYPE)
    model = page_type._data(node, *args, **kwargs)  # Вызов конструктора модели
    node._data = model

    return node


def ep_frontend_models(api, context:dict, **kwargs):
    args = api.lib.rest.get_args(context)

    repo_path = args.get('pname')

    models_api = api.find(FRONTEND_MODELS_API_PATH)
    model = models_api.get_model(repo_path, args)

    return model


def ep_frontend_actions(api, context, **kwargs):
    args = api.lib.rest.get_args(context)

    page_type = args.get('ptype')
    repo_path = args.get('pname')

    models_api = api.find(FRONTEND_MODELS_API_PATH)
    model = models_api.get_actions_list(page_type, repo_path)

    return model


def init_frontend_models():
    api.repo.add('add_frontend_model', add_frontend_model)
    api.repo.provide(['add_frontend_model'])
    
    api.add_module_api(FRONTEND_MODELS_API_PATH, ApiFrontendModels())
    api.add_url(FRONTEND_MODELS_URL, ep_frontend_models)
    api.add_url(FRONTEND_ACTIONS_URL, ep_frontend_actions)

#===========================================================================================
#   Страницы для Frontend
#===========================================================================================

class ApiPages(m_api.ModuleApi):
    def __init__(self):
        super().__init__()
        self. _items = {}
            #
            #
    def list(self):
        return ', '.join(self._items.keys())
            #
            #
    def add_page(self, page_name:str, content:str):         
        '''
        Добавляет в репозиторий содержимое страницы
            - page_name - имя страницы в дереве api
            - content - строка - содержимое страницы
        '''
        self._items[page_name] = content
            #
            #
    def add_page_file(self, page_name:str, page_file:str, path=None):  
        '''
        Добавляет в репозиторий страницу из файла 
            - page_name - имя страницы в дереве api
            - page_file - путь к файлу страницы
        '''
        if path:
            page_file = os.path.join( os.path.split(path)[0], page_file )

        self.add_page(page_name, load_content(page_file))
            #
            #
    def get_page(self, page_name:str):
        '''
        Возвращает страницу из api.pages. по имени
        '''
        return self._items.get(page_name)
            #
            #
    def get_url(self, page_name, template='default', **kwargs) -> Url:
        '''
        Возвращает URL для загрузки страницы
            - page_name - имя страницы в api.pages.
            - template - маршрут к реализации работы со страницей на frontend
        '''
        url = f'/page/{template}/{page_name}'
        if kwargs:
            params_str = '&'.join([f'{k}={v}' for k,v in kwargs.items()])
            url += '?' + params_str
        return url


class PagesNode(ApiPages, m_tree.Node): pass


def ep_pages(api, context, **kwargs):
    '''
    Обработчик /api/pages ? name=<page_name> & [html=<>]   
    возвращает страницу для загрузки
    '''
    args = api.lib.rest.get_args(context)
    
    if 'name' not in args:
        return Response('Не указан параметр name', status=404)

    name = args['name']

    content = api.find(PAGES_API_PATH).get_page(name)
    content_list = api.find(PAGES_API_PATH).list()

    if not content:
        request = context['request']
        return Response(f'Не найдено {name}\nURL={request.base_url}.\n Доступны страницы: {content_list}', status=404)

    if 'html' in args:
        return Response(content, mimetype='text/html')
    else:
        return Response(content, mimetype='application/javascript')


#===========================================================================================
#   Компоненты модальных форм для Frontend
#===========================================================================================

class Container(m_api.ModuleApi):
    def __init__(self):
        super().__init__()
        #
        #
    def list(self):
        nodes = self._node.find_down(None, node_type=MODAL_NODE_TYPE, all_levels=True, single=False)
        node_names = [node.path(relative=self._node) for node in nodes]
        return ', '.join(node_names)
            #
            #
    def item_add(self, modal_name:str, content:str=None, file_name:str=None, base_path:str=None, _path=None) -> m_tree.Node:         
        '''
        Добавляет в репозиторий модальный компонент
        '''
        if content:
            modal = {'content': content}

        elif file_name:
            file_name = get_file_path(file_name, base_path)
            modal = {'file_name': file_name}

        if _path == None:
            _path = api.modals

        _path.add(modal_name, modal, node_type=MODAL_NODE_TYPE)
        #
        #
    def item_get(self, modal_name:str, _path=None, all_levels=False):
        if _path == None:
            modal_node = self._node.find_down(modal_name, node_type=MODAL_NODE_TYPE)
        
        else:
            modal_node = api.find(_path).find_down(modal_name, all_levels=all_levels)

        if modal_node is None: 
            return None

        modal = modal_node._data

        if 'content' in modal: 
            res = modal['content']
        
        if 'file_name' in modal:
            res = load_content(modal['file_name'])

        return res
        #
        #


class ApiModals(Container):     # API для api.modals
    def add_modal(self, modal_name:str, content:str=None, file_name:str=None, base_path:str=None, _path=None) -> m_tree.Node:
        if not os.path.isfile(file_name):
            msg = f'ERROR не найден файл {file_name}'
            raise Exception(msg)
        return self.item_add(modal_name, content, file_name, base_path, _path)
            #
            #
    def get_modal(self, modal_name:str, _path=None, all_levels=False) -> m_tree.Node:
        return self.item_get(modal_name, _path, all_levels)
            #
            #
    def get_url(self, modal_name:str) -> Url:
        return f'/api/modals?name={modal_name}'


class ModalsNode(ApiModals, m_tree.Node): pass


def ep_modals(api, context, **kwargs):
    request = context['request']
    args = api.lib.rest.get_args(context)
    
    if 'name' not in args:
        return Response('Не указан параметр name', status=404)

    name = args['name']

    url = request.base_url
    content = api.find(MODALS_API_PATH).get_modal(name)
    content_list = api.find(MODALS_API_PATH).list()

    if content:
        return Response(content, mimetype='application/javascript')
    
    else:
        return Response(f'Не найдено {name}\nURL={url}.\n Доступны страницы: {content_list}', status=404)


#===========================================================================================
#   Статические ресурсы для Frontend
#===========================================================================================

def ep_lib_vue(**kwargs):
    file_name = get_file_path('./static/vue3_2_45.js', __file__)
    content = load_content(file_name)
    #api.log.msg(f'ep_lib_vue: file_name={file_name} content={content}')
    return Response(content, mimetype='application/javascript')    


def ep_lib_api2(**kwargs):
    file_name = get_file_path('./static/api2.js', __file__)
    content = load_content(file_name)
    return Response(content, mimetype='application/javascript')
    

def ep_lib_components(**kwargs):
    file_name = get_file_path('./static/components.js', __file__)
    content = load_content(file_name)
    return Response(content, mimetype='application/javascript')
