from psycopg2.extras import DictRow


import m_utils


class RecException(Exception): pass

class RecAliasException(RecException): pass


NO_DEF = 'NO_DEF'

class Record:                               # Класс хранения записи 
    '''
    Предоставляет доступ к атрибутам записи как к методам
    сканирует атрибуты полей типа JSON и строит индекс по всем вложенным атрибутам 
    '''
    def __init__(self, data, fields=[], convert=True):
        # Декларация атрибутов экземпляра
        self.__dict__['_data'] = None
        self.__dict__['_idx'] = None
        self.__dict__['_parent'] = None
        self.__dict__['_aliases'] = None
        self.__dict__['_default'] = NO_DEF

        self._data = {}                     # Атрибуты записи и их значения
        self._parent = None                 # Ссылка на родительский элемент
        self._aliases = {}                  # Список атриубутов доступных на верхнем уровне

        if data == None:
            self._data = {}
            return

        if type(data) == dict:
            self._data = dict(data)
        
        elif type(data) in (list, tuple, DictRow):
            if fields:
                self._data = dict(zip(fields, data))
            else:
                raise RecException(f'Для загрузки данных требуется определить fields\ndata={data}')

        elif isinstance(data, Record):
            self._data = dict(data._data)
            self._aliases = dict(data._aliases)

        else:
            raise RecException(f'Record: некорректный тип данных {type(data)} для {data}')

        if convert:
            self.convert_data(self._data)       
        #
        #
    def convert_data(self, data:dict):              # Преобразование данных в структуры Record и RecordSet
        for key, value in data.items():
            try:
                if type(value) == dict:
                    record = Record(value)
                    record._parent = self
                    data[key] = record

                elif type(value) == list:
                    recordset = RecordSet(value)
                    recordset._parent = self
                    data[key] = recordset

            except RecException:
                data[key] = value
        #
        #

    def __repr__(self):
        res = {}
        for field in self.keys():
            if not '.' in field:
                res[field] = self._get_attr(field, default='<ERROR>')
        
        return f'Record({res})'
        #
        #
    def keys(self):                                 # Стандартный метод keys()
        '''
        Возвращает список атрибутов, включая атрибуты являющиеся алиасами.\n
        Аналогично dict.keys()
        '''
        res = set(self._aliases.keys()) | set(self._data.keys())
        return list(res)
        #
        #
    def items(self):                                # Стандратный метод items()
        '''
        Возвращает все атрибуты записи и их значения.\n
        Аналогично dict.items()
        '''
        res = []
        for key in self.keys():
            res.append( (key, self._get_attr(key)) )
        return res
        #
        #
    def __contains__(self, name):                   # Стандартный оператор in
        if name and '.' in name:
            return self.get(name, '<NotFound>') != '<NotFound>'
        else:
            return name in self.keys()
        #
        #

    def mark_update(self, source=None):
        if self._parent:
            self._parent.mark_update(self)
        #
        #
    def _set_attr(self, attr_name, value):
        if attr_name in self._aliases:
            hash_item = self._aliases[attr_name]
            hash_item[0][hash_item[1]] = value

        elif attr_name in self._data:
            self._data[attr_name] = value

        else:
            raise RecException(f'Record.__setattr__: Атрибут {attr_name} не найден')

        self.mark_update()
        #
        #
    def _get_attr(self, attr_name, default=NO_DEF):
        if attr_name in self._data:
            return self._data[attr_name]

        elif attr_name in self._aliases and self._aliases[attr_name]:
            hash_item = self._aliases[attr_name]
            hash_data = hash_item[0]
            data_key = hash_item[1]   
            try:         
                return hash_data.get(data_key) if hash_data!=None else None
            except Exception as e:
                raise Exception(f'Не корретный хэш {hash_data} типа {type(hash_data)} в записи {self}')

        else:
            if default != NO_DEF:
                return default
            elif self._default != NO_DEF:
                return self._default
            else:
                raise AttributeError(f'Не найден атрибут "{attr_name}" для объекта {self}')
                #RecException(f'Record.__getattr__: Атрибут {attr_name} не найден, Доступны атрибуты {list(self._data.keys())}')
        #
        #
    def _delete(self, attr_name):                    # Удаление атрибута или алиаса
        if attr_name in self._aliases:
            # Удаление алиаса
            del self._aliases[attr_name]

        elif attr_name in self._data:
            # Удалени атрибута. Поиск ссылок из алиасов и клонирование данных
            to_del = []
            for alias_name, alias_data in self._aliases.items():
                if alias_data == None:
                    continue

                if attr_name == alias_data[2] or alias_data[2].startswith(attr_name+'.'):
                    self._data[alias_name] = self._get_attr(alias_name)
                    to_del.append(alias_name)
            
            del self._data[attr_name]            

            for key in to_del:
                del self._aliases[key]
        #
        #
    def __setattr__(self, name, value):
        if name in self.__dict__:
            self.__dict__[name] = value

        else:
            self._set_attr(name, value)
        #
        #
    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]

        else:
            return self._get_attr(name)
        #
        #
    def __delattr__(self, name: str) -> None:
        self._delete(name)
        #
        #
    def __setitem__(self, idx, value):
        if idx not in self.keys():
            self.__dict__['_data'][idx] = None

        self._set_attr(idx, value)
        #
        #
    def __getitem__(self, idx):                     # Обращение по индексу
        attrs = idx.split('.')
        obj = self
        for attr in attrs[:-1]:
            obj = obj._get_attr(attr, default=None)
            if not obj: return None

        return obj._get_attr(attrs[-1])
        #
        #
    def __delitem__(self, name):
        self._delete(name)
        #
        #
    def get(self, attr_name, default=None):
        '''
        Возвращает значение атрибута attr_name. При отсутствии атрибута возвращает значение defuul.\n
        Аналогично dict.get()
        '''
        try:
            return self.__getitem__(attr_name)
        except:
            return default
        #
        #

    def set_default(self, default='NO_DEF'):        # Значение возвращаемое при отсутсвии атрибута
        '''
        Устанавливает значение возвращаемое если атрибут не найден
        разрешено значение None
        для сброса установки вызвать без указания параметра
        '''
        self._default=default
            #
            #
    def is_empty(self) -> bool:
        return not bool(self._data)

    def to_dict(self) -> dict:                          # Преобразует данные в dict, клонирует алиасы
        '''
        Возвращает стурктуру типа dict содержающую пары атрибут-значение.
        '''
        def value2dict(value):
            if value == None:
                return None

            elif hasattr(value, 'to_dict'):   
                return value.to_dict()

            elif m_utils.is_datetime(value):
                return m_utils.datetime_local_str(value)

            else:                          
                return value

        res = {}
        for key, value in self._data.items():
            res[key] = value2dict(value)

        for alias_name in self._aliases.keys():
            res[alias_name] = value2dict(self._get_attr(alias_name))

        return res
        #
        #
    def get_data(self):     # Возможно не используется
        res = {}

        for attr, data in self._aliases.items():
            if '.' in attr: continue

            if data:
                value = data[attr]
            else:
                value = self[attr]

            if type(value) in (Record, RecordSet):
                value = value.get_data()

            res[attr] = value

        return res

    def is_eq(self, data: dict):                        # Сравнивает запись с шаблоном
        '''
        Сравнивает запись с шаблоном типа dict. 
        Возвращает True если все атрибуты перечисленные в data имеют аналогичное значение у записи.
        '''
        for key, value in data.items():
            attr_value = self._get_attr(key, default='Unknown')
            if attr_value != value:
                return False
        return True
        #

    def attr_delete(self, attr_name:str):               # Удаление атрибута
        '''
        Удаляет из записи атрибут attr_name
        '''
        self._delete(attr_name)
        #
        #
    def attr_rename(self, old_name:str, new_name:str):  # Переменовывает атрибут
        '''
        Переименовывает атрибут.
        '''
        if old_name == new_name: return

        if old_name in self._aliases:
            self._aliases[new_name] = self._aliases[old_name]
            del self._aliases[old_name]

        elif old_name in self._data:
            self._data[new_name] = self._data[old_name]
            del self._data[old_name]

            for alias in self._aliases.keys():
                alias_data = self._aliases[alias]
                if alias_data[1] == old_name:
                    self._aliases[alias] = (alias_data[0], new_name, alias_data[2].replace(old_name, new_name))                

        else:
            raise RecException(f'Атрибут {old_name} не найден')
        #
        #
    def attr_filter(self, attrs_list: dict):            # Удалаяет атрибуты не указанные в списке
        '''
        Удаляет атрибуты отсутствующие в списке или словаре attrs_list
        '''
        if type(attrs_list)==dict:
            attrs_list = list(attrs_list.keys())

        del_list = []
        for attr in self.keys():
            if attr not in attrs_list:
                del_list.append(attr)
        
        for attr in del_list:
            del self[attr]
        #
        #
    def add_alias(self, alias_name, data_path):         # Добавляет алиас для доступа к данным
        # Определение ссылки на данные алиаса
        attrs = data_path.split('.')
        data = self._data

        if alias_name == '':                        # Копирование атрибутов на верхний уровень
            for attr in attrs:
                if attr in data:
                    data = data[attr] 
                else:
                    raise RecAliasException(f'Не найден атрибут "{attr}". Маршрут {data_path}. Данные {self._data}')

            if not hasattr(data, 'keys') or type(data)==str:
                return
                raise Exception(f'Не корректный тип данных "{type(data)}" по маршруту {data_path} данные: {data}')

            for key in data.keys():
                self._aliases[key] = (data, key, '.'.join(attrs+[key]))
        else:
            for attr in attrs[:-1]:
                if attr not in data:
                    raise RecAliasException(f'Не найден атрибут "{attr}". Маршрут {data_path}. Данные {self._data}')

                data = data[attr]

            self._aliases[alias_name] = (data, attrs[-1], '.'.join(attrs))

        #del self._aliases[attrs[0]]
        #
        #
    def map(self, map_dict, rename=False, default=None, fill_none=None):    # Переименовывает/отображает атрибуты по словарю {new_name:old_name}
        for new_name, new_value in map_dict.items():
            if type(new_value) in (tuple, list):
                if len(new_value) > 1:
                    # Применяем преобразование
                    attr_name =  new_value[0]
                    func = new_value[1]
                    value = func( self[attr_name] )
                else:
                    # Сохраняем константу
                    value = new_value[0]

                if fill_none!=None and value == None: value = fill_none
                self[new_name] = value

            else:
                old_name = new_value

                if old_name and old_name in self:
                    if rename:
                        self.attr_rename(old_name, new_name)
                    else:
                        self.add_alias(new_name, old_name)
                else:
                    self[new_name] = default
        #
        #
    def update(self, add_dict, base_only=False, update_to_none=True, update_to_blank=True):    # Глубокое обновление атрибутов
        if type(add_dict) == Record:
            add_dict = add_dict.to_dict()

        m_utils.deep_update(self._data, add_dict, base_only=base_only, update_to_none=update_to_none, update_to_blank=update_to_blank) 

    def expand(self, attr_name):
        res = []
        attr_values = self[attr_name]
        for _line_no, row in enumerate(attr_values):
            line = dict(self._data)

            if type(row) == dict:
                line.update(row)
            elif isinstance(row, Record):
                line.update(row._data)
            else:
                line[attr_name] = row

            if '_line_no' not in line:
                line['_line_no'] = _line_no 
            res.append(line)

        return RecordSet(res)
        

    def expand_old(self, attr_name):                    # Разворачиает в RecordSet по значениям в указанном атрибуте
        res = []
        attr_values = self[attr_name]
        for _line_no, row in enumerate(attr_values):
            line = self.to_dict()
            if attr_name in line:
                del line[attr_name]

            if type(row) == dict:
                line.update(row)
            elif isinstance(row, Record):
                line.update(row.to_dict())
            else:
                line[attr_name] = row

            if '_line_no' not in line:
                line['_line_no'] = _line_no 
            res.append(line)

        return RecordSet(res)
        #
        #
    def print(self, data=None, indent='', max_data_len=80, containers=[]):
        def print_value(name, value):
            spaces = ' ' * (max_key_len-len(name))
            v_str = f'{value}'
            if len(v_str) > max_data_len:
                v_str = v_str[:max_data_len] + ' ...'

            if type(value) in (list,dict) and name not in containers: 
                print(f'{indent}{name}{spaces} :')
                self.print(value, indent+'    ', max_data_len, containers)

            else:
                print(f'{indent}{name}{spaces} : {v_str}')


        if data == None: 
            self.print(self._data)
            if self._aliases:
                print('--- aliases ---')
                self.print_aliases()

        elif type(data) == dict:
            max_key_len = 0
            for key in data.keys():
                if len(key) > max_key_len:
                    max_key_len = len(key)

            for k,v in data.items():
                print_value(k, v)


        elif type(data) == list:
            max_key_len = len(str(len(data)-1))

            for i, v in enumerate(data):
                print_value(str(i), v)
        #
        #
    def print_aliases(self, indent=''):
        max_key_len = max([len(key) for key in self._aliases.keys()])

        for attr in self._aliases.keys():
            print(f'{indent}{attr}{" "*(max_key_len-len(attr))} : {self._get_attr(attr)} (alias of "{self._aliases[attr][2]}" )')
        #
        #


class RecordSet:                            # Класс данных списка экземпляров
    def __init__(self, data, convert=True):
        self._data = []
        self._fields = {}
        self._parent = None
        self._table = None
        self._aliases = {}              # Алиасы для доступа к данным
        self._updated = []              # Ссылки на измененные записи

        self.load(data, convert)
        #
        #
    def __repr__(self):
        return f'RecordSet({len(self._data)} records)'
        #
        #
    def __getitem__(self, idx):
        return self._data[idx]
        #
        #
    def __setitem__(self, idx, value):
        self._data[idx] = value
        self.mark_update(self._data[idx])
        #
        #
    def mark_update(self, source=None):
        if source != None:
            self._updated.append(source)
        else:
            source = self

        if self._parent:
            self._parent.mark_update(self)
            #
            #
    def mark_saved(self, item=None):
        if item == None:
            self._updated = []

        elif item in self._updated:
            self._updated.remove(item)
            #
            #
    def updated(self):
        return list(self._updated)
            #
            #
    def load(self, data, convert):                  # Преобразование списка в Record[]
        '''
        Загружает данные из значения типа list of dict или tuple(list of fields, list or tuple)
        '''
        if type(data) == tuple:
            self._fields = data[0]
            rows = list(data[1])

        else:
            self._fields = {}
            rows = list(data)

        self._data = []

        if type(rows) not in (list, DictRow):
            raise RecException(f'Неправильный тип данных {type(rows)} должен быть list')

        for idx, row in enumerate(rows):
            try:
                record = Record(row, self._fields, convert=convert)
            except:
                print(f'Ошибка загрузки данных {row} в модель {self._fields} исходные данные {data}')
                record = Record({})
                raise

            record._parent = self
            record._idx = idx
            for alias, path in self._aliases.items():
                record.add_alias(alias, path)

            self._data.append(record)

        self.mark_saved()
        #
        #
    def print(self, fields=[], indent=''):          # Печать данных на консоли
        for idx, row in enumerate(self._data):
            if fields:
                row_ = []
                for field in fields:
                    row_.append(f'{field} : {row[field]}')
                row_ = ', '.join(row_) 
                print(f'{indent}{idx} : {row_}\n')
            else:
                print(f'{indent}{idx} : ')
                m_utils.print_compact_dict(row, indent=indent+(' '*4))
                print()
            
    def is_empty(self):                             # Проверка на наличие данных
        return not bool(self._data)
            #
            #
    def to_dict(self) -> list:                      # Возвращает данные в структуре list of dict
        '''
        Возвращает данные в виде list of dict
        '''
        return [rec.to_dict() for rec in self._data]
            #
            #
    def get_data(self):                             # Возвращает данные в структуре 
        res = []
        for rec in self._data:
            res.append( rec.get_data() )

        return res
        #
        #
    def first(self) -> Record:                      # Возвращает первую запись
        '''
        Возвращает первую запись
        '''
        return self[0]
        #
        #
    def count(self) -> int:                         # Кол-во записей
        '''
        Возвращает кол-во записей
        '''
        return len(self._data)
        #
        #
    def delete(self, index:int):                    # Удаление записи из списка
        '''
        Удалает запись по заданному индексу
        '''
        if type(index) == dict:
            i = 0
            while i < self.count():
                if self._data[i].is_eq(index):
                    self.delete(i)
                else:
                    i += 1
        else:
            del self._data[index]
        #
        #
    def filter(self, filter):                       # Применение функции для фильтрации записей
        '''
        Возвращает RecordSet содержащий только записи удовлетворяющие фильтру.\n
        Фильтр задается в виде dict или функции func(record:Record).
        '''
        if type(filter) == dict:
            func = lambda rec: rec.is_eq(filter)
        else:
            func = filter

        res = []
        for rec in self._data:
            if func(rec):
                res.append(rec)

        rs = RecordSet([])
        rs._data = res

        return rs
        #
        #
    def apply(self, func, **kwargs) -> None:        # Применяет функцию к каждой записи
        '''
        Применяет функцию к каждой записи.
        func(Record, **kwargs)
        '''
        for rec in self._data:
            func(rec)

    def append(self, rs) -> None:                   # Добавляет записи из другого Record, RecordSet, dict, list
        '''
        Добавляет к существующим записям записи из rs:RecordSet
        '''
        if type(rs) == Record:
            self._data.append(rs)

        elif type(rs) == dict:
            self._data.append(Record(rs))

        elif type(rs) == RecordSet: 
            self._data.extend(rs._data)

        elif type(rs) == list:
            self._data.extend(RecordSet(rs))

        else:
            raise Exception(f'Не корректный тип данных {type(rs)} допустимы dict, list, m_rec.Record, m_rec.RecordSet')
    
    # --- Операции с атрибутами записей
    def attr_delete(self, attrs) -> None:
        '''
        Удаляет атрибут или список атрибутов из всех записей.\n
        attrs: str, list
        '''
        if type(attrs) != list:
            attrs = [attrs]

        for attr in attrs:
            self.apply(lambda rec: rec.attr_delete(attr))
        #
        #
    def attr_rename(self, old_attr:str, new_attr:str) -> None:  # Переименовывает атрибут
        '''
        Переименовыевает атрибут для всех записей.
        '''
        if old_attr == new_attr: return

        for rec in self._data:
            if old_attr in rec:
                rec.attr_rename(old_attr, new_attr)
        #
        #
    def attr_set(self, attr_name:str, value) -> None:           # Устанавливает атрибут для всех записей
        '''
        Устанавливает значение атрибута для всех записей.
        '''
        for rec in self._data:
            if callable(value):
                rec[attr_name] = value(rec)
            else:
                rec[attr_name] = value
        #
        #
    def attr_copy(self, old_attr:str, new_attr:str) -> None:    # Копирует атрибут под новым именем
        '''
        Копирует атрибут для всех записей
        '''
        for rec in self._data:
            rec[new_attr] = rec[old_attr]
        #
        #
    def attr_filter(self, attrs_list):                          # Удаляет атрибуты которые отсутсвуютс в списке
        '''
        Удаляет из всех записей атрибуты указанные в списке.\n
        attrs_list: str, list, dict
        '''
        for rec in self._data:
            rec.attr_filter(attrs_list)
        #
        #
    def add_alias(self, alias_name, data_path):                 # Добавление алиаса для доступа к данным
        '''
        алиасы добавляются автоматически для всех Record внутри RecordSet
        .add_alias('', 'attrs') - добаление всех атрибутов поля attrs на верхний уровень
        .add_alias('table', 'form_table_data.main_table') - 
        '''
        self._aliases[alias_name] = data_path

        for row in self._data:
            try:
                row.add_alias(alias_name, data_path)
            except RecAliasException:
                pass

    def attr_keys(self) -> list:                                # Список атрибутов, определяется по первой записи
        if self._data and len(self._data):
            return list(self._data[0].keys())
        else:
            return []


    def attr_values(self, attr_name:str) -> set:                # Возвращает значения атрибута
        '''
        Возвращает множество значений атрибута из записей.
        '''
        res = set()
        for rec in self._data:
            res.add(rec[attr_name])
        return res
        #
        #
    def calc_sum(self, func):                                   # Рассчитывает сумму применяя функцию к каждой записи. Функция получает Record
        '''
        Расчитывает сумму применяю функцию к каждой записи.\n
        func(Record) -> float, int
        '''
        res = 0
        for rec in self._data:
            res += func(rec)
        return res

