'''
Модуль репозитория



Реализует
    api.repo        - API репозитория
    api.repo.attr_types   - дерево типов

    endpoint /repo/attr_types - возвращающий список типов

    базовые типы: date, money, quantity, slot
    тестовые типы: test_list, tast_dict
'''

STORAGE_NODE_TYPE = 'storage'
GROUP_NODE_TYPE = 'group'
CLASS_NODE_TYPE = 'class'

ATTR_TYPE_NODE_TYPE = 'attr_type'
ATTR_NODE_TYPE = 'attr'


def init(_api): 
    global api
    api = _api

    api.add_module_api('api.repo', ApiRepo())
    api.repo.provide(['add_group', 'add_class', 'add_attr_type'])
    api.repo.provide(['find_storage', 'add_datasource', 'find_datasource'])
    api.repo.provide(['add_attrs', 'add_attrs_json_file', 'gather_attrs'])

    api.add_module_api('api.repo.attr_types', ApiAttrTypes())

    # --- Базовые типы данных --------------------------------
    api.repo.add_attr_type('date', {'base_type':'DATE'})
    api.repo.add_attr_type('money', {'base_type':'BASE', 'handlers':{'before_label':['money']}, 'style':{'textAlign':'end'}})
    api.repo.add_attr_type('quantity', {'base_type':'BASE', 'handlers':{'before_label':['quantity']}, 'style':{'textAlign':'end'}})
    api.repo.add_attr_type('slot', {'base_type':'SLOT'})

    # --- Типы данных для отладки --------------------------------
    api.repo.add_attr_type('test_list', AttrTypeList(['list: Один', 'list: Два', 'list: Три']))
    api.repo.add_attr_type('test_dict', AttrTypeDict({'1':'dict: Один', '2':'dict: Два', '3':'dict: Три'}))

    api.add_url('/repo/attr_types', ep_attr_types)

api = None

import os, json
from typing import List
import m_utils, m_tree, m_rec, m_struct
import m_db, m_ds, m_api, m_rest, m_web_ds, m_pages


NODE_TYPE_DS = 'DataSource'         # значение node_type для объектов DataSource


def ep_attr_types(api, **kwargs):
    '''
    Endpoint /repo/attr_types
    '''
    res = {
        'attr_types': api.repo.attr_types.dump_models()
    }
    return m_rest.ApiRest.result(res)



class ApiRepoException(m_api.ApiException):
    pass


#=========================================================================
#   Классы элементов репозитория
#=========================================================================
class RepoNode:                         # Прототип элелементов репозитория
    def __repr__(self):
        return f'{self.__class__.__name__}'
        #
        #
    def desc(self):
        repo = self._node.find_up('repo')
        return repo._node_desc(self._node.path())
            #
            #
class RepoStorage(RepoNode):            # Описание хранилища
    pass
        #
        #
class RepoGroup(RepoNode):              # Описание группы
    pass
        #
        #
class RepoClass(RepoNode):              # Описание класса
    def get_class_attrs(self) -> dict:
        return {}
    
#====================================================================
#   Данные для IDE
#====================================================================
class Provide:
    def add_group(self, group_name:str, _path=None): pass
    def add_class(self, class_name:str, *, data=None, desc='', _path=None): pass
    def add_attr_type(self, type_name, type_data, _path=None): pass
    def find_storage(self) -> RepoStorage: pass
    def add_datasource(self, path:str, datasource:m_ds.DataSource, _path=None): pass
    def find_datasource(self, path:str, datasource:m_ds.DataSource, _path=None): pass
    def add_attrs(self, node_name: str, attrs_model, _path=None): pass
    def add_attrs_json_file(self, file_path:str, _path=None): pass
    def gather_attrs(self, local_model=None, _path:m_tree.Node=None, class_fit=False): pass

class RepoStorageNode(RepoStorage, Provide, m_tree.Node): pass
class RepoDataSourceNode(m_ds.DataSource, Provide, m_tree.Node): pass
class RepoGroupNode(RepoGroup, Provide, m_tree.Node): pass
class RepoClassNode(RepoStorage, Provide, m_tree.Node): pass

#=========================================================================
#   Api AttrTypes
#=========================================================================
class ApiAttrTypes(m_api.ModuleApi):
    def dump_models(self):
        types = self._node.find_down(None, node_type=ATTR_TYPE_NODE_TYPE, single=False)
        res = {}
        for item in types:
            if type(item._data) == dict:
                res[item._name] = item._data
            elif hasattr(item, 'frontend_model'):    
                res[item._name] = item.frontend_model()
            else:
                raise Exception(f'Не корректное описание типа {item._name} типа {type(item._data)} : {item._data}')

        return res

#=========================================================================
#   Типы атрибутов
#=========================================================================

class AttrType:                     # Прототип типов данных
    def __init__(self, base_type:str, **kwargs):
        self.name = str(id(self))
        self.base_type = base_type
        self._type_attrs = kwargs
        self.is_coded = False        # Признак что значение атрибута нужно раскодировать
            #
            #
    def check(self, value) -> bool:                 # Проверяет значение на соответствие типу
        return True
            #
            #
    def values(self) -> list:                       # Список возможных значений в формате [(pk, id, label, **attrs)]
        return []
            #
            #
    def decode(self, value_id, default='') -> str:  # Преобразование кода значения в текстовое представление
        return str(value_id)
            #
            #
    def frontend_model(self) -> dict:               # Формат описания типа для Frontend
        model = {'base_type': self.base_type}
        model.update(self._type_attrs)
        return model


class AttrTypeCalc(AttrType):       
    # Значения вводятся через калькулятор на Frontend
    def __init__(self, calc_type, **kwargs):
        super().__init__(base_type='CALC', **kwargs)
        self._calc_type = calc_type
            #
            #
    def frontend_model(self):
        res = super().frontend_model()
        res.update({
                'calc_type': self._calc_type
            } )

        return res


class AttrTypeDict(AttrType):       
    '''
    Тип атрибута - значения из словаря
    '''
    def __init__(self, items:dict, **kwargs):
        super().__init__(base_type='DICT', **kwargs)
        self._items = items
        self.is_coded = True
            #
            #
    def values(self):
        return [(value_id, value_id, label) for value_id, label in self._items.items()]
            #
            #
    def decode(self, value_id, default=''):
        return self._items.get(value_id, default)
            #
            #
    def frontend_model(self):
        res = super().frontend_model()
        res.update({
                'data_source' : self._items
            })

        return res


class AttrTypeModal(AttrType):      # Значение выбирается через модальную форму
    def __init__(self, form_model:m_pages.FrontendModel, **kwargs):
        super().__init__(base_type='MODAL')
        self._form_model = form_model
        
    def frontend_model(self) -> dict:
        res = super().frontend_model()
        res.update(dict(
            modal_url = self._form_model.get_model_url()
        ))
        return res


class AttrTypeList(AttrType):       # Значеиня из списка
    def __init__(self, items:list, **kwargs):
        super().__init__(base_type='LIST', **kwargs)
        self._items = items
        self.is_coded = False
            #
            #
    def values(self):
        return [(idx, value, value) for idx, value in enumerate(self._items)]
            #
            #
    def frontend_model(self):
        res = super().frontend_model()
        res.update({
                'data_source' : self._items
            })

        return res

    
class AttrTypeDsDict(AttrType):         
    '''
    Список значений из источника данных
    '''
    # Значения из источника данных
    def __init__(self, 
                 ds:m_tree.Node, 
                 id_attr:str, 
                 label_attr:str=None, 
                 label_func=None, 
                 cached=False, 
                 translate=False,
                 **kwargs):
        super().__init__(base_type='DICT', **kwargs)

        self.is_coded = label_attr != None and label_attr != id_attr
        self.label_func = label_func

        if type(ds) == m_tree.Node and isinstance(ds._data, m_ds.DataSource):
            ds = ds._data

        if not isinstance(ds, m_ds.DataSource):
            raise Exception(f'Не корректный источник данных {ds} type={type(ds)}')

        self._ds = ds
        self._id_attr = id_attr
        self._cached = cached
        self.translate = translate

        self._items = []

        if label_attr:
            self._label_attr = label_attr 
        else:
            self._label_attr = id_attr
                # 
                #
    def set_label_func(self, label_func):
        self.label_func = label_func
        self._items = None
            #
            #
    def load_values(self):
        rs:m_rec.RecordSet = self._ds.select()

        res = {}
        for row in rs._data:
            try:
                if self.label_func:
                    res[row[self._id_attr]] = self.label_func(row)
                else:
                    res[row[self._id_attr]] = row[self._label_attr]
            except Exception as e:
                msg = m_utils.exception_text(e)
                msg += f'self={self}  row={row}  label_attr="{self._label_attr}"'
                api.log.error(msg)

        self._items = res
            #
            #
    def values(self):
        if not self._items:
            self.load_values()

        return self._items
        #return [(k,v) for k,v in self._items.items()]
            #
            #
    def decode(self, value_id, default=None):
        if self._cached and self._items:
            pass
        else:
            self.load_values()

        return self._items.get(value_id, default)
            #
            #
    def frontend_model(self):
        self.load_values()          # Принудительня загрузка

        res = super().frontend_model()
        res.update(dict(
            data_source = self.values(),
            translate_value_id = self.translate,                
        ))
        return res


class AttrTypeDsTable(AttrType):
    '''
    Тип со список значений загружаемым из DataSource c выбором значения в модальной форме

    TODO
    список атрибутов загружаем из ds
    '''
    def __init__(self, 
                 ds:m_tree.Node, 
                 id_attr:str, 
                 label_attr:str=None, 
                 form_model:m_pages.FrontendModel=None, 
                 multiselect:bool=False, 
                 translate:bool=False,
                 modal_url:str=None,
                 **kwargs
                ):
        if modal_url:
            base_type = 'MODAL'
            self.modal_url = modal_url
        else:
            base_type = 'VOC'
            self.modal_url = ''

        super().__init__(base_type=base_type, **kwargs)
        self.ds = ds
        self.web_ds = m_web_ds.WebDs(ds, ds)
        self.id_attr = id_attr
        self.label_attr = label_attr
        self.form_model = form_model
        self.multiselect = multiselect
        self.translate = translate


    def default_form_model(self):
        attrs_model = self.ds.gather_attrs()
        model = attrs_model.frontend_model()
        headers = attrs_model.get_headers()
        if not headers: headers = [attr['field'] for attr in model]

        return dict(
                data_url = self.web_ds.get_select_url(),
                voc_name = 'title',
                edit = {"edit": False},
                attrs_wrap = False,                         # не извлекать данные из атрибута attrs
                form = {'table' : {'model': model, 'headers':headers}},
                types = {},
                lazy_load_size = 50,
                translate_value_id = self.translate,
            )


    def frontend_model(self):
        res = super().frontend_model()
        form_model = self.form_model.frontend_model() if self.form_model else self.default_form_model()

        res.update(dict(
            modal_url = self.modal_url,
            data_url = self.web_ds.get_select_url(),
            selection_mode = "multiple" if self.multiselect else "single",
            select_form = form_model
        ))
        return res


'''
DEPRECATED
Остановлено для совместимости с отчета которые делал Андрей
'''
class AttrTypeDs(AttrTypeDsTable):
    pass
#=========================================================================
#   Атрибуты
#=========================================================================
class AttrBase:                     # Прототип структуры описания атрибута
    '''
    Класс-прототип для описания атрибута

    ._options = словарь свойств атрибута

    Варианты вызова конструктора
        AttrBase( {<свойства атрибута} )
        AttrBase( <свойство>=<значение>, ... )

    '''
    def __init__(self, *args, _weak:bool=None, _replace:bool=None, _weak_options:list=None, **kwargs):
        # Декларация системных свойств
        self.__dict__['_weak'] = None
        self.__dict__['_weak_options'] = None
        self.__dict__['_replace'] = None
        self.__dict__['_options'] = None     
        self.__dict__['_db_type'] = None     

        # Определение системных ствойств
        self._weak = _weak or False
        self._weak_options = _weak_options or []
        self._replace = _replace or False
        self._options = {}

        if args and len(args)==1 and not kwargs and type(args[0]==dict):
            # Единственный аргумент типа dict
            self._options = args[0]

        elif not args and kwargs:
            # Только именованные аргументы
            self._options = kwargs

        elif not args and not kwargs:
            pass

        else:
            raise Exception(f'Ошибка описания args={args} kwargs={kwargs}. Должен быть DICT или пары <опция>=<значение>')
        
        # Переименовываем старый вариант типа атрибута type->attr_type
        if 'type' in self._options:
            attr_type = self._options['type']
            if attr_type != None:
                self._options['attr_type'] = attr_type
                if 'attr_type' in self._weak_options:
                    self._weak_options.remove('attr_type')

            del self._options['type']

        if self._weak:
            self._weak_options = list(self._options.keys())
            #
            #
    def __repr__(self):
        res = f'{self.__class__.__name__}({self._options})'
        return res
            #
            #
    def __getitem__(self, option_name):
        if option_name in self.__dict__:
            return self.__dict__[option_name]
        else:
            return self._options[option_name]
            #
            #
    def __getattr__(self, option_name):
        if option_name in self.__dict__:
            return self.__dict__[option_name]
        else:
            return self._options[option_name]
            #
            #
    def __setattr__(self, opt_name, value):
        if opt_name in self.__dict__:
            self.__dict__[opt_name] = value
        else:
            self._options[opt_name] = value
            #


    DEFAULT_DB_TYPE = 'default'

    def as_field(self, db_type=DEFAULT_DB_TYPE):
        '''
        Указывает что атрибут хранится в отдельном поле таблицы с заданным типом.
        По-умолчанию, тип поля определяется типом атрибута в описании
        '''
        if db_type not in m_db.db_types and db_type!=self.DEFAULT_DB_TYPE:
            raise Exception(f'Некорректный тип поля "{db_type}" доступны {list(m_db.db_types.keys())}')
        self._db_type = db_type

        return self


    def get_db_field(self, db_type:str=None) -> m_db.Field:
        '''
        Возвращает модель поля в таблице базы данных соответствующего атрибуту или None
         - db_type - строка с типом данных поля в БД
        '''
        if db_type!=None:
            return m_db.db_types[db_type]
        else:
            return m_db.db_types[self._db_type] if hasattr(self, '_db_type') else None


    def get(self, option_name, default=None):   # Возвращает свойство атрибута
        '''
        Возвращает свойство атрибута
        '''
        if option_name in self._options: 
            return self._options[option_name]

        else:
            return default
                #

    def is_weak(self) -> bool:
        '''
        Возвращает флаг - атрибут не содержит полезных данных для переопределения свойств
        '''
        return self._weak
    

    def is_replace(self) -> bool:
        '''
        Возвращает флаг - описание атрибута должно перекрыть все свойства        
        '''
        return self._replace


    def is_required(self) -> bool:
        '''
        Возвращает значение атрибута required - требуются данные
        '''
        return self.get('required', False)
            #

    def update(self, attr_model: dict):
        '''
        Обновляет свойства атрибута из словаря или другой модели типа AttrBase
        '''
        if not isinstance(attr_model, AttrBase):
            attr_model = AttrBase(attr_model)

        if attr_model.is_replace():  
            # Заменяем описание атрибута
            self._options = dict(attr_model._options)

        elif attr_model.is_weak():
            # Добавляем только свойства которые не были описаны ранее
            for opt_name, opt_value in attr_model._options.items():
                if opt_name not in self._options:
                    self._options[opt_name] = opt_value
                    self._weak_options.append(opt_name)

        else: 
            # Исключаем свойства которые определены в получателе и в источнике являются weak_options
            new_options = dict(attr_model._options)
            for opt_name in list(new_options.keys()):
                if opt_name in attr_model._weak_options:
                    del new_options[opt_name]
                else:
                    if opt_name in self._weak_options and opt_name not in attr_model._weak_options:
                        self._weak_options.remove(opt_name)

            # Исключаем обновленные атрибуты из weak_options
            

            self._options.update(new_options)
            #
            #
    def frontend_model(self):                   
        '''
        Возвращает описание атрибута для Frontend
        '''
        model = dict(self._options)

        # Подстановка имени типа для созданныех объектов типов
        if isinstance(model.get('attr_type'), AttrType):
            model['attr_type'] = model['attr_type'].name

        return model




class AttrModel(AttrBase):
    '''
    Модель атрибута для frontend
        - caption - заголовок поля ввода / колонки в таблице
        - attr_type - тип атрибута
        - readonly - флаг "Атрибут только для чтения"
        - required - атрибут должен иметь значение
        - replace - описание заменяет все свойства атрибута
        - weak - описание заменяет только не заданные свойства атрибута
    '''
    def __init__(self, 
                # Обазательные ствойства атрибутов
                caption:str=None, 
                attr_type:str=None, 
                *,
                readonly:bool=None, 
                required:bool=None, 

                # Системные параметры
                replace:bool=None, 
                weak=False, 
                weak_options=None,

                **kwargs
        ) -> dict:

        weak_opt = weak_options or []
        if caption == None: weak_opt.append('caption')
        if attr_type == None: weak_opt.append('attr_type')
        if readonly == None: weak_opt.append('readonly')
        if required == None: weak_opt.append('required')

        super().__init__(
            caption=caption or '<не определен>', 
            attr_type=attr_type or 'str', 
            readonly=readonly or False, 
            required=required or False, 

            _weak = weak,
            _replace = replace,
            _weak_options = weak_opt,
            **kwargs
        )


class AttrCalculated(AttrModel):
    '''
    Модель атрибута, значение которого вычисляется на backend
    '''
    def __init__(self, caption, *, slot=None, calcFront=None, **kwargs):         # Генератор вычисляемого атибута
        kwargs['slot'] = slot or 'html'
        kwargs['calcFront'] = calcFront
        super().__init__(caption, **kwargs)


class AttrsModel(RepoNode):                     # Модель набора атрибутов
    '''
    Набор атрибутов
    хранит описания атрибутов с возможностью обновления другими объектами AttrsModel

    В конструкторе можно указывать: другие модели, словари атрибутов, отдельные атрибуты
    AttrsModel(
        <AttrsModel>, ...
        <dict>, ...
        <attr_name> = AttrBase(...),
        <attr_name> = <dict>
    )
    '''
    replace_option = '_replace'                            # Опция для обновления других опций вместо замены описания атрибута
    system_options = [replace_option, '_group', '_col', 'calcFront']      # Системные опции исключаемые при генерации frontend_model
    main_options = ['field', '_col', 'caption', '_group', 'type']        # Опции выводимые в начале описания

    def from_class(data_class, strict=False):
        '''
        Создает описание атрибутов из модели
        '''
        req_value_type = AttrBase if strict else None
        attrs = m_utils.get_all_class_attrs_list(data_class, req_value_type)
        res = {}
        for attr in attrs:
            attr_name = attr[0]
            attr_desc = attr[1]

            if callable(attr_desc): continue
            if attr_name.startswith('_'): continue

            if type(attr_desc) == tuple:
                value = attr_desc[0]
            else:
                value = attr_desc

            if value == None: 
                value = AttrModel(attr_name, weak=True)

            if not isinstance(value, AttrModel): continue

            if attr_name not in res: res[attr_name] = AttrModel(attr_name, weak_options=['caption'])
            res[attr_name].update(value)

        return AttrsModel(**res)


    def __init__(self, *args, **kwargs):
        args = list(args)
        self._attrs = {}                    # Хранилище атрибутов вида <attr_name> : <AttrBase>
        self.local_types = {}              # Словарь локальных типов
        self._types = {}
        self._headers = []                  # Заголовки
        self._headers_del = []              # Список атрибутов для удаления из заголовков

        # Сохраняем заголовки
        if args and type(args[0])==dict:
            self._headers = args[0].get('headers')
            self._headers_del = args[0].get('headers_del', [])
            args[0] = args[0].get('model', [])

        else:
            if 'headers' in kwargs: self._headers = kwargs.pop('headers')
            if 'headers_del' in kwargs: self._headers_del = kwargs.pop('headers_del')

        self.update(*args, **kwargs)
            #
            #
    def __repr__(self):
        attrs = list(self._attrs.keys())
        return  'Attrs: ' + ', '.join(attrs)
            #
            #
    def _sort_options(self, data:dict):             # Сортирует опции в соответствии с main_options
        keys = list(data.keys())

        def sort_key(option):
            return self.main_options.index(option) if option in self.main_options else len(self.main_options)

        keys.sort(key=sort_key)

        res = {}
        for key in keys: res[key] = data[key]

        return res
            #
            #
    def _clear_system_options(self, data):          # Удаляет из описания системные атрибуты
        for sys_option in self.system_options:
            if sys_option in data:
                data.pop(sys_option)
            #
            #
    def print(self):                                # Вывод frontend_model для отладочной панели
        res = {}
        for k,v in self._attrs.items():
            attr_desc = self._attr_desc_dict(v)
            attr_desc = self._sort_options(attr_desc)
            res[k] = attr_desc

        api.lib.debug.debug_print(m_utils.pretty(res))

    def is_empty(self) -> bool:
        return not bool(self._attrs)

    def attrs_list(self) -> List[str]:
        '''
        Возвращает список с именами атрибутов
        '''
        return list(self._attrs.keys())

    def fit_attrs(self, attrs_list:list):
        if not attrs_list: return

        for attr_name in list(self._attrs.keys()):
            if attr_name not in attrs_list:
                del self._attrs[attr_name]

    # Работа с атрибутами
    def add_field(self, field_name:str):
        '''
        Добавляет атрибут соответсвующий указанному полю с заголовком равным имени поля
        '''
        if field_name not in self._attrs:
            self._attrs[field_name] = Attr(field_name, field_name)
            #
            #
    def _attr_desc_dict(self, desc) -> dict:        # Извлекает описание атрибута в виде DICT
        if isinstance(desc, AttrBase):
            return desc.frontend_model() 

        elif type(desc) == dict:
            return desc

        else:
            raise Exception(f'Не корректный тип описания атрибута "{type(desc)}"')
            #
            #
    def _update_attrs(self, data):                  # Обновляет атрибуты с возможностью обновления описания

        if type(data) == dict and 'field' in data:      # Преобразование старой нотации {field: ...} в {field : {...}}
            attr_name = data.pop('field')
            data = {attr_name : data}

        if isinstance(data, AttrBase):                  # Преобразование одного атрибута в словарь атрибутов
            data = {data.field : data}

        for attr, desc in data.items():
            # Игнорирование системных атрибутов
            if attr.startswith('_'): continue

            if type(desc) == dict or isinstance(desc, AttrBase):                      
                if attr not in self._attrs: self._attrs[attr] = AttrModel(attr, weak_options=['caption'])
                self._attrs[attr].update(desc)

            else:
                raise Exception(f'Не корректный тип описания для атрибута {attr}. Описание должно быть типа dict или AttrModel')
            #
            #
    def update(self, *args, **kwargs):              # Обновление модели атрибутов
        '''
        Обновляет описания атрибутов

        Если имя атрибута начинается с "+" производится обновление описания атрибута
        '''
        if args:
            for arg in args:
                if isinstance(arg, AttrsModel):
                    self._update_attrs(arg._attrs)
                    self._update_headers(arg._headers)
                    if arg._headers_del: self._headers_del = arg._headers_del
                    #self._remove_header_attrs(self._headers, arg._headers_del)

                elif isinstance(arg, AttrBase):
                    self._update_attrs( arg )

                elif type(arg) == dict:
                    self._update_attrs(arg)

                elif type(arg) == list:
                    for item in arg:
                        self._update_attrs(item)

                else:
                    raise Exception(f'Не корректный аргумент {arg}')

        self._update_attrs(kwargs)
            #
            #

    # Работа с заголовками атрибутов
    def _get_header_attrs(self, header) -> list:
        '''
        Возвращает список атрибутов используемых в header
        '''
        if isinstance(header, dict): 
            return self._get_header_attrs(list(header.values()))
        
        elif isinstance(header, list):
            res = [] 
            for item in header:
                res.extend( self._get_header_attrs(item) )
            return res
        
        return [header]
            #
            #
    def _remove_header_attrs(self, header:list, removes:list)->list:
        '''
        Удаляет из header элементы из removes
        '''
        i = 0
        while i<len(header):
            item = header[i]
            if type(item) == str:
                if item in removes: 
                    del header[i]
                    continue
            
            else:
                key = list(item.keys())[0]
                item[key] = self._remove_header_attrs(item[key], removes)
                if not item[key]: 
                    del header[i]
                    continue

            i += 1

        return header
            #
            #
    def remove_headers(self, attr_list:list):
        self._headers_del = attr_list

    def _update_headers(self, add_headers):
        '''
        Обновляет заголовки.
            - удаляет добавляемые атрибуты из существующего заголовка
            - добавляет новый заговолок к существующему
        '''
        add_attrs = self._get_header_attrs(add_headers)
        self._remove_header_attrs(self._headers, add_attrs)
        self._headers.extend(add_headers) 
            #
            #
    def _get_col_index(self, attr):
        '''
        возвращает номер колонки из описания заданного атрибута
        '''
        return (self._attrs.get(attr) or {}).get('_col', 'я')
            #
            #
    def get_headers(self):                          # Создает описание заголовков из списка атрибутов с учетом _group и _col
        '''
        Возвращает структуру заголовков для таблица на основе атрибутов _group и _col
        '''
        if self._headers: 
            self._remove_header_attrs(self._headers, self._headers_del)
            return self._headers

        if not self._attrs: return ['Данные отсутствуют']
            #raise Exception('В модели отсутствуют атрибуты')

        cols = list(self._attrs.keys())
        cols.sort(key=lambda attr : self._get_col_index(attr))

        res = []
        last_group = None

        for attr in cols:
            attr_group = self._attrs[attr].get('_group')
            if last_group and last_group==attr_group:  
                res[-1][last_group].append(attr)

            else:
                last_group = attr_group
                if last_group:
                    res.append({last_group:[attr]})
                else:
                    res.append(attr)

        self._remove_header_attrs(res, self._headers_del)
        return res
            #
            #

    # Работа с типами
    def get_types(self, custom_types:dict={}) -> dict:
        '''
        Возвращает словарь типов
        '''

        # Формирование списка типов используемых атрибутами
        types =  [attr.get('attr_type') or attr.get('type') for attr in self._attrs.values()] 

        # Заменяем ссылки на типы в дереве API на имена типов
        for i in range(len(types)):
            if isinstance(types[i], m_tree.Node):
                types[i] = types[i]._name

        # Удаление повторений в списке используемых типов
        types = list(set(types) - {None})

        print(f'Types: {types}')

        # Формирование словаря с описаними типов
        types_models = {}
        for attr_type in types:
            if type(attr_type) == str: 

                # Извлечение типа из глобального списка типов
                repo_type = api.repo.attr_types.get_attr(attr_type)
                if isinstance(repo_type, m_tree.Node):
                    repo_type = repo_type._data

                # Определяет описание типа из репозитория
                if type(repo_type) == dict:
                    front_type = repo_type
                elif hasattr(repo_type, 'frontend_model'):
                    front_type = repo_type.frontend_model()
                elif repo_type == None:
                    front_type = None
                    api.log.warning(f'Не определен тип {attr_type}')
                    continue 
                else:
                    raise Exception(f'Не корректное описание типа "{attr_type}" : "{repo_type}" : "{type(repo_type)}"')

                # Обновляем описание типа из локального описания типов
                custom_type_model = custom_types.get(attr_type, {})
                custom_type_model_dict = custom_type_model.frontend_model() if hasattr(custom_type_model, 'frontend_model') else custom_type_model 
                front_type.update(custom_type_model_dict)
                type_name = attr_type

            elif isinstance(attr_type, AttrType):
                # Извлечение описания типа из объекта AttrType
                front_type = attr_type.frontend_model()
                type_name = attr_type.name      # Подстановка имени для динамически созданных типов

            else:
                raise Exception(f'Не корректный тип данных {attr_type}')

            if type_name in self.local_types:
                front_type.update(self.local_types[type_name])

            # Сохранение модели типа для Frontend
            types_models[type_name] = front_type


        return types_models
            #
            #
    def get_type(self, type_name):
        if not self._types:
            self._types = self.get_types()

        return self._types.get(type_name)


    def frontend_model(self, attr_field='field') -> list:   # Возвращает модель описания атрибутов для Frontend
        res = []

        for attr, desc in self._attrs.items():
            # Превращяем описания типа dict атрибутов в Attr
            if type(desc) == dict: 
                desc = AttrBase(desc)

            attr_desc = desc.frontend_model()
            #attr_desc = self._attr_desc_dict(desc)
            attr_desc[attr_field] = attr 

            if 'attr_type' in attr_desc: attr_desc['type'] = attr_desc.pop('attr_type')

            # Сортировка свойств по именам
            attr_desc = self._sort_options(attr_desc)

            # Удаление системных свойств
            self._clear_system_options(attr_desc)

            res.append(attr_desc)

        return res
            #
            #

    # DEPRECATED
    def Attr(caption, attr_type=None, **kwargs):        # Генератор атрибута
        raise Exception('DEPRECATED. Use m_repo.AttrModel')
            #
    def CalcAttr():
        raise Exception('DEPRECATED. Use m_repo.AttrCalculated')


class AttrStruct(m_struct.Struct):
    '''
    Коллекция атрибутов позволяющуя генерировать модель таблицы в БД
    '''
    _table_name = None

    def __init__(self, **kwargs):
        #if 'default' not in kwargs: kwargs['default'] = False
        kwargs['ignore_default_type'] = True
        super().__init__(**kwargs)

    def make_attr_model(self) -> AttrsModel:
        '''
        Создает модель описания атрибутов
        '''
        return AttrsModel().from_class(self)
    

    def make_table_model(self, db) -> m_db.TableModel:
        '''
        Создает модель описания таблицы в БД
        '''
        fields = {}
        attr_types = self.get_attr_types()
        for attr_name, attr_value in self.get_attrs_dict().items():
            if not isinstance(attr_value, AttrModel): continue

            if attr_value._db_type == AttrModel.DEFAULT_DB_TYPE:
                fields[attr_name] = attr_value.get_db_field(attr_types[attr_name].__name__)()  
            else:
                field = attr_value.get_db_field()            
                if field != None:
                    fields[attr_name] = field()
                # Иначе, атрибут не содержит свойства _db_type и не является полем таблицы БД
            
        if self._table_name == None:
            raise Exception(f'Не определено имя таблицы. Атрибут _table_name')
        
        return m_db.TableModel(db, self._table_name, fields_map=fields, ensure=True)
    


#=========================================================================
#    API репозитория
#=========================================================================

class ApiRepo(m_api.ModuleApi):
    def __init__(self):
        super().__init__()

        self.hash = {}                  # Список классов по типам
        #
        #
    def node_desc(self, path):
        if type(path) == str:
            node = self._node[path]
        else:
            node = path

        # Список дочерних node типа RepoNode
        remove_list = [child._name for child in node._children.values() if isinstance(child._data, RepoNode)]

        return node.to_dict(exclude=remove_list)
        #
        
    # === Хранилища ==================================================
    def add_storage(self, type_name:str, data=None) -> RepoStorageNode:                   
        '''
        Добавление хранилища данных
        '''
        if type_name not in self.hash:
            self.hash[type_name] = {}
        
        node = self._node.add(type_name, data or RepoStorage(), node_type=STORAGE_NODE_TYPE)
        return node
            #
            #
    def find_storage(self, _path) -> RepoStorageNode:
        return _path.find_up(node_type=STORAGE_NODE_TYPE) 


    def add_datasource(self, path:str, datasource:m_ds.DataSource, _path=None) -> RepoDataSourceNode:
        if not _path: _path = api.repo

        return _path.add(path, datasource, node_type=NODE_TYPE_DS)
            #
            #
    def find_datasource(self, ds_name='ds', _path=None) -> RepoDataSourceNode:
        _, ds_list = _path.children_find_up(ds_name, node_type=NODE_TYPE_DS) 
        if not ds_list:
            return None
        
        return ds_list[0]

    # === Группы ======================================================
    def add_group(self, group_name:str, _path=None) -> RepoGroupNode:              
        '''
        Добавление группы для объектов
        '''
        node = _path.add(group_name, RepoGroup(), node_type=GROUP_NODE_TYPE)
        return node
        #
        #

    # === Классы ======================================================
    def add_class(self, class_name:str, *, data=None, desc='', _path=None) -> RepoClassNode: 
        '''
        Добавления класса объектов данных
        '''
        if type(_path) == str:
            _path = self._node[_path]

        storage_name = self.find_class_storage( _path )._name

        if class_name in self.hash[storage_name]:
            raise ApiRepoException(f'Класс {class_name} уже определен для типа {storage_name}')

        class_node = _path.add(class_name, data, node_type=CLASS_NODE_TYPE)
        class_node._desc = desc

        if type(self.hash[storage_name]) != dict:
            self.hash[class_name] = {}
        self.hash[storage_name][class_name] = class_node
        return class_node
        #
        #
    def get_class(self, _path:m_tree.Node) -> RepoClassNode:
        '''
        Возвращает ноду класса, к которому относится текущая нода
        '''
        return _path.find_up(node_type=CLASS_NODE_TYPE)
            #
            #
    def find_class_storage(self, path:str) -> RepoStorageNode:                                           
        '''
        Возвращает хранилище для заданного класса
        '''
        if type(path) == str:
            node = self._node.find(path)
        else:
            node = path
            path = node.path()

        if not node:
            raise ApiRepoException(f'Неизвестный путь {path}')

        while node and node._node_type != STORAGE_NODE_TYPE:
            node = node._parent

        if not node:
            raise ApiRepoException(f'Ошибка в структуре репозитория. Не определено хранилище для пути {path}')

        return node
        #
        #
    def find_node(self, repo_path:str, node_type=None, is_assert=True) -> m_tree.Node:
        '''
        Ищет node в репозитории. Маршрут может быть задан полностью 
        или в формате <storage_name>..<node_name>
        или в формате <node_name>
        '''
        node = self._node.find(repo_path)
        if node == None: node = api.find(repo_path)

        if node != None and (node_type==None or node.node_type==node_type): return node

        if '..' in repo_path:
            storage_name, node_name = repo_path.split('..')
            node = self._node.find(storage_name)
        else:
            node = self._node
            node_name = repo_path

        nodes, _ = node.children_find_down(node_name, node_type=node_type)

        if not nodes:
            if is_assert:
                raise Exception(f'Не найден элемент {repo_path}')
            else:
                return None

        return nodes[0]
            #
            #
    def find_class(self, storage_name:str, class_name:str, is_assert=True) -> RepoClassNode:
        '''
        Находит в дереве репозитория api.repo ноду соответсвующую заданному типу и классу.
        '''
        if storage_name in self.hash and class_name in self.hash[storage_name]:
            class_node = self.hash[storage_name][class_name]
            return class_node
        
        else:
            if is_assert:
                raise Exception(f'Не найден класс {class_name} для хранилища {storage_name}')
            else:
                return None
            #
            #
    def find_class_datasource(self, storage_name:str, class_name:str)->RepoDataSourceNode:
        '''
        Возвращает источник данных по-умолчанию для заданного класса
        '''
        class_node = api.repo.find_class(storage_name, class_name)
        _, ds_list = class_node.children_find_up(node_type=NODE_TYPE_DS)
        return ds_list[-1]      # Наиболее близкий источник данных к заданной ноде 

    # === Типы атрибутов ======================================================
    def add_attr_type(self, type_name, type_data: AttrType, _path=None) -> None:
        # Постановка места хранения типов атрибутов по-умолчанию
        if _path == None: _path = api.repo.attr_types

        _path.set(type_name, type_data, node_type=ATTR_TYPE_NODE_TYPE, ensure=True)
            #
            #
    def get_attr_type(self, type_name, _node=None) -> AttrType:
        return api.repo.attr_types.get_attr(type_name)

    #--------------------------------------------------------------------------
    #   Атрибуты
    #--------------------------------------------------------------------------
    def add_attrs(self, node_name: str, attrs_model: AttrsModel, _path=None) -> None:
        if _path == None: _path = api.repo

        _path.set(node_name, attrs_model, node_type=ATTR_NODE_TYPE, ensure=True)
            #        
            #
    def add_attrs_json_file(self, file_path:str, _path=None) -> None:
        '''
        Добавляет описание атрибутов и заголовков из json-файла
        '''
        file_name = os.path.split(file_path)[-1].split('.')[0]
        node_name = f'attrs_{file_name}'

        with open(file_path, encoding='utf8') as f: content = f.read()
        model = json.loads(content)
        self.add_attrs(node_name, AttrsModel(model), _path=_path)


    def gather_attrs(self, local_model:AttrsModel=None, class_fit=False, _path:m_tree.Node=None, ) -> AttrsModel:                 
        '''
        Собирает набор атрибутов для заданной ноды поднимаясь по дереву API вверх
            local_mode - локальная AttrsModel для обновления описания атрибутов
            class_fit - сокращает списко атрибутов до используемых в классе
        '''
        nodes, nodes_data = _path.children_find_up(instance=AttrsModel)
        print('\n'.join([node.path() for node in nodes]))        # TODO убрать после отладки

        res = AttrsModel()
        for attrs in nodes_data:
            res.update(attrs)

        if local_model:
            res.update(local_model)

        if class_fit:
            class_node = self.get_class(_path=_path)
            if hasattr(class_node._data, 'get_attrs_dict'):
                class_attrs = class_node._data.get_attrs_dict()
                res.fit_attrs(class_attrs.keys())

                if hasattr(class_node._data, 'set_runtime_attrs'):
                    class_node._data.set_runtime_attrs(res)

        return res


#=========================================================================
#    Deprecated
#=========================================================================

class Attr(AttrBase):               # Старая реализация для совместимости с аналитическими отчетами СГУК
    '''
    DEPRECATED (оставлен для совместимости с аналитичекими отчетами СГУК)
    Заменить на AttrModel(field=<field>, ...)

    Класс описания простого атрибута

    Конструктор:
        Attr(
            <имя атрибута в БД>,
            <заголовок для отображения>,
            <тип атрибута>,
            <значение>,
            <атрибут> = <значение>,
            ...
        )

        Attr(
            field = <имя атрибута в БД>,
            caption = <заголовок для отображения>,
            attr_type = <тип атрибута>,
            _replace = True для полной замены описания атрибута,

            required = False - флаг для проверки: атрибут должен иметь значение
            <атрибут> = <значение>,
            ...
        )
    '''
    def __init__(self, field=None, caption=None, attr_type=None, required=None, **kwargs):

        if type(field) == dict:
            field = dict(field)
            # Обязательные атрибуты
            kwargs['field'] = None

            if 'type' in field: kwargs['attr_type'] = field.pop('type') 
            elif 'attr_type' in field: kwargs['attr_type'] = field.pop('attr_type')

            kwargs.update(field)

        else:
            kwargs['field'] = field
            caption = caption or field
            if caption: kwargs['caption'] = caption
            if attr_type: kwargs['attr_type'] = attr_type 

        kwargs['required'] = required if required != None else kwargs.get('required', False)


        if isinstance(kwargs.get('attr_type'), m_tree.Node):
            kwargs['attr_type'] = kwargs['attr_type']._name

        super().__init__(**kwargs)
            #
            #


def Param(field, caption=None, param_type=None, *args, **kwargs):      
    '''
    DEPRECATED (оставлено для совместимости с аналитическими отчетами СГУК)
    Заменить на AttrModel(field=<field>, ...)

    Генератор описания параметра
    '''  
    return Attr(field, caption, param_type, *args, **kwargs).frontend_model()


