'''
URL
    /api/methods?       POST: вызывает метод из API2
        action=exec
        name=<имя метода>
        path=<маршрут в дереве API>
        context={параметры}                 параметры передаются методу как key-value arguments
'''

METHODS_URL = '/methods'
PROPS_URL   = '/props'


api = None

def init(_api):                                     # Инициализация модуля
    global api
    api = _api

    api.add_module_api('api.lib.rest', ApiRest())

    # Регистрация модуля в структуре API
    api.add_url(METHODS_URL, methods_endpoint)
    api.add_url(PROPS_URL, props_endpoint)


#=========================================================================
#   функции формирования результата
#=========================================================================
from werkzeug.wrappers.response import Response
import requests

import m_api

class ApiRest(m_api.ModuleApi):
    def host_url(self):
        request = api.context.get('request')
        if not request:
            raise Exception(f'ApiRest: context.request отсутсвует. Context={api.context}')

        url = request.scheme + '://' + request.host
        return url
    
    def get_auth_header(self):
        user = api.context.user
        # request = api.context.get('request')
        # resp.set_cookie('Auth', cookie, max_age=0)
        if user == None:
            login = 'admin'
        elif type(user)==dict:
            login = user['login']
        else:
            login = user.login

        return {'TestUser': login}

    def get_args(self, context=None, url_only=False):               # Извлекает аргументы запроса из контекста
        if not context: context = api.context
        request = context['request']
        args = request.args.to_dict()                               # Чтение параметров из url запроса

        if not url_only:
            body_data = request.get_json(force=True, silent=True)
            if body_data:
                args.update(body_data)                                  # Чтение параметров из тела запроса (context, ...)

        return args

    def get(self, url):
        if '://' not in url:
            url = self.host_url() + url

        return requests.get(url, headers=self.get_auth_header())


    def post(self, url, data={}):
        if '://' not in url:
            url = self.host_url() + url

        return requests.post(url, data, headers=self.get_auth_header())


    def download_content(content: str, mimetype='text/text', filename:str='file'):
        headers = {
            "content-length" : len(content),
            "Content-Disposition": f'attachment;filename={filename}',
        }

        return Response(content, mimetype=mimetype, headers=headers)


    def result(*args, **kwargs):              # Создает результат действия для отображения уведомления (caption, text)
        if len(args)==1:
            return args[0]
        elif args:
            return args
        else:
            return kwargs
            #
            #
    def result_notify(caption, text, refresh=None, **kwargs):
        res = {'caption':caption, 'text':text, 'refresh':refresh}
        res.update(kwargs)
        return res
            #
            #
    def result_process(proc_id):                    # Идентификатор созданного процесса
        return {'proc_id': proc_id}
            #
            #
    def result_url(url: str, inplace=False):        # Создает результат для перехода по ссылке
        if inplace:
            return {'push_url': url}
        else:
            return {'url': url}
            #
            #
    def result_file(file_name: str):                # Создает результат для загрузки файла
        return {'_send_file': file_name}
            #
            #
    def result_download(url:str):                   # Возвращает ссылку для загрузки файла
        return {'download_url': url}
            #
            #
    def result_html(html:str):
        return Response( html, mimetype='text/html' )
            #
            #
    def result_html_code(code: str):
        return ApiRest.result_html( m_utils.to_html_code( code ) )
            #
            #
    def result_modal(url, context):
        return ApiRest.result({'modal': url, 'context':context})
            #
            #
    def result_error(*args):
        args = list(args)
        for i, arg in enumerate(args):
            if type(arg) != str:
                args[i] = str(arg)

        res = '\n'.join(args)

        return m_utils.to_html(res)
            #
            #
    def result_exception(e:Exception, *args):
        return ApiRest.result_error( m_utils.exception_text(e), *args)


import json
import m_utils

#=========================================================================
#   methods REST API
#=========================================================================
class RestException(Exception):
    pass


def method_exec(path:str, method_name:str, method_context:dict):
    if not path.startswith('api'):
        path = 'api.repo.' + path

    node = api.find(path)
    if node == None :
        raise RestException( f'Маршрут {path} не найден' )

    func = node.get_attr(method_name) or getattr(node, method_name)
    if func == None:
        raise RestException( f'Метод {method_name} по маршруту {path} не найден' )

    if callable(func): 
        res = func(**method_context)
        return m_utils.simplify(res)
    else:
        raise RestException( f'{path}.{method_name} не является функцией (type={type(func)}, value {func})' )


def assert_params(args, param_names):
    if type(param_names) != list:
        param_names = [param_names]

    for param_name in param_names:
        if param_name not in args:
            raise RestException(f'Отсутствует параметр {param_name}')
    

def methods_endpoint(context, **kwargs):                      
    try:
        args = api.lib.rest.get_args(context)
        request = context['request']

        assert_params(args, 'action')     

        #api.log.msg(f'methods_endpoint: args={args}  context={api.context}')

        action = args['action']
        if action == 'exec':
            assert_params(args, ['name'])

            method_name = args['name']
            path = args.get('path', '')
            method_context = args.get('context', {})
            if type(method_context)==str:
                try:
                    method_context = json.loads(method_context)
                except:
                    print('method_context', method_context)
                    raise
                
            res = method_exec(path, method_name, method_context)
            if 'pretty' in args:
                res = ApiRest.result_html_code( m_utils.pretty(res) )

            return res

        elif action == 'tree':
            return ApiRest.result_html_code( api.tree_str()[-1] )

        return f'Неизвестное действие {action}'

    except RestException as e:
        return ApiRest.result_error( e.args, f'URL: {request.full_path}' )

    except Exception as e:
        return ApiRest.result_error( m_utils.exception_text(e), f'URL: {request.full_path}'  )

#=========================================================================
#   props REST API
#=========================================================================

def props_endpoint(api, context, **kwargs):
    args = api.lib.rest.get_args(context)
    
    action = args.get('action', 'get')

    if action == 'get':
        path = args.get('path', '')

        words = path.split('.')
        path = '.'.join( words[:-1] )
        prop_name = words[-1]

        node = api.find(path)
        if node != None:
            value = node.get_attr(prop_name)
            return m_utils.simplify(value)
        else:
            return f'Атрибут {path} не найден'


    return f'Неизвестное действие {action}'






