'''
Типизованные структуры данные

Struct - прототип класса с типизованными атрибутами
'''

import m_utils
import utils.args as args


class Struct:
    '''
    Класс с атрибутами заданного типа
        - default - загружать значения по-умолчанию, заданные при описании
        - strict - при присваивании атрибутам проверять точное соответствие типу
        - ignore_default_type - игнорировать проверку типа при присваивании значений по-умолчанию
    '''
    _attr_types: dict

    def __init__(self, init_data:dict=None, *, default=False, strict=False, ignore_default_type=False, _fake_init=False):
        # _fake_init - предотвращает бесконечную рекурсию
        self.__dict__['_attr_types'] = {}
        self._strict_type_check = strict
        if not _fake_init:
            self._gather_attrs(default=default, ignore_default_type=ignore_default_type)

        if init_data != None:
            self.from_dict(init_data)


    def __contains__(self, attr_name):
        return attr_name in self._attr_types


    def __getitem__(self, attr_name):
        if not attr_name in self:
            raise KeyError
        
        return self.__dict__[attr_name]
    

    def __repr__(self):
        data = []
        for attr_name, attr_type in self._attr_types.items():
            value = self.__dict__[attr_name]
            value_str = f"'{value}'" if type(value)==str else str(value)
            data.append(f'{attr_name}={value_str}')

        return f'Struct({", ".join(data)})'

    def print(self):
        print(self.__class__.__name__)
        for attr_name, attr_type in self._attr_types.items():
            type_name = ':'+attr_type.__name__ if attr_type else ''
            value = self.__dict__[attr_name]
            value_str = f"'{value}'" if type(value)==str else str(value)
            print(f'    {attr_name}{type_name} = {value_str}')


    def __setattr__(self, __name: str, __value) -> None:
        if __name in self._attr_types:
            attr_type = self._attr_types[__name]
            if attr_type and type(__value)!=attr_type:
                if self._strict_type_check:
                    raise Exception(f'Значение "{__value}":{type(__value)} не соответствует типу атрибута {__name}: {attr_type}')

                try:
                    __value = m_utils.to_type(__value, attr_type)
                except TypeError:
                    raise TypeError(f'Ошибка типа {__name}: {attr_type} = {__value}: {type(__value)}')
        
        object.__setattr__(self, __name, __value)


    def _gather_attrs(self, *, default, ignore_default_type):
        '''
        Заполняет _attr_types атрибутами и их типами данных
        '''
        self._attr_types = {}
        self._attr_values = {}

        attrs = m_utils.get_all_class_attrs(self.__class__, exclude=Struct, _fake_init=True)
        for attr_name, attr_desc in attrs.items():
            if attr_name.startswith('_'): continue
            if callable(attr_desc): continue

            if type(attr_desc) == tuple:
                value = attr_desc[0]
                data_type = attr_desc[1]
            else:
                value = attr_desc
                data_type = None

            self._attr_types[attr_name] = data_type
            if default:
                if ignore_default_type:
                    self.__dict__[attr_name] = value
                else:
                    self.__setattr__(attr_name, value)
            else:
                self.__dict__[attr_name] = None

    def get_attr_types(self) -> dict:
        '''
        Возвращает типы атрибутов
        '''
        return self._attr_types


    def from_dict(self, data:dict, clear_absent=True):
        '''
        Присваивает значения атрибутам класса из словаря
            - data - словарь с данными или объект с методом to_dict()
            - clear_absent - очищать атрибуты отсутствующие в словаре-аргументе
        '''
        if type(data) == dict: pass
        elif hasattr(data, 'to_dict'): data = data.to_dict()
        else: raise Exception(f'Некорректный тип данных {type(data)}. Должен быть dict или объект с методом to_dict()')

        for attr_name, attr_type in self._attr_types.items():
            if not clear_absent and attr_name not in data: continue

            value = data.get(attr_name)
            if value != None:
                if attr_type != None:
                    value = m_utils.to_type(value, attr_type)

            self.__dict__[attr_name] = value

        return self
    

    def to_dict(self, inc_empty=True, simple=False, attrs:list=None) -> dict:
        '''
        Сохраняет значения атрибутов класса в словарь
            - inc_empty - включить в результат атрибуты со значением None
            - simple - преобразовывать объекты в str
        '''
        if attrs:
            props = args.get_props(self, 'attrs')
            attrs_list = props.keys()
        else:
            attrs_list = None


        res = {}
        for attr_name in self._attr_types.keys():
            value = self.__dict__[attr_name]
            if not inc_empty and value==None: continue
            if attrs_list and attr_name not in attrs_list: continue

            if simple and type(value) not in (None.__class__, bool, int, float, str, dict, list): 
                value = m_utils.to_type(value, str)

            res[attr_name] = value

        return res


    def get_attrs_dict(self) -> dict:
        '''
        Возвращает словарь атрибутов и их значений
        используется m_repo gather_attrs(class_fit=True, ...)
        '''
        return self.to_dict(inc_empty=True, simple=False)



def test():
    class A(Struct):
        a:str = '1'
        b:int = 2
        c:str = 'var c'
        d:float = 1

    v = A(default=True)
    d = (A(default=True).to_dict(attrs=[v.a, A.b, 'c']))
    print(d)
    print(v.to_dict())


if __name__ == '__main__':
    test()