'''
Node - класс для построения иерархических деревьев

Confluence: https://robolabmsk.atlassian.net/wiki/spaces/ADM/pages/220594303/class+TreeNamedNode

'''

import re
import yaml

import m_utils

class TreeException(Exception):
    pass

class TNN_NotApplicable(Exception):                         # Исключение генерируемое конструктором данных, если данные не применимы
    pass                                                    # и требуется загрузка данных по-умолчанию


import copy

class Node:                
    PATH_DELIMITER = '.'    # Разделитель имен нодов в маршруте

    def __init__(self, name='root', data=None, src_obj=None, node_type=None, protected=False, desc=''):
        self.__dict__['_name'] = None
        self.__dict__['_node_type'] = None
        self.__dict__['_children'] = None
        self.__dict__['_parent'] = None
        self.__dict__['_data'] = None
        self.__dict__['_protected'] = None 
        self.__dict__['_proto'] = None              
        self.__dict__['_proto_update'] = None       
        self.__dict__['_src_obj'] = None
        self.__dict__['_src_name'] = None           
        self.__dict__['_child_constructor'] = None  
        self.__dict__['_owner'] = None  
        self.__dict__['_desc'] = None  
        self.__dict__['_provided'] = None  
        self.__dict__['_inherit'] = None
        self.__dict__['_show_children'] = None

        self._name = name               # Имя ноды для маршрута вызова
        self._node_type = node_type     # Тип ноды для поиска нод по типу
        self._children = {}             # Слоарь нодов потомков
        self._parent = None             # Нода-родитель
        self._data = data               # Данные ноды
        self._protected = protected     # Флаг защиты от обновления существ.нодов
        self._proto = None              # Прототип с которго сделан клон
        self._proto_update = None       # Флаг обновления прототипа при добавлении
        self._src_obj = src_obj         # Объект из которого загружен метод, для передачи как self при вызове 
        self._src_name = None           # Имя исходного атрибута или метода
        self._child_constructor = None  # Конструктор для дочерних элементов
        self._owner = m_utils.CallerInfo().outer_func(exclude_modules=['m_tree', 'm_api', 'm_actions'], exclude_starts='m_')
        self._desc = desc               # Описание
        self._provided = []             # Список атрибутов и методов предоставляемых потомкам
        self._inherit = True
        self._show_children = True         # Флаг отображения потомков для методов three_str и print
        #
        #
    
    def provide(self, name):            # Предоставляет атрибут дочерним элементам
        if type(name) == list:
            self._provided.extend(name)
        else:
            self._provided.append(name)
        #
        #
    def provided(self, show_path=True, all=False): # Возвращает словарь наследуемых атрибутов
        res = {}
        if self._parent != None:
            res = self._parent.provided(show_path, all=all)

        for attr in self._provided:
            source = self.path() if show_path else self

            if all and attr in res:
                if type(res[attr]) != list: res[attr] = [res[attr]]
                res[attr].append(source)
            else:
                res.update({attr: source})

        return res


    def is_provided(self):
        return self._parent and self._name in self._parent._provided
        #

    def clone(self, protected=False, proto_update=False, deep_copy=True):   # Клонирование ноды
        res = Node(self._name, None, protected)
        res._proto = self
        res._proto_update = proto_update
        res._src_obj = self._src_obj
        res._src_name = self._src_name
        res._node_type = self._node_type
        if deep_copy:
            try:
                if type(self._data).__name__ not in ('module'):
                    res._data = copy.copy(self._data)
                else:
                    res._data = self._data
            except Exception as e:
                raise Exception(f'Ошибка копирования данных {self._data} для элемента {self._name}\n{e} {e.args}')

        if hasattr(res._data, '_node'):
            res._data._node = res

        if deep_copy:
            for k,v in self._children.items():
                res.add(k, v.clone(protected=protected, proto_update=proto_update))
        else:
            res._children = self._children

        return res
        #
        #
    def __repr__(self):                                     # Описание ноды в виде строки
        if self._data != None:
            return f'{self._desc} Node({self._name} : {self._data})'
        else:
            return f'{self._desc} Node({self._name})'
        #
        #
    def __contains__(self, name):                           # Реализация оператор in
        not_found = '<-- NOT FOUND -->'
        value = self.get_attr(name, default=not_found)
        return value != not_found
        #
        #
    def _get_data(self):
        if not callable(self._data) and self._src_obj:      # Реализация ссылки на атрибут класса загруженного через .from_obj
            return self._src_obj.__dict__[self._src_name]
        else:
            return self.__dict__['_data']
        #
        #
    def get_data(self, path, default):                      # Возвращает данные нода по маршруту от текущей
        node = self.find(path)
        return node._data if node!=None else default
        #
        #
    def inherit_attr(self, name, not_found_key='$<ATTR NOT FOUND>'):
        if name in self._provided:
            return self.get_attr(name)
        elif self._parent != None:
            return self._parent.inherit_attr(name, not_found_key)
        else:
            return not_found_key
        #
        #
    def get_attr(self, name, default=None, inherit=True):   # Возвращает атрибут по имени
        '''
        Возвращает атрибут по его имени. Атрибутами являются (по убыванию приоритета):
        - все атрибуты и методы класса ThreeNamedNode
        - все дочерние элементы
        - атрибуты и методы свойста _data
        - атрибуты родительских элементов предоставленные через метод provide
        '''
        if type(name) != str:
            raise Exception(f'Имя атрибута должено быть строкой: name="{name}"')

        if name == 'data':
            return self._get_data()

        elif name == '_obj':
            return self

        elif type(self._data)==dict and name in self._data:
            return self._data[name]

        elif name in self.__dict__['_children']:            # Ищем среди дочерних нодов
            return  self.__dict__['_children'][name]
        
        elif hasattr(self._data, name):                     # Ищем среди методов и атрибутов данных ноды
            return getattr(self._data, name)

        elif name in self.__dict__:                         # Ищем среди методов и свойств класса
            return self.__dict__[name]        

        elif inherit:
            not_found_key = '$<ATTR NOT FOUND>'
            res = self.inherit_attr(name, not_found_key)
            if res != not_found_key:
                if type(res)==Node and callable(res._data):
                    res = res._data

                if callable(res) and type(res)!=Node:
                    def wrapper(*args, **kwargs): 
                        return res(*args, _path=self, **kwargs)
                    return wrapper
                return res

        elif hasattr(name, self):
            return getattr(self, name)

        return default
            #
            #
    def __getattribute__(self, name):                   # Для возможности перекрытия атрибутами _data атрибутов из self
        if name not in ('__dict__','_data'):      # Если обращаемся не к магическим атрибутам
            children = object.__getattribute__(self, '__dict__')['_children']
            if name in children:
                return children[name]

            data = object.__getattribute__(self, '__dict__')['_data']
            if hasattr(data, name):
                return getattr(data, name)
        
        return object.__getattribute__(self, name)
            #
            #
    def __getattr__(self, name):
        not_found_key = '<ATTR NOT FOUND>'

        res = self.get_attr(name, default=not_found_key, inherit=self._inherit) 
       
        if res != not_found_key:
            return res
        else:
            raise AttributeError(f'Node {self.path()}: не найден потомок {name}. Данные: {type(self._data)}:{self._data}')
            #
            #
    def __setattr__(self, name, value):
        if name in self.__dict__:
            self.__dict__[name] = value
        else:
            self.set(name, value)
        #
        #
    def __len__(self):
        if type(self._data) in (list, dict, str):
            return len(self._data)
        else:
            return 1
        #
        #
    def __getitem__(self, key):                             # Реализация обращения через индекс
        try:
            return self._data[key]
        except:
            pass

        if type(self._data) == list:
            return self._data[key]

        elif type(self._data) == dict and key in self._data:
            return self._data[key]

        elif type(key) == str:
            return self.find(key)

        elif type(key) == slice and type(key.start)==str and key.stop==None:
            res = {}
            lst = self.list_all()
            for k, v in lst.items():
                if k.startswith(key.start):
                    res[k] = v
                    
            return res
            
        elif type(key) == slice and key.start==None and type(key.stop)==str:
            res = {}
            lst = self.list_all()
            for k, v in lst.items():
                if k.endswith(key.stop):
                    res[k] = v
                    
            return res

        else:
            raise TreeException(f'Неверный ключ {key}')
        #
        #
    def __setitem__(self, key, value):
        if self._data != None and hasattr(self._data, '__setitem__'):
            self._data[key] = value
        else:
            Exception(f'Объект {self._data} не поддерживает присваивание элементу {self}')
        #
        #
    def __call__(self, *args, **kwargs):                    # Вызов функции сохраненной в атрибуте _data
        if callable(self._data):
            func = self._data
            if self._src_obj:
                return func(self._src_obj, *args, **kwargs)
            else:
                return func(*args, **kwargs)
        else:
            raise TreeException(f'Node {self._name}: данные {self._data} : {type(self._data)} не могут быть вызваны')
        #
        #
    def path(self, delimiter=None, relative=None):          # Строка маршрута к ноде
        '''
        Возвращает маршрут к ноде в дереве
        '''
        if not delimiter: delimiter = Node.PATH_DELIMITER

        nodes = []
        cur_node = self
        while cur_node:
            nodes.append(cur_node._name)
            cur_node = cur_node._parent

        nodes.reverse()
        path = delimiter.join(nodes)

        if relative:
            base_path = relative.path()
            if path.startswith(base_path):
                path = path[len(base_path)+1:]

        return path
        #
        #
    def is_leaf(self):                                      # Признак - нода содержит данные и нет потомков
        return len(self._children)==0 and self._data != None
        #
        #

    @property
    def root(self):                                         # Самый верхний элемент дерева
        node = self
        while node._parent:
            node = node._parent
        return node
        #
        #
    def find(self, path, ensure=False, is_assert=False):    # Поиск ноды по адресу
        if not path: return self

        if Node.PATH_DELIMITER in path:
            cur_name, rest_path = path.split(Node.PATH_DELIMITER, 1)
        else:
            cur_name = path
            rest_path = None

        if cur_name == self._name:
            node = self
        else:
            if cur_name not in self._children:
                if ensure:
                    node = self.add(cur_name)
                else:
                    if is_assert:
                        raise Exception(f'Не найден маршрут {path}')
                    return None
            else:
                node = self._children[cur_name]
            
        return node.find(rest_path, ensure)
        #
        #
    def find_up(self, node_name):                           # Поиск указанного элемента вверху
        cur_node = self
        while cur_node and cur_node._name != node_name:
            cur_node = cur_node._parent

        return cur_node
        #
        #
    def set(self, name, data=None, src_obj=None, ensure=False, node_type=None, show_children=True): # Измнение/добавление потомка
        # --- Определение параметров
        if Node.PATH_DELIMITER in name:
            last_node = self.find(name, ensure=ensure)
            if last_node == None:
                raise TreeException(f'Маршрут {name} не найден для {self.path()}')
            parent_node = last_node._parent
            child_name = last_node._name
        else:
            parent_node = self
            child_name = name

        # --- Проверка параметров
        if child_name in parent_node.__dict__:
            raise TreeException(f'Нельзя переопределить атрибут класса {name}')

        is_exist = child_name in parent_node._children
        if is_exist and self._protected:
            raise TreeException(f'Элемент {name} защищен от изменения')

        # --- Создание новой ноды
        if type(data) == Node:
            node = data.clone()
            node._name = child_name
        else:
            node = Node(child_name, data, src_obj, node_type=node_type)
            if hasattr(data, '_node'):
                data._node = node

        node._parent = parent_node
        node._show_children = show_children
        parent_node.__dict__['_children'][child_name] = node

        # --- Применение конструктора для данных
        if self._child_constructor:
            node._data = self._child_constructor(node, data)

        # --- Создание нового элемента на прототипе
        if self._proto_update and self._proto and name not in self._proto._children:
            self._proto.add(name, node)

        return node
        #
        #
    def add(self, node_name:str, data:any=None, src_obj:any=None, node_type:str=None, show_children=True):           # Добавление потомка
        if node_name in self._children:
            raise TreeException(f'Элемент с именем {node_name} уже сущестует {self._children[node_name]} src={self._children[node_name]._owner}')

        return self.set(node_name, data, src_obj, ensure=True, node_type=node_type, show_children=show_children)
        #
        #
    def find_down(self, node_name, node_type=None, instance=None, all_levels=False, single=True):  # Поиск потомков
        if all_levels:
            res = []

            nodes = self.find_down(node_name, node_type=node_type, instance=instance, all_levels=False, single=False)
            if nodes:
                res.extend(nodes)

            for child in self._children.values():
                nodes = child.find_down(node_name, node_type=node_type, instance=instance, all_levels=True, single=False)
                res.extend(nodes)

        else:
            if node_name == None:
                res = [node for node in self._children.values() if (not node_type or node._node_type==node_type) and (not instance or instance(node._data, instance))]

            else:
                res = self._children.get(node_name)
                if res and (not node_type or res._node_type == node_type) and (not instance or isinstance(res._data, instance)):
                    res = [res]
                else:
                    res = []

        if single :
            if len(res) > 1:
                nodes_list = [node.path() for node in res]
                raise Exception(f'По пути {self.path()} найдено несколько элементов с именем {nodes_list}')

            res = res[0] if res else None

        return res
        

    def child_rename(self, node_name, new_name):            # Переименование одной ноды
        child = self._children[node_name]
        child._name = new_name                  # Переименовываем ноду
        self._children[new_name] = self._children[node_name]
        self.child_delete(node_name)

        return self
        #
        #
    def child_map(self, map):                               # Переименовывает дочерние атрибуты по словарю
        for k,v in map.items():
            if k in self._children:
                self.child_rename(k, v)

        return self
        #
        #
    def child_delete(self, name, silent=False):             # Удаление потомка
        if Node.PATH_DELIMITER in name:
            last_node = self.find(name)
            if last_node != None:
                parent_node = last_node._parent
                name = last_node._name

            else:
                parent_node = self
                name = None

        else:
            parent_node = self

        if name in parent_node._children:
            del parent_node._children[name]

        else:
            if not silent:
                raise TreeException(f'Элемент {name} не найден')
        #
        #
    def child_data_set_attr(self, values, recursive=True):
        for child in self._children.values():
            for attr, value in values.items():
                if hasattr(child._data, attr):
                    setattr(child._data, attr, value)

            if recursive:
                child.child_data_set_attr(values, True)

    def extend(self, data):                                 # Добавляет потомков из ноды или dict
        if type(data) == dict:
            for k, v in data.items():
                self.set(k, v)

        elif type(data) == type(self):
            for k, v in data._children.items(): 
                self.set(k, v)
        #
        #

    # --- Работа со списком нодов
    def set_children_data(self, attr, value):
        for child in self._children.values():
            if hasattr(child._data, attr):
                child._data.__dict__[attr] = value
            child.set_children_data(attr, value)
        #
        #
    def list(self, path=None, level=1):                     # Список потомков до заданного уровня
        if path:
            node = self.find(path)
            res = node.list(None, level)
        else:
            res = {}
            for child in self._children.values():
                if child.is_leaf():
                    res[child.path()] = child._data
                else:
                    res[child.path()] = child
                    if level != 1:
                        new_level = level-1 if level!=None else None
                        res.update(child.list(level=new_level))

        return res
        #
        #
    def list_all(self, path=None):                          # Полный список потомков сл всех уровней
        return self.list(path, level=None)
        #
        #
    def list_up(self, upper_node=None):                     # Двигается от текущей ноды 
        '''
        Возвращает список нод от верхней к нижней
        upper_node - нода до которой прасматривается дерево, не входит в результат
        '''
        res = []
        cur_node = self
        while cur_node and cur_node != upper_node:
            res.append(cur_node)
            cur_node = cur_node._parent

        res.reverse()

        return res
        #
        #
    def children_find_up(self, upper_node=None, node_name=None, node_type=None, instance=None):    # Возвращает список потомков по имени, типу
        repo_nodes = self.list_up(upper_node)

        data_nodes = []
        nodes_data = []
        for node in repo_nodes:
            for child in node._children.values():           # Перебираем дочерние ноды
                if type(node_type)==str and not child._node_type == node_type: continue
                if type(node_type)==list and not child._node_type in node_type: continue
                if node_name and not child._name == node_name: continue
                if instance and not isinstance(child._data, instance): continue
                
                data_nodes.append(child)
                nodes_data.append(child._data)

        return data_nodes, nodes_data
        #
        #
    def children_find_down(self, node_name=None, node_type=None, instance=None):
        data_nodes = []
        nodes_data = []
        for child in self._children.values():
            while True:
                if node_type and not child._node_type == node_type: break
                if node_name and not child._name == node_name: break 
                if instance and not isinstance(child._data, instance): break

                data_nodes.append(child)
                nodes_data.append(child._data)
                break

            add_nodes, add_data = child.children_find_down(node_name=node_name, node_type=node_type, instance=instance)
            if add_nodes:
                data_nodes.extend(add_nodes)
                nodes_data.extend(add_data)

        return data_nodes, nodes_data
        #
        #
    def gather_data(self, path=None, node_name=None, node_type=None, instance=None):      # Собирает значения заданного атрибута в список или словарь
        if path:
            node = self.find(path)
        else:
            node = self

        if not node:
            return None

        data_nodes, nodes_data = node.children_find_up(node_name=node_name, node_type=node_type, instance=instance)

        res = None
        for node, data in zip(data_nodes, nodes_data):

            if res == None:
                if type(data) in (dict, list):
                    res = data
                else:
                    res = [data]

            elif type(res)==list:
                if type(data) == list:
                    res.extend(data)
                else:
                    res.append(data)

            elif type(res)==dict:
                if type(data) == dict:
                    m_utils.deep_update(res, data)
                elif type(data) == list:
                    for item in data:
                        if type(item)==dict:
                            m_utils.deep_update(res, item)
                        else:
                            print(f'gather_attrs: Неправильный тип данных {type(data)} для {node.path()} ожидается dict')

            #print(f'res={res} node={node} data={type(data)} : {data}')

        return res
        #
        #
    def call(self, path, *args, **kwargs):
        lst = self[path:]
        called = []
        for node_path, node_data in lst.items():
            if type(node_data)==type(lambda x:x):
                ret_val = node_data(*args, **kwargs)
                called.append({'path' : node_path, 'return':ret_val})

        return called
        #
        #
    
    def _node_sort_key(self, x):
        if x._children:
            res = '1'
        else:
            res = '0'
        return res + x._name
        #
        #
    def tree_str(self, indent='', data_max_len=80):         # Вывод структуры в виде дерева в строку
        s = ''
        if indent: s += indent[:-4] + ' +--'
        s += f'{self._name}:{self._node_type}'
        if self.is_provided(): s += ' PROVIDED' 

        # Формирование строки для отображения текущей ноды
        data_str = ''
        if self._data != None:
            if str(type(self._data)) == "<class 'module'>":
                data_str = self._data.__file__
            else:
                data_str = f'{self._data}'

            if data_str and data_str[0] == '<':     # Заменяем стандартное отображение класса
                try:
                    data_str = f'{self._data.__module__}.{self._data.__class__.__name__}'
                except:
                    data_str = ''

        if not 'src=' in data_str:              # добавляем источник
            data_str += f' src={self._owner[0]}'

        if len(data_str) > data_max_len: data_str = data_str[:data_max_len-3] + '...'
        s += f' = {data_str}'

        out = s + '\n'

        if len(self._children) and self._show_children:
            children_list = list(self._children.values())
            children_list.sort(key=self._node_sort_key)

            for child in children_list:
                is_last = child == children_list[-1]
                if is_last:
                    add_indent = '    '
                else:
                    add_indent = ' |  '
                blank_printed, child_out = child.tree_str(indent + add_indent, data_max_len)
                out += child_out 

            if not blank_printed:
                out += indent + '\n'
            
            res = True
        else:
            res = False

        if indent!='':
            return res, out
        else:
            return None, out
        #
        #
    def print(self, data_max_len=80, out=None):            # Вывод структуры в виде дерева на консоль
        _, _out = self.tree_str(data_max_len=data_max_len)
        
        if out: 
            out(_out)
            out(f'{self.root.lib.debug.debug_print}')
        elif isinstance(self.root, Node) and self.root.find('lib.debug'):
            self.root.lib.debug.debug_print(_out)
        else:
            print(_out)

    def get_globals(self):
        return globals()

    # --- Выгрузка/загрузка
    def to_dict(self, index_fields=None, exclude=[]):       # Вывод структуры в dict
        path = self.path()
        if index_fields and path in index_fields:
            field_name = index_fields[path]
            res = []
            for k, v in self._children.items():
                if k not in exclude:
                    rec = {field_name : k }
                    rec.update( v.to_dict(index_fields) )
                    res.append(rec)

        else:
            res = {}
            for k,v in self._children.items():
                if k not in exclude:
                    res[k] = v.to_dict(index_fields)

        if len(res) == 0: 
            res = self._data

        return res
        #
        #
    def _child_from_dict(self, child_name, child_data, *args):
        #print(f'_child_from_dict: {child_name}:{type(child_name)} = {child_data}')
        node = self.add(child_name)

        try:
            if self._child_constructor:
                node._data = self._child_constructor(node, child_data)
            else:
                raise TNN_NotApplicable()

        except TNN_NotApplicable as e:
            node.from_dict(child_data, *args)
        #
        #
    def from_dict(self, data, indexes=[], containers=[]):   # Загрузка структуры из dict
        if self._name in containers:
            self._data = data

        elif type(data)==dict:            
            for k,v in data.items():
                self._child_from_dict(k, v, indexes, containers)

        elif type(data)==list and len(data) and type(data[0])==dict:
            for idx, item in enumerate(data):
                key_field = None
                node_name = idx

                for k in item.keys():       # Ищем атрибут, который упомянут в indexes
                    if k in indexes:
                        key_field = k
                        node_name = item[k]
                
                if type(node_name) == int: node_name = f'_{node_name}'
                attrs = dict(item)
                if key_field:
                    del attrs[key_field]

                self._child_from_dict(node_name, attrs, indexes, containers)

        else:
            self._data = data

        return self
        #
        #
    def from_yaml(self, data, indexes=[]):
        data = yaml.load(data, Loader=yaml.FullLoader)
        self.from_dict(data, indexes=indexes)
        #
        #
    def from_obj(self, obj, attrs=[], methods=[], map={}):
        source = obj.__dict__
        source.update(obj.__class__.__dict__)

        for name, data in source.items():
            if callable(data):
                if methods == None: continue
                if type(methods)==list and not (methods == [] or name in methods): continue
                if type(methods)==str and not re.search(methods, name): continue
                if type(methods)!=str and name.startswith('_'): continue
                if len(map) and name not in map and methods==[]: continue

            else:
                if attrs == None: continue
                if type(attrs)==list and not (attrs == [] or name in attrs): continue
                if type(attrs)==str and not re.search(attrs, name): continue
                if type(attrs)!=str and name.startswith('_'): continue
                if len(map) and name not in map and attrs==[]: continue


            if name in map: 
                node_name = map[name]
            else:
                node_name = name

            node = self.add(node_name, data)
            node._src_obj = obj
            node._src_name = name

        return self
        #
        #
            

