'''
Модуль обработки событий
 - регистрирует триггеры в структуре репозитория или в ветке заданной в MODULE_PATH
 - запускает триггеры при генерации события методом api.triggers.emit(...)


'''
import pdb

import m_utils
import m_api

MODULE_PATH = 'api.triggers'

api = None
_debug = False

class TriggerException(m_api.ApiException):
    pass

class TriggerWarning(TriggerException):
    pass

class TriggerExceptionNotAllowed(TriggerException): pass
class TriggerExceptionIncorrectData(TriggerException): pass
class TriggerExceptionIncorrectLogin(TriggerException): pass

def get_caller():
    stack = m_utils.CallerInfo().outer_func(exclude_modules=['m_triggers'])
    if stack:
        return stack[0]
    else:
        return '<undefined>'


class Trigger:      # Описание триггера
    def __init__(self, trigger_func, kvargs=None):
        self.trigger_func = trigger_func
        self.kvargs = kvargs or {}
        self.source = get_caller()
        #
        #
    def __repr__(self):
        return f'Trigger({self.trigger_func.__module__}.{self.trigger_func.__name__})'
        #
        #
    def run(self, *args, **kvargs):
        kvargs.update(self.kvargs)
        self.trigger_func(*args, **kvargs)


class TriggersApi(m_api.ModuleApi):
    def __init__(self):
        super().__init__()
        #
        #
    def add_event(self, event_name):                                    # Регистрация события
        caller = get_caller()

        self._node.set(event_name, [])

        self.log(f'Triggers.add_event: added event {event_name} from {caller}')
        #
        #
    def add_trigger(self, event_name, path, trigger_func, **kwargs):    # Регистрация триггера
        '''
        Регистрация триггера.

            event_name   - имя события, должно быть зарегистрировано методом add_event
            path         - маршрут к объекту в репозитории для которого вызывается событие, обычно класс объекта
                           None для системных событий (не связанных с действиями над объектами данных)
            trigger_func - функция реализующая триггер
            **kwargs     - именованные аргументы передаваемые триггеру при вызове
        '''
        if event_name not in self._node._children:
            msg = f'Событие {event_name} должно быть зарегистрировано до регистрации триггера'
            if _debug:
                print(msg)
                pdb.set_trace()
            raise m_api.ApiException(msg)

        trigger = Trigger(trigger_func, kwargs)

        if path:
            path += '.' + event_name
            level_node = api.repo.find(path, ensure=True)
        else:
            path = MODULE_PATH + '.' + event_name
            level_node = api.find(path, ensure=True)
 
        if level_node._data == None:
            level_node._data = []

        level_node._data.append(trigger)
        if _debug: print(f'Triggers.add_trigger: added trigger {trigger} and {level_node}')
        #
        #
    def emit(self, event_name, class_path='', context=None, try_only=False, **kwargs): # Запус обработки события
        '''
        Запуск обработки события. Вызывает триггеры соответствующие событию и маршруту объекта.

            event_name - имя события, должно быть зарегистрировано методом add_event
            class_path - путь к классу объекта данных в репозитории, с которым связано событие
            context    - контекст вызова, зависит от имени события, определяется в модуле генерирующем событие
            **kwargs   - именованные агрументы передаваемые вызываемым функциям
        '''
        if event_name == 'DS_BEFORE_SELECT': return

        if event_name not in self._node._children:
            raise m_api.ApiException(f'Событие {event_name} должно быть зарегистрировано до генерации события')

        # --- Определение списка триггеров 
        tg_list = api.get_data(f'{MODULE_PATH}.{event_name}', [])

        _class_path = class_path
        if type(_class_path)==str:
            _class_path = _class_path.split('.')[0]

        repo_node = api.find('repo')
        
        if repo_node:
            tg_type_list = repo_node.gather_data(_class_path, node_name=event_name)
            if tg_type_list:
                tg_list.extend( tg_type_list )

        if api.get_config('API2_DEBUG', False) and False:
            if event_name not in ('CUSTOM_FILE_INIT'):
                api.log.msg(f'Trigger: event_name={event_name}, class_path={class_path}, tg_list={tg_list}')

        # --- Вызов триггеров
        if not tg_list: return

        for tg_obj in tg_list:
            if not try_only:
                try:
                    tg_obj.run(context=context, **kwargs)

                except TriggerWarning as e:
                    pass

                except Exception as e:
                    api.log.error(f'Exception: {e}')
                    raise e


def init(_api, debug=False):
    global api
    api = _api

    global _debug
    _debug = debug

    api.add_module_api(MODULE_PATH, TriggersApi())
