/*
Экспортируемые функции доступны в пользовательских копонентах (page, modal)
для подключения добавить в структуру Vue-компонента: inject:['api2']

Использование внутри компонента:
    this.api2.<имя функции>
*/

async function sleep(ms) {
    return new Promise(r => setTimeout(r, ms))
}


function ts_str() {
    return (new Date(Date.now())).toLocaleString()
}

function pretty(data, capt='', prefix='', addTag=true) {
    let res = prefix
    if (capt.length)
        res += String(capt) + ' : '
    
    if (Array.isArray(data)) {
        res += '[\n'
        for (let item of data) {
            res += pretty(item, '', prefix + '    ', false) + ',\n'
        }
        res += prefix + ']'
    } else if (data && typeof(data)=='object') {
        res += '{\n'
        for (let [k,v] of Object.entries(data)) {
            res += pretty(v, k, prefix+'    ', false) + ',\n'
        }
        res += prefix + '}'
    } else if (typeof(data)=='string') {
        res += '"' + data + '"'

    }  else
        res += String(data)

    if (addTag) 
        res = '<PRE>' + res + '</PRE>'
        //  style="max-height:30em;overflow:hidden;"
    
    return res
}


//===========================================================================
//   Работа с файлами
//===========================================================================

let files_api = {
    size_str : file_size_str,
    upload : file_upload,
}


function file_size_str(size) {
    if (size===null) return ''

    if (typeof(size) != 'number')
        size = size.size                // Извлекаем размер из объекта file

    return `${Math.round(size / 1024 / 1024 * 10)/10} Мб`
}


// Передает файл на endpoint /api/files, реализованный в модулей m_files.py
async function file_upload(file, host=null, progress_func=null) {
    let is_active = true
    let result = null

    if (!host) host=window.location.origin

    let url = new URL('/api/files?action=upload', host).href
    let xhr = new XMLHttpRequest();

    xhr.upload.onprogress = function(event) {
        //console.log('uploadFile2: progress', event)
        if (progress_func)
            progress_func(event)
    }

    xhr.onloadend = function() {
        if (xhr.status != 200) 
            result = `Ошибка ${xhr.status}: ${xhr.statusText}`
        else
            result = JSON.parse(xhr.response)
        
        is_active = false
    }

    let formData = new FormData();
    formData.append(file.name, file);

    xhr.open('POST', url);    
    xhr.send(formData);
    
    while (is_active) 
        await sleep(100)
 
    return result
}


function make_url(url, params) {
    const ret = [];
    for (let d in params)
      ret.push(encodeURIComponent(d) + '=' + encodeURIComponent(params[d]));
    let params_str = ret.join('&')

    if (params_str)
        url += '?' + params_str 
    return url
}

async function getJson(rest_url, params) {
    let url = make_url(rest_url, params)

    let response = await fetch(url)
    return await response.json()
}

async function postJson(rest_url, params) {
    let response = await fetch(rest_url, {
        method:'POST', 
        headers: { 'Content-Type': 'application/json;charset=utf-8' },
        body: JSON.stringify(params)
    })

    return await response.json()
}

//=============================================================================
//   Actions API
//=============================================================================
const actions_url = '/api/actions'


async function action_exec(action_name, repo_path, params) {
    let url = make_url(actions_url, { action : 'exec',  action_name : action_name, path: repo_path } )

    return await postJson(url, {context:params})
}

async function action_progress(proc_id) {
    return await getJson(actions_url, {action:'progress', proc_id})
}

async function action_cancel(proc_id) {
    return await getJson(actions_url, {action:'cancel', proc_id})
}

async function action_result(proc_id) {
    let res = null
    while (true) {
        res = await getJson(actions_url, {action:'result', proc_id})
        if (res && res.status !='active')
            break
    }

    return res
}

async function action_data_ack(proc_id, data_len) {
    return await getJson(actions_url, {action:'data_ack', proc_id, 'data_len':data_len})
}

function action_get_exec_url(action_name, path) {         // Возвращает маршрут для вызова
    let url = `/api/actions?action_name=${action_name}&action=exec`
    if (path)
        url += `&path=${path}`
    return {url}
}


function get_url_param(param_name) {
    let uri = window.location.search.substring(1)
    let params = new URLSearchParams(uri)
    return params.get(param_name)
}

let actions_api = {
    get_url_param,
    get_exec_url : action_get_exec_url,
    exec : action_exec,
    progress : action_progress,
    data_ack : action_data_ack,
    cancel : action_cancel,
    result : action_result,

    get : getJson, 
    post : postJson,
}

//=============================================================================
//   Components
//=============================================================================
export default {
    sleep,
    ts_str,
    pretty,
    action: actions_api,

    files: files_api,

    components: await import('/api/api2/components')
}
