'''
Модуль для извлечения имен переменных передаваемых в качестве параметров
'''

import inspect


def get_args(level=1):
    '''
    Извлекает имена всех аргументов переданных функции
    '''
    cur_frame = inspect.currentframe()
    call_name = inspect.getouterframes(cur_frame)[level-1].function + '('

    context_size = 0

    while True:
        # Выделяем строку вызова
        frame = inspect.getouterframes(cur_frame, context=1+context_size*2)[level]
        src = ' '.join(frame.code_context[context_size:])

        # Удаляем пробельные символы
        while ' ' in src: src = src.replace(' ', '')
        while '\n' in src: src = src.replace('\n', '')

        # Выделяем параметры вызывемой функции
        if call_name in src:
            src = '(' + src.split(call_name)[-1]        
            if src.count('(') <= src.count(')'): break
        context_size += 1

        if context_size>20:
            raise Exception(f'Ошибка определения параметров. Не найден вызов функции {call_name} в коде {src}')

    src0 = src

    # Находим конец строки параметров
    while src.count('(') < src.count(')'):
        end_pos = src.rindex(')', src.rindex(')')-1)
        src = src[:end_pos]

    # Выделяем список параметров
    try:
        src = src[src.find('(') + 1 : src.rindex(')')]
    except Exception as e:
        print(e, f'\n    src={src}\n    src0={src0}')
        raise e


    # Парсим параметры
    cur_pos = 0
    end_pos = 0

    args = []
    kwargs = {}

    while end_pos < len(src):
        end_pos = src.find(',', end_pos+1)
        if end_pos<0: end_pos = len(src)
        param = src[cur_pos: end_pos]

        if param.count('(') != param.count(')'): continue
        if param.count('[') != param.count(']'): continue
        if param.count('{') != param.count('}'): continue

        words = param.split('=')
        if len(words)>1:
            param_name = words[0]
            param_value = '='.join(words[1:])
        else:
            param_name = None
            param_value = param

        if param_value[0] == '[' and param_value[-1]==']':
            param_value = [value for value in param_value[1:-1].split(',') if value]

        if param_name:
            kwargs[param_name] = param_value
        else:
            args.append(param_value)

        cur_pos = end_pos + 1

    #print(f'{args=}\n{kwargs=}')
    
    return args, kwargs


def get_props(data, arg_name, level=2):
    '''
    Извлекает из data атрибуты перечисленные в параметре вызова функции с именем arg_name
    '''
    args, kwargs = get_args(level=level+1)

    cur_frame = inspect.currentframe()
    src_frame = inspect.getouterframes(cur_frame)[level].frame
    frame_vars = inspect.getargvalues(src_frame)

    # Извлекаем список всех базовых классов для data
    class_list = [cls.__name__ for cls in inspect.getmro(data.__class__)]

    res = {}
    for name in kwargs.get(arg_name, []):
        if name[0] in ('"',"'"): 
            class_name = data.__class__.__name__
            prop_name = name[1:-1]

        else:
            class_name = name.split('.')[0]
            prop_name = name.split('.')[-1]

            # Если указан экземпляр, заменяем его на класс
            var_value = frame_vars.locals.get(class_name)
            if var_value.__class__ != type:
                class_name = var_value.__class__.__name__

            # Проверяем что данные являются потомком класса задающего список атрибутов
            if class_name not in class_list:
                raise Exception(f'Данные типа {data.__class__.__name__} не соответствуют параметру {name}')
        
        res[prop_name] = data.__getattribute__(prop_name)

    return res

#=============================================================================================
#   Примеры использования
#=============================================================================================

def log(var):
    '''
    Пример функции выводящей имя переменной и ее значение
    '''
    args, _ = get_args(level=2)
    var_name = args[0]
    if var_name[0] not in ('"', "'"):
        print(f'{var_name} = {var}')
    else:
        print(var)


def func(data, props):
    '''
    Пример функции, которая выделяет из data атрибуты перечисленные в props
    '''
    return get_props(data, 'props')


#=============================================================================================
#   Тесты
#=============================================================================================

def test():
    class A:
        a:str = '1'
        b:int = 2

        def test(self):
            pass

    print(inspect.getsource(A))


    data = A()
    data.a = 'str'

    # Извлечение из data перечисленных атрибутов
    props = func(data, props=[
            A.a, 
            data.b
        ])

    print(props)

    log(A.a)
    log('test')


if __name__ == '__main__':
    test()