'''
Crone - Модуль для периодического вызова задач

Регистрация задачи:
    Crone.add(CroneTast(...))

Запуск планировщика:
    Crone.start()

Остановка планировщика:
    Crone.stop()
'''

import threading, datetime, time
import m_utils

class CroneTask:
    def __init__(self, period_sec, func, *args, start_now=False, **kwargs):
        self.next_exec_ts = datetime.datetime.now() - datetime.timedelta(seconds=-1 if start_now else period_sec) 
        self.period = period_sec
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.error = None


    def __repr__(self):
        return f'CroneTask({self.func.__name__}, period={self.period}, next_exec_ts={self.next_exec_ts})'


    def _exec(self) -> bool:
        '''
        Выполняет задачу
        '''
        now_ts = datetime.datetime.now()

        try:
            self.func(*self.args, crone_task=self, **self.kwargs)
            self.next_exec_ts = now_ts + datetime.timedelta(seconds=self.period) if self.period else None
            self.error = None
            return True
        
        except Exception as e:
            self.error = m_utils.exception_text(e)
            return False


    def set_period(self, period):
        '''
        Устанавилвает период для вызовов задачи
        '''
        self.period = period


    def stop(self):
        '''
        Прекращает вызов задачи
        '''
        self.set_period(0)


class Crone:
    th = None
    th_exit = False
    tasks = []
    errors = {}
    debug = False

    def _crone_func():
        '''
        Основной цикл для запуска задач внутри потока
        '''
        Crone.th_exit = False
        while not Crone.th_exit:
            now_ts =datetime.datetime.now()

            for item in list(Crone.tasks):
                task: CroneTask = item
                try:
                    if task.next_exec_ts!=None and task.next_exec_ts <= now_ts:
                        if Crone.debug: print(f'{m_utils.datetime_str()} Task {task}', flush=True)
                        task._exec()

                    if task.period == 0:
                        Crone.tasks.remove(task)

                except Exception as e:
                    err_msg = m_utils.exception_text(e)
                    task.error = err_msg

                Crone.errors[task] = task.error
                if task.error: print(task.error, flush=True)

            time.sleep(1)


    def start():
        '''
        Запустить Crone
        '''
        if Crone.th != None: return
        Crone.th = threading.Thread(target=Crone._crone_func)
        Crone.th.start()

        if Crone.debug:
            print(f'{m_utils.datetime_str()} Crone stated {len(Crone.tasks)} tasks')


    def stop():
        '''
        Остановить Crone
        '''
        Crone.th_exit = True
        Crone.th.join()
        Crone.th = None


    def add(task: CroneTask):
        '''
        Добавить задачу для периодического выполнения
        '''
        Crone.tasks.append(task)
