'''
FileObserver - модуль для наблюдения за изменением файлов в каталоге и подкаталогах
'''
import os, re, datetime
import m_utils


class FileObserver:
    debug = False

    # Список каталогов для наблюдения
    works = set()   # элементы вида (<dir_path>, <func>)

    # Список задач для отработки изменений в файле
    tasks = {}      # <task_id> : set( <file_path>, ...)

    # Состояния файлов, которые уже обработаны
    hist = {}       # <task_id> : { <file_path> : <file_info> }


    def get_file_info(file_path):
        '''
        Извлекает информацию о файле: размер, время изменения.
        Статичный метод
        '''
        fs = os.stat(file_path)
        return (fs.st_size, fs.st_mtime)


    def _check_files(task_id, dir_path, re_mask):
        '''
        Проверка файлов в соответствии с зарегистрированной задачей
        '''
        if FileObserver.hist.get(task_id) == None: FileObserver.hist[task_id] = {}
        if FileObserver.tasks.get(task_id) == None: FileObserver.tasks[task_id] = set()

        hist = FileObserver.hist[task_id]
        tasks = FileObserver.tasks[task_id]

        for r,d,files in os.walk(dir_path):
            for file_name in files:
                if re_mask and not re.findall(re_mask, file_name): continue

                file_path = os.path.join(r, file_name)
                info = FileObserver.get_file_info(file_path)                
                if hist.get(file_path) != info:
                    tasks.add(file_path)

                #print(r, fn, fs.st_size, m_utils.datetime_str(fs.st_mtime))

        if FileObserver.debug:
            print(f'{m_utils.datetime_str()} FileObserver._check_files completed {tasks}', flush=True)

        return tasks
    

    def _commit(task_id, file_path, file_info):
        '''
        Отмечает что изменения в файле отработаны.
        Вызывается автоматически после успеного завершения функции обработки изменений
        '''
        FileObserver.hist[task_id][file_path] = file_info
        FileObserver.tasks[task_id].remove(file_path)


    def add_work(dir_path, func, re_mask=None):
        '''
        Регистрация функции вызываемой при изменении файлов 
        '''
        FileObserver.works.add( (dir_path, func, re_mask) )


    def check_all(*, crone_task=None, **kwargs):
        '''
        Проверка всех файлов в каталоге
        периодически вызывается из Crone
        '''
        start_ts = datetime.datetime.now()

        for work in FileObserver.works:
            dir_path = work[0]
            func = work[1]
            re_mask = work[2]

            task_id = work.__hash__()
            tasks = FileObserver._check_files(task_id, dir_path, re_mask)

            check_ts = datetime.datetime.now()
            process_qty = len(tasks)

            for file_path in list(tasks):
                file_info = FileObserver.get_file_info(file_path)
                try:
                    func(file_path)
                    FileObserver._commit(task_id, file_path, file_info)

                except Exception as e:
                    err = m_utils.exception_text(e)                
                    print(m_utils.datetime_str(), err, flush=True)
                    
            process_ts = datetime.datetime.now() - check_ts
            check_ts = check_ts - start_ts

            process_ts = int(process_ts.microseconds/1000)
            check_ts = int(check_ts.microseconds/1000)

            if FileObserver.debug:
                print(f'dir {work[0]}  check_time {check_ts}ms  process_qty {process_qty}  process_time {process_ts}ms', flush=True)
