import threading
import time

import psycopg2


class ConnectionPool:

    def __init__(self, max_conn, *args, **kwargs) -> None:
        self.max_conn = max_conn  # максимальное кол-во подключений к бд
        self.connections_count = 0  # кол-во активных подключений
        self.connections = {}  # поделючения. user -> [connection, ...]
        self.connection_in_use = {}
        self.connections_id_query = []
        self._lock = threading.Lock()
        self.time_delta = 1
        self._args = args
        self._kwargs = kwargs

    def _update_connections_id_query(self, connection_id):
        if connection_id in self.connections_id_query:
            self.connections_id_query.remove(connection_id)
        self.connections_id_query.append(connection_id)

    def _is_connection_in_use(self, connection_id):
        return connection_id in self.connection_in_use

    def _close_and_remove_connection(self, connection_id):
        for connection_list in self.connections.values():
            for connection in connection_list:
                if id(connection) == connection_id:
                    connection_list.remove(connection)
                    return

    def _close_oldest_free_connection(self):
        while True:
            for connection_id in self.connections_id_query:
                if connection_id not in self.connection_in_use:
                    self._close_and_remove_connection(connection_id)
                    return

    def _create_new_conn(self, user, password):
        conn = psycopg2.connect(user=user, password=password, *self._args, **self._kwargs)
        self.connections[user].append(conn)
        self.connection_in_use[id(conn)] = time.time()
        self._update_connections_id_query(id(conn))
        self.connections_count += 1
        return conn

    def _get_conn(self, user, password):

        if user not in self.connections:
            self.connections[user] = []

        if not self.connections[user]:

            if self.connections_count > self.max_conn:
                self._close_oldest_free_connection()

            return self._create_new_conn(user, password)

        for conn in self.connections[user]:
            if not self._is_connection_in_use(id(conn)):
                self.connection_in_use[id(conn)] = time.time()
                self._update_connections_id_query(id(conn))
                return conn

        if self.connections_count > self.max_conn:
            self._close_oldest_free_connection()

        return self._create_new_conn(user, password)

    def _put_conn(self, user, conn):
        if len(self.connections[user]) > 1:
            conn.close()
            self.connections[user].remove(conn)
            self.connections_count -= 1
        del self.connection_in_use[id(conn)]

    def get_conn(self, user, password):
        while True:
            self._lock.acquire()
            try:
                c = self._get_conn(user, password)
                if c:
                    return c
            finally:
                self._lock.release()

    def put_conn(self, user, conn):
        self._lock.acquire()
        try:
            self._put_conn(user, conn)
        finally:
            self._lock.release()
