import copy
import datetime
import json
import sqlite3
import warnings
from dataclasses import dataclass, fields, asdict
from enum import EnumMeta
from string import Template
from typing import List, TypeVar, Generic, Any, Dict

import psycopg2
from loguru import logger
from psycopg2._psycopg import AsIs
from psycopg2.extensions import register_adapter
from psycopg2.extras import Json

import settings
from datasource.databaserepository.exceptions import UnsupportedFilterParamException, RecordNotFoundException, \
    DatabaseRepositoryException2, DatabaseRepositoryException
from services.filter import build_filter_sql_line
from settings import DEBUG

register_adapter(dict, Json)

D = TypeVar('D', bound='DatabaseRepository2.Data')

DATE_FORMATS = ("%Y-%m-%d", '%Y-%m-%d %H:%M:%S', '%Y-%m-%d %H:%M:%S.%f')


class DatabaseRepository2(Generic[D]):
    database_table_name = None

    def __init__(self, request):
        self.id_field = self.Data.ID_FIELD
        self.default_sort = f'ORDER BY "{self.id_field}" DESC'
        self.request = request
        self.conn = None
        self.api2_clone = None
        if request is not None:
            self.conn = request.storage['conn']
            self.api2_clone = request.storage['api2_clone']

    @dataclass
    class Data:
        ID_FIELD = 'ID FILED NOT SET'

        def __init__(self, **kwargs):
            self.conn = None
            self.request = None
            super().__init__(kwargs)

        def get_id(self):
            """Получить значение первичного ключа. (Получить id объекта записи)."""
            return self.__getattribute__(self.ID_FIELD)

        @classmethod
        def fields(cls):
            return {field.name: field.type for field in fields(cls)}

        @classmethod
        def fields_except_id(cls):
            return {field.name: field.type for field in fields(cls) if field.name != cls.ID_FIELD}

        @classmethod
        def empty(cls):
            data = {k: None for k in cls.fields()}
            return cls(**data)

        def as_dict(self):
            return asdict(self)

        def copy(self):
            conn = self.conn
            request = self.request
            self.conn = None
            self.request = None
            copy_obj = copy.deepcopy(self)
            copy_obj.conn = self.conn = conn
            copy_obj.request = self.request = request
            return copy_obj

        def replace(self, data):
            dict_as_obj = self.from_dict(data)
            prev_con = self.conn
            prev_request = self.request
            self.conn = None
            self.request = None
            new = copy.deepcopy(self)
            for k, v in data.items():
                if k != self.ID_FIELD:
                    try:
                        new.__setattr__(k, dict_as_obj.__getattribute__(k))
                    except AttributeError:
                        warnings.warn(
                            f'FROM DBREP2 replace(): Model {self.__class__} does not contain field --{k}--, '
                            f'which exists in data (maybe exists in database table).')

            new.conn = prev_con
            new.request = prev_request
            return new

        @classmethod
        def from_dict(cls, data):
            field_types = {field.name: field.type for field in fields(cls)}
            data_copy = copy.deepcopy(data)
            for field_, type_ in field_types.items():
                try:
                    data_field = data_copy[field_]
                except KeyError:
                    data_field = data_copy[field_] = None
                if data_field is not None and not isinstance(data_field, type_):
                    try:
                        if type_ == datetime.date:

                            if '.' in data_field:
                                data_field = data_field.split('.')[0]

                            for date_format in DATE_FORMATS:
                                try:
                                    data_copy[field_] = datetime.datetime.strptime(data_field, date_format)
                                    break
                                except ValueError as e:
                                    if date_format == DATE_FORMATS[-1]:
                                        raise e
                        elif type_ == datetime.datetime:
                            try:
                                data_copy[field_] = datetime.datetime.fromisoformat(data_field)
                            except Exception as e:
                                raise e
                        elif isinstance(type_, EnumMeta):
                            try:
                                data_copy[field_] = type_(data_field)
                            except ValueError:
                                data_copy[field_] = type_(int(data_field))
                        elif settings.DB_SYSTEM == 'sqlite' and type_ == dict:
                            data_copy[field_] = type_(json.loads(data_field))
                        else:
                            data_copy[field_] = type_(data_field)
                    except (ValueError, TypeError) as e:
                        message = f'Dataclass incorrect type. field: {field_}, type: {type_}, value: {data_copy[field_]}, class: {cls}.'
                        try:
                            message += ' ' + f'Allowed values: {[e.value for e in type_]}'
                        except TypeError:
                            pass
                        raise DatabaseRepositoryException2(message=message)

            keys_not_in_model = []
            for key in data_copy:
                if key not in cls.fields():
                    keys_not_in_model.append(key)
                    # warnings.warn(
                    #     f'Model {cls} does not contain field --{key}--, which exists in data (maby exists in database table). '
                    #     f'It will be set as None value')

            for key in keys_not_in_model:
                data_copy.pop(key)

            try:
                res = cls(**data_copy)
            except TypeError as e:
                raise Exception(f'Dataclass unexpected keyword argument. class: {cls}, message: {e}')

            return res

    def _get_cur(self, conn=None):
        if conn is None:
            conn = self.conn
        if settings.DB_SYSTEM == 'sqlite':
            conn.row_factory = sqlite3.Row
            return conn.cursor()
        elif settings.DB_SYSTEM == 'postgresql':
            return conn.cursor(cursor_factory=psycopg2.extras.DictCursor)

    def _prepare_data(self, data):
        if settings.DB_SYSTEM == 'sqlite':
            for k, v in data.items():
                if isinstance(v, dict):
                    data[k] = json.dumps(v, ensure_ascii=False)

    def _dict_fetch(self, data):
        res = dict(data)
        if settings.DB_SYSTEM == 'sqlite':
            for field, field_type in self.Data.fields_except_id().items():
                if field_type == dict and res.get(field) is not None:
                    res[field] = json.loads(res[field])
        return res

    def _catch_exceptions(self, e, sql=None):
        logger.error(f'SQL FAILED: {sql}. Error: {e}, {e.__class__}')
        if settings.DEBUG:
            raise DatabaseRepositoryException(
                message='sql failed',
                sql=sql,
                exception_class=e.__class__,
                exception_message=str(e),
            )
        raise e

    def _build_sql_filter_list_query(self, fields_, offset='', limit='', main_condition='', sort_sql='', join_line='',
                                     select_join_params_line='', select_join_tables_line='', unique='') -> str:

        if settings.DB_SYSTEM == 'sqlite' and limit in ('', '0') and offset != '':
            limit = '-1'

        return Template(
            'SELECT $fields $join_fields FROM "$table" $join_tables_line $join_line $main_condition $sort $limit $offset;').substitute(
            fields=fields_ if not unique else f'distinct({fields_})',
            join_fields=select_join_params_line,
            table=self.database_table_name,
            join_tables_line=select_join_tables_line,
            join_line=join_line,
            main_condition=main_condition,
            sort='' if unique else ' order by ' + sort_sql if sort_sql else self.default_sort,
            id_field=self.id_field,
            limit=f' limit {limit}' if limit else '',
            offset=f' offset {offset}' if offset else '',
        )

    def _build_fields_line(self, fields_):
        if fields_ == ['*'] or fields_ == '*':
            return f'"{self.database_table_name}".*'
        return ', '.join(f'"{self.database_table_name}"."{field}"' for field in fields_)

    def _read(self, record_id, fields_, select_join_params_line='', select_join_tables_line='', *args, **kwargs):
        if record_id is None:
            raise RecordNotFoundException(record_id)
        # fields_ = ', '.join(fields_)
        # fields_ = fields_.replace('*', f'"{self.database_table_name}".*')
        fields_ = self._build_fields_line(fields_)
        sql = f'SELECT {fields_} {select_join_params_line} FROM "{self.database_table_name}" ' \
              f'{select_join_tables_line} where {self.id_field} = {record_id};'

        logger.trace(f'SQL READ: {sql}')
        cur = self._get_cur()
        try:
            try:
                cur.execute(sql)
            except Exception as e:
                self._catch_exceptions(e, sql)
            record = cur.fetchone()
            if record is None:
                raise RecordNotFoundException(record_id, self.Data)
            return record
        finally:
            cur.close()

    def read(self, record_id, *args, **kwargs) -> D:
        obj = self.Data.from_dict(self._dict_fetch(self._read(record_id, fields_='*', *args, **kwargs)))
        obj.conn = self.conn
        obj.request = self.request
        return obj

    def read_raw(self, record_id, fields_='*', select_join_params_line='', select_join_tables_line='', *args,
                 **kwargs) -> Dict[str, Any]:
        return self._dict_fetch(self._read(record_id, fields_=fields_,
                                           select_join_params_line=select_join_params_line,
                                           select_join_tables_line=select_join_tables_line,
                                           *args, **kwargs))

    def _list(self, fields_, offset='', limit='', filter_sql='', sort_sql='', join_line='',
              select_join_params_line='', select_join_tables_line='', unique='',
              *args, **kwargs):
        # fields_ = ', '.join(f"{field}" for field in fields_)
        # fields_ = fields_.replace('*', f'"{self.database_table_name}".*')
        fields_ = self._build_fields_line(fields_)
        cur = self._get_cur()
        try:
            sql = self._build_sql_filter_list_query(fields_, offset, limit, '', sort_sql,
                                                    join_line, select_join_params_line, select_join_tables_line, unique)
            logger.trace(f'SQL LIST: {sql}')
            try:
                cur.execute(sql)
            except Exception as e:
                self._catch_exceptions(e, sql)
            return cur.fetchall()
        finally:
            cur.close()

    def list(self, *args, **kwargs) -> List[D]:
        objs = [self.Data.from_dict(self._dict_fetch(fetch)) for fetch in self._list('*')]
        for obj in objs:
            obj.conn = self.conn
            obj.request = self.request
        return objs

    def list_raw(self, fields_='*', offset='', limit='', filter_sql='', sort_sql='',
                 join_line='', select_join_params_line='', select_join_tables_line='', unique='',
                 *args, **kwargs) -> List[Dict[str, Any]]:
        return [self._dict_fetch(fetch) for fetch in
                self._list(fields_, offset=offset, limit=limit, filter_sql=filter_sql,
                           sort_sql=sort_sql, join_line=join_line,
                           select_join_params_line=select_join_params_line,
                           select_join_tables_line=select_join_tables_line,
                           unique=unique)]

    def _build_sql_subline_filter(self, params):
        params_lines = []
        if not params:
            return ''
        for param, value in params.items():
            if isinstance(value, list):
                none_value = False
                for i in range(len(value)):
                    if isinstance(value[i], str):
                        v = value[i]
                        value[i] = f"'{v}'"
                    elif value[i] is None:
                        none_value = True
                tp = ', '.join(str(v) for v in value if v is not None)
                if tp:
                    line = f'"{param}" in ({tp})'
                else:  # не было параметров
                    line = ''
                if none_value:
                    line_with_null = f'"{param}" is NULL'
                    if line:
                        line_with_null = ' or ' + line_with_null
                    line += line_with_null
                    # line += f' or "{param}" is NULL'
            else:
                if value is None:
                    line = f'"{param}" is NULL'
                else:
                    tp = f"'{value}'" if isinstance(value, str) else f"{value}"
                    line = f'"{param}"' + ' = ' + tp
            if line:
                params_lines.append(line)

        return ' AND '.join(params_lines)

    def _filter(self, params, fields_, offset='', limit='', filter_sql='', sort_sql='', query_condition='',
                join_line='', select_join_params_line='', select_join_tables_line='', unique='',
                *args, **kwargs):

        if params:
            if isinstance(params, list):  # rest api filter format
                filter_sql_, _, _ = build_filter_sql_line(params, self.database_table_name)
                if filter_sql:
                    filter_sql = ' AND '.join([filter_sql, filter_sql_])
                else:
                    filter_sql = filter_sql_
                params_line = ''
            else:
                unsupported_params = set(params.keys()) - set([field.name for field in fields(self.Data)])
                if unsupported_params:
                    raise UnsupportedFilterParamException(unsupported_params)
                params_line = self._build_sql_subline_filter(params)
        else:
            params_line = ''

        query_condition_line = query_condition

        main_condition = ' AND '.join(filter(lambda x: x, (query_condition_line, params_line, filter_sql)))
        if main_condition:
            main_condition = ' WHERE ' + main_condition

        fields_ = self._build_fields_line(fields_)

        cur = self._get_cur()
        try:
            sql = self._build_sql_filter_list_query(fields_, offset, limit, main_condition, sort_sql,
                                                    join_line, select_join_params_line, select_join_tables_line, unique)
            logger.trace(f'SQL FILTER: {sql}')
            try:
                cur.execute(sql)
            except (psycopg2.errors.InvalidDatetimeFormat,
                    psycopg2.errors.InvalidTextRepresentation,
                    psycopg2.errors.SyntaxError,
                    psycopg2.errors.UndefinedFunction,
                    psycopg2.errors.UndefinedColumn) as e:
                if DEBUG:
                    logger.error(f'SQL FILTER FELL WITH ERROR: {e}')
                    if settings.RAISE_EX:
                        raise e
                return []
            except Exception as e:
                self._catch_exceptions(e, sql=sql)
            return cur.fetchall()
        finally:
            cur.close()

    def filter(self, params, *args, **kwargs) -> List[D]:
        objs = [self.Data.from_dict(self._dict_fetch(fetch)) for fetch in
                self._filter(params, fields_='*', *args, **kwargs)]
        for obj in objs:
            obj.conn = self.conn
            obj.request = self.request
        return objs

    def filter_raw(self, params, fields_='*', query_condition='', offset='', limit='', filter_sql='', sort_sql='',
                   join_line='',
                   select_join_params_line='', select_join_tables_line='', unique='',
                   *args, **kwargs) -> List[Dict[str, Any]]:
        return [self._dict_fetch(fetch) for fetch in self._filter(params, fields_=fields_, offset=offset, limit=limit,
                                                                  query_condition=query_condition,
                                                                  filter_sql=filter_sql,
                                                                  sort_sql=sort_sql, join_line=join_line,
                                                                  select_join_params_line=select_join_params_line,
                                                                  select_join_tables_line=select_join_tables_line,
                                                                  unique=unique,
                                                                  *args, **kwargs)]

    def _build_sql_subline_search(self, params, strict, register_sensitive):
        params_lines = []
        for param, values in params.items():
            # todo: поддержка только строк -> raise exception

            for value in values:
                tl = f'{"lower(" if not register_sensitive else ""}' \
                     f'"{param}"' \
                     f'{")" if not register_sensitive else ""} ' \
                     f'like ' \
                     f'\'{"%" if not strict else ""}{value}{"%" if not strict else ""}\' '
                params_lines.append(tl)

        return ' OR '.join(params_lines)

    def search(self, params, fields_='*', strict=False, register_sensitive=False,
               filter_=None, limit=None,
               *args, **kwargs):
        con = self.conn
        unsupported_params = set(params.keys()) - set([field.name for field in fields(self.Data)])
        if unsupported_params:
            raise UnsupportedFilterParamException(unsupported_params)
        if not register_sensitive:
            params = {k: list(map(str.lower, v)) for k, v in params.items()}
        search_params = self._build_sql_subline_search(params, strict, register_sensitive)
        if not search_params:
            search_params = 'true'
        filter_params = self._build_sql_subline_filter(filter_)
        with con.cursor(cursor_factory=psycopg2.extras.DictCursor) as cur:
            sql = Template(
                'SELECT $fields_ from "$table" WHERE $search $filter order by $id_field DESC $limit;').substitute(
                fields_=', '.join(f"{field}" for field in fields_),
                table=self.database_table_name,
                search=search_params,
                filter=' AND ' + filter_params if filter_params else '',
                id_field=self.id_field,
                limit=f'limit {limit}' if limit else '',
            )
            cur.execute(sql)
            return [dict(fetch) for fetch in cur.fetchall()]

    def save(self, obj, commit=False, *args, **kwargs) -> D:

        def _postgres_sql():
            return f'insert into "{self.database_table_name}" (%s) values %s returning "{self.database_table_name}".*;'

        def _sqlite_sql():
            sql_ = f'insert into "{self.database_table_name}" ($params) values ($values);'
            return Template(sql_).substitute(
                params=', '.join(columns),
                values=', '.join('?' for _ in values)
            )

        def _postgres_exec():
            cur.execute(sql, (AsIs(','.join(columns)), values))

        def _sqlite_exec():
            cur.execute(sql, values)

        def _postgresql_fetch():
            data_ = self._dict_fetch(cur.fetchone())
            if kwargs.get('return_as_dict'):
                return data_
            obj_ = self.Data.from_dict(data_)
            obj_.conn = self.conn
            obj_.request = self.request
            return obj_

        def _sqlite_fetch():
            if kwargs.get('return_as_dict'):
                return self.read_raw(cur.lastrowid)
            return self.read(cur.lastrowid)

        db_system_to_sql_build = {'postgresql': _postgres_sql, 'sqlite': _sqlite_sql}
        db_system_to_exec = {'postgresql': _postgres_exec, 'sqlite': _sqlite_exec}
        db_system_to_fetch = {'postgresql': _postgresql_fetch, 'sqlite': _sqlite_fetch}

        con = self.conn

        if isinstance(obj, dict):
            data = obj
        else:
            data = obj.as_dict()

        if self.id_field in data:
            data.pop(self.id_field)

        cur = self._get_cur(con)
        try:
            self._prepare_data(data)
            columns = tuple(f'"{column}"' for column, value in data.items() if value is not None)
            values = tuple(value for value in data.values() if value is not None)
            sql = db_system_to_sql_build[settings.DB_SYSTEM]()
            try:
                db_system_to_exec[settings.DB_SYSTEM]()
            except Exception as e:
                self._catch_exceptions(e)

            if kwargs.get('no_return'):
                if commit:
                    con.commit()
                return

            obj = db_system_to_fetch[settings.DB_SYSTEM]()
            cur.close()

            if commit:
                con.commit()

            return obj
        finally:
            cur.close()

    def _build_sql_line_update(self, data, record_id):
        def _postgres_line():
            return Template('UPDATE "$table" SET $params WHERE "$id_field" = (%s) RETURNING "$table".*;').substitute(
                table=self.database_table_name,
                params=', '.join(f'"{k}" = (%s)' for k in data if k != self.id_field),
                id_field=self.id_field
            )

        def _sqlite_line():
            return Template('UPDATE "$table" SET $params WHERE "$id_field" = $record_id;').substitute(
                table=self.database_table_name,
                params=', '.join(f'"{k}" = ?' for k, v in data.items() if k != self.id_field),
                id_field=self.id_field,
                record_id=record_id,
            )

        db_system_to_funcs = {
            'postgresql': _postgres_line,
            'sqlite': _sqlite_line,
        }
        return db_system_to_funcs[settings.DB_SYSTEM]()

    def update(self, record_id, obj: Data, commit=False, *args, **kwargs) -> D:  # todo: если первичный ключ не найден

        def _postgresql_exec():
            update_data = (*tuple(v for k, v in data.items() if k != self.id_field), record_id)
            cur.execute(sql, update_data)

        def _sqlite_exec():
            cur.execute(sql, list(v for k, v in data.items() if k != self.id_field))

        def _postgresql_fetch():
            data_ = self._dict_fetch(cur.fetchone())
            if kwargs.get('return_as_dict'):
                return data_
            obj_ = self.Data.from_dict(data_)
            obj_.conn = self.conn
            obj_.request = self.request
            return obj_

        def _sqlite_fetch():
            if kwargs.get('return_as_dict'):
                return self.read_raw(record_id)
            return self.read(record_id)

        db_system_to_exec = {
            'postgresql': _postgresql_exec,
            'sqlite': _sqlite_exec,
        }

        db_system_to_fetch = {
            'postgresql': _postgresql_fetch,
            'sqlite': _sqlite_fetch,
        }

        con = self.conn
        self.read_raw(record_id)

        if isinstance(obj, dict):
            data = obj
        else:
            data = obj.as_dict()
            data.pop(self.id_field, None)

        cur = self._get_cur()
        self._prepare_data(data)

        try:
            sql = self._build_sql_line_update(data, record_id)
            logger.trace(f'SQL UPDATE: {sql}')

            try:
                db_system_to_exec[settings.DB_SYSTEM]()
                if commit:
                    con.commit()
            except Exception as e:
                self._catch_exceptions(e, sql)
            return db_system_to_fetch[settings.DB_SYSTEM]()
        finally:
            cur.close()

    def delete(self, record_id, commit=False, *args, **kwargs) -> None:
        con = self.conn
        cur = self._get_cur()
        try:
            self.read(record_id)
            sql = f'DELETE FROM "{self.database_table_name}" where "{self.id_field}" = {record_id};'
            logger.trace(f'SQL DELETE: {sql}')
            cur.execute(sql)
            if commit:
                con.commit()
        finally:
            cur.close()

    def execute(self, query, commit=False):
        conn = self.conn
        cur = self._get_cur()
        cur.execute(query)
        records = cur.fetchall()
        ans1 = []
        for row in records:
            ans1.append(dict(row))
        if commit:
            conn.commit()
        cur.close()
        return ans1


def execute(connection, query):
    cur = connection.cursor(cursor_factory=psycopg2.extras.DictCursor)
    cur.execute(query)
    records = cur.fetchall()
    ans1 = []
    for row in records:
        ans1.append(dict(row))
    connection.commit()
    cur.close()
    return ans1
