from loguru import logger

from exportmanager.templateengine.settings import (CELL_NUM, CELL_PATTERN, DEFAULT_STYLE, TABLE_NAME_PATTERN)


def get_cell_with_tag_in_tables(tables, full_tag):
    """Получить ячейку из таблиц документа.

    Args:
        tables (:obj:`iter` of :obj:`docx.table.Table`): Таблицы, по которым будет производиться поиск
        full_tag (str): Полный тег ячейки

    Returns:
        docx.table._Cell: Ячейка
        None: Если ячейка не найдена
    """
    return next(filter(
        lambda cell: full_tag in cell.text, [cell for table in tables for cell in table._cells]),
        None
    )


def get_cell_text_style(cell):
    """Получить стиль ячейки.

    Args:
        cell (docx.table._Cell): Ячейка

    Returns:
        dict: Словарь стилей ячейки или стиль по умолчанию, если стили ячейки не найдены.
        Формат словаря:

        {
            'font_name': ...
            'font_size': ...
            'alignment': ...
        }
    """
    if not cell.text or not cell.paragraphs[0].runs:
        return DEFAULT_STYLE
    par = cell.paragraphs[0]
    par_run = par.runs[0]
    return {
        'font_name': par_run.font.name,
        'font_size': par_run.font.size,
        'alignment': par.alignment,
    }


def set_cell_text(cell, new_text, style):
    """Задать текст ячейки. Полностью заменяет текст ячейки на новый.

    Args:
        cell (docx.table._Cell): Ячейка
        new_text (str): Текст
        style (dict): Словарь стилей ячейки
    """
    cell.text = str(new_text)
    para = cell.paragraphs[0]
    run = cell.paragraphs[0].runs[0]
    if style is not None:
        run.font.name = style['font_name']
        run.font.size = style['font_size']
        para.alignment = style['alignment']


def get_all_document_cells(doc):
    """Получить все ячейки документа.

    Args:
        doc (docx): Docx документ

    Returns:
        list: Список ячеек
    """
    return [cell for table in doc.tables for cell in table._cells]


def remove_tag_from_cell(cell, full_tag):
    """Удалить тег из ячейки.

    Args:
        cell (docx.table._Cell): Ячейка
        full_tag (str): Полный тег ячейки
    """
    new_cell_text = cell.text.replace(full_tag, '')
    style = get_cell_text_style(cell)
    set_cell_text(cell, new_cell_text, style)


# вероятно избыточно, можно сделать и без этого
def get_cell_order_in_row(cells, tags):
    """Получить очередность ячеек в строке таблицы.
    Используется для строк, с повторяющимися данными.

    Пояснение:
    Теги приходят не отсортированные согласно последовательности ячеек. Необходимо найти их последовательность.

    Например:
    ---------------------
    | &&one&& | &&two&& | - строка таблицы, которая будет повторяться (заполняться такими же строками с данными)
    ---------------------

    {
        "two": "two_value",
        "one": "one_value"  - последовательность данных не совпадает с последовательностью ячеек в таблице
    }

    Args:
        cells (:obj:`iter` of :obj:`docx.table._Cell`): Ячейки, среди которых будет искаться последовательность
            (как правило, все ячейки таблицы)
        tags (:obj:`iter` of :obj:`str`): Теги, очередность которых надо найти

    Raises:
        CellTagNotFoundWarning: Если не найдена ячейка, с тегом из `tags`

    Returns:
        :obj:`list` of :obj:`str`: Последовательность тегов ячеек в строке
    """
    cell_order = []
    first_cell_idx = None
    i = 0
    for cell in cells:
        if cell.text == CELL_NUM:
            cell_order.append(CELL_NUM)
            clear_cell_text(cell)
            if first_cell_idx is None:
                first_cell_idx = i
        else:
            tag = get_tag_from_list_contained_in_cell(cell, tags)
            # если тег не был найден в данных формы, но есть указание тега в ячейке, то задаем как
            # None, чтобы не было сдвига
            if tag is None and '&&' in cell.text:
                cell_order.append(tag)
            if tag is not None:
                first_tag_found = True
                cell_order.append(tag)
                clear_cell_text(cell)
                if first_cell_idx is None:
                    first_cell_idx = i
        i += 1

    # for tag in (set(tags) - set(cell_order)):
    #     warnings.warn(CellTagNotFoundWarning(CELL_PATTERN % tag))

    # if set(tags) - set(cell_order):
    #     raise CellTagNotFoundWarning(set(tags) - set(cell_order))

    return cell_order, first_cell_idx


def get_tag_from_list_contained_in_cell(cell, tags):
    return next(filter(lambda tag: CELL_PATTERN % tag in cell.text, tags), None)


def replace_cell_tag_to_text(cell, full_tag, text, style):
    """Заменить тег ячейки таблицы на текст.

    Args:
        cell (docx.table._Cell): Ячейка таблицы
        full_tag (str): Полный тек ячейки
        text (str): Текст
        style (dict): Словарь стилей ячейки
    """
    new_cell_text = cell.text.replace(full_tag, text)
    set_cell_text(cell, new_cell_text, style)


def clear_cell_text(cell):
    """Удалить текст ячейки.

    Args:
        cell (docx.table._Cell): Ячейка
    """
    style = get_cell_text_style(cell)
    set_cell_text(cell, '', style)


def get_table_cells_styles_by_tags(cells, tags):
    """Возвращает стили ячеек.

    Args:
        cells (:obj:`list` of :obj:`docx.table._Cell`): Список ячеек документа, среди которых будет производиться поиск
        tags (dict_keys): Теги элементов, по которым будут искаться ячейки

    Returns:
        Словарь: ключ - тег, значение - словарь стиля.

        {
            "<имя_элемента>": {
                "font_name": ...
                "font_size": ...
                "alignment": ...
            },
            ...
        }
    """
    return {tag: get_cell_text_style(get_cell_by_tag(cells, tag)) for tag in tags if get_cell_by_tag(cells, tag)}


def add_empty_rows_to_table(table, num_of_rows_to_add):
    """Добавить пустые строки (последовательность ячеек) в таблицу.
    (Используется, для ускорения процесса добавления данных в таблицу).

    Args:
        table (docx.table.Table): Таблица
        num_of_rows_to_add (int): Количество строк для добавления
    """
    for _ in range(num_of_rows_to_add):
        table.add_row()


def split_cells_to_rows(cells, num_of_cells_in_row):
    """Получить ячейки, разбитые на строки.

    Args:
        cells (:obj:`iter` of :obj:`docx.table._Cell`): Ячейки
        num_of_cells_in_row (int): Количество ячеек в строке

    Returns:
        :obj:`list` of :obj:`list` of :obj:`docx.table._Cell`: Список, элеметнами которого являются списки ячеек
    """
    rows = []
    rows_num = len(cells) // num_of_cells_in_row
    for l in range(rows_num):
        row = cells[l * num_of_cells_in_row: (l + 1) * num_of_cells_in_row]
        rows.append(row)
    return rows


def add_data_to_row(row, row_data, cell_order_in_row, styles, row_num):
    """Добавить данные в строку.

    Args:
        row (:obj:`iter` of :obj:`docx.table._Cell`): Строка
        row_data (dict): Данные для добавления в строку. {"tag_name": "value"}
        cell_order_in_row (:obj:`list` of :obj:`str`): Последовательность ячеек в строке (элементы списка - теги ячеек)
        styles (dict): Словарь стилей ячеек
        row_num (int): Номер строки в последовательности строк (для нумерации строк в docx докменте)
    """
    for i, cell in enumerate(row):
        tag = cell_order_in_row[i]
        if tag == CELL_NUM:
            text = str(row_num + 1)
            style = DEFAULT_STYLE
        else:
            text = row_data.get(tag, '')
            if tag is not None:
                style = styles[tag]
            else:
                style = DEFAULT_STYLE
        if text is None:
            text = ''
        print(cell, text, style)
        try:
            set_cell_text(cell, str(text), style)
        except ValueError:
            logger.warning(f'Экспорт в word. Не получилось занести значение в ячейку: {text}')


def is_row_empty(row):
    """Пустая ли строка.

    Args:
        row (:obj:`iter` of :obj:`docx.table._Cell`): Строка (последовательность ячеек)

    Returns:
        bool: True - строка пустая, False - строка не пустая
    """
    # return all(not cell.text for cell in row)
    # return not any(cell.text for cell in row)
    for cell in row:
        if cell.text != '':
            return False
    return True


def get_row_data(rows_data):
    if isinstance(rows_data, list):
        return rows_data[0].keys()
    elif isinstance(rows_data, dict):
        for key, value in rows_data.items():
            return value


def get_rows_data(rows_data):
    if isinstance(rows_data, list):
        return rows_data
    if isinstance(rows_data, dict):
        return [value for value in rows_data.values()]


def add_rows_to_table(table, rows_data):
    """Добавить строки с повторяющимися значениями в таблицу.

    Args:
        table (docx.table.Table): Таблица
        rows_data (list or dict): Данные строк (словарь или список)
    """
    # TODO: вероятно не верно, так как первая строка может быть с ошибками
    row_tags = get_row_data(rows_data)  # получаем теги строки таблицы

    table_cells_styles_by_tags = get_table_cells_styles_by_tags(table._cells, row_tags)  # получаем стили ячеек таблицы

    cell_order_in_row, first_cell_idx = get_cell_order_in_row(table._cells, row_tags)  # получаем порядок ячеек в строке

    num_of_cells_in_row = len(cell_order_in_row)  # количество ячеек в строке
    num_of_rows_to_add = len(rows_data) - 1  # потому что первая строка уже есть (где были теги, их очистили)

    first_row_idx = first_cell_idx // num_of_cells_in_row

    add_empty_rows_to_table(table, num_of_rows_to_add)  # добавляем пустые строки в таблицу

    rows = split_cells_to_rows(table._cells, num_of_cells_in_row)  # разбиваем ячейки таблицы на строки

    row_num = 0
    for row in rows[first_row_idx:]:
        if row_num < len(rows_data):
            add_data_to_row(
                row, rows_data[row_num], cell_order_in_row, table_cells_styles_by_tags, row_num)
            row_num += 1


def set_table(table, table_rows_data):
    """Заполнить таблицу на основе данных.

    Args:
        table (docx.table.Table): Таблица
        table_data_ (dict): Данные для заполнения таблицы
    """
    if table_rows_data:
        add_rows_to_table(table, table_rows_data)


def get_cell_by_tag(cells, full_tag):
    """Получить ячейку, соедержащий полный тег.

    Args:
        cells (:obj:`iter` of :obj:`docx.table._Cell`): Ячейки, по которым будет производится поиск
        full_tag (str): Полный тег

    Returns:
        docx.table._Cell: Ячейка
        None: Если ячейка не найдена
    """
    return next(filter(lambda cell: full_tag in cell.text, cells), None)


def is_tag_in_table(table, full_tag):
    """Есть ли тег в таблице.

    Args:
        table (docx.table.Table): Таблица
        full_tag (str): Полный тег

    Returns:
        bool: True - тег в таблице есть, False - тега в таблице нет
    """
    return any(full_tag in cell.text for cell in table._cells)


def get_table_by_tag(tables, full_tag, remove_tag=False):
    """Получить таблицу, содержащую тег.

    Args:
        tables (:obj:`iter` of obj:`docx.table.Table`): Таблицы, по которым будет производиться поиск
        full_tag (str): Полный тег
        remove_tag (bool, optional): Удалить ли тег таблицы после нахождения. Defaults to False.
            Как правило, функция используется для получения таблицы на основе тега, идентифицирующего таблицу.
            Данный тег не заменяется ни на какое иное значение, служит только для определения таблицы среди других.
            Если таблица не понадобится в дальнейшем -- отправить `True`.

    Returns:
        docx.table.Table: Таблица
        None: Если таблица не найдена
    """
    for table in tables:
        if is_tag_in_table(table, full_tag):
            if remove_tag:
                cell = get_cell_by_tag(table._cells, full_tag)
                remove_tag_from_cell(cell, full_tag)
            return table


def set_document_tables(tables, table_section):
    """Заполнить таблицы документа на основе данных.

    Args:
        tables (:obj:`iter` of obj:`docx.table.Table`): Таблицы
        table_section (dict): Данные таблиц
    """
    if not table_section:
        return

    for table_tag, table_data in table_section.items():
        # for row in table_section:
        full_tag = TABLE_NAME_PATTERN % table_tag
        table = get_table_by_tag(tables, full_tag, remove_tag=True)
        if table is not None:
            set_table(table, table_data)
        else:
            # warnings.warn(TableTagNotFoundWarning(full_tag))
            # raise TableTagNotFoundWarning(full_tag)
            pass


def set_remarks2(tables, remark_section):
    remarks_table = get_table_by_tag(tables, TABLE_NAME_PATTERN % 'remarks_table', remove_tag=True)
    if remarks_table is None:
        return
    for table_tag, table_remarks in remark_section.items():
        data = []
        for remark in table_remarks:
            if remark.get('ignore_at_templates'):
                continue
            try:
                col = str(int(remark['pointer']['col']))
            except ValueError:
                col = ''
            try:
                row = str(int(remark['pointer']['row']))
            except ValueError:
                row = ''
            data.append({
                'col': col,
                'row': row,
                'value': str(remark['value']['text']),
            })
        if data:
            set_table(remarks_table, data)
