import warnings
from decimal import Decimal
from typing import Any, Optional

from docx.document import Document
from docx.text.paragraph import Paragraph

from exportmanager.templateengine.docx.docx_table import get_all_document_cells, get_cell_text_style, get_cell_with_tag_in_tables, \
    get_table_cells_styles_by_tags, replace_cell_tag_to_text
from exportmanager.templateengine.settings import TEXT_PATTERN
from exportmanager.templateengine.warnings import FixedElementTagNotFoundWarning

# ParagraphStyle = dict[str, Any]
# ParagraphsStyles = dict[str, ParagraphStyle]
# Repository = dict[str, dict[str, str]]


def _get_style_of_paragraph(paragraph: Paragraph):
    """Получить стили параграфа.

    Args:
        paragraph (Paragraph): Параграф.

    Returns:
        ParagraphStyle: Стили параграфа в формате:
        {
            'font_name': ...
            'font_size': ...
            'alignment': ...
        }
    """
    return {
        'font_name': paragraph.runs[0].font.name,
        'font_size': paragraph.runs[0].font.size,
        'alignment': paragraph.alignment
    }


def _get_styles_of_paragraphs_by_tags(paragraphs, tags,
                                      repository):
    """Получить стили параграфов документа в формате словаря по тегам.

    Args:
        paragraphs (Iterable[Paragraph]): Параграфы.
        tags (Iterable[str]): Теги элементов с фиксированными значениями (text).
        repository (Repository): Дополнительная информация тегов.

    Returns:
        ParagraphsStyles: Стили параграфов по тегам
    """
    styles = {}
    for tag in tags:
        full_tag = TEXT_PATTERN % tag
        paragraph = get_paragraph_with_tag(paragraphs, full_tag)
        if paragraph:
            if paragraph.runs:
                styles[tag] = _get_style_of_paragraph(paragraph)
            # if repository.get(tag) and repository[tag].get('format'):
            #     styles[tag]['format'] = repository[tag]['format']
    return styles


def _set_paragraph_text(paragraph: Paragraph, new_text: str, style):
    """Задать текст параграфа. Полностью заменяет текст параграфа на новый.

    Args:
        paragraph (Paragraph): Параграф.
        new_text (str): Новый текст.
        style (ParagraphStyle): Стиль параграфа.
    """
    paragraph.text = new_text
    run = paragraph.runs[0]
    if style is not None:
        run.font.name = style['font_name']
        run.font.size = style['font_size']
        paragraph.alignment = style['alignment']


def convert_text_to_format(text, format_):
    return format_ % Decimal(text)


def _replace_paragraph_tag_to_text(paragraph: Paragraph, full_tag: str, text: str, style):
    """Заменить тег параграфа на текст.

    Args:
        paragraph (Paragraph): Параграф.
        full_tag (str): Полный тег.
        text (str): Текст.
        style (ParagraphStyle): Стиль параграфа.
    """
    if style.get('format'):
        text = convert_text_to_format(text, style['format'])  # TODO: форматы
    new_text = paragraph.text.replace(full_tag, str(text))
    _set_paragraph_text(paragraph, new_text, style)


def get_paragraph_with_tag(paragraphs, full_tag: str) -> Optional[Paragraph]:
    """Получить параграф docx документа, содержащий тег.

    Args:
        paragraphs (Iterable[Paragraph]): Параграфы, по которым осуществляется поиск.
        full_tag (str): Полный тег элемента.

    Returns:
        Optional[Paragraph]: Параграф, содержащий тег (None, если не найден)
    """
    return next(filter(lambda par: full_tag in par.text, paragraphs), None)


import re


def delete_paragraph(paragraph):
    p = paragraph._element
    p.getparent().remove(p)
    p._p = p._element = None


def delete_all_unused_tags(doc):
    reg = [
        r'{{\w+}}',
        r'%%\w+%%',
        r'&&\w+&&'
    ]

    for table in doc.tables:
        for cell in table._cells:
            for r in reg:
                f = re.findall(r, cell.text)
                for f_ in f:
                    style = get_cell_text_style(cell)
                    replace_cell_tag_to_text(cell, f_, '', style)

    for par in doc.paragraphs:
        # специальный тег для удаления параграфа
        # используется для того, чтобы склеить таблицы после их заполнения
        if par.text == '{{delete_paragraph}}':
            delete_paragraph(par)

        for r in reg:
            f = re.findall(r, par.text)
            for f_ in f:
                style = _get_style_of_paragraph(par)
                _replace_paragraph_tag_to_text(par, f_, '', style)


def set_document_text(doc: Document, text_section, repository_section):
    """Заполняет фиксированные значения шаблона.
    1. Сбор всех стилей параграфов и ячеек. (При изменении текста стили могут слетать.
     После изменения текстов применить стили ко всем измененным параграфам и ячейкам).
    2. Поиск фиксированных значений среди параграфов и ячеек и изменение текста с сохранением стилей.

    Args:
        doc (Document): Шаблон документа.
        text_section (dict): Фиксированные значения (text). {"<element_name>": "<element_text>"}.
        repository_section (Repository): Дополнительные данные по форматированию.
    """
    if not text_section:
        return

    all_doc_cells = get_all_document_cells(doc)
    tags = list(text_section.keys())

    paragraphs_styles = _get_styles_of_paragraphs_by_tags(doc.paragraphs, tags, repository_section)
    cells_styles = get_table_cells_styles_by_tags(all_doc_cells, tags)

    for tag, value in text_section.items():
        full_tag = TEXT_PATTERN % tag
        paragraph = get_paragraph_with_tag(doc.paragraphs, full_tag)
        if paragraph:
            _replace_paragraph_tag_to_text(paragraph, full_tag, value, paragraphs_styles[tag])
        else:
            cell = get_cell_with_tag_in_tables(doc.tables, full_tag)
            if cell:
                replace_cell_tag_to_text(cell, full_tag, value, cells_styles.get(tag))
            # else:
                # warnings.warn(FixedElementTagNotFoundWarning(full_tag))

    # delete_all_unused_tags(doc)  # TODO: работает некорректно, удаляет и таблицы, удалять неиспользуемые таблицы в docx_tables