import re
from copy import copy, deepcopy
from exportmanager.templateengine.xlsx.exceptions import XlsxExportManagerException

from loguru import logger
from openpyxl import load_workbook
from openpyxl.utils.exceptions import IllegalCharacterError

from services.exceptions import CoreBaseException
from exportmanager.templateengine.docx.docx_table import get_row_data, get_rows_data
from exportmanager.templateengine.settings import CELL_NUM, TABLE_NAME_PATTERN, TEXT_PATTERN, CELL_PATTERN


def get_tables_adding_num_of_rows(ws, tables_section):
    return dict(zip(
        [table_tag for table_tag in tables_section],
        [len(tables_section[table_tag]) for table_tag in tables_section]
    ))


def is_table_start(cell, table_tag, remove_tag):
    if cell.value is None or isinstance(cell.value, int):
        return False
    # print(TABLE_NAME_PATTERN % table_tag, cell.value)
    if TABLE_NAME_PATTERN % table_tag in str(cell.value):
        if remove_tag:
            cell.value = cell.value.replace(TABLE_NAME_PATTERN % table_tag, '')
        return True
    return False


def get_table_rows_and_cols_intervals(ws, table_tag, remove_tag=False):
    try:
        cells = []
        for row in ws:
            for cell in row:
                # print(cell.value)
                cells.append(cell)

        # rows = [row for row in ws.rows]
        # cells = [cell for cell in rows]
        table_start_row_idx = next(filter(
            lambda cell: is_table_start(cell, table_tag, remove_tag),
            # [cell for row in ws for cell in row]
            cells
            # [cell for cell in ws.rows]
        )).row
    except StopIteration:
        raise XlsxExportManagerException(f'can not find table with tag %%{table_tag}%% in xlsx document')

    table_end_row_idx = 0
    table_start_col_idx = 0
    table_end_col_idx = 0
    for cell in [cell for row in ws.iter_rows(min_row=table_start_row_idx) for cell in row]:
        if '&&' in str(cell.value) or CELL_NUM in str(cell.value):
            if not table_end_row_idx:
                table_start_col_idx = cell.column
                table_end_row_idx = cell.row
            else:
                if cell.row != table_end_row_idx:
                    break
                table_end_col_idx = cell.column

    return table_start_row_idx, table_end_row_idx, table_start_col_idx, table_end_col_idx


def get_table_rows_intervals(ws, table_tag, remove_tag=False):
    return get_table_rows_and_cols_intervals(ws, table_tag)[:2]


def get_table_first_data_row_index(ws, table_tag):
    table_first_row_idx, table_last_row_idx = get_table_rows_intervals(ws, table_tag)
    rows = ws.iter_rows(min_row=table_first_row_idx, max_row=table_last_row_idx)
    try:
        return next(filter(
            lambda cell: any(i in str(cell.value) for i in (CELL_NUM, '&&')),
            [cell for row in rows for cell in row]
        )).row
    except StopIteration:
        raise Exception(f'something wrong with tags tow in table %%{table_tag}%%. it looks like the line is missing tags defining cells')


def get_tables_first_data_row_indexes(ws, tables_tags):
    return dict(
        zip(
            tables_tags,
            [get_table_first_data_row_index(ws, table_tag) for table_tag in tables_tags]
        )
    )


def get_shift(tables_first_data_rows_indexes_before_editing, tables_adding_num_of_rows):
    total_shift = {}
    for table_tag, first_data_row in tables_first_data_rows_indexes_before_editing.items():
        shift_ = max(tables_adding_num_of_rows[table_tag] - 1, 0)
        total_shift[first_data_row] = shift_
    return total_shift


def set_cells_height(ws, row_base, row_start, row_end):
    for i in range(row_start, row_end):
        ws.row_dimensions[i].height = ws.row_dimensions[row_base].height


def unmerge_cells_inside_table_row_area(ws, merged_cells_before_editing, start_row, end_row):
    for range_ in merged_cells_before_editing:
        if range_.min_row >= start_row and range_.max_row <= end_row:
        # if range_.min_row >= start_row or range_.max_row <= end_row:
            ws.merge_cells(start_row=range_.min_row, start_column=range_.min_col,
                           end_row=range_.max_row, end_column=range_.max_col)
            ws.unmerge_cells(start_row=range_.min_row, start_column=range_.min_col,
                             end_row=range_.max_row, end_column=range_.max_col)


def count_shift(shift, row):
    return sum(value for key, value in shift.items() if row > key)


def update_merge_cells_after_editing(ws, merged_cells_before_editing, shift):
    for range_ in merged_cells_before_editing:
        shift_count = count_shift(shift, range_.min_row)
        min_row = range_.min_row + shift_count
        max_row = range_.max_row + shift_count
        ws.merge_cells(start_row=min_row, start_column=range_.min_col, end_row=max_row, end_column=range_.max_col)


def get_cell_ordering(row_tags, first_data_row):
    ordering = []
    for cell in first_data_row:
        if not cell.value:
            continue
        if str(cell.value) == CELL_NUM:
            ordering.append(CELL_NUM)
        elif '&&' in cell.value:
            ordering.append(cell.value[2:-2])
        # elif len(ordering) < len(row_tags) + 1:
        #     try:
        #         ordering.append(
        #             next(filter(lambda tag: CELL_PATTERN % tag == str(cell.value), row_tags))
        #         )  # field exists in JSON, but do not found in XLSX
        #     except StopIteration:
        #         # warnings.warn(f'field: {cell.value} exists in template, but does not found in JSON data')
        #         # raise CoreBaseException(message=f'field: {cell.value} exists in template, but does not found in JSON data')
        #         pass
    return ordering


def get_cell_style(cell):
    return {
        'font': copy(cell.font),
        'border': copy(cell.border),
        'fill': copy(cell.fill),
        'number_format': copy(cell.number_format),
        'alignment': copy(cell.alignment),
    }


def set_cell_style(cell, style):
    cell.font = style['font']
    cell.border = style['border']
    cell.alignment = style['alignment']


def get_cells_style(row, min_col_num, max_col_num):
    return [get_cell_style(row[i]) for i in range(min_col_num - 1, max_col_num)]


def _set_table_rows(ws, rows_data, ordering, min_row, max_row, min_col, max_col, cells_style):
    row_num = 1
    row_i = 0
    for row in ws.iter_rows(min_row=min_row, max_row=max_row, min_col=min_col, max_col=max_col):
        for i, cell in enumerate(row):
            row = rows_data[row_i]
            value = row.get(ordering[i], '') if ordering[i] != CELL_NUM else row_num
            try:
                cell.value = value
            except IllegalCharacterError:
                logger.warning(f'Экспорт в word. Не получилось занести значение в ячейку: {value}')
            # cell.value = ordering[i] if ordering[i] != CELL_NUM else row_num
            set_cell_style(cell, cells_style[i])
        row_i += 1
        row_num += 1


def set_table(ws, table_tag, table_data, merged_cells_before_editing):
    first_data_row_idx = get_table_first_data_row_index(ws, table_tag)
    table_start_row_idx, table_end_row_idx, table_start_col_idx, table_end_col_idx = get_table_rows_and_cols_intervals(
        ws, table_tag, remove_tag=True)

    if not table_data:
        return

    row_tags = get_row_data(table_data)
    rows_data = get_rows_data(table_data)
    num_of_rows_to_add = len(rows_data)

    ordering = get_cell_ordering(row_tags, ws[first_data_row_idx])
    cells_style = get_cells_style(ws[first_data_row_idx], table_start_col_idx, table_end_col_idx)

    if num_of_rows_to_add > 1:
        ws.insert_rows(idx=first_data_row_idx + 1, amount=num_of_rows_to_add - 1)

    set_cells_height(ws, first_data_row_idx, first_data_row_idx, first_data_row_idx + num_of_rows_to_add)
    unmerge_cells_inside_table_row_area(ws, merged_cells_before_editing,
                                        table_start_row_idx, table_end_row_idx + num_of_rows_to_add)

    _set_table_rows(ws, rows_data, ordering, first_data_row_idx, table_end_row_idx +
                    num_of_rows_to_add - 1, table_start_col_idx, table_end_col_idx, cells_style)


def set_tables(ws, wb, tables_section):
    tables_first_data_rows_indexes_before_editing = get_tables_first_data_row_indexes(ws, tables_section.keys())
    # print('tables_first_data_rows_indexes_before_editing', tables_first_data_rows_indexes_before_editing)
    tables_adding_num_of_rows = get_tables_adding_num_of_rows(ws, tables_section)  # TODO: избыточно
    #
    shift = get_shift(tables_first_data_rows_indexes_before_editing, tables_adding_num_of_rows)

    #
    merged_cells_before_editing = deepcopy(wb[wb.sheetnames[0]].merged_cells.ranges)
    #
    for table_tag, table_data in tables_section.items():
        set_table(ws, table_tag, table_data, merged_cells_before_editing)
    #
    update_merge_cells_after_editing(ws, merged_cells_before_editing, shift)


def remove_unused_tags(ws):
    reg = [
        r'{{\w+}}',
        r'%%\w+%%',
        r'&&\w+&&',
        r'###'
    ]
    # cells_with_text = list(filter(
    #     lambda cell: all(i in str(cell.value) for i in ('{{', '}}')),
    #     [cell for row in ws for cell in row]
    # ))
    for row in ws:
        for cell in row:
            for r in reg:
                if cell.value:
                    str_parts = re.findall(r, str(cell.value))
                    for str_part in str_parts:
                        cell.value = cell.value.replace(str_part, '')

        # pass


def set_text(ws, text_section):
    if not text_section:
        return
    cells_with_text = list(filter(
        lambda cell: all(i in str(cell.value) for i in ('{{', '}}')),
        [cell for row in ws for cell in row]
    ))
    for cell in cells_with_text:
        for text_tag, text_data in text_section.items():
            full_tag = TEXT_PATTERN % text_tag
            if full_tag in str(cell.value):
                cell.value = cell.value.replace(full_tag, str(text_data))


def create_xlsx_document2(data, template_file_path, result_file_path):
    wb = load_workbook(template_file_path, data_only=True)
    ws = wb.active

    set_text(ws, data.get('form_main_data'))

    data_for_tables = data.get('form_table_data')
    remark_section = data.get('form_remarks')
    remarks_data = []
    for table_tag, table_remarks in remark_section.items():
        for remark in table_remarks:
            if remark.get('ignore_at_templates'):
                continue
            remarks_data.append({
                'col': str(remark['pointer']['col']),
                'row': str(remark['pointer']['row']),
                'value': str(remark['value']['text']),
            })

    # if remarks_data:
    data_for_tables.update({'remarks_table': remarks_data})

    table_names_to_pop = []
    for table_name in data_for_tables:
        try:
            get_table_rows_and_cols_intervals(ws, table_name)
        except XlsxExportManagerException:
            table_names_to_pop.append(table_name)
    for table_name in table_names_to_pop:
        data_for_tables.pop(table_name)

    set_tables(ws, wb, data_for_tables)

    remove_unused_tags(ws)

    wb.save(result_file_path)
