import copy
import os
from typing import List

import settings

from repository.forms.exceptions import FormRepositoryNotFoundException, ReportNumNotFoundException, \
    TemplateNotFoundException, ModelItemNotFoundException, RepositoryFormsException
from repository.forms.validation import check_form_configuration, check_form_types
from repository.reports import reports_repository
from repository.types import types_repository as types_repository
from repository.utils import import_module_
from services.exceptions import CoreBaseException

forms_repository = {}  # все, что есть в файле
forms_repository_files_dirs_ = []

current_decree_year = None


def init_forms_repository(items, templates_items):
    global forms_repository_files_dirs_
    forms_repository_files_dirs_ = templates_items
    for item in items:
        form_file_data = item['file_data']
        path = item['path']
        file_name = os.path.splitext(os.path.basename(path))[0]

        check_form_configuration(form_file_data, path)
        check_form_types(path, form_file_data)

        forms_repository[file_name] = form_file_data

        if file_name.split('_')[2] != '0':
            report_num = file_name.split('_')[1]
            report_repository = reports_repository.get_report_repository_by_num(int(report_num))
            # добавление выбора версии организации в системные атрибуты
            forms_repository[file_name]['system_attrs'] = [
                {
                    'field': 'author_id',
                    'caption': 'Версия организации',
                    'type': 'organizations_voc_type_org_versions',
                }
            ]

            types_repository.types_repository['organizations_voc_type_org_versions'] = copy.deepcopy(
                types_repository.types_repository['organizations_voc_type']
            )

            # чтобы не отображался в главном меню
            types_repository.types_repository['organizations_voc_type_org_versions']['not_display_in_menu'] = True
            types_repository.types_repository['organizations_voc_type_org_versions']['voc'] = {}
            types_repository.types_repository['organizations_voc_type_org_versions']['voc']['config_url'] = {
                'base_path': '/api/vocs/organizations?',  # переопределение базового url
                'params': {
                    'action': 'list_versions'
                },
                'rename_params': {
                    # value отправляется всегда параметром, переименование параметра с value на record_id
                    'value': 'record_id',
                },
            }

            # добавление системных атрибутов из репозитория отчетов +
            if 'system_attrs' in report_repository:
                forms_repository[file_name]['system_attrs'].extend(report_repository['system_attrs'])

    # Задаем год текущего приказа.
    global current_decree_year
    for form_no in forms_repository:
        # Если form_no формата form_1_1_2020 (последняя цифра - год)
        if form_no.count('_') == 3:
            try:
                form_no_decree_year = int(form_no.split('_')[-1])
            except ValueError:
                raise RepositoryFormsException(message=f'Имя формы некорректно: {form_no}. Не удалось '
                                                       f'преобразовать год приказа в целое число (последнее число '
                                                       f'в названии имени файла репозитория формы).')
            if current_decree_year is None or form_no_decree_year > int(current_decree_year):
                current_decree_year = str(form_no_decree_year)


# def _get_decree()


def get_form_repository_f(form_name):
    form_repo = get_form_repository(form_name)
    res = {
        'form': {
            'form_main_attrs': form_repo['form_main_attrs'],
            'tables': form_repo['tables'],
            'system_attrs': form_repo['system_attrs'],
        },
        'types': types_repository.get_types_repository(),
    }
    for table_name, table in res['form']['tables'].items():
        table['options'] = {
            'server_filter_sort': False if settings.DB_SYSTEM == 'sqlite' else True,
        }
    return res


def get_clear_form_no(form_no):
    if form_no.count('_') > 2:
        return '_'.join(form_no.split('_')[:3])
    return form_no


def get_form_repository(form_no: str):
    """Получить репозиторий формы по `form_no`.
    Возвращает ссылку на репозиторий, изменения, примененные к результату функции, приведут к изменению данных
    репозитория формы во всей системе.

    Args:
        form_no (str): form_no формы (например form_1_1).

    Returns:
        Репозиторий формы в формате словаря.

    Raises:
        FormRepositoryNotFoundException(CoreBaseException) - если репозиторий формы не найден.
    """
    try:
        return forms_repository[form_no]
    except KeyError:
        # Если form_no формата form_1_1_2020 (последняя цифра - год)
        if form_no.count('_') == 3:
            form_no_without_decree_year = '_'.join(form_no.split('_')[:-1])
            form_no_with_current_decree_year = form_no_without_decree_year + f'_{current_decree_year}'
        else:
            form_no_with_current_decree_year = form_no + f'_{current_decree_year}'
        try:
            return forms_repository[form_no_with_current_decree_year]
        except KeyError:
            raise FormRepositoryNotFoundException(f'{form_no} or {form_no_with_current_decree_year}')


def get_forms_types_list_repository():
    res = [{
        'label': form_repo['form_name'],
        'value': form_no,
    } for form_no, form_repo in forms_repository.items()]
    res.sort(key=lambda item: int(item['value'].split('_')[1]) + int(item['value'].split('_')[2]))
    return res


def get_forms_types_list_repository_by_report_num(report_num, ignore_title_form=True):
    try:
        res = get_forms_types_list_repository()
        if ignore_title_form:
            return list(filter(lambda item: item['value'].split('_')[1] == str(report_num) and item['value'].split('_')[2] != '0', res))
        return list(filter(lambda item: item['value'].split('_')[1] == str(report_num), res))
    except KeyError:
        raise ReportNumNotFoundException(report_num)


def get_forms_types_list_repository_by_report_num_and_decree_year(report_num, decree_year=None):
    try:
        res = get_forms_types_list_repository_by_report_num(report_num)
        if decree_year:
            res = list(filter(lambda x: x['value'].endswith(decree_year), res))
        return res
    except KeyError:
        raise ReportNumNotFoundException(report_num)


def get_decree_year_by_form_no(form_no):
    if form_no.count('_') == 3:
        return form_no.split('_')[-1]
    return current_decree_year


def get_forms_exports_types():
    return [get_form_exports_types(form_no) for form_no in forms_repository]


def get_form_exports_types(form_name):
    res = []
    form_repo = get_form_repository(form_name)
    report_num = form_name.split('_')[1]
    for export_item in form_repo['export']:
        for ext in export_item['output_ext']:
            res.append({
                'title': export_item['form_name'] + ' ' + ext.upper(),
                'form_id': export_item['object_type'] + '.' + ext,
                'object_type': 'form',
            })

    report_repository = reports_repository.get_report_repository(int(report_num))
    if 'title_form' in report_repository:
        title_form_repository = get_form_repository(report_repository['title_form'])
        for export_item in title_form_repository['export']:
            for ext in export_item['output_ext']:
                res.insert(0, {
                    'title': export_item['form_name'] + ' ' + ext.upper(),
                    'form_id': export_item['object_type'] + '.' + ext,
                    'object_type': 'form_title',
                })
    return res


def get_form_template_full_path(template_file_name):
    for path in forms_repository_files_dirs_:
        if path.endswith(template_file_name):
            return path
    raise TemplateNotFoundException(template_file_name)


def is_form_template_exists(template_file_name):
    for dir in forms_repository_files_dirs_:
        for file_ in os.listdir(dir):
            if file_ == template_file_name:
                return True
    return False


def import_form_module(form_no):
    """Импортировать модуль PY файла отчетной формы"""
    for dir_ in forms_repository_files_dirs_:  # TODO: fix
        try:
            return import_module_(dir_, form_no)
        except Exception as e:
            pass


def get_field_order_num(form_name, table_name, field_name):
    """Получить номер порядковый номер столбца в таблице отчной формы"""
    rep = get_form_repository(form_name)

    def make_flat_headers(h, res):
        for h2 in h:
            if isinstance(h2, dict):
                make_flat_headers(h2.values(), res)
            elif isinstance(h2, list):
                make_flat_headers(h2, res)
            else:
                res.append(h2)

    flat_headers = []

    make_flat_headers(rep['tables'][table_name]['headers'], flat_headers)

    for i, model_item in enumerate(flat_headers):
        if model_item == field_name:
            return i
    raise CoreBaseException(message=f'field not found by field_name={field_name} '
                                    f'in table={table_name} in form={form_name}')


def get_model_item_by_field(form_name, table_name, field_name):
    rep = get_form_repository(form_name)
    for model_item in rep['tables'][table_name]['model']:
        if model_item['field'] == field_name:
            return model_item
    raise ModelItemNotFoundException(form_name, table_name, field_name)


def get_column_num(form_no, table_name, column_name, for_front=False):
    """Получить номер стобца отченой формы.
        1. либо явно задан в репозитории
        2. либо получаем порядковый номер"""
    model_item = get_model_item_by_field(form_no,
                                         table_name, column_name)

    if 'col_num' in model_item:
        if model_item['col_num'] == '':
            return None
        else:
            plus = 2 if for_front else 0
            return int(model_item['col_num']) + plus
    else:
        plus = 2 if for_front else 0
        return get_field_order_num(form_no, table_name, column_name) + plus


def get_form_fields(form_repo, table_name: str = 'main_table') -> List[str]:
    """Получить список имен полей формы определенной таблицы.

    Args:
        form_repo: Репозиторий формы.
        table_name: Имя таблицы. По умолчанию = main_table.

    Returns:
        List[str]: Список имен полей формы.

    """
    return [m['field'] for m in form_repo['tables'][table_name]['model']]
