import os.path
import warnings

from repository.areports.areports_repository import init_areports_repository
from repository.decrees.decrees_repository import init_decrees_repository
from repository.exceptions import ModuleTypeNotFoundException, UnknownModuleTypeException
from repository.extensions.extensions_repository import init_extensions_repository
from repository.forms.forms_repository import init_forms_repository
from repository.forms_journals.forms_journals_repository import init_forms_journals_repository
from repository.registers.registers_repository import init_registers_repository
from repository.reports.reports_repository import init_reports_repository
from repository.snk.snk import init_snk_repository
from repository.types.types_repository import init_types_repository
from repository.utils import is_py_file, import_module_
from repository.vocs.vocs_repository import init_vocs_repository


def build_context_for_to_api2(storage):
    api2_context = []
    for k, data in storage.items():
        if k == 'templates':
            continue
        for item in data:
            if k == 'areport':
                config_key = 'AREPORTS_REPOSITORY_FILES_DIRS'
            elif k == 'extensions':
                config_key = 'EXTENSIONS_FILES_DIRS'
            else:
                config_key = 'CUSTOM_FILES_DIRS'
            context = {
                'config_key': config_key,
                'file_path': item['path'],
                'data': item['file_data'],
            }
            api2_context.append(context)
    return api2_context


def _init(path, storage):
    if os.path.isdir(path):
        for sub_path in os.listdir(path):
            _init(os.path.join(path, sub_path), storage)
    elif os.path.isfile(path):
        file_name = os.path.basename(path)

        if file_name.endswith('.docx') or file_name.endswith('.xlsx'):
            storage['templates'].append(path)

        if not is_py_file(file_name):
            return
        try:
            file_name, file_ext = file_name.split('.')
        except Exception:
            return
        pathpath = os.sep.join(os.path.normpath(path).split(os.sep)[:-1])
        file_module = import_module_(pathpath, file_name)
        file_data = file_module.init()

        try:
            module_type = file_data.pop('module_type')
        except KeyError:
            raise ModuleTypeNotFoundException(path)

        if module_type not in storage:
            raise UnknownModuleTypeException(module_type)

        storage[module_type].append({
            'path': path,
            'file_data': file_data,
        })

    else:
        warnings.warn(f'is neither a folder nor a file. path: {path}')


def init(custom_files_dirs, areports_repository_files_dirs, extensions_files_dirs):
    storage = {
        'types': [],
        'areg': [],
        'report_bind': [],
        'report_form': [],
        'areport': [],
        'templates': [],
        'report_forms_journal': [],
        'decree': [],
        'snk': [],
    }
    for dir_ in custom_files_dirs:
        _init(dir_, storage)

    init_types_repository(storage['types'])
    init_reports_repository(storage['report_bind'])
    init_forms_repository(storage['report_form'], storage['templates'])
    init_vocs_repository()
    init_registers_repository(storage['areg'])
    init_forms_journals_repository(storage['report_forms_journal'])
    init_decrees_repository(storage['decree'])
    init_snk_repository(storage['snk'][0])

    return build_context_for_to_api2(storage)
