import copy

from repository.types.exceptions import CustomTypeNotFoundException, TypeRepositoryNotFoundException, \
    BaseOverrideTypeNotFoundException
from repository.types.validation import check_types_configuration

types_repository = {}  # репозиторий типов


def init_types_repository(items):
    types_override = {}

    for item in items:
        file_data = item['file_data']
        path = item['path']

        check_types_configuration(file_data, path)
        types_ = file_data['types']

        for type_, type_data in types_.items():

            if 'override_type' in type_data:
                types_override[type_] = type_data
                continue

            if type_data['base_type'] == 'BASE' and 'validation' in type_data:
                if type_data['validation'].get('type') in ('int', 'float'):
                    # разрешенные значения типов  (не являющиеся значениями этого типа)
                    type_data['validation']['allow_list'] = ['-']
                try:
                    if 'exponential' in type_data['validation']['handlers']:
                        type_data['validation']['allow_list'] = ['-']
                except KeyError:
                    pass

        types_repository.update(types_)

    for override_type, override_type_data in types_override.items():
        base_type_name = override_type_data['override_type']
        try:
            base_type = types_repository[base_type_name]
        except KeyError:
            raise BaseOverrideTypeNotFoundException(base_type_name)
        new_type = copy.deepcopy(base_type)
        new_type.update(override_type_data)
        types_repository[override_type] = new_type


def get_types_repository():
    return types_repository


def get_type_repository(type_name: str):
    """Получить репозиторий типа по `type_name`.
    Возвращает ссылку на репозиторий, изменения, примененные к результату функции, приведут к изменению данных
    репозитория типа во всей системе.

    Args:
        type_name (str): Имя типа (например, PRZN_ACTIV_type).

    Returns:
        Репозиторий типа в формате словаря.

    Raises:
        TypeRepositoryNotFoundException(CoreBaseException): Если репозиторий не найден.

    """
    try:
        return types_repository[type_name]
    except KeyError:
        raise TypeRepositoryNotFoundException(type_name)


def add_extra_types(types):
    """Функция для добавления типов с api2 (в формате словаря).

    Args:
        types: Типы в формате словаря.
    """
    types_repository.update(types)


def get_type_from_form_repository_by_field_name(form_repository, field_name, table_name='main_table'):
    for model_item in form_repository['tables'][table_name]['model']:
        if model_item['field'] == field_name and model_item.get('type'):
            try:
                return types_repository[model_item['type']]
            except KeyError:
                raise CustomTypeNotFoundException(model_item['type'])


def get_python_type_if_repo_type_is_simple(type_name: str):
    """Получить из типа репозитория `type_name` python тип, если тип репозитория является простым.
    Например, если тип репозитория `type_name` = 'int', вернется python тип int.

    Args:
        type_name: Имя типа репозитория.

    Returns:
        Python тип, если тип репозитория простой. Иначе вернется ``None``.
    """
    try:
        type_ = get_types_repository()[type_name]
    except KeyError:
        return None
    repo_types_to_python = {
        'int': int,
        'str': str,
        'float': float,
    }
    try:
        return repo_types_to_python[type_['validation']['type']]
    except KeyError:
        return None
