from typing import Dict

import settings

from repository.exceptions import UnknownSelectModeException
from repository.types import types_repository as types_repository
from repository.vocs.exceptions import CustomVocTypeKeyNotFoundException, \
    VocTypeNotFoundException, VocRepositoryNotFoundException, VocModelItemNotFound
from repository.vocs.validation import check_voc_configuration, check_voc_value_id, check_voc_label

vocs_repository = {}


def init_vocs_repository(*args, **kwargs):
    for type_, type_data in types_repository.get_types_repository().items():
        if type_data['base_type'] != 'VOC':
            continue

        # validation
        check_voc_configuration(type_data, type_)
        check_voc_value_id(type_data['value_id'], type_data['model'])
        check_voc_label(type_data['label'], type_data['model'])

        voc_type = type_data['voc_type']
        voc_name = type_data.get('voc_name')
        if voc_name is None:
            if voc_type not in ['organizations', 'users']:
                raise CustomVocTypeKeyNotFoundException(type_, voc_type)

        if type_data.get('selection_mode'):
            if type_data['selection_mode'] not in ['multiple', 'single']:
                raise UnknownSelectModeException(type_data['selection_mode'])

        # формирование filter и sort - варианты фильтрации для полей справочника (модели) и сортировка
        for model_item in type_data['model']:
            if 'filter' not in model_item:
                model_item['filter'] = ['=', 'like']
            if 'sort' not in model_item:
                model_item['sort'] = True

        if 'sort' in type_data:
            sort_field = type_data['sort'][0]
            asc_or_desc = '-' if sort_field.startswith('-') else ''
            for model_item in type_data['model']:
                if model_item['field'] == sort_field:
                    model_item['default_sort'] = asc_or_desc

        vocs_repository[voc_type] = type_data


def get_vocs_types_list():
    res = []
    for voc_type, voc_repo in vocs_repository.items():
        item = {
            'voc_type': voc_type,
            'voc_name': voc_repo['voc_name'],
            'selection_mode': voc_repo.get('selection_mode', 'single'),
            'translate_value_id': voc_repo.get('translate_value_id', False),
            'form': {
                'table': {
                    'headers': voc_repo['headers'] if voc_repo['headers'] else [m['field'] for m in voc_repo['model']],
                    'model': voc_repo['model'],
                    'options': {
                        'server_filter_sort': False if settings.DB_SYSTEM == 'sqlite' else True,
                    }
                }
            },
        }
        if not voc_repo.get('not_display_in_menu', False):
            item['group'] = voc_repo.get('voc_subtype', 'custom')
        res.append(item)
    return res


def get_vocs_types_dict():
    return {item['voc_type']: item for item in get_vocs_types_list()}


def get_voc_label_rules(voc_type):
    """Получить правила формирования имени записи типа справочника"""
    try:
        return vocs_repository[voc_type]['label']
    except KeyError:
        raise VocRepositoryNotFoundException(voc_type)


def get_voc_repository_f(voc_type):
    voc_repo = get_voc_repository(voc_type)
    res = {
        'types': types_repository.get_types_repository(),
        'form': {
            'table': {
                'headers': voc_repo['headers'] if voc_repo['headers'] else [m['field'] for m in voc_repo['model']],
                'model': voc_repo['model'],
                'options': {
                    'server_filter_sort': False if settings.DB_SYSTEM == 'sqlite' else True,
                }
            }
        },
        'edit': {
            'edit': voc_repo.get('voc_subtype') != 'settings',
        },
        'translate_value_id': voc_repo.get('translate_value_id', False),
    }
    for k in ('lazy_load_size', 'group'):
        if k in voc_repo:
            res[k] = voc_repo[k]
    return res


def get_voc_repository(voc_type: str):
    """Получить репозиторий справочника по `voc_type`.
    Возвращает ссылку на репозиторий, изменения, примененные к результату функции, приведут к изменению данных
    репозитория справочника во всей системе.

    Args:
        voc_type (str): voc_type справочника (например, 23_voc).

    Returns:
        Репозиторий справочника в формате словаря.

    Raises:
        VocRepositoryNotFoundException(CoreBaseException): Если репозиторий справочника не найден.

    """
    try:
        return vocs_repository[voc_type]
    except KeyError:
        raise VocRepositoryNotFoundException(voc_type)


def check_voc_type_existence(voc_type):
    try:
        vocs_repository[voc_type]
    except KeyError:
        raise VocTypeNotFoundException(voc_type)


def get_voc_fields(voc_type):
    try:
        return list(item['field'] for item in vocs_repository[voc_type]['model'])
    except KeyError:
        raise VocTypeNotFoundException(voc_type)


def get_voc_models(voc_type):
    try:
        return [item for item in vocs_repository[voc_type]['model']]
    except KeyError:
        raise VocTypeNotFoundException(voc_type)


def get_model_voc_model_item(voc_type, field):
    try:
        return next(filter(lambda x: x['field'] == field, get_voc_models(voc_type)))
    except StopIteration:
        raise VocModelItemNotFound(voc_type, field)


def get_voc_readonly_fields(voc_type):
    try:
        return list(filter(lambda x: x.get('readonly'),
                           [item for item in vocs_repository[voc_type]['model']]))
    except KeyError:
        raise VocTypeNotFoundException(voc_type)


def get_voc_fields_to_types(voc_repo, include_none_if_type_not_set=False) -> Dict[str, str]:
    """Получить словарь поле_справочника -> имя_типа_поля.

    Args:
        voc_repo: Репозиторий справочника.
        include_none_if_type_not_set: Включать значение ``None``, если у поля нет типа.
            По умолчанию = ``False``.

    Returns:
        Словарь, где ключи - имена полей справочника, значения - имена типов справчника.
    """
    res = {}
    for model in voc_repo['model']:
        type_ = model.get('type', None)
        if type_ is None and not include_none_if_type_not_set:
            continue
        res[model['field']] = type_
    return res


def get_system_voc_types():
    return [voc['voc_type'] for voc in vocs_repository.values() if voc.get('voc_subtype') == 'system']
