from datetime import datetime

from repository.areports import areports_repository
from repository.types import types_repository
from wsgi.apps.areports import services
from wsgi.apps.areports.models import AReportModel
from wsgi.apps.users.models import UserModel
from wsgi.serializers import BaseSerializer


class AReportsSerializer(BaseSerializer):
    model = AReportModel(None)

    auto_commit = True
    create_connect = True

    def __init__(self, request, user: UserModel.Data, areport_repository=None):
        """Сериализатор ан. отчетов.

        Args:
            areport_repository: Необязательный параметр. Если будет передан, для сохранения ан. отчета будет
                использован переданный репозиторий. Формат должен быть таким же, как и у обычных репозиториев ан.
                отчетов. (Необходим, если api2 будет явно передавать репозиторий ан. отчета, который не зарегистрирован
                из rvrao-ext).
        """
        super().__init__(request)
        self.user = user
        self.areport_repository = areport_repository

    def _prepare(self, data):
        areport_rep = self.areport_repository if self.areport_repository is not None else\
            areports_repository.get_areport_repository(data['areport_type'])
        result = data.pop('result')
        # build repository
        headers = []
        if result['data']:
            headers = [item['field'] for item in areport_rep['table']['model'] if item['field'] in result['data'][0]]
        repository = {
            'tables': {
                'form': {
                    'tab_name': areport_rep['name'],
                    'headers': headers,
                    'model': areport_rep['table']['model'],
                }
            }
        }
        # build data
        data_ = {
            'form_table_data': {
                'form': result['data']
            }
        }
        # total
        data.update(
            {
                'areport_data': {
                    'repository': repository,
                    'data': data_,
                    'areport_params': {k: v for k, v in data.items()}
                }
            }
        )

    def _response(self, resp):
        params = tuple(filter(lambda x: x != 'areport_data', resp.keys()))  # all params except 'areport_data'
        res = resp['areport_data']
        for param in params:
            res['data'][param] = resp[param]
        for param in params:
            resp.pop(param)
        res['repository']['types'] = types_repository.get_types_repository()
        res['data']['areport_params'] = res.pop('areport_params')
        res['data']['edit'] = services.calc_permissions_for_areport_actions(self.user)
        return res

    def save(self, data):
        obj = self.model.Data.from_dict(data)
        obj.areport_period = {
            'to': '',
        }
        obj.areport_creator = self.user.record_id
        obj.areport_state = self.model.Data.AReportStateType.TEMP
        obj.areport_ts = datetime.now()
        obj.areport_name = obj.areport_name + ' ' + obj.areport_ts.strftime("%d.%m.%Y %H:%M")
        return obj

    def update(self, record_id, data):
        instance = self.model.read(record_id)
        obj = instance.replace(data)
        obj.areport_ts = datetime.now()
        obj.areport_period = {
            'to': '',
        }
        obj.areport_state = self.model.Data.AReportStateType.TEMP
        obj.areport_name = obj.areport_name + ' ' + instance.areport_ts.strftime("%d.%m.%Y %H:%M")
        return obj


class AReportsListFilterSerializer(BaseSerializer):
    model = AReportModel(None)
    fields = ['areport_id', 'areport_creator', 'areport_name',
              'areport_state', 'areport_ts']

    def __init__(self, request, user: UserModel.Data):
        super().__init__(request)
        self.user = user

    def _response(self, resp):
        resp = super()._response(resp)
        resp['edit'] = services.calc_permissions_for_areport_actions(self.user)
        return resp

    def list(self):
        if self.user.is_admin or not self.user.organizations_ids:
            return super().list()

        users_ids_with_same_org_id = [user_raw['record_id'] for user_raw in
                                      UserModel(self.request).filter_raw({'org_id': self.user.org_id}, fields_=['record_id'])]
        return super().filter({'areport_creator': users_ids_with_same_org_id})
