from exportmanager.export_manager import export_areport_to_xlsx
from services.exceptions import CoreBaseException
from services.temp import TempFile
from settings import TEMP_DIR_PATH
from wsgi.apps.areports import services
from wsgi.apps.areports.models import AReportModel
from wsgi.apps.areports.permissions import check_create_areports_permission
from wsgi.apps.areports.rules import areports_app_rules
from wsgi.apps.areports.serializers import AReportsSerializer, AReportsListFilterSerializer
from wsgi.permissions import bool_permission
from wsgi.utils import response, parse_request, api_view, file_response


@parse_request(areports_app_rules)
def areports_app_view_manager(request, *args, **kwargs):
    action = kwargs['action']
    # === GET ===
    if action == 'list':
        return list_areports_view(request)
    elif action == 'read':
        return read_areports_view(request, kwargs['areport_id'])
    elif action == 'export_to_xlsx':
        return export_to_xlsx_areport_view(request, kwargs['areport_id'])
    elif action == 'get_extra_info':
        return get_extra_info_view(request)
    # === POST ===
    elif action == 'create':
        return create_areports_view(request, kwargs['body'])
    elif action == 'delete':
        return delete_areports_view(request, kwargs['areport_id'])
    elif action == 'to_save':
        return to_save_areport_view(request, kwargs['areport_id'], kwargs['body'])
    elif action == 'to_publish':
        return to_publish_areport_view(request, kwargs['areport_id'], kwargs['body'])
    elif action == 'recalculate':
        return recalculate_areport_view(request, kwargs['areport_id'], kwargs['body'])
    elif action == 'to_archive':
        return to_archive_areport_view(request, kwargs['areport_id'])


@api_view()
def list_areports_view(request, *args, **kwargs):
    """Получить список аналитических отчетов"""
    return response(request, AReportsListFilterSerializer(request, kwargs['user']).list(), many=True)


@api_view()
def read_areports_view(request, areport_id, *args, **kwargs):
    """Получить аналитический отчет по `areport_id`"""
    return response(request, AReportsSerializer(request, kwargs['user']).read(areport_id))


@api_view()
def to_save_areport_view(request, areport_id, body, *args, **kwargs):
    """-Сохранить-. Изменяет статус на -Сохранён-, доступен когда статус -Временный-"""
    areport = AReportModel(request).read(areport_id)
    if areport.areport_state != AReportModel.Data.AReportStateType.TEMP:
        raise CoreBaseException(rus_message='Действие "Сохранить" доступно только для аналитических отчетов, '
                                            'имеющих статус "Временный"')
    areport.areport_state = AReportModel.Data.AReportStateType.SAVED
    areport.areport_name = body['areport_name']
    AReportModel(request).update(areport.areport_id, areport)
    return response(request, AReportsSerializer(request, kwargs['user']).read(areport_id))


@api_view()
def to_publish_areport_view(request, areport_id, body, *args, **kwargs):
    """-Опубликовать-. Изменяет статус на -Опубликован-, доступен когда статус -Временный- или -Сохранён-"""
    areport = AReportModel(request).read(areport_id)
    st = AReportModel.Data.AReportStateType
    allowed_statuses = (st.TEMP, st.SAVED)
    if areport.areport_state not in allowed_statuses:
        raise CoreBaseException(rus_message='Действие "Опубликовать" доступно только для аналитических отчетов, '
                                            'имеющих статус "Временный" или "Сохранён"')
    areport.areport_state = st.PUBLISHED
    areport.areport_name = body['areport_name']
    AReportModel(request).update(areport.areport_id, areport)
    return response(request, AReportsSerializer(request, kwargs['user']).read(areport_id))


@api_view()
def to_archive_areport_view(request, areport_id, *args, **kwargs):
    """Действие -Архивировать-. Аналитический отчёт изменяет статус на -Архивирован-"""
    areport = AReportModel(request).read(areport_id)
    areport.areport_state = AReportModel.Data.AReportStateType.ARCHIVED
    AReportModel(request).update(areport.areport_id, areport)
    return response(request, AReportsSerializer(request, kwargs['user']).read(areport_id))


@api_view()
def create_areports_view(request, body, *args, **kwargs):
    """Сформировать аналитический отчет"""
    # permissions
    check_create_areports_permission(kwargs['user'])
    # view
    body['result'] = services.calculate_areport(body)
    return response(request, AReportsSerializer(request, kwargs['user']).save(body))


@api_view()
def delete_areports_view(request, areport_id, *args, **kwargs):
    """Удалить ананалитический отчет"""
    # permissions
    check_create_areports_permission(kwargs['user'])
    # view
    return response(request, AReportsSerializer(request, kwargs['user']).delete(areport_id))


@api_view()
def recalculate_areport_view(request, areport_id, body, *args, **kwargs):
    """Пересчитать аналитический отчет с `areport_id`"""
    # permissions
    check_create_areports_permission(kwargs['user'])
    # view
    body['result'] = services.calculate_areport(body)
    return response(request, AReportsSerializer(request, kwargs['user']).update(areport_id, body))


@api_view()
def export_to_xlsx_areport_view(request, areport_id, *args, **kwargs):
    """Экспортировать анал. отчет в xlsx файл"""
    areport = AReportsSerializer(request, kwargs['user']).read(int(areport_id))
    with TempFile(TEMP_DIR_PATH, '.xlsx') as res_file:
        export_areport_to_xlsx(areport, res_file.full_path)
        return file_response(res_file.full_path)


@api_view()
def get_extra_info_view(request, *args, **kwargs):
    """Получить доп. информацию о аналитических отчетах"""
    resp = {
        'edit': {
            'create': bool_permission(lambda: check_create_areports_permission(kwargs['user'])),
        }
    }
    return response(request, resp)
