from secrets import token_urlsafe

from werkzeug.security import check_password_hash

from services.exceptions import CoreBaseException
from wsgi.apps.auth.rules import auth_rules
from wsgi.apps.user_logs.services import UserLogsLogger
from wsgi.apps.users.models import UserModel
from wsgi.apps.users.serializers import UsersSerializer
from wsgi.utils import parse_request, api_view, response, error_response


@parse_request(auth_rules)
def auth_view_manager(request, *args, **kwargs):
    action = kwargs['action']
    # === GET ===
    if action == 'get_current_user':
        return get_current_user_view(request)
    # === POST ===
    elif action == 'login':
        return login_view(request, kwargs['body'])
    elif action == 'logout':
        return logout_view(request)


@api_view(auth_required=False)
def login_view(request, body, *args, **kwargs):
    users = UserModel(request).filter({'login': body['login']})
    if not users:
        UserLogsLogger(request, None).log_incorrect_login(body['login'])
        raise CoreBaseException(rus_message='Пользователь с таким логином не найден')
    if len(users) > 1:
        raise CoreBaseException(rus_message='Вход невозможен. В системе несколько пользоватлей с таким логином')
    user = users[0]
    if not check_password_hash(user.password, body['password']):
        UserLogsLogger(request, user).log_incorrect_password(body['password'])
        raise CoreBaseException(rus_message='Неверный пароль')

    cookie_value = token_urlsafe(120)
    user.cookie = cookie_value
    UserModel(request).update(user.record_id, user)

    resp = response(request, {'fio': user.fio})
    resp.set_cookie('Auth', cookie_value, max_age=60 * 60 * 10)  # max age of cookie = 10h
    # resp.set_cookie('Auth', cookie_value)  # max age of cookie = 10h
    UserLogsLogger(request, user).log_successful_login_action()
    return resp


@api_view(auth_required=False)
def logout_view(request, *args, **kwargs):
    cookie = request.cookies.get('Auth')
    if cookie is None:
        raise CoreBaseException(rus_message='Вы не авторизованы')
    users = UserModel(request).filter({'cookie': cookie})
    if not users:
        resp = error_response(
            {
                'rus_message': 'Ошибка при выходе. Пользователь не найден',
            }
        )
        resp.set_cookie('Auth', cookie, max_age=0)
        return resp
    user = users[0]
    user.cookie = None
    UserModel(request).update(user.record_id, user)
    resp = response(request, {'message': 'OK'})
    # resp.set_cookie('Auth', cookie, max_age=0)
    # resp.set_cookie('Auth', cookie)
    resp.delete_cookie('Auth')
    UserLogsLogger(request, user).log_logout()
    return resp


@api_view()
def get_current_user_view(request, *args, **kwargs):
    """Получить данные текущего пользователя из куков авторизации"""
    cookie = request.cookies.get('Auth')
    if cookie is None:
        raise CoreBaseException(rus_message='Не найдены cookie авторизации')
    return response(request, UsersSerializer(request).filter({'cookie': cookie}))
