import copy
import datetime as dt
import hashlib
import json
import sys
import time
from collections import defaultdict
from datetime import datetime
from typing import Dict

from loguru import logger

from datasource.databaserepository.exceptions import RecordNotFoundException
from repository.forms import forms_repository
from repository.vocs import vocs_repository
from services.exceptions import CoreBaseException
from services.services import to_exp, round_dt_by_seconds
from wsgi.apps.historical_data.services.decrees.utils import get_radionuclides_voc_data_from_csv, \
    get_table_names_from_mdb, get_reader_from_mdb_gen
from wsgi.apps.reporting_forms.models import FormModel
from wsgi.apps.services import save_import_item
from wsgi.apps.user_logs.services import UserLogsLogger
from wsgi.apps.voc.models import OrganizationVocModel, VocValueModel
from wsgi.apps.voc.services import build_voc_label


class BaseDecree:

    PROGRESS_MESSAGES = 7

    form_rows_rules = {}

    decree_year = None

    def __init__(self, thread_runner, request, user, decrees, decree_name, db_file_path, db_file_type=None, api2_obj=None, done=None):
        if db_file_type is None:
            raise CoreBaseException(message='db_file_type can not be None')

        self.done = done

        self.db_file_path = db_file_path

        self.thread_runner = thread_runner
        self.forms_map_system_to_mdb = {}

        logger.info('=========COLLECTING DATA=========')
        self.is_firebird = False
        if db_file_type == 'mdb':
            self.mdb_tables_headers = {}
            self.table_names = get_table_names_from_mdb(db_file_path)
        elif db_file_type == 'firebird':
            self.is_firebird = True
            self.table_names = list(self.form_rows_rules).extend(['ReportCollection_DbSet', 'ReportsCollection_DbSet', 'notes', 'form_10', 'form_20'])
        else:
            raise CoreBaseException(message=f'Unknown db_file_type: {db_file_type}')

        self.radios = get_radionuclides_voc_data_from_csv()

        self.request = request
        self.decree_name = decree_name
        self.decree_num = None

        self.orgs: Dict[int, dict] = {}
        self.result_forms: Dict[int, dict] = {}

        self.inited_forms_ids = {}
        self.tables_cols_order = {}

        self.org_repository = vocs_repository.get_voc_repository('organizations')
        self.user = user

        # даты действия организации относительно приказа
        self.org_period_from = None
        self.org_period_to = None

        self.api2_obj = api2_obj

        self.import_ts = round_dt_by_seconds(dt.datetime.now())
        self.result_forms_report_ids = []
        self.forms_imported_before = []
        self.forms_updated = []
        self.ignored_forms_count = 0

        self.orgs_imported_ids = []

        self.SYSTEM_FORM_NAMES_TO_MDB = {}
        self.MDB_FORM_NAMES_TO_SYSTEM = {}

        self.user_log = UserLogsLogger(self.request, self.user).log_import_historical_data(
            self.result_forms_report_ids, self.orgs_imported_ids,
        )

        self.voc_repos = {}  # remove
        self.system_vocs_types = vocs_repository.get_system_voc_types()

        self.voc_import_hash_cash = defaultdict(set)

    def after_init(self):
        for mdb_table_name, rule in self.form_rows_rules.items():
            self.MDB_FORM_NAMES_TO_SYSTEM[mdb_table_name] = rule['form_no']
        self.SYSTEM_FORM_NAMES_TO_MDB = {v: k for k, v in self.MDB_FORM_NAMES_TO_SYSTEM.items()}

    def _set_table_cols_order(self, table_name):
        if table_name not in self.tables_cols_order:
            self.tables_cols_order[table_name] = self.mdb_tables_headers[table_name][2:]

    def check_done(self):
        if self.done.wait(.1):
            logger.info('=========<завершение_выполнения_потока>=========')
            sys.exit()

    def get_voc_repo(self, voc):
        if voc not in self.voc_repos:
            self.voc_repos[voc] = vocs_repository.get_voc_repository(voc)
        return self.voc_repos[voc]

    def _to_exp(self, value):
        return to_exp(value)

    def _to_date(self, value):
        if isinstance(value, (dt.datetime, dt.date)):
            return value.strftime('%Y-%m-%d')

        if not value:
            return value

        try:
            m, d, y = value.split()[0].split('/')
        except ValueError:
            return value
        year = '19' if int(y) > 40 else '20'
        return f'{year}{y}-{m}-{d}'

    def _round_to_int_as_str(self, value):
        try:
            value = value.replace(',', '.')
            return str(int(float(value)))
        except ValueError:
            return value

    def _to_date_as_datetime_obj(self, value):
        if isinstance(value, (dt.datetime, dt.date)):
            return value
        formats = ('%m/%d/%y %H:%M:%S', )
        for f in formats:
            try:
                return datetime.strptime(value.strip(), f)
            except ValueError:
                pass
        raise ValueError('can not convert to _to_date_as_datetime_obj, value: ', value)
        # return datetime.strptime(value, '%m/%d/%y %H:%M:%S')

    def to_date_obj_if_match_format_else_none(self, value):
        v = self._to_date(value)
        try:
            dt.date.fromisoformat(v)
            return v
        except ValueError:
            return None

    def _build_str_from_dict_by_keys(self, source_dict, keys, sep=', '):
        return sep.join(str(source_dict[k]) for k in keys if source_dict[k] is not None)

    def _build_import_author_id(self, id_, cod_org, okpo):
        return f'org_id={self.decree_num}/{cod_org}/{okpo}'

    def _build_set(self, row, keys, sep=';'):
        res = set()
        for k in keys:
            for item in row[k].split(sep):
                res.add(item)
        return f'{sep} '.join(res)

    def _fill_by_dash(self, row, keys):
        for k in keys:
            row[k] = '-'

    def _convert_row_data_to_str(self, row):
        """Приводим все данные строки формы к str типу (на win данные могут иметь разные типы,
        на linux все переводится в строку, сделано, для единообразности обработки данных).
        Даты в win оставляем датами, для простоты обработки."""
        for k, v in row.items():
            if isinstance(v, (dt.datetime, dt.date)):
                continue
            if v is None:
                row[k] = ''
                continue
            row[k] = str(v)

    def _init_form(self, form_id, form, form_repository, tables_keys):
        if self.inited_forms_ids.get(int(form_id)):
            return

        form['form_state'] = FormModel.Data.FormStateStatuses.LOADED.value
        if not form.get('form_table_data'):
            form['form_table_data'] = {
                'main_table': [],
            }

        tables_keys = ('main_table',)
        if not form.get('form_remarks'):
            form['form_remarks'] = {k: [] for k in tables_keys}
        if not form.get('form_warnings'):
            form['form_warnings'] = {k: [] for k in tables_keys}
        if not form.get('form_comments'):
            form['form_comments'] = {k: [] for k in tables_keys}
        if not form.get('form_main_data'):
            form['form_main_data'] = {}

        self.inited_forms_ids[int(form_id)] = True

    def _fill_main_attrs_by_dash(self, form):
        form_repository = forms_repository.get_form_repository(form['form_no'])
        if 'form_main_attrs' in form_repository:
            for m in form_repository['form_main_attrs']:
                if not form['form_main_data'].get(m['field']):
                    form['form_main_data'][m['field']] = '-'

    def FORMP_convert(self, form, row_mdb, org: OrganizationVocModel.Data, decree_year=None):
        mdb_form_names_to_system = {
            'ZRI': 'form_1_1',
            'IOU': 'form_1_2',
            'ORI': 'form_1_3',
            'ORP': 'form_1_4',
            'ROZ': 'form_1_5',
            'RAO': 'form_1_6',
        }

        # form.form_period_start = datetime.strptime(row_mdb['F_BEG'], '%m/%d/%y %H:%M:%S')
        # form.form_period_end = datetime.strptime(row_mdb['F_END'], '%m/%d/%y %H:%M:%S')

        # form.form_period_start = self._to_date_as_datetime_obj(row_mdb['F_BEG'])
        # form.form_period_end = self._to_date_as_datetime_obj(row_mdb['F_END'])
        form['form_period_start'] = self._to_date(row_mdb['F_BEG'])
        form['form_period_end'] = self._to_date(row_mdb['F_END'])

        system_form_no = mdb_form_names_to_system[row_mdb['FORM_NAME']]

        if self.decree_year is not None:
            form['form_no'] = system_form_no + '_' + self.decree_year
            form_repository = forms_repository.get_form_repository(form['form_no'])
            form['form_name'] = form_repository['form_name']

        if decree_year is not None:
            form['form_no'] = system_form_no + '_' + decree_year
            form_repository = forms_repository.get_form_repository(form['form_no'])
            form['form_name'] = form_repository['form_name']

        self._fill_main_attrs_by_dash(form)

    def Forms_convert(self, form, row_mdb, org: OrganizationVocModel.Data, decree_year=None):
        mdb_form_names_to_system = {
            '1': 'form_2_1',
            '2': 'form_2_2',
            '3': 'form_2_3',
            '4': 'form_2_4',
            '5': 'form_2_5',
            '6': 'form_2_6',
            '7': 'form_2_7',
            '8': 'form_2_8',
            '9': 'form_2_9',
            '10': 'form_2_10',
            '11': 'form_2_11',
        }

        form['reporting_year'] = int(row_mdb['fYEAR'])

        system_form_no = mdb_form_names_to_system[row_mdb['FORM_NAME']]

        if self.decree_year is not None:
            form['form_no'] = system_form_no + '_' + self.decree_year
            form_repository = forms_repository.get_form_repository(form['form_no'])
            form['form_name'] = form_repository['form_name']

        if decree_year is not None:
            form['form_no'] = system_form_no + '_' + decree_year
            form_repository = forms_repository.get_form_repository(form['form_no'])
            form['form_name'] = form_repository['form_name']

        self._fill_main_attrs_by_dash(form)

    def _import_main_form_data_row(self, row, convert_func, user=None, decree_year=None):
        user = self.user if user is None else user

        # self.result_forms[int(row['ID'])] = form = FormModel.Data.empty()
        self.result_forms[int(row['ID'])] = form = {}

        form['adjustment_no'] = int(row['FORM_Corr'])
        form['form_main_data'] = {
            'position': row.get('FILL_DLG', ''),
            'full_name': row['FILL_FIO'],
            'phone': row['FILL_PHONE'],
            'email': row['FILL_MAIL'],
        }
        form['user_id'] = user.record_id
        # задаем id организации как указано в mdb. в конце обработки будет заменено на id в системе
        form['author_id'] = int(row['IDP'])
        # сохраняем в базе id формы из исходной базы
        form['import_src_id'] = row['ID']
        # org_id=XXX/YYY, где XXX - org_id из БД, YYY - номер импортируемого приказа (449, 600, 24, 19, 13).
        # Пример: import_author := org_id=001/24 или org_id=555/600
        # form.import_author = self._get_import_author_id(row['ID'])  # f'org_id={self.decree_num}/{row["ID"]}'

        form['form_import_ts'] = self.import_ts

        form['form_table_data'] = {'main_table': []}
        form['form_remarks'] = {'main_table': []}
        form['form_comments'] = {'main_table': []}

        org = self.orgs[int(row['IDP'])]

        if convert_func is not None:
            convert_func(form, row, org, decree_year)

    def _import_main_form_data(self, reader, table_name, convert_func=None):
        for row in get_reader_from_mdb_gen(self.db_file_path, table_name, self.mdb_tables_headers):
            self._set_table_cols_order(table_name)
            self._import_main_form_data_row(row, convert_func)

    def _import_forms_rows(self, reader, form_no, table_name,
                           # model_fields, field_to_model_item, tables_keys,
                           map_=None, convert_func=None, id_key='IDF'):

        form_repository = forms_repository.get_form_repository(form_no)
        model_fields = [model_item['field'] for model_item in
                        form_repository['tables']['main_table']['model']]
        field_to_model_item = {model_item['field']: model_item for model_item in
                               form_repository['tables']['main_table']['model']}
        tables_keys = list(form_repository['tables'].keys())

        row_count = 0

        for row in get_reader_from_mdb_gen(self.db_file_path, table_name, self.mdb_tables_headers):
            row_count += 1
            if row_count % 10000 == 0:
                self.check_done()

            self._set_table_cols_order(table_name)

            self._convert_row_data_to_str(row)

            for k, v in row.items():
                if isinstance(v, str) and 'e' in v:
                    try:
                        row[k] = float(v)
                    except:
                        pass

            idf = int(row[id_key])

            try:
                form = self.result_forms[idf]
            except KeyError:
                continue

            if form_no not in form['form_no']:
                continue

            self._init_form(row[id_key], form, form_repository, tables_keys)

            row_data = {}  # данные итоговой строки
            for field_name in model_fields:
                if map_ is not None and field_name in map_:
                    field_name_mdb = map_[field_name]
                    value = row.get(field_name_mdb, '')
                else:
                    value = row.get(field_name, '')

                if value:
                    model_item = field_to_model_item[field_name]
                    # приведение к типамvoc_filter
                    if 'type' in model_item:
                        if model_item['type'] == 'date':
                            value = self._to_date(value)
                else:
                    value = ''

                if isinstance(value, str):
                    value = value.strip()

                row_data[field_name] = value

            if convert_func is not None:
                try:
                    convert_func(row_data, row, form)
                except Exception as e:
                    raise CoreBaseException(message=f'Exception message: {e}, Class: {e.__class__}, \n'
                                                    f'row_data: {row_data},\n row: {row},\n form: {form},\n\n'
                                                    f'form_repo: {form_repository}\n'
                                                    f'convert_func: {convert_func}')

                for k, v in row_data.items():
                    if v == '':
                        row_data[k] = '-'

            try:
                # помещаем строку по индексу
                form['form_table_data'].setdefault('main_table', []).insert(int(row['Npp']) - 1, row_data)
            except (KeyError, ValueError):  # не задан порядок строки
                form['form_table_data'].setdefault('main_table', []).append(row_data)

    def _import_orgs(self, reader, table_name):
        # ДЛЯ ПРИКАЗА 13 СВОЙ МЕТОД В КЛАССЕ ПО СОЗДАНИЮ ОРГАНИЗАЦИИ
        orgs_to_set_KOD_GOL_ORG_ids = {}

        for row in get_reader_from_mdb_gen(self.db_file_path, table_name, self.mdb_tables_headers):
            self._set_table_cols_order(table_name)

            if not row['PRZN_ORG']:
                row['PRZN_ORG'] = '4'

            row.pop('PRZN_PRBL', None)

            if row['PRZN_OP']:
                row['PRZN_OP'] = int(row['PRZN_OP'])
            else:
                row['PRZN_OP'] = 0  # todo: возможно неверно

            if not row['DT_KORR']:
                row['DT_KORR'] = None
            else:
                # row['DT_KORR'] = self._to_date(row['DT_KORR'])
                row['DT_KORR'] = self.to_date_obj_if_match_format_else_none(row['DT_KORR'])

            if not row['DT_REG']:
                row['DT_REG'] = None
            else:
                row['DT_REG'] = self.to_date_obj_if_match_format_else_none(row['DT_REG'])
            if not row['DT_SU']:
                row['DT_SU'] = None
            else:
                row['DT_SU'] = self.to_date_obj_if_match_format_else_none(row['DT_SU'])
            if not row['LVL_ORG']:
                row['LVL_ORG'] = None

            row['attrs'] = {}
            not_include_fields = ('ID', 'INDEX_U', 'GOROD_U', 'ULITSA_U', 'DOM_U', 'KORPUS_U',
                                  'INDEX_P', 'GOROD_P', 'ULITSA_P', 'DOM_P', 'KORPUS_P',
                                  'DOLJN', 'FIO')  # атрибуты, которые не надо включать в attrs
            for attr, value in row.items():
                # если поле нет в системных атрибутах, записываем его в поле attrs
                # например, информационные поля - email, второй телефон и т.д.
                if attr not in not_include_fields and attr not in OrganizationVocModel.Data.fields_except_id():
                    row['attrs'][attr] = value

            org = copy.deepcopy(row)
            org['ADDRESS_U'] = self._build_str_from_dict_by_keys(row, ('INDEX_U', 'GOROD_U', 'ULITSA_U', 'DOM_U', 'KORPUS_U'))
            org['ADDRESS_P'] = self._build_str_from_dict_by_keys(row, ('INDEX_P', 'GOROD_P', 'ULITSA_P', 'DOM_P', 'KORPUS_P'))
            org['RUKOVODITE'] = self._build_str_from_dict_by_keys(row, ('DOLJN', 'FIO'))

            try:
                org['KOD_SRF'] = org['KOD_SRF'].zfill(2)
            except:
                pass

            try:
                org['KOD_IAE'] = org['KOD_IAE'].zfill(2)
            except:
                pass

            org['voc_type'] = 'organizations'
            org['label'] = build_voc_label(org, self.org_repository, attrs_wrap=False)

            org['import_src_id'] = row['ID']
            org['import_author'] = self._build_import_author_id(row['ID'], org['KOD_ORG'], org['OKPO'])

            org['valid_from'] = self.org_period_from
            org['valid_to'] = self.org_period_to
            org['value_status'] = VocValueModel.Data.ValueStatusesTypes.OUTDATED.value

            if row['PRZN_OP'] and int(row['PRZN_OP']) == 1:
                orgs_to_set_KOD_GOL_ORG_ids[int(row['ID_GOL_ORG'])] = org

            self.orgs[int(row['ID'])] = org

        for row in get_reader_from_mdb_gen(self.db_file_path, table_name, self.mdb_tables_headers):
            row_id_int = int(row['ID'])
            if row_id_int in orgs_to_set_KOD_GOL_ORG_ids:
                orgs_to_set_KOD_GOL_ORG_ids[row_id_int]['KOD_GOL_ORG'] = row['KOD_ORG']

    def _import_form_comment(self, reader, table_name):
        for row in get_reader_from_mdb_gen(self.db_file_path, table_name, self.mdb_tables_headers):
            self._set_table_cols_order(table_name)
            try:
                form = self.result_forms[int(row['IDF'])]
            except KeyError:
                # print(
                #     f'WARNING: _import_form_comments строка ссылается по несуществующему ключ в основной таблице формы\n'
                #     f'\t{int(row["IDF"])}')
                continue

            form['form_comment'] = {
                'text': row['PRIM'],
            }

    def _add_remark(self, form, remark, cols_info, col):

        form_no = form['form_no']
        if form_no.count('_') == 3:
            form_no = '_'.join(form_no.split('_')[:-1])

        mdb_table_name = self.SYSTEM_FORM_NAMES_TO_MDB[form_no]
        mdb_col_name = self.mdb_tables_headers[mdb_table_name][col - 1]
        try:
            system_col_name = self.forms_map_system_to_mdb[mdb_table_name]['map'][mdb_col_name]
        except KeyError:
            system_col_name = mdb_col_name

        # этих полей больше нет в репозитории, переносим примечания в 'Основные радионуклиды'
        if system_col_name in ('RAOCodCat', 'RAOCodDang'):
            system_col_name = 'RAOCodMax'

        remark['pointer']['col_name'] = system_col_name
        form['form_remarks']['main_table'].append(remark)

    def _import_cell_remarks(self, reader, table_name):
        # for row in self.get_reader(table_name):
        for row in get_reader_from_mdb_gen(self.db_file_path, table_name, self.mdb_tables_headers):
            self._set_table_cols_order(table_name)
            try:
                form = self.result_forms[int(row['IDF'])]
            except KeyError:
                # print(
                #     'WARNING: _import_cell_remarks строка ссылается по несуществующему ключ в основной таблице формы')
                continue

            # Ищем название таблицы в их базе по номеру формы
            mdb_table_name = None
            for k, v in self.form_rows_rules.items():
                form_no = form['form_no']
                if form_no.count('_') == 3:
                    form_no = '_'.join(form_no.split('_')[:-1])
                if v['form_no'] == form_no:
                    mdb_table_name = k
                    break

            # Порядок имен столбцов в таблице формы
            # print("\tmdb_table_name", mdb_table_name, self.tables_cols_order)
            try:
                table_cols_order = self.tables_cols_order[mdb_table_name]
            except:
                continue

            if not row['Nrow']:
                # print('_import_cell_remarks не указан Nrow, примечание не будет добавлено')
                continue

            remark = {
                'value': {
                    'text': row.get('PRIM', '')
                },
                'pointer': {}
            }

            if isinstance(row['Nrow'], str) and (',' in row['Nrow'] or '-' in row['Nrow'] or '.' in row['Nrow']):
                try:
                    if ',' in row['Nrow']:
                        rs = row['Nrow'].split(',')
                        rs = [int(r) for r in rs]
                    elif '.' in row['Nrow']:
                        rs = row['Nrow'].split('.')
                        rs = [int(r) for r in rs]
                    else:
                        from_, to_ = row['Nrow'].split('-')
                        rs = [int(r) for r in range(int(from_), int(to_))]
                    remarks = []
                    for r in rs:
                        remark_copy = copy.deepcopy(remark)
                        remark_copy['pointer']['row'] = r - 1
                        remarks.append(remark_copy)
                except ValueError:
                    pass
            else:
                remark['pointer']['row'] = int(row['Nrow']) - 1
                remarks = [remark]

            for remark in remarks:
                if row['Ncol'] and row['Ncol'] not in (1, '1'):
                    if isinstance(row['Ncol'], str) and ',' in row['Ncol']:
                        cols_int_values = row['Ncol'].split(',')
                        cols_int_values = [int(col) for col in cols_int_values]
                        for col_int_value in cols_int_values:
                            self._add_remark(form, remark, table_cols_order, col_int_value)
                    else:
                        try:
                            self._add_remark(form, remark, table_cols_order, int(row['Ncol']))
                        except (IndexError, ValueError):
                            pass
                            # print('_import_cell_remarks не найден Ncol (несуществующий номер столбца), примечание не будет добавлено')
                else:
                    # не примечание, а комментарий (для строк нет примечаний, только комменты)
                    form['form_comments']['main_table'].append(remark)

    def _save_vocs(self):
        """"""
        ...

    CUSTOM_VOC_ADDING_RULES = {
        'form_1_1': [
            {
                'voc': '27_voc',
                # form_attr -> voc_attr
                'voc_attrs': {
                    'PrName': 'x',
                    'UktPrTyp': '2',
                },
            }
        ],
        'form_1_2': [
            {
                'voc': '24_voc',
                'voc_attrs': {
                    'Typ': '6',
                    'Kg': '8',
                },
            },
            {
                'voc': '27_voc',
                'voc_attrs': {
                    'PrName': 'x',
                    'UktPrTyp': '2',
                }
            }
        ],
        'form_1_3': [
            {
                'voc': '27_voc',
                'voc_attrs': {
                    'PrName': 'x',
                    'UktPrTyp': '2',
                }
            }
        ],
        'form_1_4': [
            {
                'voc': '27_voc',
                'voc_attrs': {
                    'PrName': 'x',
                    'UktPrTyp': '2',
                }
            }
        ],
        'form_1_5': [
            {
                'voc': '21_voc',
                'voc_attrs': {
                    'PH_Name': '6',
                    'PH_Cod': '7',
                }
            },
            {
                'voc': '27_voc',
                'voc_attrs': {
                    'PrName': 'x',
                    'UktPrTyp': '2',
                }
            }
        ],
        'form_1_6': [
            {
                'voc': '21_voc',
                'voc_attrs': {
                    'PH_Name': '6',
                    'PH_Cod': '7',
                }
            },
            {
                'voc': '27_voc',
                'voc_attrs': {
                    'PrName': 'x',
                    'UktPrTyp': '2'
                }
            }
        ],
        'form_1_7': [
            {
                'voc': '21_voc',
                'voc_attrs': {
                    'g19': '6',
                    'g20': '7',
                }
            }
        ],
        'form_1_8': [
            {
                'voc': '21_voc',
                'voc_attrs': {
                    'g16': '6',
                    'g17': '7',
                }
            }
        ],
        'form_2_1': [
            {
                'voc': '28_voc',
                'voc_attrs': {
                    'g2': '1',
                    'g3': '2',
                    'g19': '3',
                }
            }
        ],
        'form_2_2': [
            {
                'voc': '27_voc',
                'voc_attrs': {
                    'g11': 'x',
                    'g12': '2'
                }
            },
            {
                'voc': '21_voc',
                'voc_attrs': {
                    'g2': '6',
                    'g3': '7',
                }
            }
        ],
    }

    def _save_data(self):
        logger.info('=========SAVE DATA=========')
        logger.info('--save orgs--')
        start_time = time.time()
        org_model = OrganizationVocModel(self.request)
        self.thread_runner.status = 'Сохранение организаций'
        count = 0

        okpos_to_orgs_for_449 = {}
        fake_org_id_to_real_for_13 = {}

        org_model_fields = OrganizationVocModel.Data.fields()

        for org_id, org in self.orgs.items():
            self.thread_runner.status = f'Сохранение организаций {count + 1}/{len(self.orgs)}'
            self._set_progress(5, 'Сохранение организаций', count, len(self.orgs))
            org['import_hash'] = hashlib.md5(json.dumps(org, default=str).encode('utf-8')).hexdigest()
            org['log_id'] = self.user_log.record_id

            keys_to_pop = []
            for k, v in org.items():
                if k not in org_model_fields:
                    keys_to_pop.append(k)
            for k in keys_to_pop:
                org.pop(k)

            res_, org_imported = save_import_item(self.request, org, OrganizationVocModel, commit=True)
            # print(f'Сохранение организации, результат: {res_}')
            if res_ != 'found':  # если произошло сохранение или обновление организации
                org = org_imported
                org['value_id'] = org['record_id']
                org = org_model.update(org['record_id'], org, commit=True, return_as_dict=True)
                self.orgs_imported_ids.append(org['record_id'])
            self.orgs[org_id] = org_imported
            # if self.decree_num == '449':
            okpos_to_orgs_for_449[org['OKPO']] = org_imported
            if self.decree_num == '13':
                fake_org_id_to_real_for_13[org_id] = org_imported
            count += 1
        logger.info("--- %s seconds ---" % (time.time() - start_time))

        if self.decree_num == '13':
            for k, v in self.pods_to_gols.items():
                pod_org = self.orgs[k]
                # ('\tПОПЫТКА_НАЙТИ_ГОЛ_ОРГ_С_ОКПО:', v)
                orgs = OrganizationVocModel(self.request).filter_raw({'OKPO': v})
                orgs = list(filter(lambda org_: org_['record_id'] > self.last_org_in_db_id, orgs))
                try:
                    gol_org = orgs[0]
                    pod_org['KOD_GOL_ORG'] = str(gol_org['record_id'])
                    OrganizationVocModel(self.request).update(pod_org['record_id'], pod_org, commit=False)
                except IndexError:
                    # print('\tНЕ_НАЙДЕНА, self.last_org_in_db_id:', self.last_org_in_db_id)
                    pass

        logger.info('--save forms--')
        self.thread_runner.status = 'Сохранение отчетных форм'

        forms_ids_not_to_save_for_13 = {}

        count = 0

        # SAVE FORMS
        for form_id, form in self.result_forms.items():
            form['form_state'] = FormModel.Data.FormStateStatuses.LOADED.value
            self.thread_runner.status = f'Сохранение отчетных форм: {count + 1}/{len(self.result_forms)}'
            self._set_progress(6, 'Сохранение отчетных форм', count, len(self.result_forms))

            if self.decree_num == '449':
                form['author_id'] = okpos_to_orgs_for_449[str(form['author_id'])].record_id
            elif self.decree_num == '13':
                if form.get('author_id') is not None:
                    form_author_obj = fake_org_id_to_real_for_13[form['author_id']]
                    form['author_id'] = form_author_obj['record_id']
                else:
                    continue
            else:
                form_author_obj = self.orgs[form['author_id']]
                form['author_id'] = self.orgs[form['author_id']]['record_id']

            form['import_author'] = self._build_import_author_id(
                form['author_id'], form_author_obj['KOD_ORG'], form_author_obj['OKPO'],
            )
            form['kod_org'] = form_author_obj['KOD_ORG']

            # сохраняем хеш всей записи
            # form_as_dict_for_hash = form.as_dict()
            form_import_ts = form['form_import_ts']
            form.pop('form_import_ts')
            form['import_hash'] = hashlib.md5(json.dumps(form, default=str).encode('utf-8')).hexdigest()
            form['form_import_ts'] = form_import_ts
            form['log_id'] = self.user_log.record_id

            form_save_res, form = save_import_item(self.request, form, FormModel, commit=True)

            if form_save_res == 'created':
                self.result_forms_report_ids.append(form['report_form_id'])

            if form_save_res == 'found':
                self.forms_imported_before.append(form['report_form_id'])

            if form_save_res == 'updated':
                self.forms_updated.append(form['report_form_id'])

            form_no_clear = '_'.join(form['form_no'].split('_')[:3])
            if form and form_no_clear in self.CUSTOM_VOC_ADDING_RULES:
                for row in form['form_table_data']['main_table']:
                    for rule in self.CUSTOM_VOC_ADDING_RULES[form_no_clear]:
                        voc_repo = vocs_repository.get_voc_repository(rule['voc'])
                        new_voc_value = {}
                        new_voc_value['voc_type'] = rule['voc']
                        new_voc_value['attrs'] = {model_item['field']: '' for model_item in voc_repo['model']}
                        new_voc_value['creator_id'] = form['author_id']
                        new_voc_value['attrs']['code_OKPO_org'] = form_author_obj['OKPO']
                        for form_attr, voc_attr in rule['voc_attrs'].items():
                            new_voc_value['attrs'][voc_attr] = row[form_attr]
                        new_voc_value['value_id'] = new_voc_value['attrs'][voc_repo['value_id']]
                        new_voc_value['label'] = ' '.join(str(new_voc_value['attrs'].get(l, '')) for l in voc_repo['label'])

                        if self.decree_num == '13':
                            new_voc_value['value_status'] = VocValueModel.Data.ValueStatusesTypes.ACTIVE.value
                        else:
                            new_voc_value['value_status'] = VocValueModel.Data.ValueStatusesTypes.OUTDATED.value

                        new_voc_value['import_hash'] = hashlib.md5(json.dumps(new_voc_value, default=str).encode('utf-8')).hexdigest()
                        new_voc_value['log_id'] = self.user_log.record_id

                        if new_voc_value['voc_type'] not in self.system_vocs_types:
                            if new_voc_value['import_hash'] in self.voc_import_hash_cash[new_voc_value['voc_type']]:
                                continue
                            else:
                                self.voc_import_hash_cash[new_voc_value['voc_type']].add(new_voc_value['import_hash'])
                            vocs_found = VocValueModel(self.request).filter({'import_hash': new_voc_value['import_hash']})
                            if not vocs_found:
                                new_voc_value['import_src_id'] = '0'
                                new_voc_value['import_author'] = self._build_import_author_id(
                                    new_voc_value['creator_id'], form_author_obj['KOD_ORG'], form_author_obj['OKPO'],
                                )
                                new_voc_value = VocValueModel(self.request).save(new_voc_value, commit=True, no_return=True)

            count += 1

        self._set_progress(7, 'Импорт данных завершен')

        logger.info("--- %s seconds ---" % (time.time() - start_time))

    def _after_all(self):
        # valid_from для головной организации - не выпускает отчетных форм, поэтому valid_from
        # ставится по самой маленькой valid_from из подособленных организаций
        logger.info('===AFTER ALL==')
        start_time = time.time()
        for org_id, org in self.orgs.items():
            if org.get('value_id') is None:
                for org_id2, org2 in self.orgs.items():
                    try:
                        if org2['KOD_GOL_ORG'] == org['KOD_ORG'] and (org.get('value_id') is None or org['valid_from'] > org2['valid_from']):
                            org['valid_from'] = org2['valid_from']
                    except:
                        pass
        logger.info("--- %s seconds ---" % (time.time() - start_time))

    def _set_progress(self, points, message, sub_point=None, total_sub_points=None):
        if self.api2_obj:
            percents = (100 / self.PROGRESS_MESSAGES) * points
            if sub_point and total_sub_points:
                next_percents = (100 / self.PROGRESS_MESSAGES) * (points + 1)
                percents += ((next_percents - percents) / total_sub_points) * sub_point
                self.api2_obj.context.action.set_progress(round(percents, 2))
            else:
                self.api2_obj.context.action.set_progress(round(percents, 2))
            self.api2_obj.context.action.set_message(message)

    def _import_data(self):
        # импорт справочника организаций
        logger.info('импорт справочника организаций')
        self.thread_runner.status = 'Импорт справочника организаций'
        self._set_progress(0, 'Импорт справочника организаций')
        start_time = time.time()
        if 'SprORG' in self.table_names:
            self._import_orgs(None, 'SprORG')
        logger.info("--- %s seconds ---" % (time.time() - start_time))

        self.check_done()

        logger.info('импорт основных таблиц форм')
        self.thread_runner.status = 'Импорт основных таблиц форм'
        self._set_progress(1, 'Импорт основных таблиц форм')
        start_time = time.time()
        # импорт основных таблиц форм
        if 'FORMP' in self.table_names:
            self._import_main_form_data(None, 'FORMP', self.FORMP_convert)
        if 'Forms' in self.table_names:
            self._import_main_form_data(None, 'Forms', self.Forms_convert)
        logger.info("--- %s seconds ---" % (time.time() - start_time))

        self.check_done()

        # импорт таблицы данных форм (строки)
        logger.info('импорт таблицы данных форм (строки)')
        # self.thread_runner.status = 'импорт таблицы данных форм (строки)'
        start_time = time.time()
        table_count = 0
        for table_name, rules in self.form_rows_rules.items():
            logger.info(f"\t=====table: {table_name}=====")
            self.thread_runner.status = f'Импорт таблицы: {table_name}'
            self._set_progress(2, f'Импорт таблицы: {table_name}', table_count, len(self.form_rows_rules))
            start_time = time.time()
            if table_name in self.table_names:
                self._import_forms_rows(
                    reader=None,
                    form_no=rules['form_no'],
                    map_=rules.get('map'),
                    convert_func=rules.get('convert'),
                    table_name=table_name,
                )

            self.check_done()

            if table_name.endswith('_24'):
                table_name_extra = table_name[:-3]
                if table_name_extra in self.table_names:
                    self._import_forms_rows(
                        reader=None,
                        form_no=rules['form_no'],
                        map_=rules.get('map'),
                        convert_func=rules.get('convert'),
                        table_name=table_name_extra,
                    )

            self.check_done()

            table_count += 1
            logger.info("\t--- %s seconds ---" % (time.time() - start_time))
            # exit()
        logger.info("--- %s seconds ---" % (time.time() - start_time))

        self.check_done()

        # импорт примечаний к форме - в нашей системе - комментарии к форме
        logger.info('импорт примечаний к форме - в нашей системе - комментарии к форме')
        self.thread_runner.status = 'Импорт комментариев к формам'
        self._set_progress(3, 'Импорт комментариев к формам')
        start_time = time.time()
        if 'NoteF' in self.table_names:
            self._import_form_comment(None, 'NoteF')
        if 'PrimF' in self.table_names:
            self._import_form_comment(None, 'PrimF')
        logger.info("--- %s seconds ---" % (time.time() - start_time))

        self.check_done()

        # импорт примечаний к ячейкам форм
        logger.info('импорт примечаний к ячейкам форм')
        self.thread_runner.status = 'Импорт примечаний к формам'
        self._set_progress(4, 'Импорт примечаний к формам')
        start_time = time.time()
        if 'PRIMCell' in self.table_names:
            self._import_cell_remarks(None, 'PRIMCell')
        # импорт примечаний к ячейкам форм
        if 'NoteCell' in self.table_names:
            self._import_cell_remarks(None, 'NoteCell')
        logger.info("--- %s seconds ---" % (time.time() - start_time))

        self.check_done()

    def _log_after(self):
        UserLogsLogger(self.request, self.user).log_import_historical_data(
            self.result_forms_report_ids, self.orgs_imported_ids, self.user_log,
        )

    def import_historical_data(self):
        start_time = time.time()
        self._import_data()
        self._after_all()
        self._save_data()
        self._log_after()
        logger.info("--- TOTAL TIME: %s seconds ---" % (time.time() - start_time))