import time
import datetime
import time
from typing import Dict

from loguru import logger

from repository.forms import forms_repository
from services.exceptions import CoreBaseException
from wsgi.apps.historical_data.services.decrees.base_decree import BaseDecree
from wsgi.apps.historical_data.services.decrees.utils import get_reader_from_firebird_gen
from wsgi.apps.reporting_forms.models import FormModel
from wsgi.apps.voc.models import OrganizationVocModel, VocValueModel
from wsgi.apps.voc.services import build_voc_label


def get_dict_from_row(cur, row):
    d = {}
    for idx, col in enumerate(cur.description):
        d[col[0]] = row[idx]
    return d


def _form_1_1_convert(obj: BaseDecree, row, row_mdb, form):
    row['Nss'] = obj._round_to_int_as_str(row['Nss'])


def _form_1_2_convert(obj: BaseDecree, row, row_mdb, form):
    row['Nss'] = obj._round_to_int_as_str(row['Nss'])


class Decree1_13_2(BaseDecree):

    decree_year = '2022'

    form_rows_rules = {
        'form_11': {
            'form_no': 'form_1_1',
            'map': {
                'PaspN': 'PassportNumber_DB', 'Typ': 'Type_DB', 'Nuclid': 'Radionuclids_DB',
                'Numb': 'FactoryNumber_DB', 'Sht': 'Quantity_DB', 'Activn': 'Activity_DB',
                'IzgotDate': 'CreationDate_DB', 'IzgotOKPO': 'CreatorOKPO_DB', 'Kateg': 'Category_DB',
                'Nss': 'SignedServicePeriod_DB', 'FormSobst': 'PropertyCode_DB', 'Pravoobl': 'Owner_DB',
                'OkpoPIP': 'ProviderOrRecieverOKPO_DB', 'OkpoPrv': 'TransporterOKPO_DB', 'PrName': 'PackName_DB',
                'UktPrTyp': 'PackType_DB', 'UktPrN': 'PackNumber_DB',
                'OpCod': 'OperationCode_DB',
                'OpDate': 'OperationDate_DB',
                'DocVid': 'DocumentVid_DB',
                'DocN': 'DocumentNumber_DB',
                'DocDate': 'DocumentDate_DB',
            },
            'exp': ['Activn'],
            'convert': _form_1_1_convert,
        },
        'form_12': {
            'form_no': 'form_1_2',
            'map': {
                'PaspN': 'PassportNumber_DB', 'Typ': 'NameIOU_DB', 'Numb': 'FactoryNumber_DB', 'Kg': 'Mass_DB',
                'IzgotOKPO': 'CreatorOKPO_DB', 'IzgotDate': 'CreationDate_DB', 'Nss': 'SignedServicePeriod_DB',
                'FormSobst': 'PropertyCode_DB', 'Pravoobl': 'Owner_DB', 'OkpoPIP': 'ProviderOrRecieverOKPO_DB',
                'OkpoPrv': 'TransporterOKPO_DB', 'PrName': 'PackName_DB', 'UktPrTyp': 'PackType_DB',
                'UktPrN': 'PackNumber_DB',
                'OpCod': 'OperationCode_DB',
                'OpDate': 'OperationDate_DB',
                'DocVid': 'DocumentVid_DB',
                'DocN': 'DocumentNumber_DB',
                'DocDate': 'DocumentDate_DB',
            },
            'exp': ['Kg'],
            'convert': _form_1_2_convert,
        },
        'form_13': {
            'form_no': 'form_1_3',
            'map': {'PaspN': 'PassportNumber_DB', 'Typ': 'Type_DB', 'Nuclid': 'Radionuclids_DB',
                    'Numb': 'FactoryNumber_DB', 'Activn': 'Activity_DB', 'IzgotOKPO': 'CreatorOKPO_DB',
                    'IzgotDate': 'CreationDate_DB', 'Agr': 'AggregateState_DB', 'FormSobst': 'PropertyCode_DB',
                    'Pravoobl': 'Owner_DB', 'OkpoPIP': 'ProviderOrRecieverOKPO_DB', 'OkpoPrv': 'TransporterOKPO_DB',
                    'PrName': 'PackName_DB', 'UktPrTyp': 'PackType_DB', 'UktPrN': 'PackNumber_DB',
                    'OpCod': 'OperationCode_DB', 'OpDate': 'DocumentDate_DB', 'DocVid': 'DocumentVid_DB',
                    'DocN': 'DocumentNumber_DB', 'DocDate': 'DocumentDate_DB'},
            'exp': ['Activn']
        },
        'form_14': {
            'form_no': 'form_1_4',
            'map': {'PaspN': 'PassportNumber_DB', 'IstName': 'Name_DB', 'Typ': 'Sort_DB',
                    'Nuclid': 'Radionuclids_DB', 'Activn': 'Activity_DB', 'ActDate': 'ActivityMeasurementDate_DB',
                    'Kbm': 'Volume_DB', 'Kg': 'Mass_DB', 'Arg': 'AggregateState_DB', 'FormSobst': 'PropertyCode_DB',
                    'Pravoobl': 'Owner_DB', 'OkpoPIP': 'ProviderOrRecieverOKPO_DB', 'OkpoPrv': 'TransporterOKPO_DB',
                    'PrName': 'PackName_DB', 'UktPrTyp': 'PackType_DB', 'UktPrN': 'PackNumber_DB',
                    'OpCod': 'OperationCode_DB', 'OpDate': 'OperationDate_DB', 'DocVid': 'DocumentVid_DB',
                    'DocN': 'DocumentNumber_DB', 'DocDate': 'DocumentDate_DB'},
            'exp': ['Activn', 'Kg', 'Kbm']
        },
        'form_15': {
            'form_no': 'form_1_5',
            'map': {'PaspN': 'PassportNumber_DB', 'Typ': 'Type_DB', 'Nuclid': 'Radionuclids_DB',
                    'Numb': 'FactoryNumber_DB', 'Sht': 'Quantity_DB', 'Activn': 'Activity_DB',
                    'IzgotDate': 'CreationDate_DB', 'BCod': 'StatusRAO_DB', 'OkpoPIP': 'ProviderOrRecieverOKPO_DB',
                    'OkpoPrv': 'TransporterOKPO_DB', 'PrName': 'PackName_DB', 'UktPrTyp': 'PackType_DB',
                    'UktPrN': 'PackNumber_DB', 'PH_Name': 'StoragePlaceName_DB', 'PH_Cod': 'StoragePlaceCode_DB',
                    'CodRAO': 'RefineOrSortRAOCode_DB', 'Subsid': 'Subsidy_DB', 'FCP': 'FcpNumber_DB',
                    'OpCod': 'OperationCode_DB', 'OpDate': 'OperationDate_DB', 'DocVid': 'DocumentVid_DB',
                    'DocN': 'DocumentNumber_DB', 'DocDate': 'DocumentDate_DB'},
            'exp': ['Activn'],
        },
        'form_16': {
            'form_no': 'form_1_6',
            'map': {'RAOCod': 'CodeRAO_DB', 'BCod': 'StatusRAO_DB', 'Kbm': 'Volume_DB', 'Tonne': 'Mass_DB',
                    'RAOCodMax': 'MainRadionuclids_DB', 'ActTR': 'TritiumActivity_DB',
                    'ActBG': 'BetaGammaActivity_DB', 'ActA': 'AlphaActivity_DB', 'ActUR': 'TransuraniumActivity_DB',
                    'ActDate': 'ActivityMeasurementDate_DB', 'Sht': 'QuantityOZIII_DB',
                    'OkpoPIP': 'ProviderOrRecieverOKPO_DB', 'OkpoPrv': 'TransporterOKPO_DB',
                    'PrName': 'PackName_DB', 'UktPrTyp': 'PackType_DB', 'UktPrN': 'PackNumber_DB',
                    'PH_Name': 'StoragePlaceName_DB', 'PH_Cod': 'StoragePlaceCode_DB', 'Subsid': 'Subsidy_DB',
                    'FCP': 'FcpNumber_DB', 'UpCod': 'RefineOrSortRAOCode_DB', 'OpCod': 'OperationCode_DB',
                    'OpDate': 'OperationDate_DB', 'DocVid': 'DocumentVid_DB', 'DocN': 'DocumentNumber_DB',
                    'DocDate': 'DocumentDate_DB'},
            'exp': ['Kbm', 'Tonne']
        },
        'form_17': {
            'form_no': 'form_1_7',
            'map': {'g4': 'PackName_DB', 'g5': 'PackType_DB', 'g6': 'PackNumber_DB', 'g7': 'PackFactoryNumber_DB',
                    'g8': 'FormingDate_DB', 'g10': 'Volume_DB', 'g11': 'Mass_DB', 'g9': 'PassportNumber_DB',
                    'g12': 'Radionuclids_DB', 'g13': 'SpecificActivity_DB', 'g17': 'ProviderOrRecieverOKPO_DB',
                    'g18': 'TransporterOKPO_DB', 'g19': 'StoragePlaceName_DB', 'g20': 'StoragePlaceCode_DB',
                    'g31': 'Subsidy_DB', 'g32': 'FcpNumber_DB', 'g21': 'CodeRAO_DB', 'g22': 'StatusRAO_DB',
                    'g23': 'VolumeOutOfPack_DB', 'g24': 'MassOutOfPack_DB', 'g25': 'Quantity_DB',
                    'g26': 'TritiumActivity_DB', 'g27': 'BetaGammaActivity_DB', 'g28': 'AlphaActivity_DB',
                    'g29': 'TransuraniumActivity_DB', 'g30': 'RefineOrSortRAOCode_DB', 'g2': 'OperationCode_DB',
                    'g3': 'OperationDate_DB', 'g14': 'DocumentVid_DB', 'g15': 'DocumentNumber_DB',
                    'g16': 'DocumentDate_DB'}
        },
        'form_18': {
            'form_no': 'form_1_8',
            'map': {'g4': 'IndividualNumberZHRO_DB', 'g5': 'PassportNumber_DB', 'g6': 'Volume6_DB',
                    'g7': 'Mass7_DB', 'g8': 'SaltConcentration_DB', 'g9': 'Radionuclids_DB',
                    'g10': 'SpecificActivity_DB', 'g14': 'ProviderOrRecieverOKPO_DB', 'g15': 'TransporterOKPO_DB',
                    'g16': 'StoragePlaceName_DB', 'g17': 'StoragePlaceCode_DB', 'g18': 'CodeRAO_DB',
                    'g19': 'StatusRAO_DB', 'g20': 'Volume20_DB', 'g21': 'Mass21_DB', 'g22': 'TritiumActivity_DB',
                    'g23': 'BetaGammaActivity_DB', 'g24': 'AlphaActivity_DB', 'g25': 'TransuraniumActivity_DB',
                    'g26': 'RefineOrSortRAOCode_DB', 'g27': 'Subsidy_DB', 'g28': 'FcpNumber_DB',
                    'g2': 'OperationCode_DB', 'g3': 'OperationDate_DB', 'g11': 'DocumentVid_DB',
                    'g12': 'DocumentNumber_DB', 'g13': 'DocumentDate_DB'}
        },
        'form_19': {
            'form_no': 'form_1_9',
            'map': {'g7': 'CodeTypeAccObject_DB', 'g8': 'Radionuclids_DB', 'g9': 'Activity_DB',
                    'g2': 'OperationCode_DB', 'g3': 'OperationDate_DB', 'g4': 'DocumentVid_DB',
                    'g5': 'DocumentNumber_DB', 'g6': 'DocumentDate_DB'}
        },
        'form_21': {
            'form_no': 'form_2_1',
            'map': {'g2': 'RefineMachineName_DB', 'g3': 'MachineCode_DB', 'g19': 'MachinePower_DB',
                    'g4': 'NumberOfHoursPerYear_DB', 'g5': 'CodeRAOIn_DB', 'g6': 'StatusRAOIn_DB',
                    'g7': 'VolumeIn_DB', 'g8': 'MassIn_DB', 'g9': 'QuantityIn_DB', 'g20': 'TritiumActivityIn_DB',
                    'g21': 'BetaGammaActivityIn_DB', 'g10': 'AlphaActivityIn_DB',
                    'g11': 'TransuraniumActivityIn_DB', 'g12': 'CodeRAOout_DB', 'g13': 'StatusRAOout_DB',
                    'g14': 'VolumeOut_DB', 'g15': 'MassOut_DB', 'g16': 'QuantityOZIIIout_DB',
                    'g22': 'TritiumActivityOut_DB', 'g23': 'BetaGammaActivityOut_DB', 'g17': 'AlphaActivityOut_DB',
                    'g18': 'TransuraniumActivityOut_DB'},
            'exp': ['g14', 'g15', 'g23', 'g17', 'g21', 'g10', 'g7', 'g8', 'g19']
        },
        'form_22': {
            'form_no': 'form_2_2',
            'map': {'g2': 'StoragePlaceName_DB', 'g3': 'StoragePlaceCode_DB', 'g11': 'PackName_DB',
                    'g12': 'PackType_DB', 'g13': 'PackQuantity_DB', 'g4': 'CodeRAO_DB', 'g5': 'StatusRAO_DB',
                    'g6': 'VolumeOutOfPack_DB', 'g7': 'VolumeInPack_DB', 'g8': 'MassOutOfPack_DB',
                    'g9': 'MassInPack_DB', 'g10': 'QuantityOZIII_DB', 'g19': 'TritiumActivity_DB',
                    'g14': 'AlphaActivity_DB', 'g15': 'BetaGammaActivity_DB', 'g20': 'TransuraniumActivity_DB',
                    'g16': 'MainRadionuclids_DB', 'g21': 'Subsidy_DB', 'g22': 'FcpNumber_DB'},
            'exp': ['g6', 'g7', 'g8', 'g9', 'g19', 'g14', 'g15', 'g20']
        },
        'form_23': {
            'form_no': 'form_2_3',
            'map': {'g2': 'StoragePlaceName_DB', 'g3': 'StoragePlaceCode_DB', 'g13': 'ProjectVolume_DB',
                    'g4': 'CodeRAO_DB', 'g5': 'Volume_DB', 'g6': 'Mass_DB', 'g7': 'QuantityOZIII_DB',
                    'g8': 'SummaryActivity_DB', 'g9': 'DocumentNumber_DB', 'g10': 'DocumentDate_DB',
                    'g11': 'ExpirationDate_DB', 'g12': 'DocumentName_DB'},
            'exp': ['g5', 'g6', 'g8']
        },
        'form_24': {
            'form_no': 'form_2_4',
            'map': {'g2': 'CodeOYAT_DB', 'g17': 'FcpNumber_DB', 'g3': 'MassCreated_DB', 'g4': 'QuantityCreated_DB',
                    'g5': 'MassFromAnothers_DB', 'g6': 'QuantityFromAnothers_DB',
                    'g7': 'MassFromAnothersImported_DB', 'g8': 'QuantityFromAnothersImported_DB',
                    'g9': 'MassAnotherReasons_DB', 'g10': 'QuantityAnotherReasons_DB',
                    'g11': 'MassTransferredToAnother_DB', 'g12': 'QuantityTransferredToAnother_DB',
                    'g13': 'MassRefined_DB', 'g14': 'QuantityRefined_DB', 'g15': 'MassRemovedFromAccount_DB',
                    'g16': 'QuantityRemovedFromAccount_DB'},
            'exp': ['g3', 'g5', 'g7', 'g9', 'g11', 'g13', 'g15']
        },
        'form_25': {
            'form_no': 'form_2_5',
            'map': {'g2': 'StoragePlaceName_DB', 'g3': 'CodeOYAT_DB', 'g4': 'StoragePlaceCode_DB',
                    'g10': 'FcpNumber_DB', 'g5': 'FuelMass_DB', 'g6': 'CellMass_DB', 'g7': 'Quantity_DB',
                    'g8': 'AlphaActivity_DB', 'g9': 'BetaGammaActivity_DB'},
            'exp': ['g5', 'g6', 'g8', 'g9']
        },
        'form_26': {
            'form_no': 'form_2_6',
            'map': {'num_s': 'ObservedSourceNumber_DB', 'zone': 'ControlledAreaName_DB',
                    'ist': 'SupposedWasteSource_DB', 'rust': 'DistanceToWasteSource_DB', 'gl': 'TestDepth_DB',
                    'nuk': 'RadionuclidName_DB', 'radio': 'AverageYearConcentration_DB'},
            'exp': ['radio']
        },
        'form_27': {
            'form_no': 'form_2_7',
            'map': {'g2': 'ObservedSourceNumber_DB', 'g3': 'RadionuclidName_DB', 'g4': 'AllowedWasteValue_DB',
                    'g5': 'FactedWasteValue_DB', 'g6': 'WasteOutbreakPreviousYear_DB'},
            'exp': ['g4', 'g5', 'g6']
        },
        'form_28': {
            'form_no': 'form_2_8',
            'map': {'g2': 'WasteSourceName_DB', 'g3': 'WasteRecieverName_DB', 'g4': 'RecieverTypeCode_DB',
                    'g5': 'PoolDistrictName_DB', 'g6': 'AllowedWasteRemovalVolume_DB',
                    'g7': 'RemovedWasteVolume_DB'},
            'exp': ['g6', 'g7']
        },
        'form_29': {
            'form_no': 'form_2_9',
            'map': {'g2': 'WasteSourceName_DB', 'g3': 'RadionuclidName_DB', 'g4': 'AllowedActivity_DB',
                    'g5': 'FactedActivity_DB'},
            'exp': ['g4', 'g5']
        },
        'form_210': {
            'form_no': 'form_2_10',
            'map': {'g2': 'IndicatorName_DB', 'g3': 'PlotName_DB', 'g4': 'PlotKadastrNumber_DB',
                    'g5': 'PlotCode_DB', 'g6': 'InfectedArea_DB', 'g7': 'AvgGammaRaysDosePower_DB',
                    'g8': 'MaxGammaRaysDosePower_DB', 'g9': 'WasteDensityAlpha_DB', 'g10': 'WasteDensityBeta_DB',
                    'g11': 'FcpNumber_DB'},
            'exp': ['g6', 'g7', 'g8', 'g9', 'g10']
        },
        'form_211': {
            'form_no': 'form_2_11',
            'map': {'g2': 'PlotName_DB', 'g3': 'PlotKadastrNumber_DB', 'g4': 'PlotCode_DB',
                    'square': 'InfectedArea_DB', 'g5': 'Radionuclids_DB', 'g6': 'SpecificActivityOfPlot_DB',
                    'g7': 'SpecificActivityOfLiquidPart_DB', 'g8': 'SpecificActivityOfDensePart_DB'},
            'exp': ['square', 'g6', 'g7', 'g8']
        },
        'form_212': {
            'form_no': 'form_2_12',
            'map': {'g2': 'OperationCode_DB', 'g3': 'ObjectTypeCode_DB', 'g4': 'Radionuclids_DB',
                    'g5': 'Activity_DB', 'g6': 'ProviderOrRecieverOKPO_DB'},
        }
    }

    def __init__(self, thread_runner, request, user, decrees, decree_name, db_file_path, db_file_type, api2_obj=None, done=None):
        super().__init__(thread_runner, request, user, decrees, decree_name, db_file_path, db_file_type,
                         api2_obj=api2_obj, done=done)

        self.decree_num = '13'

        self.org_period_from = datetime.date(2022, 1, 1)
        self.org_period_to = None

        # self.data = get_readers_from_firebird(db_file_path, self.form_rows_rules)
        # self.csv_table_files_temp = self.table_names = self.db_tables = tuple(self.readers.keys())

        self.orgs_map = {'KOD_ORG': 'RegNo_DB', 'KOD_IAE': 'OrganUprav_DB', 'KOD_SRF': 'SubjectRF_DB',
                         'NAME_FULL': 'JurLico_DB',
                         'NAME_SHORT': 'ShortJurLico_DB', 'ADDRESS_U': 'JurLicoAddress_DB',
                         'ADDRESS_P': 'JurLicoFactAddress_DB',
                         'RUKOVODITEL': 'GradeFIO_DB', 'TLFON': 'Telephone_DB', 'FAX': 'Fax_DB', 'EMAIL': 'Email_DB',
                         'OKPO': 'Okpo_DB', 'OKVED': 'Okved_DB', 'OKOGU': 'Okogu_DB', 'OKTMO': 'Oktmo_DB',
                         'INN': 'Inn_DB',
                         'KPP': 'Kpp_DB', 'OKOPF': 'Okopf_DB', 'OKFS': 'Okfs_DB'}

        self.report_collection_storage = {}
        self.reports_collection_storage = {}

        self.forms_like_orgs_to_report_collection = {}

        self.reg_nos = []
        self.okpos = []

        self.pods_to_gols = {}

        self.prev_org_ids_to_ran = {}

        try:
            self.last_org_in_db_id = OrganizationVocModel(self.request).list()[-1].record_id - 1
        except IndexError:
            self.last_org_in_db_id = -1

        self.forms_ids_to_rows = {}

        self.after_init()

        self.orgs_id_count = 0

    def _to_date(self, value):
        try:
            d, m, y = value.split('.')
            return f'{y}-{m}-{d}'
        except ValueError:
            return value

    def _convert_to_date_if_date(self, value):
        try:
            return str(datetime.datetime.strptime(value, "%d.%m.%Y").date())
        except ValueError:
            return value

    def import_historical_data(self):
        self._import_data()
        self._save_data()
        self._log_after()

    def _save_data(self):
        self.result_forms = self.report_collection_storage
        super()._save_data()

    def _check(self):
        pass

    def _import_data(self):
        # собираем данные по организациям (form_10, form_20)
        logger.info('собираем данные по организациям (form_10, form_20)')
        self.thread_runner.status = 'Импорт справочника организаций'
        self._set_progress(0, 'Импорт справочника организаций')
        t = time.time()
        self._import_formmm('form_10')
        self.check_done()
        self._import_formmm('form_20')
        logger.info(f'------{time.time() - t} s------')

        self.check_done()

        # собираем ReportsCollection
        logger.info('собираем ReportsCollection')
        self.thread_runner.status = 'Импорт отчетных форм'
        self._set_progress(1, 'Импорт отчетных форм')
        t = time.time()
        self._import_reports_collection('ReportsCollection_DbSet')
        logger.info(f'------{time.time() - t} s------')

        self.check_done()

        # собираем данные по формам
        logger.info('собираем данные по формам')
        t = time.time()
        self._import_report_collection('ReportCollection_DbSet')
        logger.info(f'------{time.time() - t} s------')

        self.check_done()

        # табличные данные форм
        logger.info('табличные данные форм')
        tables_count = 0
        for table_name, rules in self.form_rows_rules.items():
            self.thread_runner.status = f'Импорт таблицы форм: {table_name}'
            self._set_progress(2, f'Импорт таблицы: {table_name}', tables_count, len(self.form_rows_rules))
            logger.info(f'\t==={table_name}===')
            t = time.time()
            self._import_forms_rows2(
                reader=None,
                map_=rules.get('map'),
                convert_func=rules.get('convert'),
                table_name=table_name,
                id_key='ReportId',
                exp=rules.get('exp'),
            )
            tables_count += 1
            logger.info(f'------{time.time() - t} s------')

            self.check_done()

        logger.debug('Сохранение строк форм в правильном порядке')
        for form_id_fdb in self.forms_ids_to_rows:
            form = self.report_collection_storage[form_id_fdb]
            form_row_idx_to_row = self.forms_ids_to_rows[form_id_fdb]
            for idx in sorted(self.forms_ids_to_rows[form_id_fdb].keys()):
                form['form_table_data'].setdefault('main_table', []).append(form_row_idx_to_row[idx])

        # примечания к ячейкам формы
        logger.info('примечания к ячейкам формы')
        self.thread_runner.status = 'Импорт примечаний к формам'
        t = time.time()
        self._set_progress(3, 'Импорт примечаний к формам')
        # if 'notes' in self.db_tables:
        self._import_notes('notes')
        logger.info(f'------{time.time() - t} s------')

        self.check_done()

    def _create_org(self, row, RegNo, OrganUprav, przn_op, przn_org=4, przn_activ=1):
        org_data = {}  # данные орг. для сохранения
        org_fields = OrganizationVocModel.Data.fields()
        for k, v in self.orgs_map.items():  # сохраняем их атрибуты в наши
            if v == 'RegNo_DB' and not RegNo:
                continue
            if v == 'OrganUprav_DB' and not OrganUprav:
                continue
            if k in org_fields:
                org_data[k] = row[v]
            else:
                org_data.setdefault('attrs', {})[k] = row[v]

        ran = self.orgs_id_count
        self.orgs_id_count += 1

        org_data['PRZN_OP'] = przn_op
        org_data['PRZN_ORG'] = przn_org
        org_data['PRZN_ACTIV'] = przn_activ

        self.orgs[ran] = org = org_data
        org['value_status'] = VocValueModel.Data.ValueStatusesTypes.ACTIVE.value
        org['KOD_SRF'] = org['KOD_SRF'].zfill(2)
        if OrganUprav and org['KOD_IAE']:
            org['KOD_IAE'] = org['KOD_IAE'].zfill(2)
        if RegNo:
            org['KOD_ORG'] = org_data['KOD_ORG']
        else:
            org['KOD_ORG'] = None
        org['voc_type'] = 'organizations'
        org['label'] = build_voc_label(org, self.org_repository, attrs_wrap=False)

        # если у ReportCollection еще не запомнили ссылку на организацию или текущая орг. является НЕ головной,
        # то запоминаем (привязка организации к отчетным формам)
        if row['ReportId'] not in self.forms_like_orgs_to_report_collection or row['NumberInOrder_DB'] != 1:
            # Id пока их базы, потом перезаписать при сохранении организации
            self.forms_like_orgs_to_report_collection[row['ReportId']] = ran

        org['import_src_id'] = row['ReportId']
        org['import_author'] = self._build_import_author_id(row['ReportId'], org['KOD_ORG'], org['OKPO'])

        org['valid_from'] = self.org_period_from.isoformat().split('T')[0] if self.org_period_from else None
        org['valid_to'] = self.org_period_to.isoformat().split('T')[0] if self.org_period_to else None

        self.prev_org_ids_to_ran[row['Id']] = ran

        return ran

    def _import_formmm(self, table_name):
        gol = None
        pod = None
        for row in get_reader_from_firebird_gen(self.db_file_path, table_name):
            if gol is None:
                gol = row
                continue
            if pod is None:
                pod = row

                RegNo = True
                OrganUprav = True

                if not pod['RegNo_DB'] and not gol['RegNo_DB']:
                    gol = None
                    pod = None
                    continue

                if not pod['Okpo_DB']:
                    if gol['RegNo_DB'] in self.reg_nos:
                        gol = None
                        pod = None
                        continue
                    else:
                        if gol['RegNo_DB']:
                            self.reg_nos.append(gol['RegNo_DB'])
                        if gol['Okpo_DB']:
                            self.okpos.append(gol['Okpo_DB'])
                        self._create_org(gol, RegNo, OrganUprav, 0)
                else:
                    if gol['Okpo_DB'] in self.okpos:
                        pass
                    else:
                        if gol['Okpo_DB']:
                            self.okpos.append(gol['Okpo_DB'])
                        self._create_org(gol, False, False, 0)
                    if pod['RegNo_DB'] in self.reg_nos:
                        gol = None
                        pod = None
                        continue
                    else:
                        if pod['RegNo_DB']:
                            self.reg_nos.append(pod['RegNo_DB'])
                        if pod['Okpo_DB']:
                            self.okpos.append(pod['Okpo_DB'])
                        iddddd = self._create_org(pod, RegNo, OrganUprav, 1)
                        self.pods_to_gols[iddddd] = gol['Okpo_DB']

                gol = None
                pod = None

    def _import_report_collection(self, table_name):
        for row in get_reader_from_firebird_gen(self.db_file_path, table_name):
            # заносим основные данные формы
            form = {}
            form['form_no'] = 'form_' + '_'.join(row['FormNum_DB'].split('.')) + '_' + self.decree_year
            form['form_import_ts'] = self.import_ts
            form_repository = forms_repository.get_form_repository(form['form_no'])
            form['form_name'] = form_repository['form_name']
            self._init_form(row['Id'], form, forms_repository.get_form_repository(form['form_no']), None)
            form['adjustment_no'] = int(row['CorrectionNumber_DB'])
            form['form_main_data'] = {
                'position': row['GradeExecutor_DB'],
                'full_name': row['FIOexecutor_DB'],
                'phone': row['ExecPhone_DB'],
                'email': row['ExecEmail_DB'],
            }
            form['user_id'] = self.user.record_id

            if row['StartPeriod_DB']:
                try:
                    form['form_period_start'] = datetime.datetime.strptime(row['StartPeriod_DB'].strip(), '%d.%m.%Y').isoformat().split('T')[0]
                except:
                    form['form_period_start'] = None
            if row['EndPeriod_DB']:
                try:
                    form['form_period_end'] = datetime.datetime.strptime(row['EndPeriod_DB'].strip(), '%d.%m.%Y').isoformat().split('T')[0]
                except:
                    form['form_period_end'] = None
            if row['Year_DB']:
                form['reporting_year'] = row['Year_DB']

            form['import_src_id'] = row['Id']

            self.report_collection_storage[int(row['Id'])] = form

        # привязываем к организации (пока по ID в их системе) (если форма не титульная)
        for row in get_reader_from_firebird_gen(self.db_file_path, table_name):
            if row['ReportsId'] is not None:
                form = self.report_collection_storage[row['Id']]
                reports_collection_item = self.reports_collection_storage[row['ReportsId']]
                report_collection_item = self.report_collection_storage[int(reports_collection_item['Master_DBId'])]
                # находим ранее сохраненный ID организации, к данному титульнику
                try:
                    org_id = self.forms_like_orgs_to_report_collection[report_collection_item['import_src_id']]
                except KeyError:
                    org_id = None
                    # print('организация не найдена по ключу. Будет привязано к "неизвестная организация"')
                form['author_id'] = org_id

    def _import_reports_collection(self, table_name):
        # собираем ReportsCollection
        for row in get_reader_from_firebird_gen(self.db_file_path, table_name):
            self.reports_collection_storage[row['Id']] = row

    def _import_forms_rows2(self, reader, table_name, map_=None, convert_func=None,
                            id_key='IDF', exp=None):
        for row in get_reader_from_firebird_gen(self.db_file_path, table_name):
            try:
                form = self.report_collection_storage[row['ReportId']]
            except KeyError:
                # print(f'\tERR: строка ссылается на несуществующую форму по id: {row["ReportId"]}, строка не будет добавлена.')
                continue
            form_no = form['form_no']
            form_repository = forms_repository.get_form_repository(form['form_no'])
            forms_fields_names = [model_item['field'] for model_item in
                                  form_repository['tables']['main_table']['model']]
            row_data = {}  # данные итоговой строки
            for field_name in forms_fields_names:
                if map_ is not None and field_name in map_:
                    try:
                        field_name_mdb = map_[field_name]
                    except KeyError as e:
                        raise e
                    try:
                        value = row[field_name_mdb]
                    except KeyError as e:
                        raise e
                else:
                    if form_no == 'form_2_7':
                        pass
                    value = row.get(field_name, '')
                if value:
                    model_item = next(
                        filter(lambda x: x['field'] == field_name,
                               form_repository['tables']['main_table']['model']))
                    # приведение к типам
                    if 'type' in model_item:
                        if model_item['type'] == 'date':
                            value = self._to_date(value)

                    if exp and field_name in exp:
                        try:
                            value = self._to_exp(value)
                        except ValueError:
                            # print('Не удалось преобразовать в exp:', value)
                            pass

                if isinstance(value, str):
                    value = value.strip()

                if value is None:
                    value = ''

                row_data[field_name] = str(value)

            if convert_func is not None:
                try:
                    convert_func(self, row_data, row, form)
                except Exception as e:
                    raise CoreBaseException(message=f'Exception message: {e}, Class: {e.__class__}, \n'
                                                    f'row_data: {row_data},\n row: {row},\n form: {form},\n\n'
                                                    f'form_repo: {form_repository}\n'
                                                    f'convert_func: {convert_func}')

            try:
                # помещаем строку по индексу
                self.forms_ids_to_rows.setdefault(row['ReportId'], {})[int(row['NumberInOrder_DB']) - 1] = row_data
            except (KeyError, ValueError):  # не задан порядок строки
                # print('НЕ Получилось поместить по индексу строку')
                # form.form_table_data.setdefault('main_table', []).append(row_data)
                pass

            # доп данные из ReportCollection
            if form['form_no'] == 'form_2_8':
                report_collection = self.report_collection_storage[int(row['ReportId'])]
                map_2 = {'g11': 'PermissionNumber_28_DB', 'g12': 'PermissionIssueDate_28_DB',
                         'g13': 'PermissionDocumentName_28_DB', 'per1': 'ValidBegin_28_DB', 'to1': 'ValidThru_28_DB',
                         'g21': 'PermissionNumber1_28_DB', 'g22': 'PermissionIssueDate1_28_DB',
                         'g23': 'PermissionDocumentName1_28_DB', 'per2': 'ValidBegin1_28_DB', 'to2': 'ValidThru1_28_DB',
                         'g31': 'ContractNumber_28_DB', 'g32': 'ContractIssueDate2_28_DB',
                         'g33': 'OrganisationReciever_28_DB', 'per3': 'ValidBegin2_28_DB', 'to3': 'ValidThru2_28_DB'}
                for k, v in map_2.items():
                    form['form_main_data'][k] = self._convert_to_date_if_date(report_collection[v])  # report_collection[v]
            elif form['form_no'] == 'form_2_7':
                report_collection = self.report_collection_storage[int(row['ReportId'])]
                map_2 = {'g7': 'PermissionNumber27_DB', 'g8': 'PermissionIssueDate27_DB',
                         'g9': 'PermissionDocumentName27_DB', 'per_from': 'ValidBegin27_DB', 'per_to': 'ValidThru27_DB'}
                for k, v in map_2.items():
                    form['form_main_data'][k] = self._convert_to_date_if_date(report_collection[v])
            elif form['form_no'] == 'form_2_6':
                report_collection = self.report_collection_storage[int(row['ReportId'])]
                form['form_main_data']['quantity'] = report_collection['SourcesQuantity26_DB']

    def _import_form_row_comment(self, row, form: Dict, form_map, rep_fields):
        if row['RowNumber_DB'] is None:
            return

        if '-' in row['RowNumber_DB']:
            col_num_from, col_num_to = row['RowNumber_DB'].split('-')
            row_nums = [i for i in range(int(col_num_from), int(col_num_to))]
        else:
            row_nums = row['RowNumber_DB'].split(',')

        for r in row_nums:
            if not r:
                continue
            try:
                comment = {
                    'value': {
                        'text': row.get('Comment_DB', ''),
                    },
                    'pointer': {
                        'row': int(r) - 1,
                    }
                }
                form['form_comments']['main_table'].append(comment)
            except Exception as e:
                logger.error(
                    f'Error adding a comment to a line. '
                    f'Exception message: {e}. Original comment value: {row["RowNumber_DB"]}'
                )

    def _import_notes(self, table_name):
        for row in get_reader_from_firebird_gen(self.db_file_path, table_name):
            try:
                form: Dict = self.report_collection_storage[row['ReportId']]
            except KeyError:
                logger.debug(f'Примечание не будет добавлено, т.к. ссылается на несуществующую форму '
                             f'по ReportId. Данные примечания: {row}')
                continue
            form_repository = forms_repository.get_form_repository(form['form_no'])
            rep_fields = {}
            # если порядок полей задан явно в репозитории (ключ col_num в form -> model)
            for model_item in form_repository['tables']['main_table']['model']:
                if model_item.get('col_num'):
                    rep_fields[int(model_item['col_num']) - 2] = model_item['field']
            # иначе поля по порядку в репозитории
            if not rep_fields:
                rep_fields = [model_item['field'] for model_item in form_repository['tables']['main_table']['model']]

            form_map = None
            for form_table_name, rules in self.form_rows_rules.items():
                form_no_clear = '_'.join(form['form_no'].split('_')[:3])
                if rules['form_no'] == form_no_clear:
                    # table_name = form_table_name
                    form_map = rules.get('map')
                    break

            if row['RowNumber_DB'] is None:
                logger.debug(f'Примечание не будет добавлено, т.к. RowNumber_DB не задан (номер строки). '
                             f'Данные примечания: {row}')
                continue

            if row['GraphNumber_DB'] is None:
                logger.debug(f'Примечание будет добавлено как комментарий к строке, т.к. GraphNumber_DB не задан. '
                             f'Данные примечания: {row}')
                self._import_form_row_comment(row, form, None, None)
                continue
                # комментарий к строке

            fields_nums = []
            for r in row['GraphNumber_DB'].split(','):
                try:
                    if '-' in r:
                        col_num_from, col_num_to = r.split('-')
                        fields_nums.extend(i for i in range(int(col_num_from), int(col_num_to) + 1))
                    else:
                        fields_nums.append(r)
                except:
                    pass
                    # print('Не удалось добавить прим. ячейка', r, row['GraphNumber_DB'])

            row_nums = []
            for r in row['RowNumber_DB'].split(','):
                try:
                    if '-' in r:
                        col_num_from, col_num_to = r.split('-')
                        row_nums.extend(i for i in range(int(col_num_from), int(col_num_to) + 1))
                    else:
                        row_nums.append(r)
                except:
                    pass
                    # print('Не удалось добавить прим. строка', r, row['RowNumber_DB'])

            logger.debug(f'Добавление примечания. fields_nums: {fields_nums}, row_nums: {row_nums}')

            for field_num in fields_nums:
                for row_num in row_nums:
                    try:

                        if not form_map:
                            logger.debug(f'Добавление примечания. form_map is None')
                            continue

                        try:
                            our_field_name = rep_fields[int(field_num) - 2]
                        except (IndexError, KeyError):
                            # print('field not found')
                            continue

                        # этих полей больше нет в репозитории, переносим примечания в 'Основные радионуклиды'
                        if our_field_name in ('RAOCodCat', 'RAOCodDang'):
                            our_field_name = 'RAOCodMax'

                        remark = {
                            'value': {
                                'text': row['Comment_DB']
                            },
                            'pointer': {
                                'row': int(row_num) - 1,
                                'col_name': our_field_name,
                            }
                        }

                        form['form_remarks']['main_table'].append(remark)
                        # print('Примечание успещнро добавлено======================================================')
                    except ValueError as e:
                        logger.debug(f'Ошибка при добавлении примечания, сообщение исключения: {e}')
                        # print('Не удалось добавить комментарий, данные строки:', row)
                        pass
