import csv
import os
import subprocess
from datetime import datetime

import settings

from services.temp import TempFile
from settings import TEMP_DIR_DECREES_PATH, RADIO_CSV_PATH

from fdb import connect

WIN_DRV = '{Microsoft Access Driver (*.mdb, *.accdb)}'


def get_table_names_from_mdb_win(mdb_file_path):
    import pyodbc

    con = pyodbc.connect('DRIVER={};DBQ={};'.format(WIN_DRV, mdb_file_path))
    cur = con.cursor()

    table_names = [row.table_name for row in cur.tables() if not row.table_name.startswith('MSys')]
    return table_names


def get_table_names_from_mdb(mdb_file_path):
    if settings.WIN or settings.WIN_DEBUG:
        return get_table_names_from_mdb_win(mdb_file_path)

    my_env = os.environ.copy()
    my_env['MDB_JET3_CHARSET'] = 'CP1251'
    my_env['MDBICONV'] = 'UTF-8'

    table_names = subprocess.Popen(
        ['mdb-tables', '-1', mdb_file_path],
        stdout=subprocess.PIPE,
        env=my_env,
    ).communicate()[0].decode('utf8').split('\n')
    return table_names


def get_reader_from_mdb_gen(mdb_file_path, table_name, tables_headers):
    if settings.WIN or settings.WIN_DEBUG:
        import pyodbc

        con = pyodbc.connect('DRIVER={};DBQ={};'.format(WIN_DRV, mdb_file_path))
        cur = con.cursor()

        try:
            sql = f'select * from "{table_name}";'
            rows = cur.execute(sql)
            columns = [column[0] for column in cur.description]
            tables_headers[table_name] = columns[2:]
            for row in rows:
                yield dict(zip(columns, row))
        except Exception as e:
            raise e
        finally:
            cur.close()
    else:
        my_env = os.environ.copy()
        my_env['MDB_JET3_CHARSET'] = 'CP1251'
        my_env['MDBICONV'] = 'UTF-8'

        csv_table_file_temp = TempFile(
            TEMP_DIR_DECREES_PATH,
            '.csv',
            file_name=str(datetime.now().timestamp()).replace('.', '_') + '_' + table_name.replace(' ', '_'),
            close_others=False,
        )

        with open(csv_table_file_temp.full_path, 'wb') as f:
            subprocess.check_call([
                'mdb-export', mdb_file_path, table_name], stdout=f,
                env=my_env,
            )

        with open(csv_table_file_temp.full_path, encoding='utf8') as f:
            tables_headers[table_name] = csv.DictReader(f).fieldnames[2:]

        with open(csv_table_file_temp.full_path, encoding='utf8') as f:
            for row in csv.DictReader(f):
                yield row


def get_reader_from_firebird_gen(firebird_file_path, table_name):
    def get_dict_from_row(cur_, row_):
        d = {}
        for idx, col in enumerate(cur_.description):
            d[col[0]] = row_[idx]
        return d

    if settings.WIN or settings.WIN_DEBUG:
        con = connect(database=firebird_file_path, user=settings.FIREBIRD_USER, password=settings.FIREBIRD_PASSWORD,
                      charset='UTF8')
    else:
        if settings.FIREBIRD_PASSWORD:
            con = connect(database=firebird_file_path, user=settings.FIREBIRD_USER, password=settings.FIREBIRD_PASSWORD)
        else:
            con = connect(database=firebird_file_path, user=settings.FIREBIRD_USER)

    cur = con.cursor()
    for row in cur.execute(f'select * from "{table_name}";'):
        yield get_dict_from_row(cur, row)


def get_radionuclides_voc_data_from_csv():
    """Получить данные по радионуклидам из xlsx файла (задаётся в настройках)"""
    with open(RADIO_CSV_PATH, encoding='utf8') as f:
        # справочник радионуклидов из xlsx файла
        return {row['\ufeffName_RN']: row for row in csv.DictReader(f, delimiter=';')}
