import threading
import traceback

import psycopg2

import settings
from api2 import get_api2_clone
from services.exceptions import CoreBaseException

from services.temp import TempFile
from settings import TEMP_DIR_DECREES_PATH
from wsgi.apps.historical_data.services.decrees.decree_1_13 import Decree1_13_2
from wsgi.apps.historical_data.services.decrees.decree_1_19 import Decree1_19
from wsgi.apps.historical_data.services.decrees.decree_1_24 import Decree1_24
from wsgi.apps.historical_data.services.decrees.decree_24_19 import Decree_24_19
from wsgi.apps.historical_data.services.decrees.decree_449 import Decree449
from wsgi.apps.historical_data.services.decrees.decree_600 import Decree600
from wsgi.apps.historical_data.services.exceptions import UnknownDecreeException
from wsgi.apps.users.models import UserModel
from wsgi.utils import RequestCopy

DECREES = {
    'decree_1_24': {
        'label': 'Приказ 1/24',
        'class': Decree_24_19,
    },
    'decree_1_19': {
        'label': 'Приказ 1/19',
        'class': Decree_24_19,
    },
    'decree_600': {
        'label': 'Приказ 600',
        'class': Decree600,
    },
    'decree_449': {
        'label': 'Приказ 449',
        'class': Decree449,
    },
    'decree_1_13': {
        'label': 'Приказ 1/13',
        'class': Decree1_13_2,
    }
}

DECREES_F = [{'label': v['label'], 'value': k} for k, v in DECREES.items()]


class DecreeImportThreadRunner:
    def __init__(self, decree, user: UserModel.Data):
        self.decree = decree
        # self.start_time = time.time()
        self.user = user
        self.status = 'Подготовка к импорту'

    def as_dict(self):
        return {
            'decree': DECREES[self.decree]['label'],
            'user': self.user.fio,
            'status': self.status,
        }


import_decree_threads_storage = {}


def get_decree_types():
    return DECREES_F


def run_decree(decree_class, request, user, decree, db_file_path, db_file_type):
    conn = psycopg2.connect(user=settings.DB_USER['user'],
                            password=settings.DB_USER['password'],
                            host=settings.DB_HOST,
                            port=settings.DB_PORT,
                            database=settings.POSTGRES_DATABASE)
    api2_clone = get_api2_clone(conn, request, user)
    request = RequestCopy(conn, api2_clone)
    thread_runner = import_decree_threads_storage[threading.get_ident()]
    to_pop = True
    try:
        decree_class(thread_runner, request, user, DECREES, decree, db_file_path, db_file_type).import_historical_data()
    except Exception as e:
        to_pop = False
        thread_runner.status = str(e) + '' + traceback.format_exc()
    conn.close()

    if to_pop:
        import_decree_threads_storage.pop(threading.get_ident(), None)


def run_import(request, user, decree, db_file_path, db_file_type):
    try:
        decree_class = DECREES[decree]['class']
    except KeyError:
        raise UnknownDecreeException(decree)

    if import_decree_threads_storage:
        raise CoreBaseException(rus_message='Одновременно может выполняться только один импорт исторических данных. '
                                            'Дождитесь выполнения текущего импорта и повторите попытку')

    thread = threading.Thread(
        target=run_decree, args=(decree_class, request, user, decree, db_file_path, db_file_type))

    thread.start()
    import_decree_threads_storage[thread.ident] = DecreeImportThreadRunner(
        decree, user
    )


def run_import_api2(api2_obj, request, user, file_path, decree, done, connection=None):
    """Функция импорта исторических данных для Api2.

    Args:
        api2_obj: Объект api2.
        request: Запрос.
        user: Пользователь.
        file_path: Путь к файлу с историческими данными.
        decree: Название приказа.
        connection: Подключение к базе. Если =None, будет использовано из `request`.
    """
    if user is None:
        raise CoreBaseException(
            rus_message='Пользователь не авторизован. Обновите страницу для авторизации.',
            delete_auth_cookie=True,
        )

    if connection is None:
        connection = request.storage['conn']

    if decree not in DECREES:
        raise CoreBaseException(rus_message=f'Неизвестное имя приказа: "{decree}". '
                                            f'Возможные имена: {list(DECREES.keys())}')

    db_file_type = 'firebird' if decree == 'decree_1_13' else 'mdb'

    api2_clone = get_api2_clone(connection, request, user)
    fake_request = RequestCopy(connection, api2_clone, request)
    thread_runner_fake = DecreeImportThreadRunner(decree, user)  # не используется
    decree_class = DECREES[decree]['class']
    decree = decree_class(thread_runner_fake, fake_request, user, DECREES, decree, file_path, db_file_type, api2_obj=api2_obj, done=done)
    decree.import_historical_data()
    res = {
        'log_import_id': decree.user_log.record_id,
        'form_import_ts': decree.import_ts,
        'report_form_id': decree.result_forms_report_ids,
        'report_form_id_updated': decree.forms_updated,
        'forms_count_in_file': len(decree.result_forms),
        'forms_count_imported_before': len(decree.forms_imported_before),
        'forms_count_updated': len(decree.forms_updated),
        'forms_count_ignored': decree.ignored_forms_count,
    }
    return res
