import copy
import datetime
from dataclasses import dataclass
from enum import Enum

from datasource.databaserepository.database_repository2 import DatabaseRepository2, D
from repository.forms import forms_repository
from repository.forms.exceptions import ModelItemNotFoundException
from wsgi.apps.users.models import UserModel
from wsgi.apps.voc.models import OrganizationVocModel


class FormModel(DatabaseRepository2['FormModel.Data']):
    @dataclass
    class Data(DatabaseRepository2.Data):
        class FormStateStatuses(Enum):
            DRAFT = 'draft'  # черновик
            REMARKS = 'remarks'  # замечания
            VERIFIED = 'verified'  # проверена
            CONFIRMED = 'confirmed'  # подтверждена
            UNLOADED = 'unloaded'  # выгружена
            REJECTED = 'rejected'  # отклонена
            ACCEPTED = 'accepted'  # принята
            CANCELED = 'canceled'  # аннулирована
            LOADED = 'loaded'  # загружена
            CONFLICT = 'conflict'

        report_form_id: int
        report_id: int
        form_state: FormStateStatuses
        form_no: str
        form_part: str
        form_version: str
        form_name: str
        form_main_data: dict
        form_table_data: dict
        form_remarks: dict
        form_comments: dict
        form_warnings: dict
        form_notifications: dict
        sign_info: dict
        fill_info: dict
        form_period_start: datetime.date
        form_period_end: datetime.date
        adjustment_no: int
        user_id: int
        kod_org: str
        author_id: int
        form_export_ts: datetime.datetime
        form_import_ts: datetime.datetime
        reporting_year: int
        notification_date: datetime.date
        form_comment: dict
        form_errors: dict
        uniq_numb: str
        import_author: str  # код организации создателя
        import_src_id: str  # ID записи в исходной БД
        import_hash: str  # хеш данных
        log_id: int

        ID_FIELD = 'report_form_id'

        def as_dict(self):
            res = super().as_dict()
            if self.form_state is not None:
                res['form_state'] = self.form_state.value
            return res

        def replace(self, data):
            obj = super().replace(data)
            if obj.form_state is not None:
                obj.form_state = self.FormStateStatuses(obj.form_state)
            return obj

        @property
        def author_id_obj(self):
            return OrganizationVocModel(self.request).read(self.author_id)

        def get_sorted_form_errors(self):
            """Получить отсортированные по строке и столбцу ошибки формы.
            Исходный порядок ошибок формы не меняется, возвращается только копия.
            """
            sorted_form_errors = copy.deepcopy(self.form_errors)

            # Сортируем по номеру строки.
            for table_name, table_errors in sorted_form_errors.items():
                for i in range(len(table_errors) - 1):
                    for j in range(len(table_errors) - i - 1):
                        try:
                            if int(table_errors[j]['pointer']['row']) > int(table_errors[j + 1]['pointer']['row']):
                                table_errors[j], table_errors[j + 1] = table_errors[j + 1], table_errors[j]
                        except (KeyError, ValueError) as e:
                            continue

            # Сортируем по столбцу в пределах одного номера строки.
            for table_name, table_errors in sorted_form_errors.items():
                for i in range(len(table_errors) - 1):
                    for j in range(len(table_errors) - i - 1):
                        # Если номер строки еще одинаковый.
                        if table_errors[j]['pointer']['row'] == table_errors[j + 1]['pointer']['row']:
                            try:
                                # Получаем порядковый номер столбцов.
                                f_order = forms_repository.get_column_num(self.form_no, table_name,
                                                                          table_errors[j]['pointer']['col_name'])
                                s_order = forms_repository.get_column_num(self.form_no, table_name,
                                                                          table_errors[j + 1]['pointer']['col_name'])
                                if f_order > s_order:
                                    table_errors[j], table_errors[j + 1] = table_errors[j + 1], table_errors[j]
                            except ModelItemNotFoundException:
                                # Если столбец не найден в репозитории модели формы.
                                pass

            return sorted_form_errors

        @property
        def user_id_obj(self):
            return UserModel(self.request).read(self.user_id)

    database_table_name = 'ReportForms'

    @dataclass
    class FormActions:
        EDITING = 'editing'
        EDITING_REMARKS = 'editing_remarks'
        EDITING_COMMENTS = 'editing_comments'
        EXPORT_EXCHANGE_FILE = 'unload_exchange_file'
        EXPORT_DOCX_XLSX_PDF = 'export_docx_xlsx_pdf'

    actions_for_statuses = {
        Data.FormStateStatuses.DRAFT: [
            FormActions.EDITING,
            FormActions.EDITING_REMARKS,
            FormActions.EDITING_COMMENTS,
            FormActions.EXPORT_DOCX_XLSX_PDF,
        ],
        Data.FormStateStatuses.REMARKS: [
            FormActions.EDITING,
            FormActions.EDITING_REMARKS,
            FormActions.EDITING_COMMENTS,
            FormActions.EXPORT_DOCX_XLSX_PDF,
        ],
        Data.FormStateStatuses.VERIFIED: [
            FormActions.EDITING,
            FormActions.EDITING_REMARKS,
            FormActions.EDITING_COMMENTS,
            FormActions.EXPORT_DOCX_XLSX_PDF,
        ],
        Data.FormStateStatuses.CONFIRMED: [
            FormActions.EDITING_COMMENTS,
            FormActions.EXPORT_EXCHANGE_FILE,
            FormActions.EXPORT_DOCX_XLSX_PDF,
        ],
        Data.FormStateStatuses.LOADED: [
            FormActions.EDITING_COMMENTS,
            FormActions.EXPORT_DOCX_XLSX_PDF,
            FormActions.EXPORT_EXCHANGE_FILE,
        ],
        Data.FormStateStatuses.UNLOADED: [
            FormActions.EDITING_COMMENTS,
            FormActions.EXPORT_DOCX_XLSX_PDF,
            FormActions.EXPORT_EXCHANGE_FILE,
        ],
        Data.FormStateStatuses.ACCEPTED: [
            FormActions.EDITING_COMMENTS,
            FormActions.EXPORT_DOCX_XLSX_PDF,
        ],
        Data.FormStateStatuses.REJECTED: [
            FormActions.EXPORT_DOCX_XLSX_PDF,
        ],
        Data.FormStateStatuses.CANCELED: [
            FormActions.EXPORT_DOCX_XLSX_PDF,
        ],
    }

    statuses_for_action = {}
    for st, actions in actions_for_statuses.items():
        for act in actions:
            statuses_for_action.setdefault(act, [])
            if st not in statuses_for_action[act]:
                statuses_for_action[act].append(st)

    #

    def update(self, record_id, obj: Data, ignore_api2_trigger=False, *args, **kwargs) -> D:
        # если статус формы становится "Черновик", удаляем ошибки формы
        if isinstance(obj, dict):
            if obj['form_state'] == self.Data.FormStateStatuses.DRAFT.value and obj['form_errors']:
                obj['form_errors'] = {table_name: [] for table_name in obj['form_errors']}
        else:
            if obj.form_state == self.Data.FormStateStatuses.DRAFT and obj.form_errors is not None:
                obj.form_errors = {table_name: [] for table_name in obj.form_errors}
        # == api2 block ==
        old_form = self.read(record_id)
        new_form = super().update(record_id, obj, *args, **kwargs)

        if ignore_api2_trigger:
            return new_form

        if isinstance(obj, dict):
            new_form_form_no = new_form['form_no']
            new_form_dict = new_form
        else:
            new_form_form_no = new_form.form_no
            new_form_dict = new_form.as_dict()

        new_form_dict['report_form_id'] = old_form.report_form_id

        if not ignore_api2_trigger:
            api2_obj = self.api2_clone
            api2_obj.triggers.emit('DS_AFTER_UPDATE', f'ReportForms.OperForms.{new_form_form_no}',
                                   context={
                                       'OLD': old_form.as_dict(),
                                       'NEW': new_form_dict,
                                   }, core_db_connect=self.conn)
        return new_form

    def delete(self, record_id, commit=True, *args, **kwargs) -> None:
        res = super().delete(record_id, commit, *args, **kwargs)
        self.api2_clone.regs.reg_clean(True)
        return res
