import settings
from datasource.databaserepository.exceptions import RecordNotFoundException
from repository.decrees import decrees_repository
from repository.decrees.exceptions import DecreeRepositoryNotFoundByYearFromException, DecreeRepositoryNotFoundException
from repository.reports import reports_repository
from services.exceptions import CoreBaseException
from repository.forms import forms_repository
from services.filter import build_filter_in
from services.services import is_str_in_date_format
from wsgi.apps.reporting_forms.models import FormModel
from wsgi.apps.reporting_forms.services import services
from wsgi.apps.reporting_forms.services.calc_permissions import calc_permissions_for_form_actions
from wsgi.apps.reporting_forms.services.form_editor import FormEditor, RowObject, Pointer, CellObject, RowValueObject
from wsgi.apps.users.models import UserModel
from wsgi.apps.voc.models import OrganizationVocModel
from wsgi.apps.voc.services import get_all_orgs_ids_with_same_KOD_ORG
from wsgi.serializers import BaseSerializer


class FormSerializer(BaseSerializer):
    model = FormModel(None)

    select_join = [
        ('author_id', OrganizationVocModel.database_table_name, 'record_id', 'NAME_FULL', 'author_name')
    ]

    def __init__(self, request, user: UserModel.Data, *args, **kwargs):
        super().__init__(request, *args, **kwargs)
        self.user: UserModel.Data = user

    def _prepare(self, data):
        super()._prepare(data)
        if data.get('form_period_start') is not None and not is_str_in_date_format(data['form_period_start']):
            data['form_period_start'] = None
        if data.get('form_period_end') is not None and not is_str_in_date_format(data['form_period_end']):
            data['form_period_end'] = None
        if data.get('reporting_year') is not None:
            try:
                int(data['reporting_year'])
            except ValueError:
                data['reporting_year'] = None
        if data.get('notification_date') is not None and not is_str_in_date_format(data['notification_date']):
            data['notification_date'] = None
        if data.get('author_id') is not None:
            try:
                data['author_id'] = int(data['author_id'])
            except ValueError:
                data.pop('author_id')  # не изменяем организацию формы
            # if not isinstance(data['author_id'], int):
            #     data.pop('author_id')  # не изменяем организацию формы
            # else:
            #     data['author_id'] = int(data['author_id'])

    def _response(self, resp):
        resp = super()._response(resp)
        report_repository = reports_repository.get_report_repository_by_num(
            int(resp['form_no'].split('_')[1])
        )
        # todo: оптимизировать
        try:
            org = OrganizationVocModel(self.request).read(resp['author_id'])
            resp['org_name_short'] = org.NAME_SHORT
            resp['org_okpo'] = org.OKPO
            # информативная строка при в верху окна формы (основная инфа по форме)
            resp['description'] = services.build_form_description(
                resp, FormModel.Data.from_dict(resp), org.as_dict(), report_repository
            )
            resp['kod_org'] = org.KOD_ORG
        except RecordNotFoundException:
            resp['org_name_short'] = resp['org_okpo'] = None
            resp['description'] = services.build_form_description(resp, FormModel.Data.from_dict(resp), '', report_repository)
            resp['kod_org'] = None

        resp['edit'] = calc_permissions_for_form_actions(self.conn, resp['report_form_id'], self.user,
                                                         FormModel.Data.from_dict(resp))

        return resp

    def save(self, data):
        obj = self.model.Data.from_dict(data)
        form_repository = forms_repository.get_form_repository(obj.form_no)

        obj.user_id = self.user.record_id
        obj.kod_org = OrganizationVocModel(self.request).read(obj.author_id).KOD_ORG
        obj.form_state = self.model.Data.FormStateStatuses.DRAFT

        obj.form_table_data = {}
        # добавление пустой строки в каждую таблицу формы
        for table_name, table_data in form_repository['tables'].items():
            obj.form_table_data[table_name] = [{model_obj['field']: '' for model_obj in table_data['model']}]

        form_tables_to_empty_lists = {k: [] for k in form_repository['tables'].keys()}
        obj.form_remarks = form_tables_to_empty_lists
        obj.form_warnings = form_tables_to_empty_lists
        obj.form_comments = form_tables_to_empty_lists
        obj.form_errors = form_tables_to_empty_lists
        obj.form_main_data = {}

        if obj.form_no.startswith('form_1'):
            obj.form_period_start = services.get_oper_form_period_start(self.request, obj.form_no, obj.author_id)

        return obj


class FormListFilterSerializer(BaseSerializer):
    model = FormModel(None)
    fields = ['report_form_id', 'form_name', 'user_id', 'form_state',
              'report_id', 'form_period_start', 'form_period_end', 'form_no',
              'author_id', 'form_comment']

    select_join = [
        ('user_id', UserModel.database_table_name, 'record_id', 'fio', 'user'),
        ('author_id', OrganizationVocModel.database_table_name, 'record_id', 'NAME_FULL', 'author_name')
    ]

    def __init__(self, request, user: UserModel.Data):
        super().__init__(request)
        self.user: UserModel.Data = user

    def _response(self, resp):
        resp = super()._response(resp)
        resp['edit'] = calc_permissions_for_form_actions(self.request, resp['report_form_id'], self.user,
                                                         FormModel.Data.from_dict(resp))
        return resp

    def list(self):
        if self.user.is_admin or not self.user.organizations_ids:
            return super().list()
        return super().filter({'author_id': get_all_orgs_ids_with_same_KOD_ORG(self.request, self.user.organizations_ids_as_list)})


def decree_filter_override(request, filter_value):
    try:
        decree_repo = decrees_repository.get_decree_repository_by_name(filter_value)
    except DecreeRepositoryNotFoundException:
        return ''
    return f'%_{decree_repo["year_from"]}'


def author_id_filter_override(request, filter_value):
    try:
        org = OrganizationVocModel(request).read(filter_value)
    except RecordNotFoundException:
        return ''
    return [org_.record_id for org_ in org.get_all_orgs_versions_objs()]


class FormsJournalSerializer(FormListFilterSerializer):
    """Получение отчетных форм для журнала отчетных форм"""

    select_join = [
        ('user_id', UserModel.database_table_name, 'record_id', 'fio', 'user'),
        ('author_id', OrganizationVocModel.database_table_name, 'record_id', 'label', 'author_name')
    ]

    change_filter_objs = {
        ('=', 'decree'): ('like', 'form_no', decree_filter_override),
        ('=', 'author_id'): ('in', 'author_id', author_id_filter_override),
    }

    def __init__(self, request, user: UserModel.Data, journal_repository):
        super().__init__(request, user)
        self.journal_repository = journal_repository
        self.fields = ['report_form_id', 'user_id', 'author_id', 'form_comment', 'form_no', 'log_id']
        self.fields.extend([model['field'] for model in journal_repository['model']])
        self.fields = list(filter(lambda f: f not in ('author_name', 'context_menu', 'decree', 'user'), self.fields))
        self.query_condition_line = journal_repository['query_condition']  # условие выборки из журнала отчетных форм
        if settings.DB_SYSTEM == 'sqlite':
            self.query_condition_line += " ESCAPE '\\' "

    def _response(self, resp):
        super()._response(resp)
        if resp.get('author_id'):
            # тип VOC на фронте работает только со строками
            resp['author_id'] = str(resp['author_id'])

        # приказ
        decree_year_from = forms_repository.get_decree_year_by_form_no(resp.pop('form_no'))
        try:
            decree_repo = decrees_repository.get_decree_repository_by_year_from(decree_year_from)
            resp['decree'] = decree_repo["name"]
        except DecreeRepositoryNotFoundByYearFromException:
            resp['decree'] = ''

        if resp['log_id']:
            resp['log_id'] = str(resp['log_id'])

        return resp

    def _build_filter_params(self):
        # получаем все id доступных организаций, а также всех их версий
        if self.user.is_admin or not self.user.all_kod_orgs:
            params = {}
        else:
            orgs_codes = set()
            orgs_codes.add(self.user.kod_org)
            if self.user.kod_org_default:
                for code in self.user.kod_org_default.split('; '):
                    orgs_codes.add(code)
            params = build_filter_in('kod_org', orgs_codes)
        return params

    def list(self):
        filter_params = self._build_filter_params()
        return super().query_condition(self.query_condition_line, filter_params)

    def query_condition(self, condition, filter_params=None):
        filter_params = self._build_filter_params()
        condition = self.query_condition_line + f' AND {condition}'
        if settings.DB_SYSTEM == 'sqlite':
            condition += " ESCAPE '\\' "
        return super().query_condition(condition, filter_params)


class FormEditorSerializer(FormSerializer):  # todo: передлать под model object
    """Класс редактирования отчетной формы в таблице"""

    def __init__(self, request, user, record_id):
        super().__init__(request, user)
        self.record_id = record_id
        self.form_data = self.model.read(self.record_id)
        self.form_editor = FormEditor(self.form_data.as_dict(), self.user)  # todo: передлать под model object

    def change_cells(self, table_name, data):
        # Переменная, для сохранения примечаний/комментариев, которые были добавлены/изменены.
        # Необходима для фронта.
        total_changes = {}

        for item in data['table_cells']:
            cell_obj = CellObject(item)

            if 'value' in item['changes']:
                self.form_editor.change_table_cell_value(table_name, cell_obj.pointer.row, cell_obj.pointer.col_name,
                                                         cell_obj.cell.value)

            if 'comment' in item['changes']:
                if not cell_obj.cell.comment:
                    self.form_editor.delete_cell_comment(table_name, cell_obj.pointer.row, cell_obj.pointer.col_name)
                else:
                    total_changes.setdefault('form_comments', {}).setdefault(table_name, []).append(
                        self.form_editor.change_cell_comment(table_name, cell_obj.pointer.row, cell_obj.pointer.col,
                                                             cell_obj.pointer.col_name, cell_obj.cell.comment))

            if 'remark' in item['changes']:
                if not cell_obj.cell.remark:
                    self.form_editor.delete_cell_remark(table_name, cell_obj.pointer.row, cell_obj.pointer.col_name)
                else:
                    total_changes.setdefault('form_remarks', {}).setdefault(table_name, []).append(
                        self.form_editor.change_cell_remark(table_name, cell_obj.pointer.row, cell_obj.pointer.col,
                                                            cell_obj.pointer.col_name, cell_obj.cell.remark))

        self.form_editor.data['form_state'] = 'draft'
        return self.model.update(self.record_id, self.model.Data.from_dict(self.form_editor.data)), total_changes

    def change_rows(self, table_name, data):
        for item in data['table_rows']:
            cell_obj = RowValueObject(item)

            if not cell_obj.row.comment:
                self.form_editor.delete_cell_comment(table_name, cell_obj.pointer.row, cell_obj.pointer.col_name)
            elif 'comment' in item['changes']:
                self.form_editor.change_cell_comment(table_name, cell_obj.pointer.row, cell_obj.pointer.col,
                                                     cell_obj.pointer.col_name, cell_obj.row.comment)

            if not cell_obj.row.remark:
                self.form_editor.delete_cell_remark(table_name, cell_obj.pointer.row, cell_obj.pointer.col_name)
            elif 'remark' in item['changes']:
                self.form_editor.change_cell_remark(table_name, cell_obj.pointer.row, cell_obj.pointer.col,
                                                    cell_obj.pointer.col_name, cell_obj.row.remark)

        self.form_editor.data['form_state'] = 'draft'
        return self.model.update(self.record_id, self.model.Data.from_dict(self.form_editor.data))

    def add_row(self, table_name, data):
        row_obj = RowObject(data)

        self.form_editor.move_cell_comments(table_name, row_obj.pointer.row - 1, 'down')
        self.form_editor.move_cell_remarks(table_name, row_obj.pointer.row - 1, 'down')
        self.form_editor.move_cell_errors(table_name, row_obj.pointer.row - 1, 'down')

        if row_obj.row.comment:
            self.form_editor.change_cell_comment(table_name, row_obj.pointer.row, row_obj.pointer.col,
                                                 row_obj.pointer.col_name, row_obj.row.comment)

        for cell_obj in row_obj.row_cells:
            if cell_obj.cell.comment is not None:
                self.form_editor.change_cell_comment(table_name, cell_obj.pointer.row, cell_obj.pointer.col,
                                                     cell_obj.pointer.col_name,
                                                     cell_obj.cell.comment)

            if cell_obj.cell.remark is not None:
                self.form_editor.change_cell_remark(table_name, cell_obj.pointer.row, cell_obj.pointer.col,
                                                    cell_obj.pointer.col_name,
                                                    cell_obj.cell.remark)

        self.form_editor.add_row(table_name, row_obj.pointer.row, row_obj.row_data)

        self.form_editor.data['form_state'] = 'draft'
        return self.model.update(self.record_id, self.model.Data.from_dict(self.form_editor.data))

    def delete_rows(self, table_name, data):
        pointers = [Pointer(item) for item in data['pointers']]

        pointers = sorted(pointers, key=lambda p: p.row)
        delta = 0

        for pointer in pointers:
            row = pointer.row - delta
            self.form_editor.delete_all_cell_comments_in_row(table_name, row)
            self.form_editor.delete_all_cell_remarks_in_row(table_name, row)
            self.form_editor.delete_all_cell_errors_in_row(table_name, row)
            self.form_editor.move_cell_comments(table_name, row, 'up')
            self.form_editor.move_cell_remarks(table_name, row, 'up')
            self.form_editor.move_cell_errors(table_name, row, 'up')
            delta += 1

        rows_nums = [pointer.row for pointer in pointers]
        self.form_editor.delete_rows(table_name, rows_nums)

        self.form_editor.data['form_state'] = 'draft'

        return self.model.update(self.record_id, self.model.Data.from_dict(self.form_editor.data))

    def move_row(self, table_name, data):
        direction = data['move_mode']
        pointer = Pointer(data['pointer'])

        self.form_editor.switch_cell_comments(table_name, pointer.row, direction)
        self.form_editor.switch_cell_remarks(table_name, pointer.row, direction)
        self.form_editor.switch_cell_errors(table_name, pointer.row, direction)
        self.form_editor.move_row(table_name, pointer.row, direction)

        self.form_editor.data['form_state'] = 'draft'

        return self.model.update(self.record_id, self.model.Data.from_dict(self.form_editor.data))

    def update_row_indexes(self, table_name, data):
        old_idxs = [item['old_index'] for item in data['indexes']]
        new_idxs = [item['new_index'] for item in data['indexes']]

        if len(old_idxs) - len(set(old_idxs)):
            raise CoreBaseException(message='There are duplicate elements in old_indexes. Unable to update row indexes')

        if len(new_idxs) - len(set(new_idxs)):
            raise CoreBaseException(message='There are duplicate elements in new_indexes. Unable to update row indexes')

        not_pair_for_old_idxs = list(set(old_idxs) - set(new_idxs))
        not_pair_for_new_idxs = list(set(new_idxs) - set(old_idxs))

        if not_pair_for_old_idxs or not_pair_for_new_idxs:
            msg = 'Indexes are missing pairs.'
            if not_pair_for_old_idxs:
                msg += f' The following indexes in old_indexes cannot ' \
                       f'find a match in new_indexes: {not_pair_for_old_idxs}.'
            if not_pair_for_new_idxs:
                msg += f' The following indexes in new_indexes cannot ' \
                       f'find a match in old_indexes: {not_pair_for_new_idxs}.'
            raise CoreBaseException(message=msg)

        self.form_editor.update_row_indexes(table_name, data['indexes'])
        return self.model.update(self.record_id, self.model.Data.from_dict(self.form_editor.data))

    def change_main_attrs(self, body):
        for item in body['attrs']:
            if item.get('value') is not None:
                self.form_editor.change_main_attr(item['attr'], item['value'])

        self.form_editor.data['form_state'] = 'draft'

        return self.model.update(self.record_id, self.model.Data.from_dict(self.form_editor.data))
