import copy
import datetime as dt
import hashlib
import json
from datetime import datetime
from typing import List

from datasource.databaserepository.exceptions import RecordNotFoundException
from repository.forms import forms_repository
from repository.types import types_repository
from repository.vocs import vocs_repository
from services.services import round_dt_by_seconds
from wsgi.apps.reporting_forms.models import FormModel
from wsgi.apps.services import save_import_item, save_import_voc
from wsgi.apps.user_logs.services import UserLogsLogger
from wsgi.apps.users.models import UserModel
from wsgi.apps.voc.models import VocValueModel, OrganizationVocModel


class Unload:

    def __init__(self, request):
        self.request = request
        self.result_forms = []
        self.result_vocs = {}
        self.result_orgs = {}

    def _set_form(self, form):
        data = form.as_dict()
        data.pop('report_id')
        return data

    def _set_voc(self, voc_value, form_org: OrganizationVocModel.Data):
        data = voc_value.as_dict()
        data['attrs'].pop('kod_org', None)
        data['attrs'].pop('attrs1', None)
        if not data['attrs'].get('code_OKPO_org'):
            data['attrs']['code_OKPO_org'] = form_org.OKPO if form_org.OKPO is not None else ''
        return data

    def _set_org(self, org, ignore_gol_org=False):
        data = org.as_dict()
        if ignore_gol_org:
            data.pop('KOD_GOL_ORG')
        return data

    def _set_voc_item(self, voc_type, value, form_org):
        for v in str(value).split('; '):
            try:
                p = (voc_type, v)
                if p not in self.result_vocs:
                    voc = VocValueModel(self.request).filter({
                        'voc_type': voc_type,
                        'value_id': v,
                    })[0]
                    self.result_vocs[(voc_type, v)] = self._set_voc(voc, form_org)
            except IndexError:
                pass

    def _found_vocs_in_form_and_add(self, form: FormModel.Data):
        """Поиск всех ячеек в форме, которые имеют тип справочника. Сохраняем все записи, которые используются
        в отчетной форме, в контейнер обменного файла"""
        form_rep = forms_repository.get_form_repository(form.form_no)
        tables = [t for t in form_rep['tables']]
        fields = {t: {} for t in tables}  # поля, имеющие тип справочник у формы по таблицам
        form_org = form.author_id_obj
        # собираем все поля формы, которые имеют тип справочник
        for table in tables:
            for m in form_rep['tables'][table]['model']:
                if 'type' in m:
                    type_rep = types_repository.get_types_repository()[m['type']]
                    if type_rep['base_type'] == 'VOC' and type_rep.get('voc_subtype') != 'system':
                        fields[table][m['field']] = type_rep.get('voc_type_db', type_rep['voc_type'])

        # проходимся по форме и заносим значения справоника, которые в них используются
        for table in tables:
            for row in form.form_table_data[table]:
                for k, v in row.items():
                    if k in fields[table]:
                        voc_db_type = fields[table][k]
                        self._set_voc_item(voc_db_type, v, form_org)

    def _found_org_in_forms_and_add(self, form: FormModel.Data):
        org = form.author_id_obj
        if org.KOD_ORG not in self.result_orgs:
            if org.KOD_GOL_ORG and org.KOD_GOL_ORG not in self.result_orgs:
                try:
                    self.result_orgs[org.KOD_GOL_ORG_obj.KOD_ORG] = [self._set_org(org.KOD_GOL_ORG_obj)]
                except RecordNotFoundException:  # головная организация не была найдена по ссылке
                    pass

            self.result_orgs[org.KOD_ORG] = [self._set_org(o, ignore_gol_org=True) for o in org.get_all_orgs_versions_objs()]

    def unload(self, forms: List[FormModel.Data]):
        for form in forms:
            self.result_forms.append(self._set_form(form))
            self._found_vocs_in_form_and_add(form)
            self._found_org_in_forms_and_add(form)

        return {
            'vocs': list(self.result_vocs.values()),
            'forms': self.result_forms,
            'orgs': self.result_orgs,
        }


class Upload:

    def __init__(self, request, data, user: UserModel.Data):
        self.request = request
        self.data = data
        self.user = user
        self.org_ids_export_to_system = {}

        self.new_orgs_with_gol_org = {}
        self.new_gol_orgs = {}

        self.voc_orgs_repo = vocs_repository.get_voc_repository('organizations')

        self.imported_forms_ids = []
        self.imported_orgs_ids = []

        self.user_log = UserLogsLogger(self.request, self.user).log_import_exchange_file(
            self.imported_forms_ids, self.imported_orgs_ids
        )

    def _find_org_item_by_record_id(self, record_id):
        for org_code, org_versions in self.data['orgs'].items():
            for org in org_versions:
                if org['record_id'] == record_id:
                    return org
        raise Exception(f'Организация не найдена в секции "orgs" внтури обменного файла по record_id = {record_id}')

    def _set_org(self, org_code, org_versions):
        try:
            db_org_versions = OrganizationVocModel(self.request).filter({'KOD_ORG': org_code})[0].get_all_orgs_versions_objs()
        except IndexError:  # организация с таким кодом не найдена
            # создаем новые организации на основе обменного файла
            for org in org_versions:
                up_org = OrganizationVocModel(self.request).save(OrganizationVocModel.Data.from_dict(org), commit=True)
                self.imported_orgs_ids.append(up_org.record_id)
                up_org.log_id = self.user_log.record_id
                up_org.value_id = up_org.record_id
                up_org = OrganizationVocModel(self.request).update(up_org.record_id, up_org, commit=True)
                if up_org.KOD_GOL_ORG is not None:
                    self.new_orgs_with_gol_org[up_org.record_id] = (up_org.record_id, int(up_org.KOD_GOL_ORG))
                if up_org.PRZN_OP == OrganizationVocModel.Data.PrznOpTypes.ENTITY:
                    self.new_gol_orgs[org['record_id']] = up_org.record_id
                self.org_ids_export_to_system[org['record_id']] = up_org.record_id
            return

        # по каждой из версий из обменного файла
        for i, org in enumerate(org_versions):
            # если пересекается с теми версиями, что уж есть в базе
            if i < len(db_org_versions):
                # обновляем существующую версию организации на основе данных файла
                org.pop('KOD_GOL_ORG', None)  # не учитываем головную организацию из обменного файла todo: может быть ошибка
                updated_org = db_org_versions[i].replace(org)
                up_org = OrganizationVocModel(self.request).update(db_org_versions[i].record_id, updated_org, commit=True)
            else:
                # создаем новую версию
                # org = OrganizationVocModel.Data.from_dict(org)
                org['KOD_GOL_ORG'] = org_versions[0].get('KOD_GOL_ORG')  # задаем головную организацию также как и в системе
                up_org = OrganizationVocModel(self.request).save(OrganizationVocModel.Data.from_dict(org), commit=True)
                self.imported_orgs_ids.append(up_org.record_id)
                up_org = OrganizationVocModel(self.request).update(up_org.record_id, up_org, commit=True)
                if i == len(db_org_versions):
                    # последней организации из базы задаем дату окончания, делаем неактивной
                    db_org_versions[-1].valid_to = up_org.valid_from - dt.timedelta(days=1)
                    db_org_versions[-1].value_status = OrganizationVocModel.Data.ValueStatusesTypes.OUTDATED
                    OrganizationVocModel(self.request).update(db_org_versions[-1].record_id, db_org_versions[-1], commit=True)
            if i == len(org_versions) - 1:
                # делаем последную запись активной
                up_org.value_status = OrganizationVocModel.Data.ValueStatusesTypes.ACTIVE
                up_org.value_id = up_org.record_id
                OrganizationVocModel(self.request).update(up_org.record_id, up_org, commit=True)
            self.org_ids_export_to_system[org['record_id']] = up_org.record_id

    def _set_voc(self, voc_data):
        voc_data_to_set = copy.deepcopy(voc_data)
        voc_data_to_set['record_id'] = None
        voc = VocValueModel.Data.from_dict(voc_data_to_set)
        # Сохраняем id в исходной базе из обменного файла
        voc.import_src_id = voc_data.get('record_id', '')
        voc.import_author = 'kod_org=' + voc_data.get('kod_org', '')
        voc.import_hash = hashlib.md5(json.dumps(voc_data).encode('utf-8')).hexdigest()
        voc.log_id = self.user_log.record_id
        # save_import_item(self.request, voc.as_dict(), VocValueModel, fields_to_match=['voc_type', 'value_id', 'attrs'])
        if voc.voc_type not in vocs_repository.get_system_voc_types():
            save_res, voc = save_import_voc(self.request, voc.as_dict())

    def _set_form(self, form_data):
        form = FormModel.Data.from_dict(form_data)
        form.form_state = FormModel.Data.FormStateStatuses.LOADED
        form.form_export_ts = None
        form.user_id = self.request.storage['user'].record_id
        form.author_id = self.org_ids_export_to_system[form.author_id]
        # Сохраняем id в исходной базе из обменного файла
        form.import_src_id = form_data.get('report_form_id', '')
        # Получаем организацию по author_id из обменного файла и сохраняем ее код организации
        kod_org = self._find_org_item_by_record_id(form_data['author_id'])['KOD_ORG']
        if kod_org is None:
            kod_org = ''
        form.import_author = 'kod_org=' + kod_org

        # ОБНУЛЕНИЕ КОРРЕКТИРОВОК СУЩ ФОРМ В СИСТЕМЕ
        # оперативная отчетность
        form_as_dict = form.as_dict()
        if form_as_dict['form_no'].startswith('form_1'):
            form_period_params = [
                ['=', 'form_period_start', form_as_dict['form_period_start']],
                ['=', 'form_period_end', form_as_dict['form_period_end']],
            ]
        # годовая отчетность
        elif form_as_dict['form_no'].startswith('form_2'):
            form_period_params = [
                ['=', 'reporting_year', form_as_dict['reporting_year']],
            ]
        else:
            form_period_params = []

        # фильтрация по Организация, период/год/день, № формы
        forms_to_update = FormModel(self.request).filter(
            [
                'and',
                ['=', 'author_id', form_as_dict['author_id']],
                *form_period_params,
                ['=', 'form_no', form_as_dict['form_no']]
            ],
        )
        is_forms_with_bigger_or_eq_adjustment_no = any(f.adjustment_no >= form_as_dict['adjustment_no'] for f in forms_to_update)
        if is_forms_with_bigger_or_eq_adjustment_no:
            return
        for f in filter(lambda f: f.adjustment_no < form_as_dict['adjustment_no'], forms_to_update):
            f.form_state = FormModel.Data.FormStateStatuses.CANCELED
            FormModel(self.request).update(f.report_form_id, f, commit=True)

        form.import_hash = hashlib.md5(json.dumps(form_data).encode('utf-8')).hexdigest()
        form.form_import_ts = round_dt_by_seconds(datetime.now())
        form.log_id = self.user_log.record_id
        FormModel(self.request).save(form)
        self.imported_forms_ids.append(form.report_form_id)

    def upload(self):
        if 'orgs' in self.data:
            for org_code, org_versions in self.data['orgs'].items():
                self._set_org(org_code, org_versions)
            for import_sub_id, ids in self.new_orgs_with_gol_org.items():
                system_sub_id, import_gol_id = ids
                system_gol_id = self.new_gol_orgs[import_gol_id]
                org = OrganizationVocModel(self.request).read(system_sub_id)
                org.KOD_GOL_ORG = system_gol_id
                OrganizationVocModel(self.request).update(org.record_id, org)

        if 'vocs' in self.data:
            for voc in self.data['vocs']:
                self._set_voc(voc)
        if 'forms' in self.data:
            for form in self.data['forms']:
                self._set_form(form)

        UserLogsLogger(self.request, self.user).log_import_exchange_file(
            self.imported_forms_ids, self.imported_orgs_ids, self.user_log
        )
