import copy
from datetime import datetime

from services.exceptions import CoreBaseException


def _is_action_in_body(action_name, body, key):
    return any(action_name in obj['changes'] for obj in body[key])


def is_change_cells_values_action(body, key):
    return _is_action_in_body('value', body, key)


def is_change_cells_comments_action(body, key):
    return _is_action_in_body('comment', body, key)


def is_change_cells_remarks_action(body, key):
    return _is_action_in_body('remark', body, key)


def check_is_any_change_table_action(body, key):
    for obj in body[key]:
        remarks_act_in = 'remark' in obj['changes']
        comments_act_in = 'comment' in obj['changes']
        value_act_in = 'value' in obj['changes']
        if not (remarks_act_in or comments_act_in or value_act_in):
            raise CoreBaseException(rus_massage='changes должны содержать хотя бы одно из доступных действий: '
                                                '[remark, comment, value]')


class Pointer:
    def __init__(self, data):
        self.col = data.get('col')
        self.col_name = data.get('col_name')
        self.row = data.get('row')


class Cell:
    def __init__(self, data):
        self.comment = data.get('comment')  # зачение комментария
        self.remark = data.get('remark')
        self.value = data.get('value')


class Row:
    def __init__(self, data):
        self.cells_objs = [CellObject(item) for item in data]
        self.row_data = {cell_obj.pointer.col_name: cell_obj.cell.value for cell_obj in self.cells_objs}


class RowValue:
    def __init__(self, data):
        self.comment = data.get('comment')  # зачение комментария
        self.remark = data.get('remark')
        self.value = data.get('value')


class CellObject:
    def __init__(self, data):
        self.pointer = Pointer(data['pointer'])
        self.cell = Cell(data['table_cell'])


class RowObject:
    def __init__(self, data):
        self.pointer = Pointer(data['pointer'])
        self.row = RowValue(data['table_row'])
        self.row_cells = [CellObject(item) for item in data['row_cells']]
        #
        self.row_data = {cell_obj.pointer.col_name: cell_obj.cell.value for cell_obj in self.row_cells}


class RowValueObject:
    def __init__(self, data):
        self.pointer = Pointer(data['pointer'])
        self.row = RowValue(data['table_row'])


class FormEditor:
    main_data_field = 'form_main_data'
    table_data_field = 'form_table_data'
    comments_field = 'form_comments'
    remarks_field = 'form_remarks'
    errors_field = 'form_errors'

    def __init__(self, data, user):
        self.data = data
        self.user = user

    def _delete_any(self, table_name, field_name, row_num, attr):
        objects = self.data[field_name][table_name]
        objects_without_objects_belong_to_cell = []
        for i, obj_item in enumerate(objects):
            row = obj_item['pointer']['row']
            attr_ = obj_item['pointer'].get('col_name')
            if (attr_ is None or attr_ == attr) and row == row_num:
                continue
            objects_without_objects_belong_to_cell.append(obj_item)
        self.data[field_name][table_name] = objects_without_objects_belong_to_cell

    def _change_any(self, table_name, field_name, row_num, col, attr_name, value):
        """Изменение комментов, примечаний, ... Если не найдено, то добавляет новый"""
        for obj_item in self.data[field_name][table_name]:
            row = obj_item['pointer']['row']
            attr = obj_item['pointer'].get('col_name')
            if attr_name is None:  # any - для строки, не для ячейки (например, коммент для строки)
                if row == row_num and obj_item['pointer'].get('col_name') is None:
                    obj_item['value']['text'] = value
                    obj_item['value']['date'] = str(datetime.now())
                    return obj_item
            else:
                if attr == attr_name and row == row_num and obj_item['pointer'].get('col_name') is not None:
                    obj_item['value']['text'] = value
                    obj_item['value']['date'] = str(datetime.now())
                    return obj_item
        obj = {
            'value': {
                'text': value,
                'date': str(datetime.now()),
                'user': self.user.fio,
            },
            'pointer': {
                'row': row_num,
            }
        }
        if attr_name is not None:
            obj['pointer']['col_name'] = attr_name

        self.data[field_name][table_name].append(obj)
        return obj

    def _move_any(self, table_name, field_name, row_num, direction):
        """Сдвиг комментов, примечаний, ..."""
        # если есть (комменты, примечания) в форме
        if self.data[field_name] and self.data[field_name][table_name]:
            for i, obj_item in enumerate(self.data[field_name][table_name]):
                if isinstance(obj_item['pointer']['row'], int):
                    if obj_item['pointer']['row'] > row_num:
                        if direction == 'up':
                            obj_item['pointer']['row'] -= 1
                        elif direction == 'down':
                            obj_item['pointer']['row'] += 1

    def _switch_any(self, table_name, field_name, row_num, direction):
        """Сдвиг комментов, примечаний, ..."""
        # если есть (комменты, примечания) в форме
        if self.data[field_name] and self.data[field_name][table_name]:
            idx_updated_row = []
            for i, obj_item in enumerate(self.data[field_name][table_name]):
                if obj_item['pointer']['row'] == row_num:
                    if direction == 'up':
                        obj_item['pointer']['row'] -= 1
                    elif direction == 'down':
                        obj_item['pointer']['row'] += 1
                    idx_updated_row.append(i)

                if obj_item['pointer']['row'] == row_num + 1 and direction == 'down' and i not in idx_updated_row:
                    obj_item['pointer']['row'] -= 1
                if obj_item['pointer']['row'] == row_num - 1 and direction == 'up' and i not in idx_updated_row:
                    obj_item['pointer']['row'] += 1

    def _delete_all_any_in_row(self, table_name, field_name, row_num):
        """Удалить все (комментарии или примечания или ...) в строке"""
        # если есть (комменты, примечания) в форме
        if self.data[field_name] and self.data[field_name][table_name]:
            new_objs = []
            for i, comment_item in enumerate(self.data[field_name][table_name]):
                row = comment_item['pointer']['row']
                if row != row_num:
                    new_objs.append(comment_item)
            self.data[field_name][table_name] = new_objs

    def _check_is_params_exists(self, table_name, row_num, attr):
        if table_name not in self.data[self.table_data_field]:
            raise CoreBaseException(message=f'table name with name: "{table_name}" does not exists')
        try:
            self.data[self.table_data_field][table_name][row_num]
        except IndexError:
            raise CoreBaseException(message=f'row num: "{row_num}" does not exists')

    def change_table_cell_value(self, table_name, row_num, attr, value):
        self._check_is_params_exists(table_name, row_num, attr)
        self.data[self.table_data_field][table_name][row_num][attr] = value

    def delete_cell_comment(self, table_name, row_num, attr):
        self._delete_any(table_name, self.comments_field, row_num, attr)

    def change_cell_comment(self, table_name, row_num, col_num, attr, comment):
        return self._change_any(table_name, self.comments_field, row_num, col_num, attr, comment)

    def delete_cell_remark(self, table_name, row_num, attr):
        self._delete_any(table_name, self.remarks_field, row_num, attr)

    def change_cell_remark(self, table_name, row_num, col_num, attr, comment):
        return self._change_any(table_name, self.remarks_field, row_num, col_num, attr, comment)

    def switch_cell_comments(self, table_name, row_num, direction):
        self._switch_any(table_name, self.comments_field, row_num, direction)

    def switch_cell_remarks(self, table_name, row_num, direction):
        self._switch_any(table_name, self.remarks_field, row_num, direction)

    def switch_cell_errors(self, table_name, row_num, direction):
        self._switch_any(table_name, self.errors_field, row_num, direction)

    def move_cell_comments(self, table_name, row_num, direction):
        self._move_any(table_name, self.comments_field, row_num, direction)

    def move_cell_remarks(self, table_name, row_num, direction, switch=False):
        self._move_any(table_name, self.remarks_field, row_num, direction)

    def move_cell_errors(self, table_name, row_num, direction):
        self._move_any(table_name, self.errors_field, row_num, direction)

    def move_row(self, table_name, row_num, direction):
        move_to = 1 if direction == 'down' else -1
        table_data = self.data[self.table_data_field][table_name]
        table_data[row_num], table_data[row_num + move_to] = table_data[row_num + move_to], table_data[row_num]

    def update_row_indexes(self, table_name, indexes):
        table_data = self.data[self.table_data_field][table_name]
        comments_data = self.data[self.comments_field][table_name]
        remarks_data = self.data[self.remarks_field][table_name]
        errors_data = self.data[self.errors_field][table_name]

        # номера строк -> коммент с этим номером строки
        comments_row_to_idx = {com['pointer']['row']: com for com in comments_data}
        # номера строк -> примечание с этим номером строки
        remarks_row_to_idx = {rem['pointer']['row']: rem for rem in remarks_data}
        # номера строк -> ошибка с этим номером строки
        errors_row_to_idx = {err['pointer']['row']: err for err in errors_data}

        new_table_data = copy.deepcopy(table_data)

        new_comments_data = []
        new_remarks_data = []
        new_errors_data = []

        for item in indexes:
            new_table_data[item['new_index']] = table_data[item['old_index']]

            # перепривязка комментов, примечаний и ошибок к новым индексам строк
            comment = comments_row_to_idx.get(item['old_index'])  # находим коммент по страому индексу
            if comment:
                comment['pointer']['row'] = item['new_index']  # задаем новый индекс для комментария
                new_comments_data.append(comment)

            remark = remarks_row_to_idx.get(item['old_index'])  # находим коммент по страому индексу
            if remark:
                remark['pointer']['row'] = item['new_index']  # задаем новый индекс для комментария
                new_remarks_data.append(remark)

            error = errors_row_to_idx.get(item['old_index'])  # находим коммент по страому индексу
            if error:
                error['pointer']['row'] = item['new_index']  # задаем новый индекс для комментария
                new_errors_data.append(error)

        new_comments_data = sorted(new_comments_data, key=lambda x: int(x['pointer']['row']))
        new_remarks_data = sorted(new_remarks_data, key=lambda x: int(x['pointer']['row']))
        new_errors_data = sorted(new_errors_data, key=lambda x: int(x['pointer']['row']))

        self.data[self.table_data_field][table_name] = new_table_data
        self.data[self.comments_field][table_name] = new_comments_data
        self.data[self.remarks_field][table_name] = new_remarks_data
        self.data[self.errors_field][table_name] = new_errors_data

    def add_row(self, table_name, row_num, data):
        self.data[self.table_data_field][table_name].insert(row_num, data)

    def delete_all_cell_comments_in_row(self, table_name, row_num):
        self._delete_all_any_in_row(table_name, self.comments_field, row_num)

    def delete_all_cell_remarks_in_row(self, table_name, row_num):
        self._delete_all_any_in_row(table_name, self.remarks_field, row_num)

    def delete_all_cell_errors_in_row(self, table_name, row_num):
        self._delete_all_any_in_row(table_name, self.errors_field, row_num)

    def delete_rows(self, table_name, rows_nums):
        new_rows = []
        for i, item in enumerate(self.data[self.table_data_field][table_name]):
            if i not in rows_nums:
                new_rows.append(item)
        self.data[self.table_data_field][table_name] = new_rows

    def change_main_attr(self, attr_name, value):
        self.data[self.main_data_field][attr_name] = value
