from enum import Enum
from typing import List, Optional

from wsgi.apps.export_manager.services.services import build_download_name_for_form_errors_file
from wsgi.apps.reporting_forms.models import FormModel
from wsgi.apps.reporting_forms.services.forms_actions.exceptions import FormsActionsException, \
    FormsActionsPermissionsDeniedException, FormsActionsFormStateWrongException
from wsgi.apps.reporting_forms.services.forms_actions.funcs import to_confirm, to_cancel_confirm, zri_to_ozri, \
    iou_to_rao, ori_to_rao, orp_to_rao, zri_to_ori, form_f3_2_to_f1_1, export_error_to_xlsx, form_3_1_to_inform, \
    form_3_2_to_inform
from wsgi.apps.users.models import UserModel

form_statuses = FormModel.Data.FormStateStatuses
user_roles = UserModel.Data.UserRoleTypes


class FormActionsResponseTypes(Enum):
    """Типы ответов действий отчетной формы."""
    JSON = 'json'
    FILE = 'file'
    XLSX_WORKBOOK = 'xlsx_workbook'


class FormActionTabs(Enum):
    """Имена табов в нижней части страницы отчетной формы."""
    form_remarks = 'form_remarks'
    form_comments = 'form_comments'
    form_errors = 'form_errors'


class FormAction:
    """Класс универсального действия (кнопки) над отчетной формой."""

    base_url = '/api/forms/actions?action=exec&path=forms_actions&action_name='

    def __init__(self, action_name, caption, func, allowed_form_statuses=None,
                 text=None, allowed_user_roles=None, form_nos=None, response_type=FormActionsResponseTypes.JSON,
                 bottom_tab: Optional[FormActionTabs] = None, extra=None, extra_kwargs=None):
        """Init.

        Args:
            action_name: Имя действия (англ.).
            caption: Текст кнопки на фронтенде.
            func: Функция-обработчик, которая будет вызываться при нажатии на кнопку (отправке запроса)
            allowed_form_statuses: Разрешенные статусы форм для данного действия. Только для форм с такими
                статусами будет доступно действие. Если None - без ограничений (разрешены для всех).
            text: Текст для отображения на фронтенде, после успешного выполнения действия.
            allowed_user_roles: Разрешенные роли пользователей для данного действия. Только для пользователей с такими
                ролями будет доступно действие. Если None - без ограничений (разрешены для всех).
            form_nos:  Номера форм (в формате `form_1_1`), для которых будет доступно действие. Если None -
                без ограничений (разрешены для всех).
            response_type: Тип ответа. По умолчанию - json
            bottom_tab: Имя таба, в котором будет отображаться действие (ошибки, примечания, комментарии). По умолчанию
                действие располагается в верху страницы формы.
            extra: Дополнительные аргументы в формате словаря.
            extra_kwargs: Доп свойства для внутреннего использования.
        """
        self.action_name = action_name
        self.caption = caption
        self.allowed_form_statuses = allowed_form_statuses
        self.func = func

        self.form_nos = form_nos
        self.response_type = response_type
        self.bottom_tab = bottom_tab
        self.extra = extra
        self.extra_kwargs = extra_kwargs

        self.text = text
        self.allowed_user_roles = allowed_user_roles

    def checking_permissions_before_executing(self, request, user: UserModel.Data, form_id):
        """Проверка прав пользователя на выполнение действия над отчетной формой."""
        if self.allowed_user_roles is not None:
            if user.user_role not in self.allowed_user_roles:
                raise FormsActionsPermissionsDeniedException(self.allowed_user_roles)
        if self.allowed_form_statuses is not None:
            if FormModel(request).read(form_id).form_state not in self.allowed_form_statuses:
                raise FormsActionsFormStateWrongException(self.allowed_form_statuses)

    def result(self):
        """Результат выполнения отчетной формы для response."""
        return {
            'caption': 'Выполнение действия отчетной формы',
            'text': self.text,
            'refresh': 'record'
        }

    def error_result(self, e):
        """response - ошибка выполнения действия."""
        try:
            return {
                'caption': 'Ошибка выполнения действия отчетной формы',
                'text': e.rus_message,
            }
        except:
            raise e

    def dict_(self):
        """Получить информацию о действии в формате словаря для передачи на фронтенд."""
        res = {
            'action_name': self.action_name,
            'url': self.base_url + self.action_name,
        }
        if self.caption is not None:
            res['caption'] = self.caption
        if self.extra is not None:
            res.update(self.extra)
        return res


class DropDownFormAction:
    """Класс элемента выпадающего меню, в котором содержатся кнопки отчетной формы."""

    def __init__(self, caption: str, action_name: str = None, url: str = None, icon: str = None,
                 actions: List[FormAction] = None):
        """Init.

        Args:
            caption: Текст, который будет отображаться на элементе выпадающего меню.
            action_name: Имя (англ.) - пока не исп.
            url: URL - пока не исп.
            icon: Иконка, которая будет отображаться на элементе. Должен содержать ссылку на одну из картинок внутри
                frontend/interface/...
            actions: Список действий, которые будут включены в выпадающий список.
        """
        self.caption = caption
        self.action_name = action_name
        self.url = url
        self.icon = icon
        self.actions = actions

        # поля ниже будут высчитаны на основе self.actions
        self.allowed_form_statuses = None
        self.allowed_user_roles = None
        self.form_nos = None
        for action in self.actions:
            if action.allowed_form_statuses is not None:
                if self.allowed_form_statuses is None:
                    self.allowed_form_statuses = set()
                self.allowed_form_statuses.update(action.allowed_form_statuses)

            if action.allowed_user_roles is not None:
                if self.allowed_user_roles is None:
                    self.allowed_user_roles = set()
                self.allowed_user_roles.update(action.allowed_user_roles)

            if action.form_nos is not None:
                if self.form_nos is None:
                    self.form_nos = set()
                self.form_nos.update(action.form_nos)

    def dict_(self):
        """Получить информацию о действии в формате словаря для передачи на фронтенд."""
        res = {
            'caption': self.caption,
        }
        if self.action_name is not None:
            res['action_name'] = self.action_name
        if self.url is not None:
            res['url'] = self.url
        if self.icon is not None:
            res['icon'] = self.icon
        if self.actions is not None:
            res['actions'] = [action.dict_() for action in self.actions]
        return res


class FormActions:
    """Общий класс для взаимодействия с действиями над отчетными формами."""

    def __init__(self, actions: List[FormAction]):
        self.actions = actions

        self.form_statuses_to_actions = {}
        for action in self.actions:
            if action.allowed_form_statuses is None:
                for form_status in FormModel.Data.FormStateStatuses:
                    self.form_statuses_to_actions.setdefault(form_status, []).append(action)
            else:
                for form_status in action.allowed_form_statuses:
                    self.form_statuses_to_actions.setdefault(form_status, []).append(action)

        self.actions_names_to_actions = {}
        for action in self.actions:
            if isinstance(action, DropDownFormAction):
                for sub_action in action.actions:
                    self.actions_names_to_actions[sub_action.action_name] = sub_action
            else:
                self.actions_names_to_actions[action.action_name] = action

    def dict_(self, form, user, api2_actions=None):
        """Получить информацию о всех в формате словаря для передачи на фронтенд. (Аналог репозитория)."""
        actions = self.get_form_actions(form, user)
        res = {}
        for action in actions:
            if isinstance(action, FormAction):
                if action.bottom_tab is None:
                    res.setdefault('toolbar', []).append(action.dict_())
                else:
                    res.setdefault('bottom_tabs', {}).setdefault(action.bottom_tab.value, []).append(action.dict_())
            elif isinstance(action, DropDownFormAction):
                res.setdefault('toolbar', []).append(action.dict_())
        if api2_actions is not None:
            res.setdefault('toolbar', []).extend(api2_actions)
        return res

    def get_form_actions(self, form, user):
        """Получить действия, на основе пользователя и формы. (Учитываются статусы, роли пользователя и т.д.)."""
        actions_for_form = self.form_statuses_to_actions.get(form.form_state, [])
        if not actions_for_form:
            return actions_for_form

        actions = list(filter(lambda act: act.allowed_user_roles is None or user.user_role in act.allowed_user_roles,
                              actions_for_form))

        res_actions = []
        for act in actions:
            if not act.form_nos:
                res_actions.append(act)
                continue
            for form_no in act.form_nos:
                if form_no in form.form_no:
                    res_actions.append(act)

        return res_actions

    def get_func_by_action_name(self, action_name):
        """Получить функцию обработчик, привязанную к данному действию."""
        return self.actions_names_to_actions[action_name]


FORMS_ACTIONS = FormActions(
    actions=[
        FormAction('to_confirm', 'Подтвердить', to_confirm,
                   allowed_form_statuses=[form_statuses.VERIFIED, form_statuses.REMARKS],
                   allowed_user_roles=[
                       user_roles.CHIEF_OPERATOR,
                       user_roles.ADMINISTRATOR,
                       user_roles.VOC_OPERATOR,
                   ],
                   text='Форма подтверждена'),
        FormAction('to_cancel_confirm', 'Отменить подтверждение', to_cancel_confirm,
                   allowed_form_statuses=[form_statuses.CONFIRMED],
                   allowed_user_roles=[
                       user_roles.CHIEF_OPERATOR,
                       user_roles.ADMINISTRATOR,
                       user_roles.VOC_OPERATOR,
                   ],
                   text='Отменено подтверждение формы'),
        # == переводы ==
        FormAction('iou_to_rao', 'ИОУ в РАО', func=iou_to_rao, form_nos=['form_1_2']),
        FormAction('ori_to_rao', 'ОРИ в РАО', func=ori_to_rao, form_nos=['form_1_3']),
        FormAction('orp_to_rao', 'ОРП в РАО', func=orp_to_rao, form_nos=['form_1_4']),
        FormAction('form_f3_2_to_f1_1', 'в форму 1.1', func=form_f3_2_to_f1_1, form_nos=['form_3_2']),
        FormAction('form_3_1_to_inform', 'отобразить в уведомление', func=form_3_1_to_inform, form_nos=['form_3_1']),
        FormAction('form_3_2_to_inform', 'отобразить в уведомление', func=form_3_2_to_inform, form_nos=['form_3_2']),
        # == экспорт ошибок xlsx как икнока в табе "ошибки"
        FormAction('export_errors_to_xlsx', None, func=export_error_to_xlsx,
                   response_type=FormActionsResponseTypes.XLSX_WORKBOOK, bottom_tab=FormActionTabs.form_errors,
                   extra={
                       'icon': '/interface/upload.svg',
                       'class': 'button-opacity',  # иконка
                       'title': 'Экспортировать ошибки в xlsx',  # подсказка при наведении
                   },
                   extra_kwargs={
                       'download_name_gen_func': build_download_name_for_form_errors_file,
                   }),
    ],
)


def exec_action(request, action_name_, user, form_id):
    action: FormAction = FORMS_ACTIONS.get_func_by_action_name(action_name_)
    try:
        action.checking_permissions_before_executing(request, user, form_id)
        res = action.func(request, user, form_id)
        if res is not None:
            return action.response_type, res, action.extra_kwargs
        return action.response_type, action.result(), action.extra_kwargs
    except FormsActionsException as e:
        return FormActionsResponseTypes.JSON, action.error_result(e), None
