import datetime

from openpyxl import Workbook

from exportmanager.templateengine.xlsx.utils import thin_border
from repository.forms import forms_repository
from repository.forms.exceptions import ModelItemNotFoundException
from services.services import to_exp
from wsgi.apps.reporting_forms.models import FormModel
from wsgi.apps.reporting_forms.services.forms_actions.exceptions import FormsActionsException
from wsgi.apps.reporting_forms.services.forms_actions.services import try_to_convert, get_empty_row, fill_with_dash, \
    set_nucleoids, \
    copy_form_to_another, form_to_form, form_response, get_source_form_row
from wsgi.apps.users.models import UserModel
from wsgi.utils import api2_func_wrapper


def to_confirm(conn, user: UserModel.Data, form_id):
    form = FormModel(conn).read(form_id)
    form.form_state = FormModel.Data.FormStateStatuses.CONFIRMED
    FormModel(conn).update(form.report_form_id, form)


def to_cancel_confirm(conn, user: UserModel.Data, form_id):
    form = FormModel(conn).read(form_id)
    form.form_state = FormModel.Data.FormStateStatuses.VERIFIED
    FormModel(conn).update(form.report_form_id, form)


@api2_func_wrapper
def zri_to_ozri(request, form_id):
    """Перевод ЗРИ в ОЗРИ. 1.1 -> 1.5. Возвращает объект созданной отчетной формы в формате словаря."""
    return form_to_form(request, form_id, 'OpCod', '41', 'form_1_5_2022').res_form.as_dict()


@api2_func_wrapper
def zri_to_ori(request, form_id):
    """Из формы 1.1 в 1.3 (ЗРИ в ОРИ). Возвращает объект созданной отчетной формы в формате словаря."""
    form_to_form_res = form_to_form(request, form_id, 'OpCod', '65', 'form_1_3_2022')
    for i, row in enumerate(form_to_form_res.res_form.form_table_data['main_table']):
        row['Arg'] = '2'
    return FormModel(request).update(form_to_form_res.res_form.report_form_id, form_to_form_res.res_form).as_dict()


@form_response
def iou_to_rao(request, user: UserModel.Data, form_id):
    """Перевод ИОУ в РАО. 1.2 -> 1.6"""
    source_form = FormModel(request).read(form_id)
    form_to_form_res = form_to_form(request, form_id, 'OpCod', '41', 'form_1_6_2022')
    res_form = form_to_form_res.res_form
    for i, row in enumerate(res_form.form_table_data['main_table']):
        source_row = get_source_form_row(form_to_form_res, i)
        row['RAOCod'] = '22511302522'
        fill_with_dash(row, ('Sht', 'ActTR', 'ActUR', 'OkpoPrv', 'UpCod'))
        row['RAOCodMax'] = 'уран-238; уран-234; торий-234'
        row['OkpoPIP'] = source_form.author_id_obj.OKPO
        try_to_convert(row, source_row, 'Tonne', lambda r: float(r['Kg']) / 1000)
        try_to_convert(
            row, source_row, 'ActBG',
            lambda r: '(' + to_exp(
                float(r['Kg']) * float('2.50E+07')
            ) + ')'
        )
        try_to_convert(
            row, source_row, 'ActA',
            lambda r: '(' + to_exp(
                float(r['Kg']) * float('1.61E+07')
            ) + ')'
        )
        row['OkpoPIP'] = source_form.author_id_obj.OKPO

    res_form = FormModel(request).update(res_form.report_form_id, res_form)
    return res_form


@form_response
def ori_to_rao(request, user: UserModel.Data, form_id):
    """Перевод ОРИ в РАО. 1.3 -> 1.6"""
    form_to_form_res = form_to_form(request, form_id, 'OpCod', '41', 'form_1_6_2022')
    for i, row in enumerate(form_to_form_res.res_form.form_table_data['main_table']):
        fill_with_dash(row, ('Sht', 'OkpoPrv', 'UpCod'))
        row['BCod'] = row['OkpoPIP'] = form_to_form_res.source_form.author_id_obj.OKPO
        source_row = get_source_form_row(form_to_form_res, i)
        set_nucleoids(request, row, source_row, 'Nuclid', 'Activn')
    res_form = FormModel(request).update(form_to_form_res.res_form.report_form_id, form_to_form_res.res_form)
    return res_form


@form_response
def orp_to_rao(request, user: UserModel.Data, form_id):
    """Из формы 1.4 в 1.6 (ОРП в РАО)"""
    form_to_form_res = form_to_form(request, form_id, 'OpCod', '41', 'form_1_6_2022')
    for i, row in enumerate(form_to_form_res.res_form.form_table_data['main_table']):
        source_row = get_source_form_row(form_to_form_res, i)
        fill_with_dash(row, ('Sht', 'OkpoPrv', 'UpCod'))
        try_to_convert(row, source_row, 'Tonne', lambda r: float(r['Kg']) / 1000)
        row['OkpoPIP'] = form_to_form_res.source_form.author_id_obj.OKPO
        row['RAOCodMax'] = source_row['Nuclid']
        set_nucleoids(request, row, source_row, 'Nuclid', 'Activn')

    return FormModel(request).update(form_to_form_res.res_form.report_form_id, form_to_form_res.res_form)


@form_response
def form_f3_2_to_f1_1(request, user: UserModel.Data, form_id):
    """Создание формы из 3.2 в 1.1"""
    form = FormModel(request).read(form_id)
    new_form = copy_form_to_another(form, 'form_1_1_2022')
    new_form.form_period_start = form.notification_date
    if form.notification_date is not None:
        new_form.form_period_end = form.notification_date + datetime.timedelta(days=10)
    else:
        new_form.form_period_end = None
    new_form.form_table_data = {'main_table': []}
    new_form.form_remarks = {'main_table': []}
    new_form.form_comments = {'main_table': []}
    new_form.form_warnings = {'main_table': []}

    map_ = {
        'main_table': {
            'q13': 'PaspN',
            'q14': 'Typ',
            'q15': 'Numb',
            'q16': 'Nuclid',
            'q17': 'IzgotDate',
            'q18': 'Activn',
            'q20': 'Kateg',
            'q21': 'IzgotOKPO',
        },
        'second_table': {
            'q24': 'PrName',
            'q25': 'UktPrTyp',
            'q26': 'UktPrN',
        }
    }

    form_tables = [k for k in forms_repository.get_form_repository('form_3_2_2022')['tables']]
    for table in filter(lambda t: t in map_, form_tables):
        for i, row in enumerate(form.form_table_data[table]):
            try:
                new_row = new_form.form_table_data['main_table'][i]
                found = True
            except IndexError:
                new_row = get_empty_row('form_1_1_2022')
                found = False
            for f_from, f_to in map_[table].items():
                new_row[f_to] = row[f_from]
                new_row['OpCod'] = '81'
                new_row['Sht'] = len(form.form_table_data['main_table'])
                new_row['OpDate'] = str(form.notification_date) if form.notification_date is not None else ''
            if not found:
                new_form.form_table_data['main_table'].append(new_row)

        for i, row in enumerate(form.form_comments[table]):
            if row['pointer']['col_name'] in map_[table]:
                row['pointer']['col_name'] = map_[table][row['pointer']['col_name']]
                new_form.form_comments['main_table'].append(row)

        for i, row in enumerate(form.form_remarks[table]):
            if row['pointer']['col_name'] in map_[table]:
                row['pointer']['col_name'] = map_[table][row['pointer']['col_name']]
                new_form.form_remarks['main_table'].append(row)

    new_form = FormModel(request).save(new_form)

    return new_form


@form_response
def form_3_1_to_inform(request, user: UserModel.Data, form_id):
    """Отображение формы 3.1 в уведомление."""
    source_form = FormModel(request).read(form_id)
    res_form = source_form.copy()
    res_form.form_no = 'form_30_1_2022'
    res_form_repo = forms_repository.get_form_repository(res_form.form_no)
    res_form.form_name = res_form_repo['form_name']

    res_form.form_main_data = {}
    res_form.form_table_data = {table_name: [] for table_name in res_form_repo['tables']}
    res_form.form_remarks = {table_name: [] for table_name in res_form_repo['tables']}
    res_form.form_comments = {table_name: [] for table_name in res_form_repo['tables']}
    res_form.form_errors = {table_name: [] for table_name in res_form_repo['tables']}

    third_table_attrs = ['g2', 'g4', 'g5', 'g6', 'q7', 'q8', 'q10', 'q11', 'q12', 'q13', 'q14', 'q15', 'q16']
    third_table_attrs_source_to_res = {
        'g7': 'q7',
        'g8': 'q8',
        'g10': 'q10',
        'g11': 'q11',
        'g12': 'q12',
        'g13': 'q13',
        'g14': 'q14',
        'g15': 'q15',
        'g16': 'q16',
    }
    third_table_row = {}
    for attr in third_table_attrs:
        third_table_row[attr] = source_form.form_main_data.get(attr, '')
    for source_k, res_k in third_table_attrs_source_to_res.items():
        third_table_row[res_k] = source_form.form_main_data.get(source_k, '')

    res_form.form_table_data['third_table'].append(third_table_row)

    res_form.form_table_data['fourth_table'] = source_form.form_table_data['main_table']

    res_form = FormModel(request).save(res_form)

    return res_form


@form_response
def form_3_2_to_inform(request, user: UserModel.Data, form_id):
    """Отображение формы 3.2 в уведомление."""
    source_form = FormModel(request).read(form_id)
    res_form = source_form.copy()
    res_form.form_no = 'form_30_2_2022'
    res_form_repo = forms_repository.get_form_repository(res_form.form_no)
    res_form.form_name = res_form_repo['form_name']

    res_form.form_main_data = {}
    res_form.form_table_data = {table_name: [] for table_name in res_form_repo['tables']}
    res_form.form_remarks = {table_name: [] for table_name in res_form_repo['tables']}
    res_form.form_comments = {table_name: [] for table_name in res_form_repo['tables']}
    res_form.form_errors = {table_name: [] for table_name in res_form_repo['tables']}

    third_table_attrs = ['q2', 'q3', 'q4', 'q5', 'q6', 'q7', 'q8', 'q9', 'q10', 'q11', 'q12', 'q13']
    third_table_row = {}
    for attr in third_table_attrs:
        third_table_row[attr] = source_form.form_main_data.get(attr, '')
    res_form.form_table_data['third_table'].append(third_table_row)

    res_form.form_table_data['fourth_table'] = source_form.form_table_data['main_table']
    res_form.form_table_data['fifth_table'] = source_form.form_table_data['second_table']

    res_form = FormModel(request).save(res_form)

    return res_form


def export_error_to_xlsx(request, user: UserModel.Data, form_id):
    """Экспортировать ошибки отчетной формы в xlsx файл"""
    wb = Workbook()
    ws = wb.active

    form = FormModel(request).read(form_id)
    form_repo = forms_repository.get_form_repository(form.form_no)

    table_num = 0
    errors_found = False

    if form.form_errors is not None:
        for table_name, table_errors in form.form_errors.items():
            title = f'Таблица {table_num + 1}'
            if table_num == 0:
                ws.title = title
            else:
                ws = wb.create_sheet(title)

            row_cell = ws.cell(row=1, column=1)
            row_cell.value = 'строка'
            col_cell = ws.cell(row=1, column=2)
            col_cell.value = 'столбец'
            value_cell = ws.cell(row=1, column=3)
            value_cell.value = 'ошибка'

            for cell in (row_cell, col_cell, value_cell):
                cell.border = thin_border

            row = 2
            for error in form.get_sorted_form_errors()[table_name]:
                errors_found = True
                try:
                    row_value = str(int(error['pointer']['row']) + 1)
                except ValueError:
                    # Не указана строка - значит ошибка в атрибутах.
                    col_name = error['pointer']['col_name']
                    try:
                        # Ищем в репозитории формы form_main_attrs по col_name
                        repo_main_attr = next(filter(lambda x: x['field'] == col_name,  form_repo['form_main_attrs']))
                        row_value = repo_main_attr['caption']
                    except StopIteration:
                        row_value = ''

                ws.cell(row=row, column=1).value = row_value
                try:
                    column_num = forms_repository.get_column_num(
                        form.form_no, table_name, error['pointer']['col_name'], for_front=True,
                    )
                except ModelItemNotFoundException:
                    column_num = ''

                ws.cell(row=row, column=1).border = thin_border
                column_num = str(column_num) if column_num is not None else ''
                ws.cell(row=row, column=2).value = column_num
                ws.cell(row=row, column=2).border = thin_border
                ws.cell(row=row, column=3).value = error['value'].get('text', '')
                ws.cell(row=row, column=3).border = thin_border
                row += 1

    if not errors_found:
        raise FormsActionsException(rus_message='Форма не содержит ошибок')

    return wb
