import functools
from dataclasses import dataclass
from typing import Dict

from repository.forms import forms_repository
from wsgi.apps.reporting_forms.models import FormModel
from wsgi.apps.reporting_forms.services.forms_actions.exceptions import FormsActionsException
from wsgi.apps.reporting_forms.services.services import get_form_items_by_row_num
from wsgi.apps.users.models import UserModel
from wsgi.apps.voc.models import VocValueModel


def smth_to_smth(func):
    def inner(request, user: UserModel.Data, form_id):
        new_form: FormModel.Data = func(request, user, form_id)
        new_form = FormModel(request).save(new_form, commit=True)
        return {'url': f'/editor-forms/{new_form.report_form_id}'}
    return inner


def form_response(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
        form = func(*args, **kwargs)
        return {'url': f'/editor-forms/{form.report_form_id}'}
    return inner


@dataclass
class FormToFormStructure:
    source_form: FormModel.Data
    res_form: FormModel.Data
    res_form_row_num_to_source_form_row_num: Dict[int, int]


def form_to_form(request, source_form_id, source_code_field_name, code_value,  res_form_no, form_map_source_to_res=None) -> FormToFormStructure:
    """Перевести форму в форму.

    Args:
        request: Запрос.
        source_form_id: ID исходной формы.
        source_code_field_name: Атрибут в строках таблицы формы, по которому получить код операции.
        code_value: Значение кода операции в строках формы, которые необходимо перевести
            (например, перевести все строки с кодом 65).
        res_form_no: form_no результирующей формы в формате 13 приказа (*_2022)ю
        form_map_source_to_res: Map для переноса. Соответсвие полей исходной формы -> поля результирующей.

    Returns:
        Результирующая форма.

    """
    if form_map_source_to_res is None:
        form_map_source_to_res = {}

    source_form = FormModel(request).read(source_form_id)
    res_form_repo = forms_repository.get_form_repository(res_form_no)
    res_form_fields = forms_repository.get_form_fields(res_form_repo)

    rows_to_add = []
    items_to_add = {
        'form_comments': [],
        'form_remarks': [],
    }

    count_row_source = 0
    count_row_res = 0

    res_form_row_num_to_source_form_row_num = {}

    for row in source_form.form_table_data['main_table']:
        if row[source_code_field_name] == code_value:

            new_res_row = {}
            for k, v in row.items():
                if k in res_form_fields:
                    new_res_row[k] = v
                elif k in form_map_source_to_res:
                    new_res_row[form_map_source_to_res[k]] = v
            rows_to_add.append(new_res_row)
            res_form_row_num_to_source_form_row_num[count_row_res] = count_row_source

            for k in items_to_add:
                row_items = get_form_items_by_row_num(source_form, k, count_row_source)
                for item in row_items:
                    try:
                        col_name_source = item['pointer']['col_name']
                    except KeyError:
                        continue
                    col_name_res = col_name_source
                    if col_name_source in res_form_fields:
                        pass
                    elif col_name_source in form_map_source_to_res:
                        col_name_res = form_map_source_to_res[col_name_source]
                    else:
                        continue
                    item['pointer']['col_name'] = col_name_res
                    item['pointer']['row'] = count_row_res
                    items_to_add[k].append(item)

            count_row_res += 1
        count_row_source += 1

    if not rows_to_add:
        raise FormsActionsException(rus_message=f'Нет учётных единиц с кодом операции {code_value}')

    res_form = source_form.copy()
    res_form.form_state = FormModel.Data.FormStateStatuses.DRAFT
    res_form.form_no = res_form_no
    res_form.form_name = res_form_repo['form_name']
    res_form.form_table_data['main_table'] = rows_to_add
    res_form.form_comments['main_table'] = items_to_add['form_comments']
    res_form.form_remarks['main_table'] = items_to_add['form_remarks']
    res_form.form_errors = {'main_table': []}
    res_form.form_notifications = {'main_table': []}
    res_form = FormModel(request).save(res_form)
    return FormToFormStructure(
        source_form=source_form,
        res_form=res_form,
        res_form_row_num_to_source_form_row_num=res_form_row_num_to_source_form_row_num,
    )


def get_source_form_row(ftf: FormToFormStructure, res_form_i):
    return ftf.source_form.form_table_data['main_table'][ftf.res_form_row_num_to_source_form_row_num[res_form_i]]


def copy_form_to_another(form, form_no) -> FormModel.Data:
    new_form = form.copy()
    new_form.form_no = form_no
    new_form.form_name = forms_repository.get_form_repository(new_form.form_no)['form_name']
    return new_form


def try_to_convert(new_row, row, key, func):
    try:
        new_row[key] = func(row)
    except Exception as e:
        pass


def get_empty_row(form_no):
    form_rep = forms_repository.get_form_repository(form_no)
    return {model_item['field']: '' for model_item in form_rep['tables']['main_table']['model']}


def set_same_fields(new_row, row):
    for f in row:
        if f in new_row:
            print('f exists in new row', f)
            new_row[f] = row[f]


def set_new_row_by_map(new_row, row, map_, with_tables=False):
    for k, v in map_.items():
        new_row[k] = row[v]


def fill_with_dash(new_row, fields):
    for f in fields:
        new_row[f] = '-'


def set_nucleoids(request, new_row, row, field_nuc, field_active):
    nucleoids = []

    for v in row[field_nuc].split('; '):
        for voc_value in VocValueModel(request).filter({'voc_type': 'radionuclide_voc1'}):
            if voc_value.attrs['Name_RN'] == v:
                nucleoids.append(voc_value)

    if len(set(n.attrs.get('Kod_gruppy') for n in nucleoids)) == 1:
        vi = nucleoids[0].attrs.get('Kod_gruppy')
        if vi == 'а':
            new_row['ActA'] = row[field_active]
            fill_with_dash(new_row, ('ActTR', 'ActBG', 'ActUR'))
        elif vi == 'б':
            new_row['ActBG'] = row[field_active]
            fill_with_dash(new_row, ('ActTR', 'ActA', 'ActUR'))
        elif vi == 'Трансурановый':
            new_row['ActUR'] = row[field_active]
            fill_with_dash(new_row, ('ActTR', 'ActA', 'ActBG'))
        elif vi == 'Тритий':
            new_row['ActTR'] = row[field_active]
            fill_with_dash(new_row, ('ActUR', 'ActA', 'ActBG'))
    else:
        fill_with_dash(new_row, ('ActTR', 'ActUR'))
        new_row['ActA'] = new_row['ActBG'] = f'({row[field_active]})'
