import datetime
import re
from typing import List, Dict, Any

from repository.extensions.extensions_repository import get_method_to_files
from repository.utils import import_module_
from wsgi.apps.reporting_forms.models import FormModel
from wsgi.apps.voc.models import OrganizationVocModel
from wsgi.utils import api2_func_wrapper


def exec(method, params):
    """Выполнить действия отчетной для отчетной формы"""
    path, module_name = get_method_to_files(method)
    path = path[0]
    module_name = module_name.split('.')[0]
    module = import_module_(path, module_name)
    return module.exec(method, params)


def get_oper_form_period_start(request, form_no, org_id):
    """Получить дату начала периода отчетной формы. Берутся все формы с номером form_no принадлежащие организации
    и выбирается максимальная дата конца периода среди этих форм. Если дата не найдена, вернуть 01.01.<current_year>"""
    filter_params = {'form_no': form_no}
    org = OrganizationVocModel(request).read(org_id)
    if org:
        filter_params['kod_org'] = org.KOD_ORG

    forms = FormModel(request).filter(filter_params)
    st = FormModel.Data.FormStateStatuses
    forms = list(filter(lambda x: x.form_state != st.CANCELED and x.form_period_end is not None, forms))

    try:
        form_with_max_period_end_date = max(forms, key=lambda x: x.form_period_end)
        if form_with_max_period_end_date is None:
            raise ValueError
    except ValueError:
        return f'{datetime.datetime.now().year}-01-01'

    return form_with_max_period_end_date.form_period_end


def build_download_name_for_C_CH_file(file_path, forms: List[FormModel.Data]):
    """Сгенерировать имя при скачивании обменного файла"""
    ext = file_path.split('.')[-1]
    cd = datetime.datetime.now().date()

    kod_org = None
    for form in forms:
        if form.author_id_obj.KOD_ORG is not None:
            if kod_org is not None and form.author_id_obj.KOD_ORG != kod_org:
                kod_org = 'many'
                break
            kod_org = form.author_id_obj.KOD_ORG

    if kod_org is None:
        kod_org = 'none'

    return f'A_{kod_org}_{cd.day}_{cd.month}_{cd.year}' + '.' + ext


def build_form_description(form_data, form_data_obj,  org_data, report_rep):
    """Сформировать label для отчетный форм. Содержит информацию об орг. и т.д.
    Формируется на основе репозитория отчета"""
    if 'description' not in report_rep:
        return ''

    label = report_rep['description']
    regex = r'{\w+}'
    fields_in_label = re.findall(regex, label)
    fields_clear = [f[1:-1] for f in fields_in_label]
    fields_zip = zip(fields_in_label, fields_clear)
    for f_l, f_c in fields_zip:
        value = ''
        if f_c in form_data:
            value = form_data_obj.__getattribute__(f_c)
        if f_c in org_data:
            value = org_data[f_c]
        if value is None:
            value = ''
        if isinstance(value, datetime.date) or isinstance(value, datetime.datetime):
            value = value.strftime('%d.%m.%Y')
        elif not isinstance(value, str):
            value = str(value)
        label = label.replace(f_l, value)
    return label


def build_response_for_table_actions_changes(form_before: FormModel.Data, form_after: FormModel.Data, form_total_changes=None) -> Dict[str, Any]:
    """Сформировать response после действий по изменению отченой формы.
    Экшены:
        | "change_cells"
        | "add_row"
        | "delete_rows"
        | "move_row"

    Args:
        form_before: Форма до изменений.
        form_after: Форма после изменений.
        form_total_changes: Изменения формы по примечаниям/комментариям. Содержит только те, которые были
            сохранены/изменены.

    Returns:
        Данные для response.
    """
    resp = {}
    if form_total_changes is not None:
        resp['merge'] = form_total_changes
    if form_before.form_state != form_after.form_state:
        resp.setdefault('changes', {})['form_state'] = form_after.form_state.value
    return resp


@api2_func_wrapper
def validate_form(request, form_id: int) -> Dict[str, Any]:
    """Проверить отчетную форму.

    Args:
        request: Запрос.
        form_id: ID формы в БД.

    Returns:
        Словарь с ошибками по каждой таблице.

    Raises:
        CoreBaseException (или класс наследник от CoreBaseException) при возникновении ошибки.
    """
    form = FormModel(request).read(int(form_id))

    api2_obj = request.storage['api2_clone']

    api2_check_res = api2_obj.check_forms.check(form.as_dict())

    if api2_check_res.get('notifications'):
        res_notifications = {'main_table': []}
        counter = 0
        for note in api2_check_res['notifications']:
            res_notifications['main_table'].append({
                'value': {
                    'text': note,
                },
                # Генерируем уникальный pointer не как объект, а как строку. Необходимо для фронта.
                'pointer': 'notification ' + str(counter),
            })
            counter += 1
        form.form_notifications = res_notifications
        form = FormModel(request).update(form.report_form_id, form)
    if api2_check_res.get('errors'):
        res_errors = {}
        is_errors_found = False
        for table_name, errors in api2_check_res['errors'].items():
            res_errors[table_name] = []
            for error in errors:
                is_errors_found = True
                res_errors[table_name].append({
                    'pointer': error['pointer'],
                    'source': error.get('source'),
                    'value': {
                        'text': error['value'],
                    }
                })
        form.form_errors = res_errors
        form.form_state = FormModel.Data.FormStateStatuses.REMARKS if is_errors_found else FormModel.Data.FormStateStatuses.VERIFIED
        FormModel(request).update(form.report_form_id, form)
        return form.form_errors

    form.form_state = FormModel.Data.FormStateStatuses.VERIFIED
    FormModel(request).update(form.report_form_id, form)

    return form.form_errors


def get_form_items_by_row_num(
        form: FormModel.Data,
        form_item_name,
        row_num: int,
        table_name: str = 'main_table'
):
    """Получить строки примечания, коммента, ... по номеру строки, к которой относится.

    Args:
        form: Форма.
        form_item_name: примечания, коммента, ...
            Доступные значения: 'form_remarks', 'form_comments', 'form_errors'.
        row_num: Номер строки.
        table_name: Имя таблицы.

    Returns:
        Если не найдено - None.

    """
    form_item = form.__getattribute__(form_item_name)
    return list(filter(lambda form_item_: form_item_['pointer']['row'] == row_num, form_item[table_name]))
