import datetime
import re
from datetime import date

from services.exceptions import CoreBaseException
from repository.forms import forms_repository


class ValidationException(CoreBaseException):
    ...


class ValueErrorTypeException(ValidationException):
    type_to_rus = {
        int: 'целочисленный',
        float: 'вещественный',
        'date': 'дата'
    }

    def __init__(self, value, type_, *args: object, **kwargs) -> None:
        super().__init__(
            type='ValueErrorTypeException',
            message=f'ValueErrorTypeException: Value: {value} Expected: {type_} ',
            rus_message=f'Значение имеет некорректный тип. Значение должно иметь {self.type_to_rus.get(type_)} тип'
        )


class RegexDoesNotMathTypeException(ValidationException):
    def __init__(self, help_text, value, *args: object, **kwargs) -> None:
        if not help_text:
            help_text = 'Неверный формат'
        super().__init__(
            type='RegexDoesNotMathTypeException',
            message=f'RegexDoesNotMathTypeException: help_text: {help_text} value: {value} ',
            rus_message=f'{help_text}: {value}'
        )


class ValueRequiredException(ValidationException):
    def __init__(self, *args: object, **kwargs) -> None:
        super().__init__(
            type='ValueRequiredException',
            message=f'ValueRequiredException',
            rus_message=f'Поле обязательно к заполнению'
        )


class ExpTypeException(ValidationException):
    def __init__(self, value, *args: object, **kwargs) -> None:
        super().__init__(
            type='ExpTypeException',
            message=f'ExpTypeException',
            rus_message=f'Значение должно быть экспоненциальным'
        )


def _validate_type_with_handler(value, type_):
    try:
        if 'exponential' in type_['validation']['handlers']:
            value = str(value)
            value = value.strip()
            value = value.replace('\n', '')
            value = value.lower()
            value = value.replace('е', 'e')  # замена русской на англ.
            value = value.replace(',', '.')
            if 'e-' not in value:
                value = value.replace('-', 'e-')
            if 'e+' not in value:
                value = value.replace('+', 'e+')
            if value.startswith('(') and value.endswith(')'):
                value = value[1:-1]
            try:
                float(value)
                print('\t\t\tCOR', value)
            except (ValueError, TypeError):
                print('\t\t\tINCORRETC', value)
                raise ExpTypeException(value)
    except KeyError:  # не exp тип
        return


def _validate_by_base_type(value, type_):
    repository_base_type_to_python_type = {
        'int': int,
        'str': str,
        'float': float,
    }

    try:
        if value in type_['validation']['allow_list']:
            return
    except KeyError:
        pass

    if 'handlers' in type_:
        _validate_type_with_handler(value, type_)
        return

    repository_base_type = type_['validation'].get('type', 'str')

    python_type = repository_base_type_to_python_type[repository_base_type]
    try:
        python_type(value)
        # regex
        if python_type == str and type_['validation'].get('regex'):
            reg = type_['validation']['regex']
            if re.match(reg, value) is None:
                if 'даты' not in type_['help']:  # todo: убрать валидацию даты
                    raise RegexDoesNotMathTypeException(type_['help'], value)
        # required = True
        if 'required' in type_['validation'] and type_['validation']['required']:
            if not value:
                raise ValueRequiredException()
    except (ValueError, TypeError):
        raise ValueErrorTypeException(value, python_type)


def _validate_date(value):
    if value is None:
        raise ValueErrorTypeException('', 'date')
    if 'T' in value:  # todo: убрать
        value = value.split('T')[0]
    try:
        date.fromisoformat(value)
    except ValueError:
        raise ValueErrorTypeException(value, 'date')


def _validate_by_type(value, type_):
    if type_['base_type'] == 'BASE':
        _validate_by_base_type(value, type_)
    elif type_['base_type'] == 'DATE':
        _validate_date(value)


def _get_table_attr_repository_and_col_index(attr, table_model_repository):
    for i, item in enumerate(table_model_repository):
        if item['field'] == attr:
            return i, item


class ValidationError:
    def __init__(self, message, row, col, col_name):
        self.message = message
        self.row = row
        self.col = col
        self.col_name = col_name


class FormValidator:
    def __init__(self, form_data, form_repository, rep_types_, form_no):
        self.form_data = form_data
        self.form_repository = form_repository
        self.rep_types_ = rep_types_
        self.form_no = form_no

        self.form_table_data = self.form_data['form_table_data']
        self.form_repository_tables = self.form_repository['tables']

        self.form_module = forms_repository.import_form_module(self.form_no)

        self.errors = {}

    def validate(self):
        self.errors.update(self._validate_tables())
        self._validate_tables()

    def _validate_tables(self):
        tables_errors = {}
        for table_key, table_rows_data in self.form_table_data.items():
            tables_errors[table_key] = self._validate_table(table_rows_data, table_key,
                                                            self.form_repository_tables[table_key])
        return tables_errors

    def _validate_table(self, table_rows_data, table_key, table_repository):
        table_errors = []
        table_errors += self._validate_table_attrs(table_rows_data, table_repository)
        # if not table_errors:
        #     table_errors += self._validate_table_funcs(table_key, table_rows_data)
        return table_errors

    def _validate_table_attrs(self, table_rows_data, table_repository):
        errors = []
        for i, table_row in enumerate(table_rows_data):
            for attr_key, attr_value in table_row.items():
                try:
                    col_idx, attr_repository = _get_table_attr_repository_and_col_index(attr_key,
                                                                                        table_repository['model'])
                except TypeError:
                    continue
                    # raise CoreBaseException(rus_message=f'Поле {attr_key} есть в данных, но не найдено в репозитории. Невозможно проверить форму')
                try:
                    if 'type' in attr_repository:
                        if self.rep_types_.get(
                                attr_repository['type']) is None:  # todo: fix, no type in repository defined
                            continue
                        _validate_by_type(attr_value, self.rep_types_[attr_repository['type']])
                except ValidationException as e:
                    errors.append(
                        {
                            'value': {
                                'text': e.__dict__['rus_message'],
                            },
                            'pointer': {
                                'row': i,
                                'col': col_idx,
                                'col_name': attr_key,
                            }
                        }
                    )
        return errors
