import json
from datetime import datetime

from repository.forms import forms_repository
from repository.forms.exceptions import FormRepositoryNotFoundException
from repository.forms_journals import forms_journals_repository
from repository.types import types_repository
from services.exceptions import CanNotReadJsonExportFileException, CoreBaseException
from services.temp import TempFile
from settings import TEMP_DIR_PATH
from wsgi.apps.hints.models import HintModel
from wsgi.apps.reporting_forms.models import FormModel
from wsgi.apps.reporting_forms.permissions import check_edit_report_form_permission, check_unload_upload_exchange_files, \
    check_add_remarks_to_reports_forms_permission, check_add_comments_to_reports_forms_permission, \
    check_change_form_hint_permission, check_read_form_permission
from wsgi.apps.reporting_forms.rules import forms_rules
from wsgi.apps.reporting_forms.serializers import FormEditorSerializer, FormSerializer, FormListFilterSerializer, \
    FormsJournalSerializer
from wsgi.apps.reporting_forms.services import file_hosting, services
from wsgi.apps.reporting_forms.services.form_editor import is_change_cells_values_action, \
    is_change_cells_comments_action, is_change_cells_remarks_action, \
    check_is_any_change_table_action
from wsgi.apps.reporting_forms.services.services import build_download_name_for_C_CH_file, \
    build_response_for_table_actions_changes, validate_form
from wsgi.apps.reporting_forms.services.validation import FormValidator
from wsgi.apps.users.models import UserModel
from wsgi.permissions import bool_permission
from wsgi.utils import parse_request, response, api_view, file_response


@parse_request(forms_rules)
def reporting_forms_view_manager(request, *args, **kwargs):
    action = kwargs['action']
    # === GET ===
    if kwargs['action'] == 'read':
        return read_reporting_forms_view(request, kwargs['form_id'])
    elif kwargs['action'] == 'list':
        return list_reporting_forms_view(request)
    elif kwargs['action'] == 'filter':
        return filter_reporting_forms_view(request)
    elif action == 'unload':
        return unload_forms_to_file_view(request, request.args.getlist('form_id'))
    elif action == 'set_forms_expt_time_unloaded_status':
        return set_forms_expt_time_unloaded_status_view(request, request.args.getlist('form_id'))
    elif action == 'unload_cp_file':
        return unload_cp_file_view(request, request.args.getlist('form_id'))
    elif action == 'list_for_journal':
        return list_for_journal_view(request, kwargs['journal_name'], kwargs.get('preset'))
    elif action == 'get_extra_info':
        return get_extra_info_view(request)
    # === POST ===
    elif kwargs['action'] == 'save':
        return save_reporting_forms_view(request, kwargs['body'])
    elif kwargs['action'] == 'update':
        return reporting_forms_update_view(request, kwargs['form_id'], kwargs['body'])
    elif kwargs['action'] == 'delete':
        return reporting_forms_delete_view(request, kwargs['form_id'])
    elif action == 'upload':
        return upload_forms_from_file_view(request)
    elif action == 'exec':
        return exec_forms_view(request, kwargs['body'])
    elif action == 'change_comment':
        return change_comment_view(request, kwargs['form_id'], kwargs['body'])
    # - update table -
    elif action == 'change_cells':
        return change_cells_view(request, kwargs['form_id'], kwargs['table_name'], kwargs['body'])
    elif action == 'change_rows':
        return change_rows_view(request, kwargs['form_id'], kwargs['table_name'], kwargs['body'])
    elif action == 'add_row':
        return add_row_view(request, kwargs['form_id'], kwargs['table_name'], kwargs['body'])
    elif action == 'delete_rows':
        return delete_rows_view(request, kwargs['form_id'], kwargs['table_name'], kwargs['body'])
    elif action == 'move_row':
        return move_row_view(request, kwargs['form_id'], kwargs['table_name'], kwargs['body'])
    elif action == 'update_row_indexes':
        return update_row_indexes_view(request, kwargs['form_id'], kwargs['table_name'], kwargs['body'])
    elif action == 'change_hint':
        return change_hint_view(request, kwargs['form_no'], kwargs['attr'], kwargs.get('table_name'), kwargs['body'])
    # ==CHANGE MAIN ATTRS==
    elif action == 'change_attrs':
        return change_main_attrs_view(request, kwargs['form_id'], kwargs['body'])
    # ==VALIDATION==
    elif action == 'validate':
        return validation_view(request, int(kwargs['form_id']))  # TODO: делать приведение типов в parse_request


@api_view()
def list_reporting_forms_view(request, *args, **kwargs):
    """Вернуть список отчетных форм из бд"""
    return response(request, FormListFilterSerializer(request, kwargs['user']).list(), many=True)


@api_view()
def filter_reporting_forms_view(request, *args, **kwargs):
    """Вернуть список отчетных форм из бд по фильтру"""
    return response(request, FormListFilterSerializer(request, kwargs['user']).filter(request.args), many=True)


@api_view()
def read_reporting_forms_view(request, form_id, *args, **kwargs):
    """Вернуть отчетную форму по `form_id` в базе"""
    check_read_form_permission(kwargs['user'], FormModel(request).read(form_id))
    return response(request, FormSerializer(request, user=kwargs['user']).read(form_id))


@api_view()
def save_reporting_forms_view(request, body, *args, **kwargs):
    """Создать отчетные формы"""
    forms = [FormSerializer(request, kwargs['user']).save(form_data) for form_data in body]
    return response(request, forms, many=True)


@api_view()
def reporting_forms_update_view(request, record_id, body, *args, **kwargs):
    """Обновить отчетную фому в бд по id"""
    check_edit_report_form_permission(kwargs['user'], FormModel(request).read(record_id))
    return response(request, FormSerializer(request, kwargs['user']).update(record_id, body))


@api_view()
def reporting_forms_delete_view(request, record_id, *args, **kwargs):
    """Удалить отчетную фому в бд по id"""
    check_edit_report_form_permission(kwargs['user'], FormModel(request).read(record_id))
    return response(request, FormSerializer(request, kwargs['user']).delete(record_id))


@api_view()
def exec_forms_view(request, body, *args, **kwargs):
    """Выполнить действия отчетной для отчетной формы"""
    return response(request, services.exec(body.pop('method'), body))


@api_view()
def change_comment_view(request, form_id, body, *args, **kwargs):
    form = FormModel(request).read(form_id)
    form.form_comment = {
        'value': body['value']
    }
    FormModel(request).update(form.report_form_id, form)
    return response(request, None)


# == UPLOAD/UNLOAD FILES ==


@api_view()
def unload_forms_to_file_view(request, forms_id, *args, **kwargs):
    """Выгрузить отчетные формы в json файл"""
    # view
    forms = [FormModel(request).read(id_) for id_ in forms_id]
    data_to_unload = file_hosting.Unload(request).unload(forms)

    with TempFile(TEMP_DIR_PATH, '.json') as res_file:
        json.dump(data_to_unload, open(res_file.full_path, 'w', encoding='UTF8'), ensure_ascii=False, default=str)
        return file_response(res_file.full_path,
                             download_name=build_download_name_for_C_CH_file(res_file.full_path, forms))


@api_view()
def set_forms_expt_time_unloaded_status_view(request, forms_id, *args, **kwargs):
    """Задать формам дату выгрузки и поменять их статус на 'выгружены'.
    Эндпоинт для Api2.

    Args:
        forms: Список id форм.
    """
    # validation
    # проверяем права и статус форм
    check_unload_upload_exchange_files(kwargs['user'])
    forms = [FormModel(request).read(id_) for id_ in forms_id]
    allowed_statuses = FormModel.statuses_for_action[FormModel.FormActions.EXPORT_EXCHANGE_FILE]
    if not all(form.form_state in allowed_statuses for form in forms):
        raise CoreBaseException(rus_message=f'Невозможно изменить статус форм на "Выгружены". '
                                            f'Не все формы имеют разрешенный статус.')
    # view
    # изменяем статус форм на "Выгружена", задаем время выгрузки
    for form_id_ in forms_id:
        form = FormModel(request).read(form_id_)
        form.form_state = FormModel.Data.FormStateStatuses.UNLOADED
        form.form_export_ts = datetime.now()
        FormModel(request).update(form.report_form_id, form)
    # ответ для api2 - ошибок при обработке endpoint нет
    return response(request, {'errors': None})


@api_view()
def upload_forms_from_file_view(request, *args, **kwargs):  #
    """Выгрузить данные из файла в систему"""
    # permissions
    check_unload_upload_exchange_files(kwargs['user'])
    # view
    file_ = request.files['file']
    try:
        data = json.load(file_)
    except Exception:
        raise CanNotReadJsonExportFileException()
    file_hosting.Upload(request, data, kwargs['user']).upload()
    return response(request, None)


@api_view()
def unload_cp_file_view(request, form_ids, *args, **kwargs):
    """Получить обменный файл ЦЧ"""
    return response(request, {'message': 'method does not supported now'})


@api_view()
def list_for_journal_view(request, journal_name, preset, *args, **kwargs):
    """Получить список отчетных форм для журнала отчетных форм"""
    journal_repository = forms_journals_repository.get_forms_journal(journal_name)
    if preset is not None:
        preset_query = forms_journals_repository.get_preset_query(journal_name, preset)
        return response(
            request,
            FormsJournalSerializer(request, kwargs['user'], journal_repository).query_condition(preset_query),
            many=True
        )
    return response(request, FormsJournalSerializer(request, kwargs['user'], journal_repository).list(), many=True)


@api_view()
def get_extra_info_view(request, *args, **kwargs):
    """Получить доп информацию для пользователя (в контексте отчетных форм)"""
    resp = {
        'right_to_act': {
            'upload_exchange_file': bool_permission(lambda: check_unload_upload_exchange_files(kwargs['user'])),
            'unload_exchange_file': bool_permission(lambda: check_unload_upload_exchange_files(kwargs['user'])),
        }
    }
    return response(request, resp)


# === CHANGE TABLE ===

@api_view(db_connection_=True)
def change_cells_view(request, form_id, table_name, body, *args, **kwargs):
    """Изменить данные ячеек таблицы"""
    form = FormModel(request).read(form_id)
    check_is_any_change_table_action(body, 'table_cells')
    # permissions
    if is_change_cells_values_action(body, 'table_cells'):
        check_edit_report_form_permission(kwargs['user'], form)
    if is_change_cells_comments_action(body, 'table_cells'):
        check_add_comments_to_reports_forms_permission(kwargs['user'])
    if is_change_cells_remarks_action(body, 'table_cells'):
        check_add_remarks_to_reports_forms_permission(kwargs['user'])
    # view
    form_editor_serializer = FormEditorSerializer(request, kwargs['user'], form_id)
    form_after_changes, form_total_changes = form_editor_serializer.change_cells(table_name, body)
    resp = build_response_for_table_actions_changes(form, form_after_changes, form_total_changes)
    return response(request, resp)


@api_view()
def change_rows_view(request, form_id, table_name, body, *args, **kwargs):
    """Изменить данные строк таблицы. Комменты, примечания, и т.д."""
    form = FormModel(request).read(form_id)
    check_is_any_change_table_action(body, 'table_rows')
    # permissions
    if is_change_cells_values_action(body, 'table_rows'):
        check_edit_report_form_permission(kwargs['user'], form)
    if is_change_cells_comments_action(body, 'table_rows'):
        check_add_comments_to_reports_forms_permission(kwargs['user'])
    if is_change_cells_remarks_action(body, 'table_rows'):
        check_add_remarks_to_reports_forms_permission(kwargs['user'])
    # view
    form_editor_serializer = FormEditorSerializer(request, kwargs['user'], form_id)
    form_after_changes = form_editor_serializer.change_rows(table_name, body)
    resp = build_response_for_table_actions_changes(form, form_after_changes)
    return response(request, resp)


@api_view()
def add_row_view(request, form_id, table_name, body, *args, **kwargs):
    """Добавить строку по индексу"""
    form = FormModel(request).read(form_id)
    # permissions
    check_edit_report_form_permission(kwargs['user'], form)
    # view
    form_editor_serializer = FormEditorSerializer(request, kwargs['user'], form_id)
    form_after_changes = form_editor_serializer.add_row(table_name, body)
    resp = build_response_for_table_actions_changes(form, form_after_changes)
    return response(request, resp)


@api_view()
def delete_rows_view(request, form_id, table_name, body, *args, **kwargs):
    """Удалить строки по индексам"""
    form = FormModel(request).read(form_id)
    # permissions
    check_edit_report_form_permission(kwargs['user'], form)
    # view
    form_editor_serializer = FormEditorSerializer(request, kwargs['user'], form_id)
    form_after_changes = form_editor_serializer.delete_rows(table_name, body)
    resp = build_response_for_table_actions_changes(form, form_after_changes)
    return response(request, resp)


@api_view()
def move_row_view(request, form_id, table_name, body, *args, **kwargs):
    """Переместить строку по индексу вверх или вниз"""
    # permissions
    check_edit_report_form_permission(kwargs['user'], FormModel(request).read(form_id))
    # views
    form_editor_serializer = FormEditorSerializer(request, kwargs['user'], form_id)
    form_editor_serializer.move_row(table_name, body)
    return response(request, {})


@api_view()
def update_row_indexes_view(request, form_id, table_name, body, *args, **kwargs):
    """Обновить индексы строк таблицы с новых на старые"""
    # permissions
    check_edit_report_form_permission(kwargs['user'], FormModel(request).read(form_id))
    # view
    form_editor_serializer = FormEditorSerializer(request, kwargs['user'], form_id)
    form_editor_serializer.update_row_indexes(table_name, body)
    return response(request, None)


@api_view()
def change_hint_view(request, form_no, attr, table_name, body, *args, **kwargs):
    """Изменить подсказку для ячеек таблицы или атрибутов отчетной формы"""
    # permissions
    check_change_form_hint_permission(kwargs['user'])
    # view
    try:
        form_repository = forms_repository.get_form_repository(form_no)
    except FormRepositoryNotFoundException:
        raise CoreBaseException(message=f'form repository not found with form_no: {form_no}')

    if table_name is not None:
        if table_name not in form_repository['tables']:
            raise CoreBaseException(message=f'table_name: {table_name} not found in form_no: {form_no}')
        model_fields = tuple(item['field'] for item in form_repository['tables'][table_name]['model'])
        if attr not in model_fields:
            raise CoreBaseException(message=f'attr with name: {attr} not found in table_name: {table_name} '
                                            f'of form: {form_no}. allowed fields: {model_fields}')
    else:
        attr_fields = tuple(item['field'] for item in form_repository['form_main_attrs'])
        if attr not in attr_fields:
            raise CoreBaseException(message=f'attr: {attr} not found in form_main_attrs of form with '
                                            f'form_no: {form_no}. allowed fields: {attr_fields}')

    hints = HintModel(request).filter({
        'item_type': [HintModel.Data.ItemTypeTypes.FORM.value],
        'item_class': [form_no],
    })

    if hints:  # подсказка уже существует
        hint = hints[0]
        if table_name is not None:
            hint.attrs['tables'][table_name][attr] = body['text']
        else:
            hint.attrs['form_main_attrs'][attr] = body['text']
        HintModel(request).update(hint.record_id, hint)
    else:
        hint_data = {
            'item_type': HintModel.Data.ItemTypeTypes.FORM.value,
            'item_class': form_no,
            'attrs': {
                'form_main_attrs': {},
                'tables': {k: {} for k in form_repository['tables']}
            },
        }
        if table_name is not None:
            hint_data['attrs']['tables'][table_name][attr] = body['text']
        else:
            hint_data['attrs']['form_main_attrs'][attr] = body['text']
        HintModel(request).save(HintModel.Data.from_dict(hint_data))

    return response(request, None)


# ==CHANGE MAIN ATTRS==

@api_view(db_connection_=True)
def change_main_attrs_view(request, form_id, body, *args, **kwargs):
    # permissions
    check_edit_report_form_permission(kwargs['user'], FormModel(request).read(form_id))
    # views
    form_editor_serializer = FormEditorSerializer(request, kwargs['user'], form_id)
    form_editor_serializer.change_main_attrs(body)
    return response(request, None)


# ==VALIDATION==

@api_view()
def validation_view(request, form_id, *args, **kwargs):
    """Провалидировать отчетную форму"""
    validation_res = validate_form(request, form_id)
    errors_not_found = True
    for table_name, errors in validation_res.items():
        if errors:
            errors_not_found = False
    if errors_not_found:
        return response(request, [])
    return response(request, validation_res)
