from datetime import datetime
from enum import Enum

from wsgi.apps.reporting_forms.models import FormModel
from wsgi.apps.user_logs.models import UserLogsModel
from wsgi.apps.users.models import UserModel
from wsgi.apps.voc.models import VocValueModel
from wsgi.utils import api2_func_wrapper


class UserLogsActions(Enum):
    LOGIN = 'login'
    LOGOUT = 'logout'
    EXPORT = 'export'
    IMPORT_HISTORICAL_DATA = 'import_historical_data'
    IMPORT_EXCHANGE_FILE = 'import_exchange_file'


class UserLogsLogger:

    def __init__(self, request, user: UserModel.Data):
        self.user_log_obj = UserLogsModel.Data.empty()
        self.user_log_obj.attrs = {
            'ip': str(request.environ.get('HTTP_X_FORWARDED_FOR', 'не найден')),
            'user_agent': str(request.__dict__.get('user_agent', '')),
            'date_time': str(datetime.now())
        }
        self.request = request
        self.user = user
        self.conn = request.storage['conn']

        if self.user:
            self.user_log_obj.attrs['login'] = self.user.login

    def log_successful_login_action(self):
        self.user_log_obj.attrs['action'] = UserLogsActions.LOGIN.value
        self.user_log_obj.attrs['info'] = 'Успешный вход в систему'
        UserLogsModel(self.request).save(self.user_log_obj)

    def log_incorrect_login(self, login):
        self.user_log_obj.attrs['action'] = UserLogsActions.LOGIN.value
        self.user_log_obj.attrs['info'] = f'Попытка входа с несуществующим логином: "{login}"'
        UserLogsModel(self.request).save(self.user_log_obj)

    def log_incorrect_password(self, password):
        self.user_log_obj.attrs['action'] = UserLogsActions.LOGIN.value
        self.user_log_obj.attrs['info'] = f'Попытка входа с неверным паролем: "{password}"'
        UserLogsModel(self.request).save(self.user_log_obj)

    def log_logout(self):
        self.user_log_obj.attrs['action'] = UserLogsActions.LOGIN.value
        UserLogsModel(self.request).save(self.user_log_obj)

    def log_export_form(self, export_form_id, format_):
        self.user_log_obj.attrs['action'] = UserLogsActions.EXPORT.value
        self.user_log_obj.attrs[
            'info'] = f'Экспорт отчетной формы. Идентификатор формы: {export_form_id}, формат: {format_}'
        UserLogsModel(self.request).save(self.user_log_obj)

    def log_import_historical_data(self, imported_forms_ids, imported_orgs_ids, log=None):
        log_obj = log if log is not None else self.user_log_obj
        log_obj.attrs['action'] = UserLogsActions.IMPORT_HISTORICAL_DATA.value
        log_obj.attrs['info'] = f"""
        Импорт исторических данных.
        
        Всего форм в файле: {len(imported_forms_ids)},
        ID импортированных форм: {imported_forms_ids},
        
        Всего организаций в файле: {len(imported_orgs_ids)},
        ID импортированных организаций: {imported_orgs_ids}.
        """
        if log:
            return UserLogsModel(self.request).update(log_obj.record_id, log_obj, commit=True)
        return UserLogsModel(self.request).save(log_obj, commit=True)

    def log_import_exchange_file(self, imported_forms_ids, imported_orgs_ids, log=None):
        log_obj = log if log is not None else self.user_log_obj
        log_obj.attrs['action'] = UserLogsActions.IMPORT_EXCHANGE_FILE.value
        log_obj.attrs['info'] = f"""
                Импорт из обменного файла.

                Всего форм в файле: {len(imported_forms_ids)},
                ID импортированных форм: {imported_forms_ids},

                Всего организаций в файле: {len(imported_orgs_ids)},
                ID импортированных организаций: {imported_orgs_ids}.
                """
        if log:
            return UserLogsModel(self.request).update(log_obj.record_id, log_obj, commit=True)
        return UserLogsModel(self.request).save(log_obj, commit=True)


@api2_func_wrapper
def delete_db_items_with_log_id(request, log_id: int) -> None:
    """Удалить записи ReportForms, VocValues, в которых log_id=`log_id`.

    Args:
        request: Запрос.
        log_id (int): ID записи лога (таблица user_logs).

    Returns:
        None

    """
    for form in FormModel(request).filter({'log_id': log_id}):
        FormModel(request).delete(form.report_form_id)
    for voc in VocValueModel(request).filter({'log_id': log_id}):
        VocValueModel(request).delete(voc.record_id)
