from dataclasses import dataclass
from enum import Enum
from typing import List

from loguru import logger

import settings
from datasource.databaserepository.database_repository2 import DatabaseRepository2, D
from wsgi.apps.voc.models import OrganizationVocModel


class UserModel(DatabaseRepository2['UserModel.Data']):
    @dataclass
    class Data(DatabaseRepository2.Data):
        class UserRoleTypes(Enum):  # todo: выносить в репозиторий? в базе данных должен быть enum
            OPERATOR = 'operator'
            CHIEF_OPERATOR = 'chief_operator'
            ADMINISTRATOR = 'administrator'
            VOC_OPERATOR = 'voc_operator'
            ANALYST = 'analyst'

        record_id: int
        fio: str
        is_blocked: bool
        login: str
        password: str
        cookie: str
        user_role: UserRoleTypes
        permissions: int
        creator_id: str
        kod_org: str
        kod_org_default: str
        org_id: int
        organizations_ids: str
        attrs: dict

        ID_FIELD = 'record_id'

        @property
        def permission_as_binary(self):
            return "{0:b}".format(self.permissions).zfill(13)

        @property
        def is_admin(self):
            return self.user_role == self.UserRoleTypes.ADMINISTRATOR

        @property
        def kod_orgs_default_as_list(self):
            return self.kod_org_default.split('; ')

        @property
        def org_active_version_by_kod_org_obj(self) -> OrganizationVocModel.Data:
            """Получить активную версию организации по коду пользователя.
            Если активной не будет найдено, вернется последняя.
            Если организаций с таким кодом найдено не будет - вернется None.

            """
            orgs = OrganizationVocModel(self.request).filter(['=', 'KOD_ORG', self.kod_org])
            for org in orgs:
                if not org.valid_to:
                    return org
            if orgs:
                return orgs[-1]

        @property
        def all_kod_orgs(self) -> List[str]:
            """Все коды организаций, к которым привязан пользователь (доступная и основная)"""
            all_kod_orgs = set()
            if self.kod_org and self.kod_org != '0':
                all_kod_orgs.add(self.kod_org)
            if self.kod_org_default:
                for kod in self.kod_org_default.split('; '):
                    all_kod_orgs.add(kod)
            return list(all_kod_orgs)

        def as_dict(self):
            res = super().as_dict()
            if self.user_role:
                res['user_role'] = self.user_role.value
            return res

    database_table_name = 'users'

    def update(self, record_id, obj: Data, commit=False, *args, **kwargs) -> D:
        user_before = self.read(record_id)
        res = super().update(record_id, obj, commit, args, kwargs)
        if settings.DEBUG:
            if user_before.cookie and res.cookie and user_before.cookie != res.cookie:
                logger.warning(f'Cookie пользователя поменялись. Cookie before update: {user_before.cookie}, '
                               f'cookie after update: {res.cookie}. '
                               f'User before: {user_before}, '
                               f'User after: {res}')
        return res

    @dataclass
    class GlobalActions:
        CONFIRM_CANCEL_CONFIRM_FORM = 'подтверждение/отмена подтверждения отчётной формы'

    ACTIONS_TO_ROLES = {
        GlobalActions.CONFIRM_CANCEL_CONFIRM_FORM: [  # todo: remove
            Data.UserRoleTypes.CHIEF_OPERATOR,
            Data.UserRoleTypes.ADMINISTRATOR,
            Data.UserRoleTypes.VOC_OPERATOR,
        ]
    }
