import enum
import itertools
import re
from dataclasses import fields

from werkzeug.security import generate_password_hash

from datasource.databaserepository.exceptions import RecordNotFoundException
from repository.types import types_repository
from repository.vocs import vocs_repository
from services.exceptions import CoreBaseException
from wsgi.apps.users.models import UserModel
from wsgi.apps.users.services import get_org_active_version, get_orgs_ids_line
from wsgi.apps.voc.models import OrganizationVocModel
from wsgi.serializers import BaseSerializer


def org_id2_filter_override(request, value):
    res = '0'
    try:
        if value:
            res = str(OrganizationVocModel(request).read(value).KOD_ORG)
    finally:
        return res


def organizations_ids2_filter_override(request, value):
    """Фильтруем по кодам привязанных организаций, поле kod_org_default"""
    try:
        res = []
        kod_orgs = set()
        for id_ in value.split('; '):
            if id_:
                kod_orgs.add(str(OrganizationVocModel(request).read(id_).KOD_ORG))
        for subset in itertools.permutations(kod_orgs):
            res.append('; '.join(subset))
        return res
    except Exception:
        return ''


class UsersSerializer(BaseSerializer):
    model = UserModel(None)

    class PermissionActions(enum.Enum):
        # модуль отчетности
        EDIT_ORG = (0, 'Редактирование информации об организации')
        EDIT_CREATED_REPORT_FORMS = (1, 'Редактирование отчетных форм, созданных пользователем')
        EDIT_ALL_REPORT_FORMS = (2, 'Редактирование отчетных форм, созданных любым пользователем')
        UPLOAD_UNLOAD_FILE = (3, 'Загрузка и выгрузка форм отчетности в формате обменного файла')
        EDIT_REMAKRS_reports_forms = (4, 'Заполнение примечаний к отчетных формам')
        EDIT_COMMENTS_REPORTS_FORMS = (5, 'Заполнение комментариев к отчетным формам')
        # модуль справочников
        EDIT_SYSTEM_VOCS = (6, 'Редактирование системных справочников')
        EDIT_NOT_SYSTEM_VOCS = (7, 'Редактирование общих справочников')
        # модуль управления
        EDIT_USERS_VOCS = (8, 'Создание и редактирование учетных записей пользователей')
        EDIT_INTERFACE = (9, 'Изменение внешнего вида интерфейса')
        DATABASE_CREATION = (10, 'Создание базы данных')
        # модуль импорта данных
        LOAD_ARCHIVE_DATA = (11, 'Загрузка архивных данных (базы данных оперативного учета)')
        # модуль аналитической отчётности
        CREATION_AREPORTS = (12, 'Формирование аналитических отчетов')

    roles = {
        'operator': 'оператор',
        'chief_operator': 'главный оператор',
        'administrator': 'администратор',
        'voc_operator': 'оператор справочников',
        'analyst': 'аналитик',
    }

    roles_to_permissions = {
        'operator': '010011_01_010_0_0',
        'chief_operator': '111111_01_010_0_1',
        'administrator': '111111_11_111_1_1',
        'voc_operator': '111111_01_010_0_1',
        'analyst': '000000_00_000_0_1',
    }

    change_filter_objs = {
        ('=', 'attrs.org_id2'): ('=', 'attrs.kod_org', org_id2_filter_override),
        ('=', 'attrs.organizations_ids2'): ('in', 'attrs.kod_org_default', organizations_ids2_filter_override),
    }

    def __init__(self, request, user=None):
        super(UsersSerializer, self).__init__(request)
        self.user: UserModel.Data = user
        self.repository = vocs_repository.get_voc_repository('users')

    def _response(self, resp):
        resp = super(UsersSerializer, self)._response(resp)
        resp.pop('cookie', None)
        resp.pop('password', None)
        resp.pop('permissions', None)
        if resp['kod_org']:
            try:
                org = get_org_active_version(self.request, resp['kod_org'])
                if org:
                    resp['default_org'] = {
                        'label': org.NAME_SHORT,
                        'value': org.record_id,
                    }
                else:
                    pass
            except RecordNotFoundException:
                pass  # не найдена организация или пользователь не привязан к организации

        return resp

    def _validate(self, data, instance=None):
        if not instance:
            if not data.get('login'):
                raise CoreBaseException(rus_message='Логин пользователя обязателен к заполнению')
            if not data.get('password'):
                raise CoreBaseException(rus_message='Пароль пользователя обязателен к заполнению')

        if self.model.filter({'login': data['login']}):
            raise CoreBaseException(rus_message='Невозможно создать пользователя.'
                                                'Пользователь с таким логином уже существует')

        # Проверка пароля на правила регулярки, заданные в репозитории
        if data.get('password'):
            # print('\t\tPASSWORD', data['password'])
            login_model = next(filter(lambda m: m['field'] == 'password', [model_item for model_item in self.repository['model']]))
            if 'type' in login_model:
                type_repo = types_repository.get_types_repository()[login_model['type']]
                if 'validation' in type_repo and 'regex' in type_repo['validation']:
                    if not re.match(type_repo['validation']['regex'], data['password']):
                        raise CoreBaseException(rus_message=type_repo.get('help', 'Недопустимый пароль'))

    def _prepare(self, data):
        super()._prepare(data)
        data.pop('cookie', None)
        return data

    def _common(self, data, obj: 'model.Data'):
        if data['attrs'].get('org_id2'):
            try:
                obj.kod_org = OrganizationVocModel(self.request).read(data['attrs']['org_id2']).KOD_ORG
            except:
                pass
        else:
            obj.kod_org = 0
        if data['attrs'].get('organizations_ids2'):
            res_orgs = []
            for org_id_ in data['attrs']['organizations_ids2'].split('; '):
                if org_id_:
                    try:
                        kod_org = OrganizationVocModel(self.request).read(org_id_).KOD_ORG
                        if kod_org:
                            res_orgs.append(kod_org)
                    except:
                        pass
            obj.kod_org_default = '; '.join(res_orgs)
        else:
            obj.kod_org_default = ''

    def update(self, record_id, data):
        instance = self.model.read(record_id)
        obj = instance.replace(data)

        if data.get('password'):
            obj.password = generate_password_hash(data['password'])

        obj.permissions = int(self.roles_to_permissions[obj.user_role.value], 2)

        return obj

    def save(self, data):
        obj = self.model.Data.from_dict(data)
        obj.is_blocked = False
        obj.password = generate_password_hash(data['password'])
        try:
            user_role = data.get('user_role', self.model.Data.UserRoleTypes.OPERATOR.value)
            user_role = self.model.Data.UserRoleTypes(user_role)
            obj.permissions = int(self.roles_to_permissions[user_role.value], 2)
        except KeyError:
            raise CoreBaseException(rus_message=f'Роль пользователя не найдена: {obj.user_role.value}')
        obj.creator_id = self.user.login
        obj.user_role = user_role

        return obj


class UsersVocSerializer(UsersSerializer):
    write_only_fields = ['password']
    not_include_in_resp_attrs = ['record_id', 'cookie', 'value_id', 'label', 'attrs', 'kod_org', 'kod_org_default']

    def _prepare(self, data):
        super()._prepare(data)
        if data.get('attrs'):
            to_pop = []
            for attr, value in data['attrs'].items():
                if attr in {field.name: field.type for field in fields(self.model.Data)}:
                    data[attr] = value
                    to_pop.append(attr)
            for attr in to_pop:
                data['attrs'].pop(attr)

    def _response(self, resp):
        resp = super()._response(resp)
        for field in self.model.Data.fields():
            if field not in self.not_include_in_resp_attrs and field in resp:
                resp['attrs'][field] = resp[field]
                resp.pop(field, None)
        resp['voc_type'] = 'users'
        resp['edit'] = {
            'edit': self.user.is_admin if self.user else False,
        }
        resp['value_id'] = resp['attrs']['login']
        resp['label'] = resp['attrs']['fio']
        if resp['kod_org']:
            org_active_version = get_org_active_version(self.request, resp['kod_org'])
            resp['attrs']['org_id2'] = str(org_active_version.record_id) if org_active_version else ''
        if resp['kod_org_default']:
            resp['attrs']['organizations_ids2'] = get_orgs_ids_line(self.request, resp['kod_org_default'])
        return resp

    def list(self):
        # todo: отдельный сериализатор для админа
        if self.user.is_admin:
            return super().list()
        return self.filter({'org_id': self.user.org_id})
