import datetime
from dataclasses import dataclass
from enum import Enum
from typing import List

from datasource.databaserepository.database_repository2 import DatabaseRepository2


class VocModel(DatabaseRepository2['VocModel.Data']):
    @dataclass
    class Data(DatabaseRepository2.Data):
        voc_id: int
        voc_name: str
        voc_type: str

        ID_FIELD = 'voc_id'

    database_table_name = 'Vocs'


class VocValueModel(DatabaseRepository2['VocValueModel.Data']):
    @dataclass
    class Data(DatabaseRepository2.Data):
        class ValueStatusesTypes(Enum):
            DRAFT = 0
            ACTIVE = 1
            OUTDATED = 2

        record_id: int
        voc_type: str
        value_id: str
        label: str
        attrs: dict
        valid_from: datetime.date
        valid_to: datetime.date
        value_status: ValueStatusesTypes
        creator_id: str
        # kod_org: str
        import_author: str  # код организации создателя
        import_src_id: str  # ID записи в исходной БД
        import_hash: str  # хеш данных
        log_id: int

        ID_FIELD = 'record_id'

        def as_dict(self):
            res = super().as_dict()
            if self.value_status:
                res['value_status'] = self.value_status.value
            return res

    database_table_name = 'VocValues'


class OrganizationVocModel(DatabaseRepository2['OrganizationVocModel.Data']):
    @dataclass
    class Data(VocValueModel.Data):
        class PrznOpTypes(Enum):
            ENTITY = 0
            BRANCH_REPRESENTATIVE_OFFICE = 1

        class OrgTypeTypes(Enum):

            RIAC = '1'
            VIAC = '2'
            CIAC = '3'

        class PrznActvnTypes(Enum):
            NONE = None
            NOT_SHOW_IN_ORGS_LIST = 0
            SHOW_IN_ORGS_LIST = 1

        KOD_ORG: str
        ORG_TYPE: OrgTypeTypes
        KOD_IAE: str
        KOD_SRF: str
        NAME_FULL: str
        NAME_SHORT: str
        ADDRESS_U: str
        ADDRESS_P: str
        RUKOVODITEL: str
        OKPO: str
        OKVED: str
        OKATO: str
        OKTMO: str
        INN: str
        KPP: str
        OKOPF: str
        OKFS: str
        PRIM: str
        PRZN_ORG: int
        PRZN_OP: PrznOpTypes
        KOD_GOL_ORG: str
        PRZN_ACTIV: PrznActvnTypes
        LVL_ORG: int
        OKPO_PRP: str
        RIAC: str
        VIAC: str
        OUIAE: str
        DT_KORR: datetime.date
        USER_KORR: str
        DT_REG: datetime.date
        DT_SU: datetime.date
        category1: str
        category2: str
        category3: str
        category4: str
        category5: str
        YROO: bool

        def as_dict(self):
            res = super().as_dict()
            if self.PRZN_OP is not None:
                res['PRZN_OP'] = self.PRZN_OP.value
            if self.ORG_TYPE is not None:
                res['ORG_TYPE'] = self.ORG_TYPE.value
            if self.PRZN_ACTIV is not None:
                res['PRZN_ACTIV'] = self.PRZN_ACTIV.value
            return res

        @property
        def is_ciac(self):
            return self.ORG_TYPE == OrganizationVocModel.Data.OrgTypeTypes.CIAC

        @property
        def is_viac(self):
            return self.ORG_TYPE == OrganizationVocModel.Data.OrgTypeTypes.VIAC

        @property
        def is_riac(self):
            return self.ORG_TYPE == OrganizationVocModel.Data.OrgTypeTypes.RIAC

        def get_all_orgs_versions_objs(self, reverse_sort=False, except_self=False) -> List['OrganizationVocModel.Data']:
            orgs = OrganizationVocModel(self.request).filter({'KOD_ORG': self.KOD_ORG})
            if except_self:
                orgs = list(filter(lambda x: x != self, orgs))

            orgs_with_valid_from_not_none = list(filter(lambda org: org.valid_from, orgs))
            orgs_with_valid_from_none = list(filter(lambda org: not org.valid_from, orgs))
            sorted_orgs_with_valid_from_not_none = sorted(
                orgs_with_valid_from_not_none,
                key=lambda org: org.valid_from,
            )
            total = sorted_orgs_with_valid_from_not_none + orgs_with_valid_from_none
            return total

        @property
        def KOD_GOL_ORG_obj(self):
            return OrganizationVocModel(self.request).read(self.KOD_GOL_ORG)

    database_table_name = 'VocOrgs'
