import datetime

import settings

from repository.types import types_repository
from repository.vocs import vocs_repository
from repository.vocs.exceptions import VocRepositoryNotFoundException
from services.exceptions import CoreBaseException
from wsgi.apps.users.models import UserModel
from wsgi.apps.voc import services
from wsgi.apps.voc.models import VocValueModel, OrganizationVocModel
from wsgi.apps.users.services import bind_new_valid_org_version_to_users
from wsgi.serializers import BaseSerializer


class VocValuesSerializer(BaseSerializer):
    model = VocValueModel(None)
    model_fields = model.Data.fields()

    change_request_args_filter_sort_extract = {
        'attrs.status': 'value_status',
    }

    def __init__(self, request, user: UserModel.Data, voc_type):
        super().__init__(request)
        self.voc_type = voc_type
        self.voc_repository = vocs_repository.get_voc_repository(self.voc_type)
        self.user = user

        if 'sort' in self.voc_repository and settings.DB_SYSTEM != 'sqlite':
            self.default_sort = ','.join('attrs.' + field for field in self.voc_repository['sort'])

        # получаем из репозитория список разрешенных операций для фильтрации полей.
        # если в запросе содержится неподдерж. операция, она не будет включена в итоговый sql.
        self.allowed_filter_opers_for_fields.update(
            {m['field']: m.get('filter', None) for m in self.voc_repository['model']})
        # возможная замена полей внутри attrs на системные поля
        # пример - value_status - систменое поле, см. change_request_args_filter_sort_extract
        keys_to_pop = []
        for k, v in self.change_request_args_filter_sort_extract.items():
            if '.' in k:
                k = k.split('.')[1]
            if k in self.allowed_filter_opers_for_fields:
                self.allowed_filter_opers_for_fields[v] = self.allowed_filter_opers_for_fields[k]
                keys_to_pop.append(k)
        for k in keys_to_pop:
            self.allowed_filter_opers_for_fields.pop(k)

    def _after_init(self):
        super()._after_init()
        if settings.DB_SYSTEM == 'sqlite':
            if (self.extract_field and 'json_extract' in self.extract_field) or \
                    (self._sort and 'json_extract' in self._sort) or \
                    (self._filter and 'json_extract' in self._filter):
                raise CoreBaseException(
                    type='not_implemented',
                    rus_message='Фильтрация и сортировка не поддерживается на сервере',
                )

    def _validate(self, data, instance=None):
        voc_repository = vocs_repository.get_voc_repository(data.get('voc_type'))
        if voc_repository.get('voc_subtype', 'custom') == 'system':
            voc_type = data.get('voc_type')
            value_id = str(data['attrs'].get(voc_repository['value_id'], ''))
            filter_ = [
                'and',
                ['=', 'voc_type', voc_type],
                ['=', 'value_id', value_id]
            ]
            # если редактирование существующей записи - не включаем данную запись в поиск одинаковых значений
            if instance:
                filter_.append(['=', 'record_id', instance.record_id, 'not'])
            voc_value_with_same_voc_type_value_id = self.model.filter(filter_)
            if voc_value_with_same_voc_type_value_id:
                field_caption = vocs_repository.get_model_voc_model_item(voc_type, voc_repository['value_id'])[
                    'caption']
                raise CoreBaseException(
                    rus_message=f'Невозможно создать запись для системного справочника, значение '
                                f'"{field_caption}" = "{value_id}" уже существует',
                    not_repeat=True,  # для фронта, чтобы запросы не слались повторно
                )

    def _response(self, resp):
        attrs = resp.get('attrs') if resp.get('attrs') is not None else {}
        not_system_attrs = []
        for k, v in resp.items():
            if k not in self.model_fields and k not in attrs:
                attrs[k] = v
                not_system_attrs.append(k)
        for attr in not_system_attrs:
            resp.pop(attr)
        resp['attrs'] = attrs

        # добавление статуса записи справочника в attrs для отображения в таблице (readonly)
        resp['attrs']['status'] = str(resp['value_status']) if resp['value_status'] is not None else ''
        # разрешено ли редактирования справочника для пользователя
        try:
            resp['edit'] = services.calc_permissions_for_voc_record_edit(voc_type=resp['voc_type'],
                                                                         voc_creator_id=resp['creator_id'],
                                                                         user=self.user,
                                                                         voc_obj=self.model.Data.from_dict(resp))
        except VocRepositoryNotFoundException:
            resp['edit'] = True

        if self.voc_type is not None:
            resp['voc_type'] = self.voc_type

        # если указано в репозитории - обновляем label на лету
        if self.voc_repository.get('refresh_label'):
            resp['label'] = services.build_voc_label(resp, self.voc_repository)

        # формирование value_id на лету, если указан параметр в репозитории
        if 'value_id_return' in self.voc_repository:
            if self.voc_repository['value_id_return'] == 'label':
                resp['value_id'] = resp['label']
            else:
                resp['value_id'] = resp['attrs'][self.voc_repository['value_id_return']]

        return super()._response(resp)

    def _prepare(self, data):
        super()._prepare(data)
        if 'voc_type_db' in self.voc_repository:
            data['voc_type'] = self.voc_repository['voc_type_db']

        # Попытка приведения тела строки к типам (нужно, для SQL сортировки)
        for field, type_name in vocs_repository.get_voc_fields_to_types(self.voc_repository).items():
            python_type = types_repository.get_python_type_if_repo_type_is_simple(type_name)
            if python_type:
                try:
                    if python_type == str and data['attrs'][field] is None:
                        data['attrs'][field] = ''
                    else:
                        data['attrs'][field] = python_type(data['attrs'][field])
                except (ValueError, KeyError):
                    pass

    def _common(self, data, obj: model.Data):
        obj.label = services.build_voc_label(obj.as_dict(), self.voc_repository)
        obj.value_id = str(obj.attrs.get(self.voc_repository['value_id']))

    def save(self, data):
        obj = self.model.Data.from_dict(data)
        obj.creator_id = self.user.login

        # task RVRAO-757
        voc_fields = [m['field'] for m in self.voc_repository['model']]
        user_org = self.user.org_active_version_by_kod_org_obj
        if 'code_OKPO_org' in voc_fields and user_org:
            obj.attrs['code_OKPO_org'] = user_org.OKPO if user_org.OKPO is not None else ''
        if 'adding_date' in voc_fields:
            obj.attrs['adding_date'] = str(datetime.datetime.today().date())

        return obj

    def list(self, *args, **kwargs):
        voc_type = self.voc_repository.get('voc_type_db', self.voc_type)
        filter_params = {'voc_type': voc_type}
        if kwargs.get('only_active'):  # отображать только активные записи справочника
            filter_params['value_status'] = self.model.Data.ValueStatusesTypes.ACTIVE.value
        return self.filter(filter_params)


class OrganizationsVocSerializer(VocValuesSerializer):
    model = OrganizationVocModel(None)

    read_only_fields = ['valid_to', 'value_id', 'label', 'creator_id']

    def __init__(self, request, user: UserModel.Data, voc_type):
        super().__init__(request, user, voc_type)
        if self.extract_field:
            if self.extract_field == 'value_id' and 'value_id_return' in self.voc_repository:
                self.extract_field = self.voc_repository['value_id_return']

    def _prepare(self, data):
        super()._prepare(data)
        if data.get('attrs'):
            for attr, value in data['attrs'].items():
                if attr in self.model.Data.fields_except_id():
                    if value == '':
                        value = None
                    data[attr] = value
            for attr in self.model.Data.fields_except_id():
                data['attrs'].pop(attr, None)
        # работа с типами организации
        if data.get('ORG_TYPE'):
            org_type = data['ORG_TYPE']
            if org_type == '0':
                data['PRZN_OP'] = 0  # Юридическое лицо
                data['ORG_TYPE'] = None
            elif org_type == '1':
                data['PRZN_OP'] = 1  # Филиал, представительство, обособленное подразделение
                data['ORG_TYPE'] = None
            elif org_type == '2':
                data['ORG_TYPE'] = '1'  # РИАЦ
            elif org_type == '3':
                data['ORG_TYPE'] = '2'  # ВИАЦ
            elif org_type == '4':
                data['ORG_TYPE'] = '3'  # ЦИАЦ
        # ЯРОО
        if data.get('YROO'):
            if data['YROO'] == 'false':
                data['YROO'] = False
            elif data['YROO'] == 'true':
                data['YROO'] = True

    def _response(self, resp):
        resp = super()._response(resp)
        if resp['attrs']['PRZN_OP'] == 0:
            resp['attrs']['ORG_TYPE'] = '0'
        elif resp['attrs']['PRZN_OP'] == 1:
            resp['attrs']['ORG_TYPE'] = '1'
        elif resp['attrs']['ORG_TYPE'] == '1':
            resp['attrs']['ORG_TYPE'] = '2'
        elif resp['attrs']['ORG_TYPE'] == '2':
            resp['attrs']['ORG_TYPE'] = '3'
        elif resp['attrs']['ORG_TYPE'] == '3':
            resp['attrs']['ORG_TYPE'] = '4'

        if resp['attrs']['YROO'] is None:
            pass
        elif not resp['attrs']['YROO']:
            resp['attrs']['YROO'] = 'false'
        else:
            resp['attrs']['YROO'] = 'true'

        # права на действие "Архивирование записи справочника"
        resp['edit']['to_archive'] = (
                resp['edit']['edit'] and
                resp['value_status'] == OrganizationVocModel.Data.ValueStatusesTypes.ACTIVE.value
        )
        # readonly поля, если поле имеет статус отличный от "Активна"
        resp['edit']['readonly'] = {
            'valid_to': True,
            # 'attrs': {
            #     'KOD_ORG': True,
            # },
        }

        resp['kod_org'] = resp['attrs']['KOD_ORG']

        return resp

    def _common(self, data, obj: model.Data):
        BaseSerializer._common(self, data, obj)

    def save(self, data) -> model.Data:
        obj = self.model.Data.from_dict(data)
        obj.valid_from = datetime.datetime.now().date()
        obj.valid_to = None
        obj.creator_id = self.user.login
        if obj.value_status is None:
            obj.value_status = self.model.Data.ValueStatusesTypes.ACTIVE
        return obj

    def _after_save(self, obj: model.Data):
        voc_repository = vocs_repository.get_voc_repository(obj.voc_type)
        d = obj.as_dict()
        obj.label = services.build_voc_label(self._response(obj.as_dict()), voc_repository)
        obj.value_id = str(d.get(voc_repository['value_id']))
        return obj

    def update(self, record_id, data):
        instance = self.model.read(record_id)
        obj = instance.replace(data)
        # obj.KOD_ORG = instance.KOD_ORG  # запрет на редактирование кода организации
        obj.value_status = instance.value_status

        # перерасчет дат организаций
        if obj.valid_from != instance.valid_from:
            all_orgs_versions = obj.get_all_orgs_versions_objs()
            # заменяем текущую запись организации на obj
            if not all_orgs_versions:
                return obj
            all_orgs_versions = [obj if v.record_id == obj.record_id else v for v in all_orgs_versions]
            # есть ли активная версия организации
            is_any_active_org_version = any(org.value_status == OrganizationVocModel.Data.ValueStatusesTypes.ACTIVE for org in all_orgs_versions)
            # ошибка при сортировке по дате начала, явное приведение к типу date
            for org in all_orgs_versions:
                if isinstance(org.valid_from, datetime.datetime):
                    org.valid_from = org.valid_from.date()
            # сортируем по дате начала периода действия
            all_orgs_versions.sort(key=lambda v: v.valid_from)
            # все организациям кроме последней рассчитываем даты заново, задаем статус "Устаревшая"
            for i in range(len(all_orgs_versions) - 1):
                all_orgs_versions[i].valid_to = all_orgs_versions[i + 1].valid_from - datetime.timedelta(days=1)
                all_orgs_versions[i].value_status = OrganizationVocModel.Data.ValueStatusesTypes.OUTDATED
            # последнюю запись делаем активной
            if is_any_active_org_version:
                current_active_version = all_orgs_versions[-1]
                current_active_version.value_status = OrganizationVocModel.Data.ValueStatusesTypes.ACTIVE
                for org in all_orgs_versions:
                    self.model.update(org.record_id, org)

        return obj


class OrganizationsVocVersionsSerializer(OrganizationsVocSerializer):

    def __init__(self, request, user: UserModel.Data, record_id):
        super().__init__(request, user, 'organizations')
        self.org = self.model.read(record_id)

    def _validate(self, data, instance=None):
        super()._validate(data)
        if self.org.voc_type != 'organizations':
            raise CoreBaseException(rus_message='Создание новой версии не поддерживается для данного справочника')
        if self.org.valid_to is not None and self.org.value_status != OrganizationVocModel.Data.ValueStatusesTypes.ACTIVE:
            raise CoreBaseException(rus_message='Невозможно создать новую версию для данной записи, так как '
                                                'запись не является активной',
                                    message='valid_to should be equal None')
        if not self.org.KOD_ORG:
            raise CoreBaseException(rus_message='Невозможно создать новую версию организации, т.к. текущая версия '
                                                'имеет пустой регистрационный номер.')

        filter_params = {
            'KOD_ORG': self.org.KOD_ORG,
            'valid_to': None,
        }
        orgs = OrganizationVocModel(self.request).filter(filter_params)

        if len(orgs) > 1:
            raise CoreBaseException(
                message=f'more than one record of organizations with KOD_ORG={self.org.KOD_ORG} and valid_to=None')
        elif not orgs:
            raise CoreBaseException(
                message=f'no record of organizations with KOD_ORG={self.org.KOD_ORG} and valid_to=None')

    def save(self, data):
        """Сохранение новой версии организации"""
        # обновление текущей версии
        self.org.valid_to = datetime.datetime.today() - datetime.timedelta(
            days=1)  # окончание действия: тек. дата - 1 день
        self.org.value_status = VocValueModel.Data.ValueStatusesTypes.OUTDATED
        OrganizationVocModel(self.request).update(self.org.record_id, self.org)
        # создание новой версии
        obj = self.model.Data.from_dict(data)
        obj.valid_from = datetime.datetime.now().date()
        obj.valid_to = None
        obj.creator_id = self.user.login
        obj.KOD_ORG = self.org.KOD_ORG  # запрет на изменение кода организации
        obj.value_status = self.model.Data.ValueStatusesTypes.ACTIVE
        return obj

    def _after_save(self, obj):
        obj = super(OrganizationsVocVersionsSerializer, self)._after_save(obj)
        return obj

    def list(self):
        """Получить версии организации"""
        try:
            return sorted(self.filter({'KOD_ORG': self.org.KOD_ORG}), key=lambda org: org['valid_from'], reverse=True)
        except:
            return self.filter({'KOD_ORG': self.org.KOD_ORG})
