import re
from typing import List

from datasource.databaserepository.database_repository2 import DatabaseRepository2
from repository.vocs import vocs_repository
from services.exceptions import PermissionDeniedException
from wsgi.apps.users.models import UserModel
from wsgi.apps.voc.models import OrganizationVocModel, VocValueModel
from wsgi.apps.voc.permissions import check_edit_system_vocs_permission, check_edit_vocs_permissions, \
    check_edit_users_voc_permissions


def build_voc_label(voc, voc_rep, attrs_wrap=True):
    label = voc_rep['label']
    if isinstance(label, str):
        regex = r'{\w+}'
        fields_in_label = re.findall(regex, label)
        fields_clear = [f[1:-1] for f in fields_in_label]
        fields_zip = zip(fields_in_label, fields_clear)
        for f_l, f_c in fields_zip:
            if attrs_wrap:
                f_c = '' if voc['attrs'][f_c] is None else voc['attrs'][f_c]
                label = label.replace(f_l, str(f_c))
            else:
                f_c = '' if voc[f_c] is None else voc[f_c]
                label = label.replace(f_l, f_c)
    elif isinstance(label, list):
        label = ' '.join(str(voc['attrs'].get(attr, '')) for attr in voc_rep['label'])
    return label


def get_stock_list(request, form_no, org_id, date, *args, **kwargs):
    """Получить список наличного количества по номеру формы (целое число), id орг., дате"""
    sql_line = request.storage['api2_clone'].regs.get_snk_sql(form_no, date, org_id)
    return DatabaseRepository2(request).execute(sql_line, commit=True)


def calc_permissions_for_voc_edit(voc_type, user: UserModel.Data):
    """Может ли пользователь редактировать или создавать записи у справочника с типом <voc_type>"""
    voc_rep = vocs_repository.get_voc_repository(voc_type)
    voc_subtype = voc_rep.get('voc_subtype', 'custom')

    if voc_type == 'user':
        try:
            check_edit_users_voc_permissions(user)
            return True
        except PermissionDeniedException:
            return False

    if voc_subtype in ('system', 'settings'):
        if voc_type == 'bugreports':
            return True
        try:
            check_edit_system_vocs_permission(user)
        except PermissionDeniedException:
            return False
    else:
        try:
            check_edit_vocs_permissions(user)
        except PermissionDeniedException:
            return False

    return True


def calc_permissions_for_voc_record_edit(voc_type, voc_creator_id, user: UserModel.Data, voc_obj: VocValueModel.Data):
    """Может ли пользователь редактировать запись справочника"""
    edit = {
        'edit': False,
    }

    voc_rep = vocs_repository.get_voc_repository(voc_type)
    voc_subtype = voc_rep.get('voc_subtype', 'custom')
    calc_perm = calc_permissions_for_voc_edit(voc_type, user)
    if calc_perm and voc_subtype not in ('system', 'settings'):
        if voc_creator_id != user.login and user.user_role != user.UserRoleTypes.ADMINISTRATOR:
            edit['edit'] = False

    edit['edit'] = calc_perm

    st = OrganizationVocModel.Data.ValueStatusesTypes
    edit['to_archive'] = edit['edit'] and voc_obj.value_status == st.ACTIVE
    edit['to_make_active'] = edit['edit'] and voc_obj.value_status in (st.DRAFT, st.OUTDATED)

    return edit


def get_all_orgs_ids_with_same_KOD_ORG(request, orgs: List[OrganizationVocModel.Data]):
    """Получить список org_id организаций, имеющих тот же KOD_ORG.
    Организации с тем же KOD_ORG - разные версии одной и той же организации"""
    total_orgs_ids = set()
    for org_id in orgs:
        org = OrganizationVocModel(request).read(org_id)
        total_orgs_ids.update(org_.record_id for org_ in org.get_all_orgs_versions_objs())
    return list(total_orgs_ids)
