import json

import werkzeug

from datasource.databaserepository.database_repository2 import DatabaseRepository2
from datasource.datasource import get_db_connection
from services.exceptions import CoreBaseException
from services.filter import build_filter_sql_line
from services.sort import build_sort_sql_line


class MetaBaseSerializer(type):
    def __init__(cls, name, bases, clsdict):
        if 'save' in clsdict:
            def new_save(self, data):
                if not self._aft_init:
                    self._after_init()
                    self._aft_init = True
                if not self._prep:
                    self._prepare(data)
                    self._prep = True
                if not self._val:
                    self._validate(data)
                    self._val = True
                obj = clsdict['save'](self, data)
                if not self._com:
                    self._common(data, obj)
                    self._com = True
                if not self._sav:
                    if self.create_connect:
                        self.model.conn = get_db_connection()
                    saved_obj = self.model.save(obj, commit=self.auto_commit)
                    self._sav = True
                if not self._aft_save:
                    obj_after_save = self._after_save(saved_obj)
                    if obj_after_save is not None:
                        saved_obj = self.model.update(obj_after_save.get_id(),
                                                      obj_after_save)
                    self._aft_save = True
                if not self._resp:
                    self._resp = True
                    return self._response(saved_obj.as_dict())
                return obj

            setattr(cls, 'save', new_save)
        if 'update' in clsdict:
            def new_update(self, record_id, data):
                if not self._aft_init:
                    self._after_init()
                    self._aft_init = True
                if not self._prep:
                    self._prepare(data)
                    self._prep = True
                if not self._val:
                    self._validate(data, self.model.read(record_id))
                    self._val = True
                obj = clsdict['update'](self, record_id, data)
                if not self._com:
                    self._common(data, obj)
                    self._com = True
                if not self._upd:
                    saved_obj = self.model.update(record_id, obj, commit=self.auto_commit)
                    self._upd = True
                if not self._aft_save:
                    res = self._after_save(obj)
                    if res is not None:
                        obj = self.model.update(record_id, res)
                        saved_obj = self.model.update(record_id, obj)
                    self._aft_save = True
                if not self._resp:
                    self._resp = True
                    return self._response(saved_obj.as_dict())
                return obj

            setattr(cls, 'update', new_update)
        if 'read' in clsdict:
            def new_read(self, record_id):
                if not self._aft_init:
                    self._after_init()
                    self._aft_init = True
                obj = clsdict['read'](self, record_id)
                if not self._resp:
                    self._resp = True
                    return self._response(obj)
                return obj

            setattr(cls, 'read', new_read)
        if 'list' in clsdict:
            def new_list(self, *args, **kwargs):
                if not self._aft_init:
                    self._after_init()
                    self._aft_init = True
                objs = clsdict['list'](self, *args, **kwargs)
                if not self._resp:
                    self._resp = True
                    if self.extract_field:
                        return self._extract(objs)
                    resp = [self._response(obj) for obj in objs]
                    return resp
                return objs

            setattr(cls, 'list', new_list)
        if 'filter' in clsdict:
            def new_filter(self, params):
                if not self._aft_init:
                    self._after_init()
                    self._aft_init = True
                objs = clsdict['filter'](self, params)
                if not self._resp:
                    self._resp = True
                    if self.extract_field:
                        return self._extract(objs)
                    resp = [self._response(obj) for obj in objs]
                    return resp
                return objs

            setattr(cls, 'filter', new_filter)
        if 'query_condition' in clsdict:
            def new_query_condition(self, condition, filter_params=None):
                if not self._aft_init:
                    self._after_init()
                    self._aft_init = True
                objs = clsdict['query_condition'](self, condition, filter_params)
                if not self._resp:
                    self._resp = True
                    if self.extract_field:
                        return self._extract(objs)
                    resp = [self._response(obj) for obj in objs]
                    if self.extract_field:
                        return self._extract(resp)
                    return resp
                return objs

            setattr(cls, 'query_condition', new_query_condition)

        super().__init__(name, bases, clsdict)


class BaseSerializer(metaclass=MetaBaseSerializer):
    model = DatabaseRepository2(None)
    data = model.Data

    read_only_fields = []
    write_only_fields = []

    default_sort = ''
    select_join = []

    fields = ['*']

    change_request_args_filter_sort_extract = {}
    change_filter_objs = {}

    auto_commit = False
    create_connect = False

    def __init__(self, request, connection=None):
        self.request = request
        self.conn = request.storage['conn'] if connection is None else connection
        if connection is not None:
            self.request.storage['conn'] = connection

        self.model.conn = self.conn
        self.model.request = self.request
        self.model.api2_clone = self.request.storage['api2_clone']

        # список разрешенных операция для фильтрации. по умолчанию поле модели -> 'all' (разрешены все операции)
        self.allowed_filter_opers_for_fields = {k: 'all' for k in self.model.Data.fields()}

        self.request_args = request_args = dict(request.args)
        for k, v in self.change_request_args_filter_sort_extract.items():
            if k in request_args:
                request_args[k] = v

        # переопределение имен параметров при фильтрации, сортировке и т.д.
        # например, если в таблицу справочника помещен системный атрибут (обернут в attrs)
        # 'attrs.status' -> 'value_status' (системное поле)

        # PAGINATION
        self.offset = request_args.get('offset', '')

        if isinstance(self.offset, list):
            self.offset = self.offset[0]

        self.limit = request_args.get('limit', '')
        if isinstance(self.limit, list):
            self.limit = self.limit[0]

        self.extract_field = request_args.get('extract', '')
        if isinstance(self.extract_field, list):
            self.extract_field = self.extract_field[0]

        self.unique = request_args.get('unique', '')
        if isinstance(self.unique, list):
            self.unique = self.unique[0]

        if self.unique == 'true':
            self.unique = True
        else:
            self.unique = False
        if not self.extract_field:
            self.unique = ''

        for k in ('extract', 'filter', 'sort'):
            if k in request_args:
                for req_arg_original, change_to_arg in self.change_request_args_filter_sort_extract.items():
                    request_arg_value = request_args[k]
                    if isinstance(request_arg_value, list):
                        request_arg_value = request_arg_value[0]
                    request_args[k] = request_arg_value.replace(req_arg_original, change_to_arg)

        self.join_line = []

        # SELECT JOIN
        # self.select_join_lines = []
        self.select_join_params_lines = []
        self.select_join_tables_lines = []
        # имя параметра ответа в join строке. нужен для extract
        self.response_params_to_extract_join_param_line = {}
        self.total_select_join_params_line = ''
        self.total_select_join_tables_line = ''
        if self.select_join:
            for r in self.select_join:
                join_param_line = f'"{r[1]}"."{r[3]}" as "{r[4]}"'
                self.response_params_to_extract_join_param_line[r[4]] = f'"{r[1]}"."{r[3]}"'
                select_join_line = f'left join "{r[1]}" on "{self.model.database_table_name}"."{r[0]}" = "{r[1]}"."{r[2]}"'
                self.select_join_params_lines.append(join_param_line)
                self.select_join_tables_lines.append(select_join_line)
            self.total_select_join_params_line = ', ' + ', '.join(self.select_join_params_lines)
            self.total_select_join_tables_line = ' '.join(self.select_join_tables_lines)

        if self.offset:

            if not self.limit:
                self.limit = ''

            try:
                int(self.offset)
            except ValueError:
                raise CoreBaseException(message='offset param should be int')

        if self.limit:
            try:
                int(self.limit)
            except ValueError:
                raise CoreBaseException(message='offset param should be int')

            self.limit = str(int(self.limit) + 1)

        # FILTERING
        # todo: вынести в BaseModel
        self._filter = request_args.get('filter', '')
        self.filter_warnings = None

        # SORTING
        self._sort = None

        self.join_line = ' '.join(self.join_line)

        self._prep = False
        self._val = False
        self._resp = False
        self._com = False
        self._aft_save = False

        self._sav = False
        self._upd = False

        self._aft_init = False

        self.storage = {}

    def _after_init(self):
        if self.extract_field:
            self.available_params_for_extract = list(self.model.Data.fields().keys()) + [sj[4] for sj in self.select_join]
            if '.' in self.extract_field:
                if self.extract_field.split('.')[0] not in self.available_params_for_extract:
                    raise CoreBaseException(message=f'can not extract by field {self.extract_field}, field does not exists')
            elif self.extract_field not in self.available_params_for_extract:
                raise CoreBaseException(message=f'can not extract by field {self.extract_field}, field does not exists')

        # EXTRACT
        if self.extract_field:
            self.extract_field_line = self.extract_field
            if self.extract_field in self.response_params_to_extract_join_param_line:
                self.extract_field_line = self.response_params_to_extract_join_param_line[self.extract_field]
            if '.' in self.extract_field:
                json_field, json_par = self.extract_field.split('.')
                self.extract_field_line = f'"{json_field}"->>\'{json_par}\''

        if self._filter:
            try:
                # получение sql строки фильтрации
                # второй параметр - как указывать параметр, если параметра нет в исходной таблицы (join таблиц)
                self._filter, filter_fields_to_values, self.filter_warnings = build_filter_sql_line(
                    self._filter, self.model.database_table_name, self.allowed_filter_opers_for_fields, {x[4]: f'"{x[1]}"."{x[3]}"' for x in self.select_join},
                    fields_to_types=self.model.Data.fields(),
                    change_filter_objs=self.change_filter_objs,
                    request=self.request,
                )
            except json.decoder.JSONDecodeError:
                raise CoreBaseException(message=f'incorrect filter format, can not parse it. Filter: {self._filter}')

            # проверка разрешения сортировки по полю
            fields_from_join_tables = [item[4] for item in self.select_join]
            for filter_field, filter_value in filter_fields_to_values.items():
                # если поле не находится в модели и в полях сериализации
                if filter_field not in self.model.Data.fields_except_id() and filter_field not in fields_from_join_tables:
                    # и поле не находится в объекте для join с другой таблицей
                    raise CoreBaseException(message=f'it is impossible to filter by this field: {filter_field}, '
                                                    f'perhaps the field is inside the json attribute')

        self.request.storage['filter_warnings'] = self.filter_warnings

        self._sort = self.request_args.get('sort', self.default_sort)
        if self._sort:
            try:
                self._sort = build_sort_sql_line(self._sort)
            except:
                raise CoreBaseException(message='incorrect sort format, can not parse it')

    def _extract(self, resp):
        """Метод для получения только одного поля в ответе (для каждого объекта)"""
        if not resp:
            return []
        return [list(r.values())[0] for r in resp]

    def _validate(self, data, instance=None):
        pass

    def _prepare(self, data):
        for field in self.read_only_fields:
            data.pop(field, None)

    def _response(self, resp):
        for field in self.write_only_fields:
            resp.pop(field, None)

        return resp

    def _common(self, data, obj: 'model.Data'):
        pass

    def _after_save(self, obj: 'model.Data'):
        pass

    def save(self, data):
        return self.model.Data.from_dict(data)

    def update(self, record_id, data):
        instance = self.model.read(record_id)
        obj = instance.replace(data)
        return obj

    def delete(self, record_id):
        return self.model.delete(record_id)

    def read(self, record_id):
        return self.model.read_raw(
            record_id,
            fields_=self.fields,
            select_join_params_line=self.total_select_join_params_line,
            select_join_tables_line=self.total_select_join_tables_line,
        )

    def list(self, *args, **kwargs):
        if self._filter:
            return self.filter({})
        return self.model.list_raw(
            fields_=self.fields if not self.extract_field else (self.extract_field_line, ),
            limit=self.limit,
            offset=self.offset,
            filter_sql=self._filter,
            sort_sql=self._sort,
            select_join_params_line=self.total_select_join_params_line if not self.extract_field else '',
            select_join_tables_line=self.total_select_join_tables_line,
            unique=self.unique,
        )

    def filter(self, args):
        not_filter_params = ('action', 'offset', 'limit', 'extract', 'unique')
        if isinstance(args, werkzeug.datastructures.ImmutableMultiDict):
            args = {arg: args.getlist(arg) for arg in args if arg not in not_filter_params}

        return self.model.filter_raw(
            args,
            fields_=self.fields if not self.extract_field else (self.extract_field_line, ),
            limit=self.limit,
            offset=self.offset,
            filter_sql=self._filter,
            sort_sql=self._sort,
            select_join_params_line=self.total_select_join_params_line if not self.extract_field else '',
            select_join_tables_line=self.total_select_join_tables_line,
            unique=self.unique,
        )

    def query_condition(self, condition, filter_params=None):
        if filter_params is None:
            filter_params = {}
        return self.model.filter_raw(
            params=filter_params,
            fields_=self.fields if not self.extract_field else (self.extract_field_line, ),
            query_condition=condition,
            filter_sql=self._filter,
            sort_sql=self._sort,
            select_join_params_line=self.total_select_join_params_line if not self.extract_field else '',
            select_join_tables_line=self.total_select_join_tables_line,
            unique=self.unique,
            limit=self.limit,
            offset=self.offset,
        )
