import functools
import json
import os
import traceback

from loguru import logger
from werkzeug.local import Local, LocalManager
from werkzeug.utils import send_file
from werkzeug.wrappers import Request as BaseRequest
from werkzeug.wrappers.response import Response as Rr

import settings
from services.exceptions import CoreBaseException
from settings import DEBUG, RAISE_EX
from wsgi.apps.users.models import UserModel
from wsgi.exceptions import (ActionNotAllowedException,
                             ActionRequiredException, ApiException,
                             ArgRequired, BodyParamsRequired, BodyRequiredException)

local = Local()
local_manager = LocalManager([local])
exported_views = {}


class RequestCopy:
    def __init__(self, conn_, api2_clone=None, request=None):
        self.storage = {
            'conn': conn_,
            'api2_clone': api2_clone,
        }
        self.args = {}
        if request:
            for k, v in request.__dict__.items():
                if k not in ('storage', ):
                    self.__dict__[k] = v


def api2_func_wrapper(func):
    @functools.wraps(func)
    def inner(request, *args, **kwargs):
        if 'api2_clone' in kwargs or 'connection' in kwargs:
            request = RequestCopy(
                kwargs.pop('connection', request.storage['conn']),
                kwargs.pop('api2_clone', request.storage['api2_clone']),
                request
            )
        return func(request, *args, **kwargs)

    return inner


def print_func_info(func):
    print(func.__doc__)


def api_view(db_connection_=False, auth_required=True):
    def _api_view(func):
        @functools.wraps(func)
        def inner(request, *args, **kwargs):
            if auth_required:

                if 'TestUser' in request.headers and request.headers['TestUser'] and settings.DEBUG or request.full_path.startswith('/forms?action=set_forms_expt_time_unloaded_status'):
                    login = request.headers['TestUser']
                    try:
                        kwargs['user'] = UserModel(request).filter({'login': login})[0]
                    except IndexError:
                        return error_response(
                            {'rus_message': f'Не найден тестовый пользователь с логином {login}. См Headers'})
                else:
                    cookie = request.cookies.get('Auth')
                    if cookie is None:
                        return error_response({
                            'type': 'auth_required',
                            'rus_message': 'Требуется авторизация'
                        })
                    users = UserModel(request).filter({'cookie': cookie})
                    if not users:
                        resp = error_response({
                            'type': 'auth_failed',
                            'rus_message': 'Неверные данные авторизации'
                        })
                        resp.set_cookie('Auth', cookie, max_age=0)
                        return resp

                    kwargs['user'] = users[0]

            try:
                res = func(request, *args, **kwargs)
            except CoreBaseException as e:
                if RAISE_EX:
                    raise e
                return error_response(e.__dict__)
            except Exception as e:
                logger.error(traceback.format_exc())
                if DEBUG:
                    raise e
                return error_response(
                    {
                        'message':  traceback.format_exc(),
                    }
                )

            return res

        return inner

    return _api_view


def parse_request(rules):
    def _parse_request(f):
        @functools.wraps(f)
        def inner(request, *args, **kwargs):
            try:
                method = request.method
                opposite_method = 'POST' if method == 'GET' else 'GET'

                action = request.args.get('action')
                if action is None:
                    raise ActionRequiredException()

                if action not in rules[method]:
                    if opposite_method not in rules or action not in rules[opposite_method]:
                        raise ActionNotAllowedException(action=action, method=method)
                    else:
                        method = opposite_method

                query_params = rules[method][action].get('query_params')
                request_args = dict(request.args)
                # проверка параметров запроса
                if query_params is not None:
                    query_params_required = []
                    for query_param, query_param_type in query_params.items():
                        if not query_param.endswith('_?') and request.args.get(query_param) is None:
                            query_params_required.append(query_param)
                        query_param_ = query_param
                        if query_param.endswith('_?'):
                            query_param_ = query_param[:-2]
                        if request.args.get(query_param_) is not None:
                            if query_param_type == bool:
                                request_args[query_param_] = True if request.args[query_param_] == 'true' else False
                            else:
                                request_args[query_param_] = query_param_type(request.args[query_param_])
                    if query_params_required:
                        raise ArgRequired(arg=query_params_required)

                def validate_body_params(body_, body_rules_, body_params_required_):
                    for body_rules_param, body_rules_param_type in body_rules_.items():
                        if body_rules_param.endswith('_?'):
                            continue
                        if body_.get(body_rules_param) is None:
                            body_params_required_.append(body_rules_param)
                        else:

                            # if isinstance(body_rules_param_type, type):  # TODO
                            # validate_body_params(body_[body_rules_param], body_rules_[body_rules_param], body_params_required_)

                            if isinstance(body_rules_param_type, dict):
                                validate_body_params(body_[body_rules_param], body_rules_[body_rules_param],
                                                     body_params_required_)
                            elif isinstance(body_rules_param_type, list):
                                validate_body_params(body_[body_rules_param][0], body_rules_[body_rules_param][0],
                                                     body_params_required_)

                body_rules = rules[method][action].get('body')
                body = request.json

                if body_rules is not None:
                    if request.json is None:
                        raise BodyRequiredException()
                    body_params_required = []
                    if isinstance(body, list):
                        for body_item in body:
                            validate_body_params(body_item, body_rules, body_params_required)
                    else:
                        validate_body_params(body, body_rules, body_params_required)
                    if body_params_required:
                        raise BodyParamsRequired(body_params=body_params_required, required_structure=body_rules)

                # По какой-то причине, на windows с python3.7 заносится как list
                for par in ('action', 'limit', 'offset', 'extract', 'unique'):
                    if par in request_args and isinstance(request_args[par], list):
                        request_args[par] = request_args[par][0]

                res = f(request, action, *args, **request_args, body=request.json)
                if res is None:
                    return error_response({'msg': 'not implemented'})
                return res
            except ApiException as e:
                return error_response(e.__dict__)

        return inner

    return _parse_request


def set_exported_views2():
    from wsgi.urls import urls as base_urls

    for base_url in base_urls:
        if len(base_url) == 2:
            url_string, app = base_url
            logger.info(f'Регистрация views. Регистрация app: {app}')
            try:
                module = __import__(f"wsgi.apps.{app}.urls", globals(), locals(), ['urls'])
            except ModuleNotFoundError as e:
                logger.error(f'Не удалось импортировать app: {app}. Исключение: {e}')
                continue
            for app_url_string, methods, app_fun in module.urls:
                endpoint = '.'.join(['wsgi.apps', app, 'views', app_fun.__name__])
                total_url_string = url_string + app_url_string
                exported_views[endpoint] = (app_fun, methods, total_url_string)
        else:
            url_string, methods, fun = base_url
            endpoint = '.'.join(['wsgi.views', fun.__name__])
            total_url_string = url_string
            exported_views[endpoint] = (fun, methods, total_url_string)


class Request(BaseRequest):
    """
    The concrete request object used in the WSGI application.
    It has some helper functions that can be used to build URLs.
    """

    charset = "utf-8"

    def __init__(self, environ, url_adapter):
        super().__init__(environ)
        self.url_adapter = url_adapter
        local.request = self


def error_response(data, **kwargs):
    status = kwargs.pop('status', 400)
    errors_list = data
    if isinstance(data, dict):
        errors_list = [data]
    data = {
        'errors': errors_list
    }
    return Rr(json.dumps(data, default=str, ensure_ascii=False), content_type='application/json', status=status)


def response(request, data, many=False, ignore_format=False, **kwargs):
    status = kwargs.pop('status', 200)

    if ignore_format:
        response_data = json.dumps(data, default=str)
        return Rr(response_data, content_type='application/json', status=status)

    if request is not None:
        offset = request.args.get('offset', '')
        limit = request.args.get('limit', '')

        if offset or limit:
            if offset:
                offset = int(request.args.get('offset', ''))
            if limit:
                limit = int(request.args.get('limit', ''))

                try:
                    data.pop(limit)
                    values_remain = True
                except IndexError:
                    values_remain = False
            else:
                values_remain = False

            response_data = {
                'values_remain': values_remain,
                'params': {},
                'result': data,
            }
            if request.storage.get('filter_warnings'):
                response_data['filter_warnings'] = request.storage['filter_warnings']

            if offset != '':
                response_data['params']['offset'] = offset
            if limit != '':
                response_data['params']['limit'] = limit

            response_data = json.dumps(response_data, default=str, ensure_ascii=False)
            return Rr(response_data, content_type='application/json', status=status)

    response_data = json.dumps(data, default=str, ensure_ascii=False)
    if many:
        response_data = {
            'count': len(data),
            'result': data,
        }
        if request.storage.get('filter_warnings'):
            response_data['filter_warnings'] = request.storage['filter_warnings']
        response_data = json.dumps(response_data, default=str, ensure_ascii=False)
    return Rr(response_data, content_type='application/json', status=status)


def split_all(path):
    all_parts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            all_parts.insert(0, parts[0])
            break
        elif parts[1] == path:  # sentinel for relative paths
            all_parts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            all_parts.insert(0, parts[1])
    return all_parts


def file_response(dir_, file_=None, download_name=None):
    if not file_:
        all_ = split_all(dir_)
        dir_ = os.path.join(*all_[:-1])
        file_ = all_[-1]
    path = dir_ + '/' + file_
    return send_file(path, {'REQUEST_METHOD': 'GET', 'RESPONSE_TYPE': 'FILE'}, as_attachment=True,
                     download_name=download_name)


class CoreRequest(Request):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.storage = {}

    @property
    def full_path_with_action(self):
        return self.full_path.split('&')[0]


def get_user(request):
    cookie = request.cookies.get('Auth')
    if cookie is None:
        return None

    users = UserModel(request).filter({'cookie': cookie})
    if not users:
        return None

    return users[0]
