import copy

import werkzeug
from loguru import logger
from werkzeug.exceptions import HTTPException, NotFound
from werkzeug.routing import Map, RequestRedirect, Rule
from werkzeug.utils import send_file

import api2
from datasource.datasource import get_db_connection, close_connection
from web import extra_views, extra_urls_d
from wsgi.utils import local_manager, set_exported_views2, response, CoreRequest, get_user, error_response

import settings
import os


class WSGIApp:
    def __init__(self, config):
        self.config = config

        set_exported_views2()
        from wsgi.utils import exported_views

        self.rules = []
        self.rules_only = []
        self.views = {}

        for endpoint, (func, methods, rule) in exported_views.items():
            if rule is not None:
                self.rules_only.append(Rule(rule, endpoint=endpoint, methods=methods))
                self.rules.append(Rule(rule, endpoint=endpoint, methods=methods))
            for method in methods:
                if endpoint not in self.views:
                    self.views[endpoint] = {}
                self.views[endpoint][method] = func

        self.url_map = Map(self.rules)

    def update_extra_url(self, extra_url):
        self.views.update(extra_views)
        self.rules_only.append(Rule(extra_url['url'],
                                    endpoint=extra_url['endp_get'],
                                    methods=['GET']))
        self.rules_only.append(Rule(extra_url['url'],
                                    endpoint=extra_url['endp_post'],
                                    methods=['POST']))
        rc = copy.deepcopy(self.rules_only)
        self.url_map = Map(rc)

    # @logger.catch
    def __call__(self, environ, start_response, *args, **kwargs):
        urls = self.url_map.bind_to_environ(environ)
        req = CoreRequest(environ, urls)
        conn = get_db_connection()
        try:
            serve_res = self.serve_call(environ, start_response, urls, req, conn)
            conn.commit()
        finally:
            close_connection(conn)
        return serve_res

    def serve_call(self, environ, start_response, urls, req, conn):
        req.storage['conn'] = conn
        req.storage['api2_clone'] = None
        user = get_user(req)
        req.storage['user'] = user
        req.storage['api2_clone'] = new_api = api2.get_api2_clone(conn, req, user)

        try:
            endpoint, args = urls.match(req.path, method=req.method)
            if endpoint.startswith('extra'):
                no_cookie = False
                if 'action_name=import_hist_file&action=exec' in req.full_path:
                    cookie = req.cookies.get('Auth')
                    if not cookie:
                        resp = error_response({
                                'type': 'auth_required',
                                'rus_message': 'Требуется авторизация'
                            })
                        resp.set_cookie('Auth', '', max_age=0)
                        no_cookie = True
                if not no_cookie:
                    endpoint_obj = extra_urls_d[req.path]['endpoint_obj']
                    context = {
                        'args': args,
                        'request': req,
                    }
                    if endpoint_obj is not None:
                        resp = self.views[endpoint][req.method](endpoint_obj, api=new_api, context=context)
                    else:
                        resp = self.views[endpoint][req.method](_api=None, api=new_api, context=context)
                    if not isinstance(resp, werkzeug.Response):
                        resp = response(None, resp)
            else:
                resp = self.views[endpoint][req.method](req, **args, conn=conn)
        except NotFound:
            from wsgi.views import not_found
            resp = not_found()
        except (HTTPException, RequestRedirect) as e:
            resp = e
        try:
            api2.api2.clear_context()
        except:
            pass

        return resp(environ, start_response)


class WSGIAppServeStatic(WSGIApp):

    def serve_call(self, environ, start_response, urls, req, conn):
        if not req.path.startswith('/api') and not req.path.startswith('/forms'):
            from werkzeug.wrappers import Response
            from jinja2 import Environment
            from jinja2 import FileSystemLoader
            jinja_env = Environment(loader=FileSystemLoader(settings.FRONT_BUILD_PATH))

            try:
                sep = '\\' if settings.WIN else '/'
                return send_file(settings.FRONT_BUILD_PATH + sep + req.path,
                                  {'REQUEST_METHOD': 'GET', 'RESPONSE_TYPE': 'FILE'})(environ, start_response)
            except (FileNotFoundError, IsADirectoryError, PermissionError):
                return Response(jinja_env.get_template('index.html').render(**{}), mimetype="text/html")(environ,
                                                                                                         start_response)
        elif req.path.startswith('/api'):
            req.path = req.path[4:]  # убираем префикс /api
        return super().serve_call(environ, start_response, urls, req, conn)


def make_app(config=None, serve_static=False):
    if serve_static:
        app_obj = WSGIAppServeStatic(config)
    else:
        app_obj = WSGIApp(config)

    app_f = local_manager.make_middleware(app_obj)
    return app_obj, app_f
